# -*- coding: utf-8 -*-
#
#  Copyright 2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

#---------- Module Imports -------------#

# Standard Libraries
import sys
import os
import traceback
import re
import json
import base64
import unicodedata
from cStringIO import StringIO
from imp import load_source
from optparse import OptionParser
from ConfigParser import ConfigParser
from ast import literal_eval
from datetime import datetime, timedelta
import hashlib

# Custom module imports
from storage import Storage
from scheduler import Scheduler
from modules.base.mailer import SpeechMailer
from modules.base.media import media2mp4, media2ogv, media2pcm, audio2mp4, make_thumbnail, get_media_length, check_media_audio
from modules.base.format_adapters.dfxp.dfxp_tools import srt2dfxp, trs2dfxp, create_empty_dfxp, dfxp2txt, txt2dfxp
from modules.base.urldownloader import UrlDownloader

#---------- VARIABLES -------------#

OP_CODE_NEW = 0
OP_CODE_UPDATE = 1
OP_CODE_DELETE = 2
OP_CODE_CANCEL = 3

DELETE_MODE_SOFT = "soft"
DELETE_MODE_HARD = "hard"

STCODE_INGESTED = 0
STCODE_MP_PROC = 1
STCODE_TX_PROC = 2
STCODE_TL_PROC = 3
STCODE_TTS_PROC = 4
STCODE_PL_PROC = 5
STCODE_NEW_FINISHED = 6
STCODE_UPD_FINISHED = 20
STCODE_DEL_FINISHED = 30
STCODE_CANCEL_FINISHED = 40
STCODE_CANCELLED = 90
STCODE_UNK_ERROR = 100
STCODE_MP_ERROR = 101
STCODE_TX_ERROR = 102
STCODE_TL_ERROR = 103
STCODE_TTS_ERROR = 104
STCODE_PL_ERROR = 105
STCODE_DB_ERROR = 106
STCODE_UPD_ERROR = 200
STCODE_DEL_ERROR = 300
STCODE_CANCEL_ERROR = 400

CAPTION_TYPE_TX = "tx"
CAPTION_TYPE_TL = "tl"

VOICE_TYPE_TTS = "tts"
VOICE_TYPE_RECORDING = "rec"

CONST_IS_TIME = 300 # 5 min

#---------- Logging functions -------------#


def log(msg):
    sys.stderr.write("[LL] (%s) %s\n" %(now(), msg))


def error(msg, exit_code=1):
    sys.stderr.write("\n[EE] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" % now())
    sys.stderr.write("[EE] (%s) |                         FATAL ERROR!!                     |\n" % now())
    sys.stderr.write("[EE] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" % now())
    sys.stderr.write("[EE] (%s) %s\n" %(now(), msg))
    sys.stderr.write("[EE] (%s) ------------------------- ERROR CODE %d ---------------------\n" %(now(), exit_code))
    print_excpt() 
    sys.exit(exit_code)


def bug(msg=None, exit=False):
    sys.stderr.write("\n[BB] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" % now())
    sys.stderr.write("[BB] (%s) |                     UNCAUGHT EXCEPTION!!                  |\n" % now())
    sys.stderr.write("[BB] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" % now())
    if msg != None:
        sys.stderr.write("[BB] (%s) %s\n" % (now(), msg))
    print_excpt()
    if exit:
         sys.exit(1)


def print_excpt(string=None):
    if string != None:
        excpt_str = string.strip()
    else:
        out = StringIO()
        traceback.print_exc(file=out)
        excpt_str = out.getvalue().strip()
    if excpt_str != "None":
        sys.stderr.write("\n[TT] (%s) ++++++++++++++++++++++++ ERROR TRACE ++++++++++++++++++++++++\n" % now())
        sys.stderr.write("%s\n" % excpt_str)
        sys.stderr.write("[TT] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" % now())
        sys.stderr.write("\n")


def warn(msg):
    sys.stderr.write("[WW] (%s) %s\n" %(now(), msg))


def now():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


#---------- Ingest status update functions -------------#


def send_mail_processing(mailer, manif, CFG, up_id, user_email, eta=None):
    if manif['operation_code'] != CFG['OPC']['NEW'] or not(manif.get('from_webui', False)): return
    v_id = manif['metadata']['external_id']
    language = manif['metadata'].get('language', "")
    title = manif['metadata'].get('title', "")
    dbname = CFG['STO']['DB_NAME']
    username =  manif.get('username', None)
    if user_email != None and CFG['MAI']['SCSM']: 
        res = mailer.send_mail_user_processing(up_id, v_id, title, user_email, eta=eta)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()
    if CFG['MAI']['SASM'] and CFG['MAI']['AAS'] != None: 
        res = mailer.send_mail_admin_processing(up_id, v_id, title, language, dbname, user_email, username, eta=eta)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()


def send_mail_error(mailer, manif, CFG, up_id, user_email, st_code, err_code, info, to_user=False):
    if manif == None:
        v_id = "-"
        language = "-"
        title = "-"
        username = "-"
    else:
        v_id = manif['metadata']['external_id']
        language = manif['metadata'].get('language', "")
        title = manif['metadata'].get('title', "")
        username =  manif.get('username', None)
    dbname = CFG['STO']['DB_NAME']
    if user_email != None and CFG['MAI']['SCEM'] and to_user: 
        res = mailer.send_mail_user_error(up_id, v_id, title, err_code, info, user_email)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()
    if CFG['MAI']['SAEM'] and CFG['MAI']['AAE'] != None: 
        res = mailer.send_mail_admin_error(up_id, v_id, title, language, dbname, user_email, st_code, err_code, info, username)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()


def send_mail_finished(mailer, manif, CFG, up_id, user_email):
    if manif['operation_code'] != CFG['OPC']['NEW'] or not(manif.get('from_webui', False)): return 
    v_id = manif['metadata']['external_id']
    language = manif['metadata'].get('language', "")
    title = manif['metadata'].get('title', "")
    dbname = CFG['STO']['DB_NAME']
    username =  manif.get('username', None)
    if user_email != None and CFG['MAI']['SCFM']: 
        res = mailer.send_mail_user_finished(up_id, v_id, title, user_email)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()
    if CFG['MAI']['SAFM'] and CFG['MAI']['AAF'] != None: 
        res = mailer.send_mail_admin_finished(up_id, v_id, title, language, dbname, user_email, username)
        if res['code'] != 0:
            warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
            print_excpt()


def ingest_error(stg, manif, CFG, up_id, ext_id, lang, user_id, status_code, error_code, info, mailer, user_email, debug=False, inform_user=False, session_id=None, commit_errors=True):
    sys.stderr.write("[IE] (%s) [Error Code %d] %s\n" %(now(), error_code, info.encode('utf-8')))
    print_excpt()
    if commit_errors:
        if user_id != None:
            res = stg.restore_user_quota(up_id, user_id)
            if res['code'] != 0:
                warn("Cannot restore user quota (Error Code: %d, Info: %s)." % (res['code'], res['info']))
        if CFG['SES']['ENABLED'] and session_id != None:
            res = stg.delete_session(session_id)
            if res['code'] != 0:
                error("Cannot delete session.", exit_code=res['code'])
        res = stg.update_upload(up_id, ext_id, lang, user_id, status_code, info, datetime.now(), error_code=error_code) 
        if res['code'] != 0:
            error("Cannot update upload.", exit_code=res['code'])
        if mailer != None:
            send_mail_error(mailer, manif, CFG, up_id, user_email, status_code, error_code, info, to_user=inform_user)


def ingest_update(stg, up_id, ext_id, lang, user_id, status_code, info, jobs=None, session_id=None, eta=None):
    if jobs != None:
        jobs = [ int(x) for x in jobs ]
    res = stg.update_upload(up_id, ext_id, lang, user_id, status_code, info, datetime.now(), session_id=session_id, jobs=jobs, eta=eta)
    if res['code'] != 0:
        error("Cannot update upload.", exit_code=res['code'])
    sys.stderr.write("[IU] (%s) %s\n" %(now(), info))


#---------- Config file parse function -------------#


def parse_config(cfg_fd): 
    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}
    # General
    CFG['GEN'] = {}
    CFG['GEN']['HOSTNAME'] = config.get('general', 'hostname')
    CFG['GEN']['TL_USER'] = config.get('general', 'tl_user')
    CFG['GEN']['TL_GROUP'] = config.get('general', 'tl_group')
    CFG['GEN']['RM_FIN_UP_DAYS'] = int(config.get('general', 'rm_finished_up_days'))
    CFG['GEN']['RM_ERR_UP_DAYS'] = int(config.get('general', 'rm_error_up_days'))
    CFG['GEN']['LOCAL_REPO'] = config.getboolean('general', 'local_repository')
    # Authentication
    CFG['AUT'] = {}
    CFG['AUT']['ENABLED'] = True
    # Storage
    CFG['STO'] = {}
    CFG['STO']['DB_NAME'] = config.get('storage', 'db_name')
    CFG['STO']['DB_USER'] = config.get('storage', 'db_user')
    CFG['STO']['DB_PASS'] = config.get('storage', 'db_passwd')
    if CFG['STO']['DB_PASS'] != None:
        if CFG['STO']['DB_PASS'].strip() == "": CFG['STO']['DB_PASS'] = None
    CFG['STO']['DB_HOST'] = config.get('storage', 'db_host')
    # Scheduler
    CFG['SCH'] = {}
    CFG['SCH']['LOCALHOST'] = config.get('scheduler', 'localhost')
    CFG['SCH']['STATUS_CMD'] = config.get('scheduler', 'status_cmd')
    CFG['SCH']['SUBMIT_SCR'] = config.get('scheduler', 'submit_scr')
    CFG['SCH']['SUBMIT_OPTS'] = config.get('scheduler', 'submit_opts')
    CFG['SCH']['JOB_NAME_PREFIX'] = config.get('scheduler', 'job_name_prefix') + ".speech"
    # SESSIONS
    CFG['SES'] = {}
    CFG['SES']['ENABLED'] = config.getboolean('sessions', 'enabled')
    CFG['SES']['AID'] = config.get('sessions', 'author_id')
    CFG['SES']['ANM'] = config.get('sessions', 'author_name')
    CFG['SES']['ACF'] = config.getint('sessions', 'author_conf')
    # Mailing
    CFG['MAI'] = {}
    try:
        CFG['MAI']['ENABLED'] = config.getboolean('mailing', 'enabled')
        CFG['MAI']['SRC_ADDRESS'] = config.get('mailing', 'from_address')
        CFG['MAI']['SMTP_SERVER'] = config.get('mailing', 'smtp_server')
    except:
        warn("Mailing configuration variables are not defined. Mailing module will be disabled.")
        CFG['MAI']['ENABLED'] = False
    CFG['MAI']['SCSM'] = config.getboolean('mailing', 'send_client_started_mail')
    CFG['MAI']['SCEM'] = config.getboolean('mailing', 'send_client_error_mail')
    CFG['MAI']['SCFM'] = config.getboolean('mailing', 'send_client_finished_mail')
    CFG['MAI']['SASM'] = config.getboolean('mailing', 'send_admin_started_mail')
    try:
        CFG['MAI']['AAS'] = config.get('mailing', 'admin_address_started').replace(' ', '').split(',')
    except:
        CFG['MAI']['AAS'] = None
    CFG['MAI']['SAEM'] = config.getboolean('mailing', 'send_admin_error_mail')
    try:
        CFG['MAI']['AAE'] = config.get('mailing', 'admin_address_error').replace(' ', '').split(',')
    except:
        CFG['MAI']['AAE'] = None
    CFG['MAI']['SAFM'] = config.getboolean('mailing', 'send_admin_finished_mail')
    try:
        CFG['MAI']['AAF'] = config.get('mailing', 'admin_address_finished').replace(' ', '').split(',')
    except:
        CFG['MAI']['AAF'] = None
    # Operation Codes
    CFG['OPC'] = {}
    CFG['OPC']['NEW'] =  OP_CODE_NEW
    CFG['OPC']['UPDATE'] =  OP_CODE_UPDATE
    CFG['OPC']['DELETE'] =  OP_CODE_DELETE
    CFG['OPC']['CANCEL'] =  OP_CODE_CANCEL
    # Status Codes
    CFG['STC'] = {}
    CFG['STC']['ST_INGESTED'] = STCODE_INGESTED
    CFG['STC']['ST_MP_PROCESSING'] = STCODE_MP_PROC
    CFG['STC']['ST_TX_PROCESSING'] = STCODE_TX_PROC
    CFG['STC']['ST_TTS_PROCESSING'] = STCODE_TTS_PROC
    CFG['STC']['ST_TL_PROCESSING'] = STCODE_TL_PROC
    CFG['STC']['ST_PL_PROCESSING'] = STCODE_PL_PROC
    CFG['STC']['PROCESSING_CODES'] = [ CFG['STC']['ST_MP_PROCESSING'], CFG['STC']['ST_TX_PROCESSING'],
                                       CFG['STC']['ST_TL_PROCESSING'], CFG['STC']['ST_TTS_PROCESSING'], 
                                       CFG['STC']['ST_PL_PROCESSING'] ]
    CFG['STC']['ST_NEW_FINISHED'] = STCODE_NEW_FINISHED
    CFG['STC']['ST_UP_FINISHED'] = STCODE_UPD_FINISHED
    CFG['STC']['ST_DEL_FINISHED'] = STCODE_DEL_FINISHED
    CFG['STC']['ST_CANCEL_FINISHED'] = STCODE_CANCEL_FINISHED
    CFG['STC']['ST_CANCELLED'] = STCODE_CANCELLED
    CFG['STC']['FINISHED_CODES'] = [ CFG['STC']['ST_NEW_FINISHED'], CFG['STC']['ST_UP_FINISHED'],
                                     CFG['STC']['ST_DEL_FINISHED'], CFG['STC']['ST_CANCEL_FINISHED'],
                                     CFG['STC']['ST_CANCELLED'] ]
    CFG['STC']['ST_UNK_ERROR'] = STCODE_UNK_ERROR
    CFG['STC']['ST_MP_ERROR'] = STCODE_MP_ERROR
    CFG['STC']['ST_TX_ERROR'] = STCODE_TX_ERROR
    CFG['STC']['ST_TL_ERROR'] = STCODE_TL_ERROR
    CFG['STC']['ST_TTS_ERROR'] = STCODE_TTS_ERROR
    CFG['STC']['ST_PL_ERROR'] = STCODE_PL_ERROR
    CFG['STC']['ST_DB_ERROR'] = STCODE_DB_ERROR
    CFG['STC']['ST_UP_ERROR'] = STCODE_UPD_ERROR
    CFG['STC']['ST_DEL_ERROR'] = STCODE_DEL_ERROR
    CFG['STC']['ST_CANCEL_ERROR'] = STCODE_CANCEL_ERROR
    CFG['STC']['ERROR_CODES'] = [ CFG['STC']['ST_UNK_ERROR'], CFG['STC']['ST_MP_ERROR'], 
                                  CFG['STC']['ST_TX_ERROR'], CFG['STC']['ST_TL_ERROR'], 
                                  CFG['STC']['ST_PL_ERROR'], CFG['STC']['ST_DB_ERROR'], 
                                  CFG['STC']['ST_UP_ERROR'], CFG['STC']['ST_DEL_ERROR'] ]
    # Required Media Formats
    CFG['FFT'] = {}
    CFG['FFT']['MTL'] = int(config.get('file_formats', 'max_audio_track_length'))
    CFG['FFT']['PCM'] = config.getboolean('file_formats', 'generate_pcm_stream')
    CFG['FFT']['RVF'] = config.get('file_formats', 'required_video_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['AVF'] = config.get('file_formats', 'allowed_video_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['AAF'] = config.get('file_formats', 'allowed_audio_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['ASTF'] = config.get('file_formats', 'allowed_slides_text_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['ASVF'] = config.get('file_formats', 'allowed_slides_video_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['ADF'] = config.get('file_formats', 'allowed_docs_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['ACF'] = config.get('file_formats', 'allowed_caption_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['ATF'] = config.get('file_formats', 'allowed_thumbnail_formats').replace(' ', '').lower().split(',')
    CFG['FFT']['APK'] = config.get('file_formats', 'allowed_packages').replace(' ', '').lower().split(',')
 
    # Audio Background
    CFG['ABG'] = None
    try:
        CFG['ABG'] = config.get('data', 'audio_background_img')
        if CFG['ABG'].strip() == "":
            CFG['ABG'] = None
        elif config.get('data', 'audio_background_img')[0] != '/':
            CFG['ABG'] = SCRIPT_DIR + '/' + config.get('data', 'audio_background_img')
    except:
        warn("Audio background file not provided.")
    CFG['TEST_THUMB'] = None
    try:
        CFG['TEST_THUMB'] = config.get('data', 'test_thumbnail_img')
        if CFG['TEST_THUMB'].strip() == "":
            CFG['TEST_THUMB'] = None
        elif config.get('data', 'test_thumbnail_img')[0] != '/':
            CFG['TEST_THUMB'] = SCRIPT_DIR + '/' + config.get('data', 'test_thumbnail_img')
    except:
        warn("Audio background file not provided.")
    CFG['DEF_THUMB'] = None
    try:
        CFG['DEF_THUMB'] = config.get('data', 'default_thumbnail_img')
        if CFG['DEF_THUMB'].strip() == "":
            CFG['DEF_THUMB'] = None
        elif config.get('data', 'default_thumbnail_img')[0] != '/':
            CFG['DEF_THUMB'] = SCRIPT_DIR + '/' + config.get('data', 'default_thumbnail_img')
    except:
        warn("Audio background file not provided.")

    # Text Retrieval Module
    try:
        CFG['TXT'] = config.get('text_retrieval_module', 'module_path')
        if CFG['TXT'].strip() == "":
            CFG['TXT'] = None
        elif config.get('text_retrieval_module', 'module_path')[0] != '/':
            CFG['TXT'] = SCRIPT_DIR + '/' + config.get('text_retrieval_module', 'module_path')
    except:
        warn("Text Retrieval Module not available")
        CFG['TXT'] = None
    # URL decoders
    CFG['URLDC'] = {}
    for x in config.items('url_decoders'):
        if x[1][0] == '/':
            CFG['URLDC'][x[0]] = x[1]
        else:
            CFG['URLDC'][x[0]] = SCRIPT_DIR + '/' + x[1]
    return CFG


#---------- Media Package Processing functions -------------#


def sts(s, m=0):
    # sanitize string
    if m == 0: # general
        return re.subn(r'[\r\b\n$&`´|{}()\[\];/\\"\'#*~^<>=+!?¡¿]|--', '', unicode(s))
    elif m == 1: # fileformats, allow / +
        return re.subn(r'[\r\b\n$&`´|{}()\[\];\\"\'#*~^<>=!?¡¿]|--', '', unicode(s))
    elif m == 2: # filenames, allow ( ) +
        return re.subn(r'[\r\b\n$&`´|{}\[\];/\\"\'#*~^<>=!?¡¿]|--', '', unicode(s)) 
    elif m == 3: # urls
        return re.subn(r'[\r\b\n`´|{}\\"\'^<>¡¿]|--', '', unicode(s)) 


def rename_filename(stg, up_id, fn):
    fn_new = ''.join(c for c in unicodedata.normalize('NFD',unicode(fn)) if unicodedata.category(c)!='Mn').replace(' ', '_')
    if unicode(fn_new) != unicode(fn):
        res = stg.rename_mp_file(up_id, fn, fn_new)
        if res['code'] != 0:
            return res
    return {'code':0, 'data':fn_new}

def sanitize_manifest(stg, jbs, up_id, m):
    res = sts(m['operation_code'])
    if res[1] > 0:
        return {'code':1000, 'info':'Invalid operation code (contains forbidden characters)' }
    for k in m['metadata'].keys():
        if k == "external_id":
            res = sts(m['metadata'][k])
            if res[1] > 0:
                return {'code':1001, 'info':'Invalid Video ID (contains forbidden characters)' }
            if len(m['metadata'][k]) > 200:
                return {'code':1091, 'info':'Invalid Video ID (exceeds 200 characters length)' }
        elif k == "language":
            res = sts(m['metadata'][k])
            if res[1] > 0:
                return {'code':1002, 'info':'Invalid language code (contains forbidden characters)' }
        elif k == "speakers":
            if not(isinstance(m['metadata'][k], list)):
                return {'code':1043, 'info':'"metadata" -> "speakers" must be an array.' }
            for s in m['metadata'][k]:
                for k2 in s.keys():
                    if k2 == "speaker_id":
                        res = sts(s[k2])
                        if res[1] > 0:
                            return {'code':1003, 'info':'Invalid speaker ID (contains forbidden characters)' }
                    elif k2 == "speaker_name":
                        res = sts(s[k2])
                        if res[1] > 0:
                            return {'code':1034, 'info':'Invalid speaker name (contains forbidden characters)' }
                    elif k2 == "speaker_gender" and s[k2] != None:
                        res = sts(s[k2])
                        if res[1] > 0:
                            return {'code':1035, 'info':'Invalid speaker gender (contains forbidden characters)' }
                    elif k2 == "speaker_email" and s[k2] != None:
                        res = sts(s[k2])
                        if res[1] > 0:
                            return {'code':1036, 'info':'Invalid speaker email (contains forbidden characters)' }
        elif k == "date" and m['metadata'][k] != None:
            res = sts(m['metadata'][k], m=1)
            if res[1] > 0:
                return {'code':1031, 'info':'Invalid lecture date (contains forbidden characters)' }
        else:
            if m['metadata'][k] != None:
                m['metadata'][k] = sts(m['metadata'][k])[0]
    if 'media' in m:
        if m['media'].get('url', None) != None:
            res = sts(m['media']['url'], m=3)
            if res[1] > 0:
                return {'code':1006, 'info':'Invalid url (contains forbidden characters).' }
        else:
            res = sts(m['media']['md5'])
            if res[1] > 0:
                return {'code':1005, 'info':'Invalid md5sum for media file (contains forbidden characters).' }
            res = sts(m['media']['filename'], m=2)
            if res[1] > 0:
                return {'code':1006, 'info':'Invalid media file name (contains forbidden characters).' }
            res = rename_filename(stg, up_id, m['media']['filename'])
            if res['code'] > 0:
                res['email'] = email
                return res
            m['media']['filename'] = res['data'] 
            res = sts(m['media']['fileformat'], m=1)
            if res[1] > 0:
                return {'code':1007, 'info':'Invalid media file format (contains forbidden characters).' }
            m['media']['fileformat'] = m['media']['fileformat'].lower()
    
    if 'attachments' in m: 
        if len(m['attachments']) > 20:
            return {'code':1008, 'info':'Too many attachments (max. 20).' }
        for i in xrange(len(m['attachments'])):
            res = sts(m['attachments'][i]['md5'])
            if res[1] > 0:
                return {'code':1009, 'info':'Invalid md5sum for attachment file no.%d (contains forbidden characters).' % i }
            res = sts(m['attachments'][i]['filename'], m=2)
            if res[1] > 0:
                return {'code':1010, 'info':'Invalid file name of attachment no.%d (contains forbidden characters).' % i }
            res = rename_filename(stg, up_id, m['attachments'][i]['filename'])
            if res['code'] > 0:
                res['email'] = email
                return res
            m['attachments'][i]['filename'] = res['data'] 
            res = sts(m['attachments'][i]['fileformat'], m=1)
            if res[1] > 0:
                return {'code':1011, 'info':'Invalid file format of attachment no.%d (contains forbidden characters).' % i }
            m['attachments'][i]['fileformat'] = m['attachments'][i]['fileformat'].lower()
            res = sts(m['attachments'][i]['type_code'])
            if res[1] > 0:
                return {'code':1012, 'info':'Invalid type code of attachment no.%d (contains forbidden characters).' % i }
    return {'code':0, 'data':m }


def update_manifest(CFG, stg, jbs, manif, up_id, lec_hash):
    lec_id = manif['metadata']['external_id']
    res_li = stg.get_lecture_info(lec_hash)
    if res_li['code'] != 0:
        return res_li
    if manif['operation_code'] == CFG['OPC']['UPDATE'] and len(res_li['data']) == 0:
        return {'code':3000, 'info':"External ID '%s' does not exist in Database." % lec_id }
    # Media duration
    if 'media' in manif:
        if manif['media'].get('url', None) == None:
            media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
            res = get_media_length(jbs, media_fp)
            if res['code'] != 0:
                return res
            duration = int(res['data'])
            manif['metadata']['duration'] = duration         
    elif manif['operation_code'] == CFG['OPC']['UPDATE']:
        duration = res_li['data'][0]['duration']
        manif['metadata']['duration'] = duration
    # Media language and title
    if 'media' not in manif and manif['operation_code'] == CFG['OPC']['UPDATE']:
        if manif['metadata'].get('language', None) == None:
            language = res_li['data'][0]['language']
            manif['metadata']['language'] = language
        if manif['metadata'].get('title', None) == None:
            title = res_li['data'][0]['title'].decode('utf-8')
            manif['metadata']['title'] = title
    return {'code':0, 'data':manif}
   

def process_media_package(CFG, stg, jbs, up_id, lec_hash, manif):
    """Unzip, extract and check data from media package.

    Arguments:

     mp_fp         --  Media Package file path.
     dest_dir      --  Directory to extract Media Package contents.
     op_codes      --  Dict containing available operating codes.

    Returns:
     
     Dict -> {'code':INT, 'info':STR, 'data':DICT }
     where 'data' contains the parsed JSON data in a Dict.

    """
    req_pkg = False
    if 'media' in manif:
        if 'url' not in manif['media']:
            req_pkg = True
    if len(manif.get('attachments', [])) > 0:
        req_pkg = True
    if req_pkg:
        res = stg.extract_media_package(up_id)
        if res['code'] != 0: 
            return res
    else:
        res = stg.create_media_package_extraction_dir(up_id)
        if res['code'] != 0: 
            return res
    try:
        res = sanitize_manifest(stg, jbs, up_id, manif)
    except:
        print_excpt()
        return {'code':1013, 'info':'Cannot sanitize manifest.json data. Please contact system administrator.'}
    if res['code'] != 0:
        return res
    manif = res['data']
    # Check 'external_id' key
    try:
        ext_id = manif['metadata']['external_id']
    except KeyError:
        return {'code':1016, 'info':'"external_id" key not found in "metadata" in manifest.json file.'}
    # Update manifest data
    res = update_manifest(CFG, stg, jbs, manif, up_id, lec_hash)
    if res['code'] != 0:
        return res 
    manif = res['data']
    # Store sanitized manifest
    res = stg.store_manifest(up_id, manif, backup=False)
    if res['code'] != 0:
        return res 
    # Check Operation Code
    try:
        opc = manif['operation_code']
    except KeyError:
        info = 'Invalid JSON format: operation_code key not found.'
        return {'code':1014, 'info':info}
    # New or Update operations
    if opc == CFG['OPC']['NEW'] or opc == CFG['OPC']['UPDATE']:
        # Check Metadata
        res = check_metadata(manif, opc)
        if res['code'] != 0:
            return res
        # If New, check media
        if opc == CFG['OPC']['NEW'] or (opc == CFG['OPC']['UPDATE'] and 'media' in manif):
            res = check_media(CFG, stg, jbs, up_id, manif) 
            if res['code'] != 0:
                return res
        # Check media duration
        if manif['metadata']['duration'] > CFG['FFT']['MTL']:
            return {'code':1088, 'info':'Media length (%d sec.) exceeds the maximum length allowed (%d sec.).' % (manif['metadata']['duration'], CFG['FFT']['MTL'])}
        # Check attachments
        res = check_attachs(CFG, stg, up_id, manif) 
        if res['code'] != 0:
            return res
        # Check media as slides
        res = check_media_as_slides(CFG, stg, up_id, manif)
        if res['code'] != 0:
            return res
        res = stg.store_manifest(up_id, manif, backup=False)
        if res['code'] != 0:
            return res 
        # Return JSON data
        return {'code':0, 'data':manif}
    # Delete operation
    elif opc == CFG['OPC']['DELETE'] or opc == CFG['OPC']['CANCEL']:
        # Check metadata (just external_id)
        res = check_metadata(manif, opc)
        if res['code'] != 0:
            return res
        # Return JSON data
        return {'code':0, 'data':manif}
    # Unknown operation
    else:
        info = 'Operation code %s not supported' % opc
        return {'code':2, 'info':info}


def check_metadata(manif, opc):
    # Check and get 'metadata' key
    try:
        meta = manif['metadata']
    except KeyError:
        return {'code':1015, 'info':'"metadata" key not found in manifest.json file.'}
    # Check 'external_id' key
    try:
        ext_id = meta['external_id']
    except KeyError:
        return {'code':1016, 'info':'"external_id" key not found in "metadata" in manifest.json file.'}
    # Optional checks:
    if opc in (OP_CODE_NEW, OP_CODE_UPDATE):
        # Check 'language' key
        try:
            lang = meta['language']
        except KeyError:
            return {'code':1017, 'info':'"language" key not found in "metadata" in manifest.json file.'}
        if 'media' in manif:
            try:
                title = meta['title']
            except KeyError:
                return {'code':1082, 'info':'"title" key not found in "metadata" in manifest.json file.'}
            if 'speakers' in meta.keys():
                speakers = meta['speakers']
                if not(isinstance(speakers, list)):
                    return {'code':1039, 'info':'"speakers" key must be an array.'}
                if len(speakers) == 0:
                    #return {'code':1044, 'info':'"speakers" key must be a non-empty array.'}
                    pass
                for s in speakers:
                    if 'speaker_id' not in s.keys():
                        return {'code':1045, 'info':'"speaker_id" key not found in "metadata" -> "speakers" in manifest.json file.'}
                    if 'speaker_name' not in s.keys():
                        return {'code':1037, 'info':'"speaker_name" key not found in "metadata" -> "speakers" in manifest.json file.'}
                    if 'speaker_gender' in s.keys():
                        if s['speaker_gender'] != None and s['speaker_gender'] not in [ "M", "F" ]:
                            return {'code':1038, 'info':'"speaker_gender" key must be "M" (male) or "F" (female).'}
    return {'code':0}

def check_media(CFG, stg, jbs, up_id, manif):
    # Check and get 'media' key
    try:
        media = manif['media']
    except KeyError:
        return {'code':1019, 'info':'"media" key not found in the manifest.json file.'}
    if media.get('url', None) != None:
        url = media['url']
        fn = "mainmedia"
        res = CFG['URLDW'].download_url(url, stg.media_package_file_abspath(fn, up_id), "mp4")
        if res['code'] != 0:
            return res 
        ext, website, decoded = res['data']
        fn = "%s.%s" % (fn, ext)
        ff = ext
        media['filename'] = fn
        media['fileformat'] = ext
        media['store'] = True if decoded else False
        type_code = CFG['FTC']['media']
        media_fp = stg.media_package_file_abspath(fn, up_id)
        res = get_media_length(jbs, media_fp)
        if res['code'] != 0:
            return res
        duration = int(res['data'])
        manif['metadata']['duration'] = duration
    else:
        # Check and get 'media' sub-keys
        try:
            fn = media['filename']
            ff = media['fileformat']
            md5 = media['md5']
            type_code = CFG['FTC']['media']
            media['store'] = True
        except KeyError:
            return {'code':1020, 'info':'"filename", "fileformat" and/or "md5" keys not found in "media" key in manifest.json file.'}
        # Check if main media file is actually a video/audio file
        media_fp = stg.media_package_file_abspath(fn, up_id)
        # Check if file exists and md5sum matching
        res = stg.check_mp_file(up_id, fn, md5)
        if res['code'] != 0:
            return res 
    res = check_media_audio(jbs, media_fp)
    if res['code'] != 0:
        return res 
    # Check media fileformat
    if (type_code, ff) not in CFG['ATT']:
        return {'code':1021, 'info':'File format not allowed for main media file.'}
    return {'code':0}

def check_media_as_slides(CFG, stg, up_id, manif):
    if 'media' in manif:
        if manif['media'].get('slides', False):
            slides = {'filename':manif['media']['filename'], 'fileformat':manif['media']['fileformat'], 'type_code':CFG['FTC']['slides'], 'store':False}
            if 'attachments' in manif:
                for a in manif['attachments']:
                    if a['type_code'] == CFG['FTC']['slides']:
                        return {'code':0} # Ignore 'slides' option if slides were attached
                manif['attachments'].append(slides)
            else:
                manif['attachments'] = [slides]
    return {'code':0}


def check_attachs(CFG, stg, up_id, manif):
    # Check and get 'attachments' key
    attachs = manif.get('attachments', [])
    # For each attachment
    for att in attachs:
        # Check and get attachment sub-keys
        try:
            fn = att['filename']
            ff = att['fileformat']
            md5 = att['md5']
            type_code = att['type_code']
        except KeyError:
            return {'code':1022, 'info':'"filename", "fileformat", "md5" and/or "type_code" keys not found in one attachment in manifest.json file.'}
        # Check if file exists and md5sum matching
        res = stg.check_mp_file(up_id, fn, md5)
        if res['code'] != 0:
            return res 
        # Check media fileformat
        if ff in CFG['FFT']['APK']: # Package file
            if type_code == CFG['FTC']['slides']:
                res = stg.unzip_attached_package_file(up_id, fn)
                if res['code'] != 0:
                    return res
            else:
                return {'code':1023, 'info':'File format not allowed for file type code of attachment "%s".' % fn}
 
        else:
            if (type_code, ff) not in CFG['ATT']:
                return {'code':1023, 'info':'File format not allowed for file type code of attachment "%s".' % fn}
            if type_code in (CFG['FTC']['captions'], CFG['FTC']['audiotrack']) and 'language' not in att.keys():
                return {'code':1033, 'info':'"language" key not found for subtitles file (attachment) "%s".' % fn}
            # Rename caption files
            if type_code == CFG['FTC']['captions']:
                media_lang = manif['metadata']['language']
                ext = os.path.splitext(fn)[1]
                if media_lang == att['language']:
                    fn_new = "subs.%s.tx%s" % (att['language'], ext)
                else:
                    fn_new = "subs.%s-%s.tl%s" % (media_lang, att['language'], ext)
                res = stg.rename_mp_file(up_id, fn, fn_new)
                if res['code'] != 0:
                    return res
                att['filename'] = fn_new
            elif type_code == CFG['FTC']['audiotrack']:
                ext = os.path.splitext(fn)[1]
                fn_new = "audiotrack.%s%s" % (att['language'], ext)
                res = stg.rename_mp_file(up_id, fn, fn_new)
                if res['code'] != 0:
                    return res
                att['filename'] = fn_new
    
    return {'code':0}


def get_dfxp_file(CFG, stg, lang, up_id, manif, lec_hash):
    dfxp_fp = None
    # 1. Look in the media package
    for att in manif.get('attachments', []):
        if att['type_code'] == CFG['FTC']['captions'] and att['language'] == lang and att['fileformat'] == 'dfxp':
            dfxp_fp = stg.media_package_file_abspath(att['filename'], up_id)
            return {'code':0, 'data':dfxp_fp}
    # 2. Look in outputs
    res = stg.load_outputs_manifest(up_id)
    if res['code'] != 0:
        return res
    outputs_manif = res['data']
    for c in outputs_manif['captions']:
        if c['language'] == lang:
            dfxp_fp = stg.output_file_abspath(c['filename'], up_id)
            return {'code':0, 'data':dfxp_fp}
    # 3. Look in DB
    res = stg.get_dfxp_file_abspath(lec_hash, lang)
    if res['code'] != 0:
        return res
    dfxp_fp = res['data']
    return {'code':0, 'data':dfxp_fp}


def get_user_custom_default_systems(stg, user_id):
    res = stg.get_user_default_systems(user_id)
    if res['code'] != 0:
        return res
    lsys = res['data']
    ret = { 'asr':{}, 'mt':{}, 'tts':{} }
    for s in lsys:
        if s['type'] == "asr":
            lang = s["lang1"]
            ret["asr"][lang] = s["id"]
        elif s['type'] == "mt":
            lang = "%s-%s" % (s["lang1"], s["lang2"])
            ret["mt"][lang] = s["id"]
        elif s['type'] == "tts":
            lang = s["lang1"]
            ret["tts"][lang] = s["id"]
    return {'code':0, 'data':ret}

def check_manual_captions(CFG, stg, up_id, manif, lec_hash):
    langs = []
    auto_langs = {}
    # Look for attached captions
    for att in manif.get('attachments', []):
        if att['type_code'] == CFG['FTC']['captions']:
            try:
                lang = att['language']
            except KeyError:
                info = "Attached subtitles file '%s' has no 'language' key." % att['filename']
                return {'code':1024, 'info':info}
            else:
                langs.append(lang)
    # Look for existing manual captions only in update operation, except if it's a re-recording
    if manif['operation_code'] == CFG['OPC']['UPDATE'] and 'media' not in manif.keys():
        res = stg.get_captions(lec_hash)
        if res['code'] != 0:
            return res
        for c in res['data']:
            if c['type'] == stg.caption_type['fully_human'] or c['type'] == stg.caption_type['partially_human']: ########## POSSIBLE IMPROVEMENT -> THIS IS BD/STORAGE DEPENDENT
                if c['language'] not in langs:
                    langs.append(c['language'])
            elif c['type'] == stg.caption_type['fully_automatic']:
                auto_langs[c['language']] = (c['system_id'],c['from_language'])
            
    return {'code':0, 'data':(langs, auto_langs)}


def check_tts_audiotracks(CFG, stg, up_id, manif, lec_hash):
    langs = []
    auto_langs = {}
    # Look for attached captions
    for att in manif.get('attachments', []):
        if att['type_code'] == CFG['FTC']['audiotrack']:
            try:
                lang = att['language']
            except KeyError:
                info = "Attached audiotrack file '%s' has no 'language' key." % att['filename']
                return {'code':1024, 'info':info}
            langs.append(lang)
    # Look for existing manual captions only in update operation, except if it's a re-recording
    if manif['operation_code'] == CFG['OPC']['UPDATE'] and 'media' not in manif.keys():
        res = stg.get_audiotracks(lec_hash)
        if res['code'] != 0:
            return res
        for c in res['data']:
            if c['voice_type'] == VOICE_TYPE_RECORDING:
                if c['language'] not in langs:
                    langs.append(c['language'])
            elif c['voice_type'] == VOICE_TYPE_TTS:
                auto_langs[c['language']] = c['system_id']
    return {'code':0, 'data':(langs, auto_langs)}




def check_caption_conversion(CFG, manif):
    attachs = manif.get('attachments', [])
    conv = []
    for att in attachs:
        type_code = att['type_code']
        if type_code ==  CFG['FTC']['captions']:
            fn = att['filename']
            ff = att['fileformat']
            lang = att['language']
            if ff != 'dfxp': 
                conv.append({'filename':fn, 'fileformat':ff, 'language':lang}) 
    return conv


def check_media_formats_conversion(CFG, manif):
    if 'media' not in manif:
        return { 'MP4':False, 'OGV':False, 'THUMB':False, 'FROM_AUDIO':False, 'PCM':False }
    else:
        res = { 'MP4':('mp4' in CFG['FFT']['RVF']), 'OGV':('ogv' in CFG['FFT']['RVF']), 'THUMB':True, 'FROM_AUDIO':False, 'PCM':CFG['FFT']['PCM'] }
        ffs = []
        ffs.append(manif['media']['fileformat'])
        for att in manif.get('attachments',[]):
            if att['type_code'] == CFG['FTC']['media']:
                ffs.append(att['fileformat'])
            elif att['type_code'] == CFG['FTC']['thumbnail']:
                res['THUMB'] = False
        for ff in ffs:
            if ff == 'mp4':
                res['MP4'] = False
            elif ff == 'ogv':
                res['OGV'] = False
            elif ff in CFG['FFT']['AAF']:
                res['FROM_AUDIO'] = True
        return res


def check_txt_resources(CFG, manif, stg=None):
    """if stg object is supplied, then it will look for existing resources"""
    ret = {}
    ret['slides'] = None
    ret['docs'] = []
    best_sl = None, None, 99, False # att_obj, format, precedence_level, existing_file
    for att in manif.get('attachments', []):
        ff = att['fileformat']
        if att['type_code'] == CFG['FTC']['slides']:
            # Determine precedence level and determine best option
            try:
                prec = CFG['FFT']['ASTF'].index(ff) # allowed_slides_text_formats is an ordered list by precedence
            except ValueError:
                prec = len(CFG['FFT']['ASTF']) # Video slides -> worst precedence value
            if best_sl[2] > prec:
                best_sl = att['filename'], ff, prec, False
        elif att['type_code'] == CFG['FTC']['docs']:
            ret['docs'].append({'fileformat':ff, 'filename':att['filename'], 'existing':False})
    # Check for existing attachments in DB (update operation)
    if stg != None:
        res = stg.get_files(manif['metadata']['external_id'])
        if res['code'] != 0:
            return res
        files = res['data']
        for f in files:
            ft = f['file_type_id']
            ff = f['file_format']
            if ft == CFG['FTC']['slides']:
                # Determine precedence level and determine best option
                try:
                    prec = CFG['FFT']['ASTF'].index(ff) # allowed_slides_text_formats is an ordered list by precedence
                except ValueError:
                    prec = len(CFG['FFT']['ASTF']) # Video slides -> worst precedence value
                if best_sl[2] > prec:
                    best_sl = f['file_path'], ff, prec, True
            elif ft == CFG['FTC']['docs']:
                ret['docs'].append({'fileformat':ff, 'filename':f['file_path'], 'existing':True})
    if best_sl[0] != None: 
        ret['slides'] = {'fileformat':best_sl[1], 'filename':best_sl[0], 'existing':best_sl[3]}
    return ret


def has_txt_resources(CFG, manif):
    res = check_txt_resources(CFG, manif)
    if res['slides'] != None or len(res['docs']) > 0: 
        return True
    else:
        return False

 
def determine_tL_tasks(CFG, stg, manif, up_id, lec_hash, lec_lang, user_id, user_is_admin, verb):
    use_system_ids = { 'asr': None, 'mt': {}, 'tts': {} }
    eta = None
    tL = manif.get('generate', True)
    failsafe = manif.get('failsafe', False) # Not being used, actually
    if tL:
        res = check_manual_captions(CFG, stg, up_id, manif, lec_hash)
        if res['code'] != 0:
            return res
        man_subs_langs = res['data'][0]
        auto_subs_langs = res['data'][1]
        res = check_tts_audiotracks(CFG, stg, up_id, manif, lec_hash)
        if res['code'] != 0:
            return res
        man_at_langs = res['data'][0]
        auto_at_langs = res['data'][1]
        res = get_user_custom_default_systems(stg, user_id)
        if res['code'] != 0:
            return res
        duser_def_sys = res['data']
        regen_opt = manif.get("re-generate", [])
        force_opt = manif.get("force", 0)
        req_langs_opt = manif.get('requested_langs', None)
        if req_langs_opt != None:
            if not(isinstance(req_langs_opt, dict)):
                return {'code':3070, 'info':"Bad format for 'requested_langs' option (Must be a JSON dict)"}
            for rl in req_langs_opt.keys():
                d = manif['requested_langs'][rl]
                if not(isinstance(d,dict)):
                    return {'code':3070, 'info':"Bad format for 'requested_langs' option (language key value must be a JSON dict)"}
                if len(d.keys()) == 0:
                    return {'code':3070, 'info':"No actions specified for lang '%s' in 'requested_langs' option" % (rl)}
                for k in d.keys():
                    if d[k] == None:
                        d[k] = {}
                    elif not(isinstance(d[k], dict)):
                        return {'code':3070, 'info':"Bad format for 'requested_langs' option (operation key value must be a JSON dict)"}
                if rl == lec_lang: # ASR
                    if "sub" in d: # Client requested subtitles
                        if rl in man_subs_langs and not(force_opt):
                            if verb: log("Manual '%s' transcription existing or provided. Will not automatically transcribe media." % rl)
                        elif rl not in CFG['ASR'].keys(): # No available ASR system
                            code = 3070
                            info = "There are not ASR systems available for media language %s" % rl
                            return {'code':code, 'info':info}
                        else:
                            clsid = d['sub'].get('sid', None)
                            if clsid != None: # Client has specfied a System ID
                                try:
                                    clsid = int(clsid)
                                except:
                                    code = 3096
                                    info = "Invalid System ID %s (must be int)" % (str(clsid))
                                    return {'code':code, 'info':info}
                                dsys = CFG['ASR'][rl]['dsys']
                                if clsid not in dsys:
                                    code = 3071
                                    info = "ASR System ID %s does not exist for media language %s" % (str(clsid), rl)
                                    return {'code':code, 'info':info}
                                s = dsys[clsid]
                                if not(s['public']) and CFG['AUT']['ENABLED']: # System ID is not public: check permissions
                                    if not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                        code = 3072
                                        info = "You are not allowed to explicitly make use of the %s ASR System ID %s" % (rl, str(clsid))
                                        return {'code':code, 'info':info}
                                use_system_ids['asr'] = {}
                                use_system_ids['asr']['sid'] = clsid
                            else:
                                use_system_ids['asr'] = {}
                                if rl in duser_def_sys['asr'].keys():
                                    use_system_ids['asr']['sid'] = duser_def_sys['asr'][rl]
                                else:
                                    use_system_ids['asr']['sid'] = CFG['ASR'][rl]['default']
                            use_system_ids['asr']['lma'] =  d['sub'].get('lma', True) # LM Adaptation
                else: # MT / TTS
                    if "sub" in d: # Client requested subtitles
                        if rl in man_subs_langs and not(force_opt):
                            if verb: log("Manual '%s' translation existing or provided. Will not automatically translate to '%s' language." % (rl, rl))
                        elif 'tlpath' in d['sub']: # No available MT System
                            pair = "%s-%s" %(lec_lang, rl)
                            lpath = []
                            sl = lec_lang
                            for ds in d['sub']['tlpath']:
                                if not(isinstance(ds, dict)):
                                    code = 3076
                                    info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                                    return {'code':code, 'info':info}                           
                                if 'l' not in ds.keys():
                                    code = 3076
                                    info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                                    return {'code':code, 'info':info}                           
                                ppair = "%s-%s" % (sl, ds['l'])
                                if ds['l'] in man_subs_langs and not(force_opt):
                                    if verb: log("Manual '%s' translation existing or provided. Will not automatically translate to '%s' language." % (ds['l'], ds['l']))
                                    sl = ds['l']
                                    lpath.append({ 'p':ppair, 'launch':False})
                                    continue
                                if ppair not in CFG['MT'].keys():
                                    code = 3077
                                    info = "There are not %s MT systems available to indirectly translate media file into %s language" % (ppair, rl)
                                    return {'code':code, 'info':info}
                                clsid = ds.get('sid', None)
                                if clsid != None:
                                    try:
                                        clsid = int(clsid)
                                    except:
                                        code = 3097
                                        info = "Invalid System ID %s (must be int)" % (str(clsid))
                                        return {'code':code, 'info':info}
                                    dsys = CFG['MT'][ppair]['dsys']
                                    if clsid not in dsys:
                                        code = 3078
                                        info = "MT System ID %s does not exist for translation direction %s" % (str(clsid), ppair)
                                        return {'code':code, 'info':info}
                                    s = dsys[clsid]
                                    if not(s['public']) and CFG['AUT']['ENABLED']: # System ID is not public: check permissions
                                        if not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                            code = 3079
                                            info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (ppair, str(clsid))
                                            return {'code':code, 'info':info}
                                    lpath.append({ 'p':ppair, 'id':clsid})
                                else:
                                    if ppair in duser_def_sys['mt'].keys():
                                        lpath.append({ 'p':ppair, 'id':duser_def_sys['mt'][ppair]})
                                    else:
                                        lpath.append({ 'p':ppair, 'id':CFG['MT'][ppair]['default']})
                                sl = ds['l']
                            use_system_ids['mt'][pair] = lpath
                        else:
                            clsid = d['sub'].get('sid', None)
                            if clsid != None:
                                try:
                                    clsid = int(clsid)
                                except:
                                    code = 3098
                                    info = "Invalid System ID %s (must be int)" % (str(clsid))
                                    return {'code':code, 'info':info}
                                try:
                                    pair = CFG['MT']['_ids'][clsid]
                                except:
                                    code = 3075
                                    info = "MT System ID %s does not exist" % (str(clsid))
                                    return {'code':code, 'info':info}
                                tl = pair.split("-")[1]
                                if rl != tl:
                                    code = 3075
                                    info = "MT System ID %s (%s) does not generate translations into the requested language %s" % (str(clsid), pair, rl)
                                    return {'code':code, 'info':info}
                            else:
                                pair = "%s-%s" %(lec_lang, rl)
                            if pair not in CFG['MT'].keys():
                                code = 3073
                                info = "There are not %s MT systems available to directly translate media file into %s language" % (pair, rl)
                                return {'code':code, 'info':info}
                            else:
                                if clsid != None: # Client has specfied a System ID
                                    dsys = CFG['MT'][pair]['dsys']
                                    if clsid not in dsys:
                                        code = 3075
                                        info = "MT System ID %s does not exist for translation direction %s" % (str(clsid), pair)
                                        return {'code':code, 'info':info}
                                    s = dsys[clsid]
                                    if not(s['public']) and CFG['AUT']['ENABLED']: # System ID is not public: check permissions
                                        if not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                            code = 3074
                                            info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (pair, str(clsid))
                                            return {'code':code, 'info':info}
                                    use_system_ids['mt'][pair] = [ { 'p':pair, 'id':clsid } ]
                                else:
                                    if pair in duser_def_sys['mt'].keys():
                                        use_system_ids['mt'][pair] = [ { 'p':pair, 'id':duser_def_sys['mt'][pair] } ]
                                    else:
                                        use_system_ids['mt'][pair] = [ { 'p':pair, 'id':CFG['MT'][pair]['default'] } ]
                    if "tts" in d: # Client requested TTS track
                        if rl not in CFG['TTS'].keys(): # No available ASR system
                            code = 3070
                            info = "There are not TTS systems available for media language %s" % rl
                            return {'code':code, 'info':info}
                        else:
                            clsid = d['tts'].get('sid', None)
                            if clsid != None: # Client has specfied a System ID
                                try:
                                    clsid = int(clsid)
                                except:
                                    code = 3099
                                    info = "Invalid System ID %s (must be int)" % (str(clsid))
                                    return {'code':code, 'info':info}
                                dsys = CFG['TTS'][rl]['dsys']
                                if clsid not in dsys:
                                    code = 3071
                                    info = "TTS System ID %s does not exist for media language %s" % (str(clsid), rl)
                                    return {'code':code, 'info':info}
                                s = dsys[clsid]
                                if not(s['public']) and CFG['AUT']['ENABLED']: # System ID is not public: check permissions
                                    if not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                        code = 3072
                                        info = "You are not allowed to explicitly make use of the %s TTS System ID %s" % (rl, str(clsid))
                                        return {'code':code, 'info':info}
                                use_system_ids['tts'][rl] = {}
                                use_system_ids['tts'][rl]['sid'] = clsid
                            else:
                                use_system_ids['tts'][rl] = {}
                                if rl in duser_def_sys['tts'].keys():
                                    use_system_ids['tts'][rl]['sid'] = duser_def_sys['tts'][rl]
                                else:
                                    use_system_ids['tts'][rl]['sid'] = CFG['TTS'][rl]['default']

            if manif['operation_code'] == CFG['OPC']['NEW'] and use_system_ids['asr'] == None and lec_lang not in man_subs_langs:
                if lec_lang not in CFG['ASR'].keys(): # No available ASR system
                    code = 3070
                    info = "There are not ASR systems available for media language %s" % lec_lang
                    return {'code':code, 'info':info}
                else:
                    use_system_ids['asr'] = {}
                    if lec_lang in duser_def_sys['asr'].keys():
                        use_system_ids['asr']['sid'] = duser_def_sys['asr'][lec_lang]
                    else:
                        use_system_ids['asr']['sid'] = CFG['ASR'][lec_lang]['default']
                    use_system_ids['asr']['lma'] = True
         
        else: # Requested Langs no present. Time to guess user thoughts!
            if manif['operation_code'] == CFG['OPC']['NEW']:
                if lec_lang not in man_subs_langs:
                    if lec_lang not in CFG['ASR'].keys(): # No available ASR system
                        code = 3070
                        info = "There are not ASR systems available for media language %s" % lec_lang
                        return {'code':code, 'info':info}
                    else:
                        use_system_ids['asr'] = {}
                        if lec_lang in duser_def_sys['asr'].keys():
                            use_system_ids['asr']['sid'] = duser_def_sys['asr'][lec_lang]
                        else:
                            use_system_ids['asr']['sid'] = CFG['ASR'][lec_lang]['default']
                        use_system_ids['asr']['lma'] = True
                
            elif manif['operation_code'] == CFG['OPC']['UPDATE']:
                retx = ('media' in manif.keys()) or (has_txt_resources(CFG, manif)) or ("tx" in regen_opt)
                retl = ('media' in manif.keys()) or (has_txt_resources(CFG, manif)) or ("tl" in regen_opt)
                retts = ('media' in manif.keys()) or (has_txt_resources(CFG, manif)) or ("tts" in regen_opt)
                if retx:
                    if 'media' not in manif.keys() and lec_lang in man_subs_langs and not(force_opt):
                        if verb: log("Manual '%s' transcription existing or provided. Will not automatically transcribe media." % lec_lang)
                    elif lec_lang not in CFG['ASR'].keys(): # No available ASR system
                        #code = 3070
                        #info = "There are not ASR systems available for media language %s" % lec_lang
                        #return {'code':code, 'info':info}
                        warn("There are not ASR systems available for media language %s" % lec_lang)
                    else:
                        use_system_ids['asr'] = {}
                        use_system_ids['asr']['lma'] = True
                        if lec_lang in auto_subs_langs:
                            system_id, from_lang = auto_subs_langs[lec_lang]
                            if system_id == None:
                                if lec_lang in duser_def_sys['asr'].keys():
                                    system_id = duser_def_sys['asr'][lec_lang]
                                else:
                                    system_id = CFG['ASR'][lec_lang]['default']
                                use_system_ids['asr']['sid'] = system_id
                                warn("On regenerate '%s' transcription: System ID was not specified. Using default ASR system." %(lec_lang))
                            elif CFG['ASR'][lec_lang]['dsys'].get(system_id, None) != None:
                                use_system_ids['asr']['sid'] = system_id
                            else:
                                if lec_lang in duser_def_sys['asr'].keys():
                                    use_system_ids['asr']['sid'] = duser_def_sys['asr'][lec_lang]
                                else:
                                    use_system_ids['asr']['sid'] = CFG['ASR'][lec_lang]['default']
                                warn("On regenerate '%s' transcription: System ID %d does not exist. Using default ASR system." %(lec_lang, system_id))
                        else:
                            if lec_lang in duser_def_sys['asr'].keys():
                                use_system_ids['asr']['sid'] = duser_def_sys['asr'][lec_lang]
                            else:
                                use_system_ids['asr']['sid'] = CFG['ASR'][lec_lang]['default']
                if retl:
                    all_sub_langs = man_subs_langs + auto_subs_langs.keys()
                    dp = {}
                    for sublang in all_sub_langs:
                        if sublang != lec_lang:
                            if 'media' not in manif.keys() and sublang in man_subs_langs and not(force_opt):
                                if verb: log("Manual '%s' translation existing or provided. Will not automatically translate to '%s' language." % (sublang, sublang))
                            else:
                                try:
                                    system_id, from_lang = auto_subs_langs[sublang]
                                except:
                                    system_id = None
                                    from_lang = None
                                if system_id in CFG['MT']['_ids'].keys():
                                    pair = CFG['MT']['_ids'][system_id]
                                elif from_lang != None:
                                    pair = "%s-%s" % (from_lang, sublang)
                                else:
                                    pair = "%s-%s" % (lec_lang, sublang)
                                    warn("On regenerate '%s' translation: Cannot get system ID or source language. Assuming %s translation pair." % (sublang, pair))
                                if pair not in CFG['MT'].keys():
                                    #code = 3073
                                    #info = "There are not %s MT systems available to directly translate media file into %s language" % (pair, sublang)
                                    #return {'code':code, 'info':info}
                                    warn("There are not %s MT systems available to directly translate media file into %s language. Skipping this target language." % (pair, sublang))
                                else:
                                    if system_id == None or CFG['MT'][pair]['dsys'].get(system_id, None) == None:
                                        warn("On regenerate '%s' translation: Using default system..." %(sublang))
                                        if pair in duser_def_sys['mt'].keys():
                                            system_id = duser_def_sys['mt'][pair]
                                        else:
                                            system_id = CFG['MT'][pair]['default']
                                    dp[pair] = { 'p':pair, 'id':system_id }
                    pathlist = []
                    do_infer_tlpaths(dp.keys(), pathlist)
                    for path in pathlist:
                        sl = path[0].split("-")[0]
                        tl = path[-1].split("-")[1]
                        pair = "%s-%s" % (sl, tl)
                        pairpath = []
                        for spair in path:
                            pairpath.append(dp[spair])
                        use_system_ids['mt'][pair] = pairpath
                if retts:
                    all_at_langs = man_at_langs + auto_at_langs.keys()
                    dp = {}
                    for atlang in all_at_langs:
                        if atlang != lec_lang:
                            if atlang in man_at_langs and not(force_opt):
                                if verb: log("TTS '%s' audiotrack provided. Will not automatically generate TTS track." % atlang)
                            elif atlang not in CFG['TTS'].keys(): # No available ASR system
                                #code = 3070
                                #info = "There are not TTS systems available for language %s" % atlang
                                #return {'code':code, 'info':info}
                                warn("There are not TTS systems available for language %s. Skipping..." % atlang)
                            else:
                                use_system_ids['tts'][atlang] = {}
                                if atlang in auto_at_langs:
                                    system_id = auto_at_langs[atlang]
                                    if system_id == None:
                                        if atlang in duser_def_sys['tts'].keys():
                                            system_id = duser_def_sys['tts'][atlang]
                                        else:
                                            system_id = CFG['TTS'][atlang]['default']
                                        warn("On regenerate '%s' audiotrack: System ID was not specified. Using default TTS system." %(atlang))
                                        use_system_ids['tts'][atlang]['sid'] = system_id
                                    elif CFG['TTS'][atlang]['dsys'].get(system_id, None) != None:
                                        use_system_ids['tts'][atlang]['sid'] = system_id
                                    else:
                                        if atlang in duser_def_sys['tts'].keys():
                                            use_system_ids['tts'][atlang]['sid'] = duser_def_sys['tts'][atlang]
                                        else:
                                            use_system_ids['tts'][atlang]['sid'] = CFG['TTS'][atlang]['default']
                                        warn("On regenerate '%s' audiotrack: System ID %d does not exist. Using default TTS system." %(atlang, system_id))
                                else:
                                    if atlang in duser_def_sys['tts'].keys():
                                        use_system_ids['tts'][atlang]['sid'] = duser_def_sys['tts'][atlang]
                                    else:
                                        use_system_ids['tts'][atlang]['sid'] = CFG['TTS'][atlang]['default']

        # Check if translations and TTS tracks will have the required inputs
    
        # Gather Ingest subtitle outputs
        ingest_sub_outputs = []
        if use_system_ids['asr'] != None: 
            ingest_sub_outputs.append(lec_lang)
        for pair in use_system_ids['mt'].keys():
            tl = pair.split("-")[1]
            ingest_sub_outputs.append(tl)
        all_subs = man_subs_langs + auto_subs_langs.keys() + ingest_sub_outputs
        # MT
        for pair in use_system_ids['mt'].keys():
            sl = pair.split("-")[0]
            if sl not in all_subs:
                code = 3070
                info = "Impossible to launch %s MT system due to the lack of input %s subtitles." % (pair, sl)
                return {'code':code, 'info':info}
        # TTS
        for lang in use_system_ids['tts'].keys():
            if lang not in all_subs:
                code = 3070
                info = "Impossible to generate %s TTS audiotrack due to the lack of input %s subtitles." % (lang, lang)
                return {'code':code, 'info':info}

        # Compute ETA
        time_asr = None
        # ASR
        if use_system_ids['asr'] != None:
            sysinfo = CFG['ASR'][lec_lang]['dsys'][use_system_ids['asr']['sid']]
            if sysinfo['rtf'] != None:
                time_asr = sysinfo['rtf'] * manif['metadata']['duration']
                if use_system_ids['asr']['lma'] and sysinfo['features_tadapt']: 
                    if sysinfo['const_time'] != None:
                        time_asr += sysinfo['const_time']
                    else:
                        time_asr = None
            else:
                time_asr = None
        else:
            time_asr = 0

        # MT
        if len(use_system_ids['mt'].keys()) > 0:
            maxtime_tl = -1
            for pair in use_system_ids['mt'].keys():
                lpath = use_system_ids['mt'][pair]
                acctime = 0
                for lp in lpath:
                    if lp.get('launch', True):
                        tpair = lp['p']
                        tsid = lp['id']
                        sysinfo = CFG['MT'][tpair]['dsys'][tsid]
                        if sysinfo['const_time'] != None and acctime != None:
                            acctime += sysinfo['const_time']
                        else:
                            acctime = None
                if maxtime_tl != None:
                    if acctime != None:
                        maxtime_tl = acctime if acctime > maxtime_tl else maxtime_tl
                    else:
                        maxtime_tl = None
        else:
            maxtime_tl = 0

        # TTS
        if len(use_system_ids['tts'].keys()) > 0:
            maxtime_tts = -1
            for lang in use_system_ids['tts'].keys():
                sid = use_system_ids['tts'][lang]['sid']
                sysinfo = CFG['TTS'][lang]['dsys'][sid]
                acctime = sysinfo['const_time']
                if maxtime_tts != None:
                    if acctime != None:
                        maxtime_tts = acctime if acctime > maxtime_tts else maxtime_tts
                    else:
                        maxtime_tts = None
        else:
            maxtime_tts = 0

        if time_asr == None or maxtime_tl == None or maxtime_tts == None:
            eta = None
        else:
            eta = time_asr + maxtime_tl + maxtime_tts 

    else:
        if verb: log("transLectures option disabled.")
   
    return {'code':0, 'data':use_system_ids, 'eta':eta}

   
def do_infer_tlpaths(plist, slist):
    return infer_tlpaths(plist, [], slist, True, True)

def infer_tlpaths(plist, clist, slist, lb, la):
    if len(plist) == 0: 
        if len(clist) > 0:
            slist.append(clist)
        return
    if len(clist) == 0:
        clist.insert(0, plist.pop())
    # Search input
    if lb:
        slang = clist[0].split("-")[0]
        i = 0; fb = False
        while i <len(plist) and not(fb):
            tlang = plist[i].split("-")[1]
            if tlang == slang:
                fb = True
            i += 1
        if fb:
            clist.insert(0, plist.pop(i-1))
    else:
        fb = False
    # Search output
    if la:
        tlang = clist[-1].split("-")[1]
        i = 0; fa = False
        while i <len(plist) and not(fa):
            slang = plist[i].split("-")[0]
            if slang == tlang:
                fa = True
            i += 1
        if fa:
            clist.append(plist.pop(i-1))
    else:
        fa = False
    if not(fb) and not(fa):
        slist.append(clist)
        clist = []
        fb = True; fa = True
    return infer_tlpaths(plist, clist, slist, fb, fa)
 

#---------- Ingest Output management functions -------------#


def init_output_manifest(stg, up_id):
    D = {}
    D['media'] = []
    D['slides'] = []
    D['attachments'] = []
    D['captions'] = []
    D['audiotracks'] = []
    D['waveform'] = None
    res = stg.create_outputs_dir(up_id)
    if res['code'] != 0:
        return res
    res = stg.store_outputs_manifest(up_id, D)
    if res['code'] != 0:
        return res
    return {'code':0}

def add_file_output_manifest(stg, up_id, fn, filetype, fileformat, key, lang=None, human=False, otype=None, system_id=None, from_subs=False, from_language=None, voice_gender=None):
    res = stg.load_outputs_manifest(up_id)
    if res['code'] != 0:
        return res
    output_manif = res['data']
    outf = {}
    outf['filename'] = fn
    outf['filetype'] = filetype
    outf['fileformat'] = fileformat
    if lang != None:
        outf['language'] = lang
        outf['human'] = human
        outf['type'] = otype
        outf['system_id'] = system_id
        outf['from_subs'] = from_subs
        outf['from_language'] = from_language
        outf['voice_gender'] = voice_gender
    outf['output_file'] = True
    if key == 'waveform':
        output_manif[key] = outf
    else:
        if outf not in output_manif[key]:
            try:
                output_manif[key].append(outf)
            except:
                print_excpt()
                return {'code':1025, 'info': 'Cannot append output file to manifest. Please contact system administrator'}
    res = stg.store_outputs_manifest(up_id, output_manif)
    if res['code'] != 0:
        return res
    return {'code':0}


def output_manifest2storage(output_manif):
    media = output_manif['media']
    slides = output_manif['slides']
    attachments = output_manif['attachments']
    captions = output_manif['captions']
    audiotracks = output_manif['audiotracks']
    waveform = output_manif['waveform']
    return {'code':0, 'data':(media, slides, attachments, captions, audiotracks, waveform) }


def data_for_storage(CFG, stg, up_id, manif):
    res = manifest2storage(CFG, manif)
    if res['code'] != 0:
        return res
    metadata, media, slides, attachments, captions, audiotracks = res['data']
    res = stg.load_outputs_manifest(up_id)
    if res['code'] != 0:
        return res
    output_manif = res['data']
    res = output_manifest2storage(output_manif)
    if res['code'] != 0:
        return res
    o_media, o_slides, o_attachments, o_captions, o_audiotracks, o_waveform = res['data']
    if media != None:
        media = media + o_media
    slides = slides + o_slides
    attachments = attachments + o_attachments
    captions = captions + o_captions
    audiotracks = audiotracks + o_audiotracks
    return {'code':0, 'data':(metadata, media, slides, attachments, captions, audiotracks, o_waveform) }


def manifest2storage(CFG, manif):
    media = []
    slides = []
    attachments = []
    captions = []
    audiotracks = []
    metadata = manif['metadata']
    # Media file
    if 'media' in manif.keys():
        try:
            media_fn = manif['media']['filename']
        except KeyError:
            return {'code':1026, 'info':'"filename" key of media file not found in manifest.json file.'}
        try:
            ff = manif['media']['fileformat']
        except KeyError:
            return {'code':1026, 'info':'"fileformat" key of media file not found in manifest.json file.'}
        store = True if CFG['GEN']['LOCAL_REPO'] else manif['media'].get('store', True)
        url = manif['media'].get('url', None)
        if not(store) and url == None:
            return {'code':1026, 'info':'"url" key of media file not found in manifest.json file.'}
        media.append( {'filename':media_fn, 'filetype':CFG['FTC']['media'], 'fileformat':ff, 'store':store, 'url':url } )
    else:
        media = None
    # Attachments + media
    attachs = manif.get('attachments',[])
    for att in attachs:
        try:
            fn = att['filename']
            ff = att['fileformat']
            type_code = att['type_code']
            store = att.get('store', True)
        except KeyError:
            return {'code':1028, 'info':'"filename", "fileformat", and/or "type_code" keys not found in an attachment in manifest.json file.'}
        if att['type_code'] == CFG['FTC']['media']:
            if 'media' in manif.keys():
                media.append( {'filename':fn, 'filetype':type_code, 'fileformat':ff } )
        elif att['type_code'] == CFG['FTC']['slides']:
            slides.append( {'filename':fn, 'filetype':type_code, 'fileformat':ff, 'store':store } )
        elif att['type_code'] == CFG['FTC']['captions']:
            try:
                lang = att['language']
            except KeyError:
                return {'code':1030, 'info':'"language" key of caption file not found in manifest.json file.'}
            sub_type = "tx"
            if lang != manif['metadata']['language']:
                sub_type = "tl"
            human = att.get('human', True)
            captions.append( {'filename':fn, 'filetype':type_code, 'fileformat':ff, 'language':lang, 'type':sub_type, 'human':human, 'system_id':None, 'from_language':None } )
        elif att['type_code'] == CFG['FTC']['audiotrack']:
            try:
                lang = att['language']
            except KeyError:
                return {'code':1030, 'info':'"language" key of audiotrack file not found in manifest.json file.'}
            human = att.get('human', True)
            from_subs = False
            audiotracks.append( {'filename':fn, 'filetype':type_code, 'fileformat':ff, 'language':lang, 'type':att['type'], 'human':human, 'from_subs':from_subs, 'system_id':None, 'voice_gender':None } )

        else:
            attachments.append( {'filename':fn, 'filetype':type_code, 'fileformat':ff } )
    return {'code':0, 'data':(metadata, media, slides, attachments, captions, audiotracks) }



#------- TL functions ---------------#

def mtsort(keys):
    skeys=[]
    skeys.append(keys.pop())
    _mtsort(keys,[],skeys)
    return skeys

def _mtsort(keys,akeys,skeys):
    if len(keys) == 0:
        if len(akeys) > 0:
            skeys.append(akeys.pop())
            return _mtsort(akeys,[],skeys)
    else:
        p = keys.pop()
        tsrc = p.split('-')[0]
        ttrg = p.split('-')[1]
        i=0; f=False
        while i < len(skeys) and not(f):
            src = skeys[i].split('-')[0]
            trg = skeys[i].split('-')[1]
            if trg == tsrc:
                skeys.insert(i+1,p)
                f=True
            elif src == ttrg:
                skeys.insert(i,p)
                f=True
            i+=1
        if f:
            return _mtsort(keys,akeys,skeys)
        else:
            akeys.append(p)
            return _mtsort(keys,akeys,skeys)
 
    
def prepare_and_launch_translations(CFG, stg, jbs, manif, up_id, lec_hash, verb):
    jobs = []
    pair_jobids = {}
    target_jobids = {}
    for pair in mtsort(manif['use_system_ids']['mt'].keys()):
        sl = pair.split('-')[0]
        tl = pair.split('-')[1]
        tlpath = manif['use_system_ids']['mt'][pair]
        if len(tlpath) == 1:
            if pair in pair_jobids:
                if verb: log("%s captions were explicitly requested but '%s' translation system was already launched." % (tl, pair))
            else:
                system_id = manif['use_system_ids']["mt"][pair][0]['id']
                wait_jobs = target_jobids.get(sl, [])
                if verb: log("Launching '%s' translation system (System ID %d) ..." % (pair, system_id))
                res = launch_translation(CFG, stg, jbs, system_id, sl, tl, up_id, manif, lec_hash, verb, wait_jobs)
                if res['code'] != 0:
                    return res
                jobs = jobs + res['jobs']
                pair_jobids[pair] = res['jobs']
                target_jobids[tl] = res['jobs']
        else:
            if verb: log("Will Launch several MT systems to generate a '%s' translation..." % (pair))
            wait_jobs = target_jobids.get(sl, [])
            for step in tlpath:
                if step.get('launch', True):
                    spair = step['p']
                    ssl = spair.split('-')[0]
                    stl = spair.split('-')[1]
                    if spair in pair_jobids:
                        if verb: log(" - '%s' translation system was already launched." % (spair))
                        wait_jobs = pair_jobids[spair]
                    else:
                        system_id = step['id']
                        if verb: log(" - Launching '%s' translation system (System ID %d) ..." % (spair, system_id))
                        res = launch_translation(CFG, stg, jbs, system_id, ssl, stl, up_id, manif, lec_hash, verb, wait_jobs)
                        if res['code'] != 0:
                            return res
                        jobs = jobs + res['jobs']
                        wait_jobs = res['jobs']         
                        pair_jobids[spair] = res['jobs']
                        target_jobids[stl] = res['jobs']
    info = "Translation(s) in progress."
    scode = CFG['STC']['ST_TL_PROCESSING']
    return {'code':0, 'info':info, 'scode':scode, 'jobs':jobs}

 
def prepare_and_launch_tts(CFG, stg, jbs, manif, up_id, lec_hash, verb):
    jobs = []
    for l in manif['use_system_ids']['tts']:
        system_id = manif['use_system_ids']['tts'][l]['sid']
        if verb: log("Launching '%s' TTS system (System ID %d) ..." % (l, system_id))
        res = launch_tts(CFG, stg, jbs, system_id, l, up_id, manif, lec_hash, verb)
        if res['code'] != 0:
            return res
        jobs = jobs + res['jobs']
    info = "Text-to-Speech audiotracks generation in progress."
    scode = CFG['STC']['ST_TTS_PROCESSING']
    return {'code':0, 'info':info, 'scode':scode, 'jobs':jobs}


def launch_tts(CFG, stg, jbs, system_id, lang, up_id, manif, lec_hash, verb):
    tts = CFG['TTS'][lang]['dsys'][system_id]['module'] # TTS module
    vgender = CFG['TTS'][lang]['dsys'][system_id]['voice_gender'] # TTS module
    res = get_media_file_abspath(CFG, stg, manif, lec_hash, up_id)
    if res['code'] != 0:
        return res
    media_fp = res['data']
    res = get_dfxp_file(CFG, stg, lang, up_id, manif, lec_hash)
    if res['code'] != 0:
        return res
    dfxp_fp = res['data']
    res = stg.create_tmp_tts_lang_dir(up_id, lang)
    if res['code'] != 0:
        return res
    track_fn =  "audiotrack.%s.mp3" % lang
    track_fp = stg.output_file_abspath(track_fn, up_id)
    tmp_dir = stg.tmp_tts_lang_dir(up_id, lang)
    job_name = "%s.%s.%s.tts_%s" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id, lang)
    res = tts.synthesize_track(jbs, media_fp, dfxp_fp, track_fp, tmp_dir, job_name=job_name)
    if res['code'] != 0:
        return res
    jids = res['jids']
    if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
    res = add_file_output_manifest(stg, up_id, track_fn, CFG['FTC']['audiotrack'], 'mp3', 'audiotracks', lang=lang, human=False, otype=VOICE_TYPE_TTS, system_id=system_id, from_subs=True, voice_gender=vgender)
    if res['code'] != 0:
        return res
    return {'code':0, 'jobs':jids}



def transLectures_processing(CFG, stg, jbs, manif, up_id, user_id, lec_hash, verb, mailer, email, session_id=None): 
    # Checking lecture language, attached subtitles, and transLectures option
    lec_id = manif['metadata']['external_id']
    lec_lang = manif['metadata']['language']
    if manif['use_system_ids']['asr'] != None:
        system_id = manif['use_system_ids']['asr']['sid']
        lm_adapt_opt = manif['use_system_ids']['asr']['lma']
        res = launch_transcription(CFG, stg, jbs, system_id, lec_lang, lm_adapt_opt, up_id, manif, lec_hash, verb)
        if res['code'] != 0:
            return res
        info = "Transcription in progress."
        scode = CFG['STC']['ST_TX_PROCESSING']
        return {'code':0, 'info':info, 'scode':scode, 'jobs':res['jobs']}
    elif len(manif['use_system_ids']['mt'].keys()) != 0:
        if verb: log("No transcription processes will be launched.")
        return prepare_and_launch_translations(CFG, stg, jbs, manif, up_id, lec_hash, verb)
    elif len(manif['use_system_ids']['tts'].keys()) != 0:
        if verb: log("No transcription nor translation processes will be launched.")
        return prepare_and_launch_tts(CFG, stg, jbs, manif, up_id, lec_hash, verb)
    else: 
        if verb: log("No transcription nor translation nor TTS processes will be launched.")
        res = media_conversion(CFG, stg, jbs, up_id, manif, verb)
        if res['code'] != 0:
            return res
        if len(res['jobs']) > 0:
            info = "Media conversion for tL Player in progress."
            scode = CFG['STC']['ST_PL_PROCESSING']
            return {'code':0, 'info':info, 'scode':scode, 'jobs':res['jobs']}
        else: 
            return commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=session_id)


def get_media_file_abspath(CFG, stg, manif, lec_hash, up_id):
    if 'media' in manif.keys():
        media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
    else:
        res = stg.get_media_file_abspath(lec_hash)
        if res['code'] != 0:
            return res
        if res['is_url']: 
            url = res['data']
            fn = "mainmedia"
            res = CFG['URLDW'].download_url(url, stg.media_package_file_abspath(fn, up_id), "mp4")
            if res['code'] != 0:
                return res 
            ext, website, decoded = res['data']
            fn = "%s.%s" % (fn, ext)
            media_fp = stg.media_package_file_abspath(fn, up_id)
        else:
            media_fp = res['data']
    return {'code':0, 'data':media_fp}


def launch_translation(CFG, stg, jbs, system_id, src_lang, trg_lang, up_id, manif, lec_hash, verb, wait_jobs=[]):
    tmp_dir = stg.tmp_tl_pair_dir(up_id, src_lang, trg_lang)
    res = stg.create_tmp_tl_pair_object_dir(up_id, src_lang, trg_lang, up_id)
    if res['code'] != 0:
        return res
    cjobs = []
    pair = "%s-%s" % (src_lang, trg_lang)
    mt = CFG['MT'][pair]['dsys'][system_id]['module'] # MT module
    system_label = mt.ID
    # input DFXP
    res = get_dfxp_file(CFG, stg, src_lang, up_id, manif, lec_hash)
    if res['code'] != 0:
        return res
    src_dfxp_fp = res['data']
    in_dfxp_lst = "%s/in_dfxp.lst" % tmp_dir
    stg.write_list([src_dfxp_fp], in_dfxp_lst)
    # input TXT
    src_txt_fp = stg.tmp_tl_pair_object_file_abspath(up_id, src_lang, trg_lang, up_id, "in.txt")
    in_txt_lst = "%s/in_txt.lst" % tmp_dir
    stg.write_list([src_txt_fp], in_txt_lst)
    # output TXT
    trg_txt_fp = stg.tmp_tl_pair_object_file_abspath(up_id, src_lang, trg_lang, up_id, "out.txt")
    out_txt_lst = "%s/out_txt.lst" % tmp_dir
    stg.write_list([trg_txt_fp], out_txt_lst)
    # output DFXP
    trg_dfxp_fn = "subs.%s-%s.tl.dfxp" % (src_lang, trg_lang)
    trg_dfxp_fp = stg.output_file_abspath(trg_dfxp_fn, up_id)
    out_dfxp_lst = "%s/out_dfxp.lst" % tmp_dir
    stg.write_list([trg_dfxp_fp], out_dfxp_lst)
    job_prefix = "%s.%s.%s.tl_%s-%s" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id, src_lang, trg_lang)
    # DFXP2TXT
    dfxp2txt_tmpdir = "%s/dfxp2txt" % tmp_dir
    my_job_name = "%s.dfxp2txt" % job_prefix
    mylog = "%s/dfxp2txt.log" % tmp_dir
    res = dfxp2txt(jbs, in_dfxp_lst, in_txt_lst, dfxp2txt_tmpdir, log_file=mylog, job_name=my_job_name, hold_job_ids=wait_jobs)
    if res['code'] != 0:
        return res
    jids = res['jids']
    cjobs += jids
    # TRANSLATION
    my_job_name = "%s.translate" % job_prefix
    mylog = "%s/translate.log" % tmp_dir
    res = mt.translate(jbs, in_txt_lst, out_txt_lst, tmp_dir, log_file=mylog, job_name=my_job_name, hold_job_ids=jids)
    if res['code'] != 0:
        return res
    jids = res['jids']
    cjobs += jids
    # TXT2DFXP
    txt2dfxp_tmpdir = "%s/txt2dfxp" % tmp_dir
    my_job_name = "%s.txt2dfxp" % job_prefix
    mylog = "%s/txt2dfxp.log" % tmp_dir
    res = txt2dfxp(jbs, trg_lang, dfxp2txt_tmpdir, out_txt_lst, in_dfxp_lst, out_dfxp_lst, txt2dfxp_tmpdir, log_file=mylog, job_name=my_job_name, hold_job_ids=jids, system_label=system_label)
    if res['code'] != 0:
        return res
    jids = res['jids']
    cjobs += jids
    if verb and cjobs != []: log("Job(s) sent to cluster with ID(s) %s" % (str(cjobs)))
    res = add_file_output_manifest(stg, up_id, trg_dfxp_fn, CFG['FTC']['captions'], 'dfxp', 'captions', lang=trg_lang, human=False, otype=CAPTION_TYPE_TL, system_id=system_id, from_language=src_lang)
    if res['code'] != 0:
        return res
    return {'code':0, 'jobs':cjobs}


def launch_transcription(CFG, stg, jbs, system_id, lang, lm_adapt_opt, up_id, manif, lec_hash, verb):
    trm = CFG['TXT'] # Text retrieval module
    asr = CFG['ASR'][lang]['dsys'][system_id]['module'] # ASR module
    asr_LM_Adapt = True
    try:
        asr_LM_Adapt = asr.FEATURE_LM_ADAPT
    except:
        pass
    if trm != None and asr_LM_Adapt and lm_adapt_opt == 1:
        if verb: log("Transcription preparation...")
        res = transcription_preparation(CFG, trm, stg, jbs, up_id, manif, verb)
        if res['code'] != 0:
            return res
        slides, ext_res, cjobs = res['data']
    else:
        if verb: log("LM Adaptation disabled (TR module not available, ASR Module without LM Adapt, or LM Adapt manually disabled).")
        slides, ext_res, cjobs = None, None, []
    if verb: log("Launching '%s' transcription system (System ID %d) ..." % (lang, system_id))
    dfxp_fn = "subs.%s.tx.dfxp" % lang
    res = get_media_file_abspath(CFG, stg, manif, lec_hash, up_id)
    if res['code'] != 0:
        return res
    media_fp = res['data']
    dfxp_fp = stg.output_file_abspath(dfxp_fn, up_id)
    tmp_dir = stg.tmp_tx_dir(up_id)
    job_name = "%s.%s.%s.tx_%s" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id, lang)
    res = asr.transcribe_lecture(jbs, media_fp, dfxp_fp, tmp_dir, slides=slides, ext_res=ext_res, job_name=job_name, hold_job_ids=cjobs)
    if res['code'] != 0:
        return res
    jids = res['jids']
    if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
    cjobs = cjobs + jids
    res = add_file_output_manifest(stg, up_id, dfxp_fn, CFG['FTC']['captions'], 'dfxp', 'captions', lang=lang, human=False, otype=CAPTION_TYPE_TX, system_id=system_id)
    if res['code'] != 0:
        return res
    return {'code':0, 'jobs':cjobs}


def transcription_preparation(CFG, trm, stg, jbs, up_id, manif, verb):
    cjobs = []
    slides = None
    ext_res = []
    ext_res_txt = None
    txt_res = check_txt_resources(CFG, manif, stg=stg)
    tmp_dir = stg.tmp_media_dir(up_id)
    # Slides
    if txt_res['slides'] != None:
        if txt_res['slides']['existing']:
            slides_fp = stg.get_data_file_abspath(txt_res['slides']['filename'])
        else:
            slides_fp = stg.media_package_file_abspath(txt_res['slides']['filename'], up_id)
        txt_slides_fn = "%s.txt" % os.path.basename(txt_res['slides']['filename'])
        txt_slides_fp = stg.output_file_abspath(txt_slides_fn, up_id)
        ff = txt_res['slides']['fileformat']
        if ff in CFG['FFT']['ASVF'] or ff in CFG['FFT']['APK']:
            if ff in CFG['FFT']['ASVF']:
              if verb: log("Found slides in video format. Launching OCR text extraction...")
              job_name = "%s.%s.%s.videoslides2text" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
              res = trm.video_slides2text(jbs, slides_fp, txt_slides_fp, manif['metadata']['language'], tmp_dir, job_name=job_name)
            else:
              if verb: log("Found JPG slides bundled into a compressed package file. Launching OCR text extraction...")
              job_name = "%s.%s.%s.jpgslides2text" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
              jpg_folder = slides_fp + ".ex"
              res = trm.jpg_slides2text(jbs, jpg_folder, txt_slides_fp, manif['metadata']['language'], tmp_dir, job_name=job_name)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, txt_slides_fn, CFG['FTC']['slides'], 'txt', 'slides')
            if res['code'] != 0:
                return res
            slides = txt_slides_fp
        elif ff == 'txt':
            if verb: log("Found slides in TEXT format.")
            slides = slides_fp
        elif ff in CFG['FFT']['ASTF']:
            if verb: log("Found slides in PPT/PDF format. Launching text extraction...")
            job_name = "%s.%s.%s.slides2text" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
            res = trm.slides2text(jbs, slides_fp, txt_slides_fp, tmp_dir, job_name=job_name)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            # Do not store it in DB since it's easy and fast to obtain text from PPT/PDF
            #res = add_file_output_manifest(stg, up_id, txt_slides_fn, CFG['FTC']['slides'], 'txt', 'slides')
            #if res['code'] != 0:
            #    return res
            slides = txt_slides_fp
        else:
            return {'code':1, 'info':'Unexpected slide file type.'}
    # External Resources
    if len(txt_res['docs']) > 0:
        for er in txt_res['docs']:
            if er['existing']:
                ext_res_fp = stg.get_data_file_abspath(er['filename'])
            else:
                ext_res_fp = stg.media_package_file_abspath(er['filename'], up_id)
            txt_ext_res_fn = "%s.txt" % os.path.basename(er['filename'])
            txt_ext_res_fp = stg.output_file_abspath(txt_ext_res_fn, up_id)
            ff = er['fileformat']
            if ff == 'txt':
                if verb: log("Found external resource in TEXT format.")
                ext_res.append(ext_res_fp)
            elif ff in CFG['FFT']['ADF']:
                if (er['existing'] and not(os.path.exists("%s.txt" % stg.get_data_file_abspath(os.path.basename(er['filename']))))) or not(er['existing']):
                    if verb: log("Found external resource in PDF/DOC format. Launching text extraction...")
                    job_name = "%s.%s.%s.ext_res2text" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
                    res = trm.ext_res2text(jbs, ext_res_fp, txt_ext_res_fp, tmp_dir, job_name=job_name)
                    if res['code'] != 0:
                        return res
                    jids = res['jids']
                    if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
                    cjobs = cjobs + jids
                    # Do not store it in DB since it's easy and fast to obtain text from PPT/PDF
                    #res = add_file_output_manifest(stg, up_id, txt_ext_res_fn, CFG['FTC']['docs'], 'txt', 'attachments')
                    #if res['code'] != 0:
                    #    return res
                    ext_res.append(txt_ext_res_fp)
            else:
                return {'code':1, 'info':'Unexpected external resource file type.'}
    else:
        lec_title = manif['metadata'].get('title', None)
        if lec_title != None:
            if verb: log("No document attached. Launching external resources retrieval...")
            txt_ext_res_fn = "retrieved_ext_res.txt"
            txt_ext_res_fp = stg.output_file_abspath(txt_ext_res_fn, up_id)
            lang = manif['metadata']['language']
            job_name = "%s.%s.%s.ext_res_retrieval" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
            res = trm.retrieve_external_resources(jbs, lec_title, lang, txt_ext_res_fp, tmp_dir, job_name=job_name)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, txt_ext_res_fn, CFG['FTC']['docs'], 'txt', 'attachments')
            if res['code'] != 0:
                return res
            ext_res.append(txt_ext_res_fp)
    if len(ext_res) > 1:
        cat_ext_res_fp = "%s/ext_res.txt" % stg.tmp_tx_dir(up_id)
        log_file = "%s/cat_ext_res.log" % stg.tmp_tx_dir(up_id)
        cmd = 'cat %s > %s' % (' '.join(ext_res), cat_ext_res_fp)
        job_name = "%s.%s.%s.cp_ext_res" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id)
        res = jbs.submit_job(cmd, output=log_file, job_name=job_name, local=True, hold_jid=cjobs)
        if res['code'] != 0:
            return res
        jid = res['jid']
        cjobs = cjobs + [jid]
        ext_res_txt = cat_ext_res_fp
    elif len(ext_res) == 1:
        ext_res_txt = ext_res[0]
    return {'code':0, 'data':(slides, ext_res_txt, cjobs)}


def media_conversion(CFG, stg, jbs, up_id, manif, verb):
    cjobs = []
    tmp_dir = stg.tmp_media_dir(up_id)
    # Video
    conv = check_media_formats_conversion(CFG, manif)
    if conv['FROM_AUDIO']:
        if verb: log("Media file provided in audio format.")
        # to MP4
        if conv['MP4']:
            if verb: log("Audio file will be converted into MP4 format.")
            media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
            mp4_fn = "%s.mp4" % manif['media']['filename']
            mp4_fp = stg.output_file_abspath(mp4_fn, up_id)
            job_name = "%s.%s.%s.audio2mp4" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
            bg_image = CFG['ABG']
            res = audio2mp4(jbs, media_fp, mp4_fp, tmp_dir=tmp_dir, job_name=job_name, bg_image=bg_image)
            if res['code'] != 0:
                return res
            jids = res['jids']
            a2v_jids = jids
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, mp4_fn, CFG['FTC']['media'], 'mp4', 'media')
            if res['code'] != 0:
                return res
        ## to OGV, must wait previous process
        if conv['OGV']:
            if verb: log("Audio file will be converted into OGV format.")
            media_fp = mp4_fp
            ogv_fn = "%s.ogv" % mp4_fn
            ogv_fp = stg.output_file_abspath(ogv_fn, up_id)
            job_name = "%s.%s.%s.media2ogv" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
            res = media2ogv(jbs, media_fp, ogv_fp, tmp_dir, job_name=job_name, hold_job_ids=a2v_jids)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, ogv_fn, CFG['FTC']['media'], 'ogv', 'media')
            if res['code'] != 0:
                return res
        ## Thumbnail
        if verb: log("A thumbnail image will be automatically generated from the video.")
        media_fp = mp4_fp
        thumb_fn = "%s.jpg" % mp4_fn
        thumb_fp = stg.output_file_abspath(thumb_fn, up_id)
        job_name = "%s.%s.%s.thumbnail" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
        res = make_thumbnail(jbs, media_fp, thumb_fp, tmp_dir=tmp_dir, job_name=job_name, hold_job_ids=a2v_jids, default_thumbnail=CFG['DEF_THUMB'])
        if res['code'] != 0:
            return res
        jids = res['jids']
        if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
        cjobs = cjobs + jids
        res = add_file_output_manifest(stg, up_id, thumb_fn, CFG['FTC']['thumbnail'], 'jpg', 'attachments')
        if res['code'] != 0:
            warn('Cannot generate thumbnail. Info: %s' % res['info']) 
    else:
        if conv['MP4']:
            if verb: log("Media file will be converted into MP4 format.")
            media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
            mp4_fn = "%s.mp4" % manif['media']['filename']
            mp4_fp = stg.output_file_abspath(mp4_fn, up_id)
            job_name = "%s.%s.%s.media2mp4" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
            res = media2mp4(jbs, media_fp, mp4_fp, tmp_dir, job_name=job_name)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, mp4_fn, CFG['FTC']['media'], 'mp4', 'media')
            if res['code'] != 0:
                return res
        if conv['OGV']:
            if verb: log("Media file will be converted into OGV format.")
            media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
            ogv_fn = "%s.ogv" % manif['media']['filename']
            ogv_fp = stg.output_file_abspath(ogv_fn, up_id)
            job_name = "%s.%s.%s.media2ogv" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
            res = media2ogv(jbs, media_fp, ogv_fp, tmp_dir, job_name=job_name)
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, ogv_fn, CFG['FTC']['media'], 'ogv', 'media')
            if res['code'] != 0:
                return res
        if conv['THUMB']:
            if verb: log("A thumbnail image from the video will be automatically generated.")
            media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
            thumb_fn = "%s.jpg" % manif['media']['filename']
            thumb_fp = stg.output_file_abspath(thumb_fn, up_id)
            job_name = "%s.%s.%s.thumbnail" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
            res = make_thumbnail(jbs, media_fp, thumb_fp, tmp_dir=tmp_dir, job_name=job_name, default_thumbnail=CFG['DEF_THUMB'])
            if res['code'] != 0:
                return res
            jids = res['jids']
            if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
            cjobs = cjobs + jids
            res = add_file_output_manifest(stg, up_id, thumb_fn, CFG['FTC']['thumbnail'], 'jpg', 'attachments')
            if res['code'] != 0:
                warn('Cannot generate thumbnail. Info: %s' % res['info']) 

    if conv['PCM']:
        if verb: log("A PCM stream file from the audio will be automatically generated.")
        media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
        pcm_fn = "%s.pcm" % manif['media']['filename']
        pcm_fp = stg.output_file_abspath(pcm_fn, up_id)
        job_name = "%s.%s.%s.pcm_stream" % (CFG['SCH']['JOB_NAME_PREFIX'], CFG['STO']['DB_NAME'], up_id) 
        res = media2pcm(jbs, media_fp, pcm_fp, tmp_dir=tmp_dir, job_name=job_name)
        if res['code'] != 0:
            return res
        jids = res['jids']
        if verb and jids != []: log("Job(s) sent to cluster with ID(s) %s" % str(jids))
        cjobs = cjobs + jids
        res = add_file_output_manifest(stg, up_id, pcm_fn, CFG['FTC']['waveform'], 'pcm', 'waveform')
        if res['code'] != 0:
            warn('Cannot generate PCM Stream. Info: %s' % res['info']) 


    return {'code':0, 'jobs':cjobs}


def caption_conversion(CFG, stg, jbs, lec_lang, up_id, manif, verb):
    conv = check_caption_conversion(CFG, manif)
    for c in conv:
        src_fp = stg.media_package_file_abspath(c['filename'], up_id)
        dfxp_fn = "%s.dfxp" % c['filename']
        dfxp_fp = stg.output_file_abspath(dfxp_fn, up_id)
        if c['fileformat'] == 'srt':
            if verb: log("Subtitles file in srt format found. Converting into dfxp... ")
            res = srt2dfxp(src_fp, dfxp_fp)
        elif c['fileformat'] == 'trs':
            if verb: log("Subtitles file in trs format found. Converting into dfxp... ")
            res = trs2dfxp(src_fp, dfxp_fp)
        else:
            return {'code':1, 'info':'Unexpected caption format'}
        if res['code'] != 0:
            return res
        human = c.get('human', True)
        otype = "tl"
        if c['language'] == lec_lang:
            otype = "tx"
        res = add_file_output_manifest(stg, up_id, dfxp_fn, CFG['FTC']['captions'], 'dfxp', 'captions', lang=c['language'], otype=otype, human=human)
        if res['code'] != 0:
            return res
        if verb: log("Subtitles file succesfully converted into dfxp format and registered as ingest output.")
    return {'code':0}


def commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=None):

    if manif['operation_code'] == CFG['OPC']['NEW']:
        if verb: log("Inserting new lecture into DB and copying files...")
        # prepare insert info
        res = data_for_storage(CFG, stg, up_id, manif) 
        if res['code'] != 0:
            return res
        metadata, media, slides, attachments, captions, audiotracks, waveform = res['data']
        if metadata.get('date', None) in (None, ""):
            metadata['date'] = datetime.now()
        # insert lecture
        res = stg.insert_lecture(up_id, metadata, media, slides, attachments, captions, audiotracks, waveform, user_id)
        if res['code'] != 0:
            return res
        info = "Lecture succesfully inserted into DB with ID '%s'" % lec_id
        scode = CFG['STC']['ST_NEW_FINISHED']
        return {'code':0, 'info':info, 'scode':scode}

    elif manif['operation_code'] == CFG['OPC']['UPDATE']:
        if verb: log("Updating lecture...")
        # prepare insert info
        res = data_for_storage(CFG, stg, up_id, manif) 
        if res['code'] != 0:
            return res
        metadata, media, slides, attachments, captions, audiotracks, waveform = res['data']
        # Do not update with automatic subs that have been supervised in the meantime.
        new_captions = []
        for c in captions:
            if not(c['human']) and not(manif.get('force', False)):
                clang = c['language']
                res = stg.has_supervised_subs(lec_hash, clang)
                if res['code'] != 0:
                    return res
                has_sup_subs = res['data']
                if not(has_sup_subs):
                    new_captions.append(c)
                else:
                    warn("Won't update '%s' subtitles, they have been edited by a user during the upload processing." % clang)
            else:
                new_captions.append(c)
        regen = bool(len(manif.get('re-generate', [])))
        if metadata.get('date', None) in (None, ""):
            metadata['date'] = datetime.now()
        # update lecture
        res = stg.update_lecture(up_id, metadata, media, slides, attachments, new_captions, audiotracks, waveform, user_id, regen)
        if res['code'] != 0:
            return res
        # End session
        if CFG['SES']['ENABLED'] and session_id != None:
            res = stg.end_session(session_id, CFG['SES']['AID'], ended_by_name=CFG['SES']['ANM'])
            if res['code'] != 0:
                return res
        info = "Lecture with ID '%s' succesfully updated." % lec_id
        scode = CFG['STC']['ST_UP_FINISHED']
        return {'code':0, 'info':info, 'scode':scode}

    else:
        info = "Unknown Operation Code."
        scode = CFG['STC']['ST_MP_ERROR']
        return {'code':4000, 'info':info, 'scode':scode}

    
def check_running_processes(rjobs, jobs):
    for jid in jobs:
        try:
            jst = rjobs[jid]
        except KeyError:
            continue
        if jst == Scheduler.ST_RUNNING:
            return {'code':0, 'data':Scheduler.ST_RUNNING}
        elif jst == Scheduler.ST_QUEUED:
            return {'code':0, 'data':Scheduler.ST_QUEUED}
        elif jst == Scheduler.ST_ERROR:
            return {'code':0, 'data':Scheduler.ST_ERROR}
    return {'code':0, 'data':Scheduler.ST_FINISHED}


def check_user_quota(stg, jbs, up_id, user_id, manif):
    res = stg.get_user_quota(user_id)
    if res['code'] != 0:
        return res
    lim_sec = res['data']['seconds']
    lim_nvid = res['data']['num_videos']
    if lim_nvid != None and lim_nvid <= 0:
        info = "You have reached your video uploads allowance."
        return {'code':3012, 'info':info}
    if lim_sec != None:
        if lim_sec <= 0:
            info = "You have reached your time allowance."
            return {'code':3010, 'info':info}
        media_fp = stg.media_package_file_abspath(manif['media']['filename'], up_id)
        duration = manif['metadata']['duration']
        if (lim_sec - duration) < 0:
            info = "You have %d seconds left, which is not enough to process this media (%d seconds length)." % (lim_sec, duration)
            return {'code':3011, 'info':info}
    return {'code':0}


#---------- Upload processing functions -------------#


def process_failed_uploads(CFG, stg, uploads, verb, debug, commit_errors):
    for up in uploads:
        try:
            if up['up_status'] in CFG['STC']['ERROR_CODES']:
                td = (datetime.now() - up['last_ts']).days
                if td >= CFG['GEN']['RM_ERR_UP_DAYS']:
                    up_id = up['id']
                    if verb: log("[*] Deleting upload ID %s ..." % up_id)
                    up_dir = "%s/%s" % (stg.uploads_mp, up_id)
                    res = stg.delete_upload(up_id, up_dir)
                    if res['code'] != 0:
                        error(res['info'], exit_code=res['code'])
                    if verb: log("Upload ID %s succesfully deleted." % up_id)
        except:
            bug("Upload ID: %s" % up['id'])
            continue


def process_finished_uploads(CFG, stg, uploads, verb, debug, commit_errors):
    for up in uploads:
        try:
            if up['up_status'] in CFG['STC']['FINISHED_CODES']:
                td = (datetime.now() - up['last_ts']).days
                if td >= CFG['GEN']['RM_FIN_UP_DAYS']:
                    up_id = up['id']
                    if verb: log("[*] Deleting upload ID %s ..." % up_id)
                    up_dir = "%s/%s" % (stg.uploads_mp, up_id)
                    res = stg.delete_upload(up_id, up_dir)
                    if res['code'] != 0:
                        error(res['info'], exit_code=res['code'])
                    if verb: log("Upload ID %s succesfully deleted." % up_id)
        except:
            bug("Upload ID: %s" % up['id'])
            continue


def process_running_uploads(CFG, stg, jbs, uploads, verb, mailer, debug, commit_errors):
    # Get Running jobs
    res = jbs.get_running_jobs()
    if res['code'] != 0:
        error(res['info'], exit_code=res['code'])
    rjobs = res['data']
    # For each upload
    for up in uploads:
        try:
            if up['up_status'] in CFG['STC']['PROCESSING_CODES']:
                up_id = up['id']
                if verb: log("[*] Checking running upload ID %s ..." % up_id)
                res = stg.load_manifest(up_id)
                if res['code'] != 0:
                    ingest_error(stg, None, CFG, up_id, up['external_id'], None, up['user_id'], CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, None, debug, session_id=up['session_id'], commit_errors=commit_errors)
                manif = res['data']
                email = manif.get('email', None)
                lec_id = manif['metadata']['external_id']
                lec_lang = manif['metadata']['language']
                user_id = up['user_id']
                lec_hash = up['hash']
                session_id = up['session_id']
                if CFG['SES']['ENABLED'] and session_id != None:
                    res = stg.update_session(session_id)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                res = check_running_processes(rjobs, up['jobs'])
                if res['code'] != 0:
                    error(res['info'], exit_code=res['code'])
                job_status = res['data']
                if job_status in (Scheduler.ST_QUEUED, Scheduler.ST_RUNNING):
                    if verb: log("Processes still queued/running.")
                    continue
                elif job_status == Scheduler.ST_ERROR:
                    if up['up_status'] == CFG['STC']['ST_TX_PROCESSING']:
                        info = "Transcription Error. Please contact system administrator."
                        error_code = 2000
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TX_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    elif up['up_status'] == CFG['STC']['ST_TL_PROCESSING']:
                        info = "Translation Error. Please contact system administrator."
                        error_code = 2001
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TL_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    elif up['up_status'] == CFG['STC']['ST_TTS_PROCESSING']:
                        info = "Text-To-Speech Error. Please contact system administrator."
                        error_code = 2001
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TL_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    elif up['up_status'] == CFG['STC']['ST_PL_PROCESSING']:
                        info = "Media Conversion Error. Please contact system administrator."
                        error_code = 2002
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    else:
                        error_code = 2003
                        error("Unknown Upload Status Code %d" % up['up_status'], exit_code=error_code)
                elif job_status == Scheduler.ST_FINISHED:

                    if up['up_status'] == CFG['STC']['ST_TX_PROCESSING']:
                        if verb: log("Transcription processes already finished.")
                        if len(manif['use_system_ids']['mt'].keys()) != 0:
                            res = prepare_and_launch_translations(CFG, stg, jbs, manif, up_id, lec_hash, verb)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TL_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue
                            else:
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TL_PROCESSING'], res['info'], jobs=res['jobs'])
                        elif len(manif['use_system_ids']['tts'].keys()) != 0: 
                            if verb: log("No translation processes will be launched.")
                            res = prepare_and_launch_tts(CFG, stg, jbs, manif, up_id, lec_hash, verb)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TTS_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue
                            else:
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TTS_PROCESSING'], res['info'], jobs=res['jobs'])
                        else:
                            if verb: log("No translation nor TTS processes will be launched.")
                            res = media_conversion(CFG, stg, jbs, up_id, manif, verb)
                            if res['code'] != 0:
                                return res
                            if len(res['jobs']) > 0:
                                info = "Media conversion for tL Player in progress."
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_PROCESSING'], info, jobs=res['jobs'])
                            else: 
                                res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=session_id)
                                if res['code'] != 0:
                                    ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                    continue
                                scode = res['scode']
                                info = res['info']
                                jobs = res.get('jobs',[1])
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                                if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)

                    elif up['up_status'] == CFG['STC']['ST_TL_PROCESSING']:
                        if verb: log("Translation processes already finished.")
                        if len(manif['use_system_ids']['tts'].keys()) != 0: 
                            res = prepare_and_launch_tts(CFG, stg, jbs, manif, up_id, lec_hash, verb)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TTS_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue
                            else:
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_TTS_PROCESSING'], res['info'], jobs=res['jobs'])
                        else:
                            if verb: log("No TTS processes will be launched.")
                            res = media_conversion(CFG, stg, jbs, up_id, manif, verb)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue
                            if len(res['jobs']) > 0:
                                info = "Media conversion for tL Player in progress."
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_PROCESSING'], info, jobs=res['jobs'])
                            else:
                                res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=session_id)
                                if res['code'] != 0:
                                    ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                    continue
                                scode = res['scode']
                                info = res['info']
                                jobs = res.get('jobs',[1])
                                ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                                if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
    
                    elif up['up_status'] == CFG['STC']['ST_TTS_PROCESSING']:
                        if verb: log("TTS processes already finished.")
                        res = media_conversion(CFG, stg, jbs, up_id, manif, verb)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        if len(res['jobs']) > 0:
                            info = "Media conversion for tL Player in progress."
                            ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_PL_PROCESSING'], info, jobs=res['jobs'])
                        else:
                            res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=session_id)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue
                            scode = res['scode']
                            info = res['info']
                            jobs = res.get('jobs',[1])
                            ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                            if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
 
                    elif up['up_status'] == CFG['STC']['ST_PL_PROCESSING']:
                        if verb: log("Media conversion processes already finished.")
                        res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb, session_id=session_id)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        scode = res['scode']
                        info = res['info']
                        jobs = res.get('jobs',[1])
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                        if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
        except:
            bug("Upload ID: %s" % up['id'])
            continue


def process_new_uploads(CFG, stg, jbs, uploads, verb, mailer, debug, commit_errors):
    for up in uploads:
        try:
            if up['up_status'] == CFG['STC']['ST_INGESTED']:
                up_id = up['id']
                manif = json.loads(base64.b64decode(up['manifest']), encoding='utf-8')
                manif['operation_code'] = up['op_code']
                session_id=None
                email = manif.get('email', None)
                if email != None:
                    for e in email:
                        res = sts(e)
                        if res[1] > 0:
                            scode = 1004
                            info = 'Invalid e-mail (contains forbidden characters)'
                            ingest_error(stg, None, CFG, up_id, up['external_id'], None, up['user_id'], CFG['STC']['ST_MP_ERROR'], scode, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                if verb: log("[*] Processing new upload ID %s ..." % up_id)
                res = process_media_package(CFG, stg, jbs, up_id, up['hash'], manif)
                if res['code'] != 0:
                    ingest_error(stg, None, CFG, up_id, up['external_id'], None, up['user_id'], CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                    continue

                lec_id = manif['metadata']['external_id']
                lec_lang = manif['metadata'].get('language', None)
                user_id = up['user_id']
                lec_hash = stg.get_hash(lec_id, user_id)

                if lec_id != up['external_id'].decode('utf-8'):
                    scode = 3005
                    info = 'Mismatch between external_id given to web service and external_id given in the Manifest.JSON file'
                    ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], scode, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                    continue

                if lec_hash != up['hash']:
                    scode = 3006
                    info = 'Non-matching hash values.'
                    ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], scode, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                    continue

                if manif['operation_code'] != up['op_code']:
                    scode = 3007
                    info = 'Mismatch between "opc" parameter given to web service and "operation_code" given in the Manifest.JSON file'
                    ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], scode, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                    continue

                user_is_admin = None
                username = None
                realusername = None
                usermail = None
                if CFG['AUT']['ENABLED']:
                    if user_id == None:
                        code = 3006
                        info = 'User ID from uploads row is NULL and authentication is enabled!'
                        error(info, exit_code=code)
                        continue
                    res = stg.get_user(user_id)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    user_is_admin = res['data'][0]['is_admin']
                    username = res['data'][0]['username']
                    realusername = res['data'][0]['name']
                    usermail = res['data'][0]['email']

                manif['username'] = username
             
                tL = manif.get('generate', True)

                eta = None; etas = None

                if manif['operation_code'] in (CFG['OPC']['NEW'], CFG['OPC']['UPDATE']):
                    if CFG['AUT']['ENABLED'] and not(user_is_admin) and not(manif.get('test_mode', False)) and tL:
                        # Check and update quota if new of re-recording
                        if 'media' in manif:
                            # Check quota
                            res = check_user_quota(stg, jbs, up_id, user_id, manif)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue
                            # Update quota
                            res = stg.update_user_quota(up_id, user_id, manif['metadata']['duration'])
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue

                    # Default Speaker 
                    if 'speakers' not in manif['metadata']:
                        def_speaker = { "speaker_id":"default-speaker", "speaker_name":"Default Speaker", "speaker_gender":None, "speaker_email":None }
                        if username != None: 
                            def_speaker["speaker_id"] = username
                            def_speaker["speaker_name"] = realusername
                            def_speaker["speaker_email"] = usermail
                        manif['metadata']['speakers'] = [ def_speaker ]

                    # determine tasks
                    res = determine_tL_tasks(CFG, stg, manif, up_id, lec_hash, lec_lang, user_id, user_is_admin, verb)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    use_system_ids = res['data']
                    etas = res['eta'] 
                    manif['use_system_ids'] = use_system_ids
                    res = stg.store_manifest(up_id, manif, backup=False)
                    if res['code'] != 0:
                        return res

                    # Time estimation
                    if etas != None:
                        conv = check_media_formats_conversion(CFG, manif)
                        if conv['FROM_AUDIO'] or conv['MP4']:
                            etas += manif['metadata']['duration']
                        etas += CONST_IS_TIME
                        eta = datetime.now() + timedelta(seconds=etas)

                    # Create Outputs JSON File
                    res = init_output_manifest(stg, up_id)
                    if res['code'] != 0:
                        res2 = stg.restore_user_quota(up_id, user_id)
                        if res2['code'] != 0:
                             warn("Cannot restore user quota (Error Code: %d, Info: %s)." % (res2['code'], res2['info']))
                        error(res['info'], exit_code=res['code'])
     
                    # Caption Conversion
                    res = caption_conversion(CFG, stg, jbs, lec_lang, up_id, manif, verb)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                        continue

                    # Create tmp dirs
                    res = stg.create_tmp_dirs(up_id)
                    if res['code'] != 0:
                        res2 = stg.restore_user_quota(up_id, user_id)
                        if res2['code'] != 0:
                             warn("Cannot restore user quota (Error Code: %d, Info: %s)." % (res2['code'], res2['info']))
                        error(res['info'], exit_code=res['code'])

                # Cancel Operation
                if manif['operation_code'] == CFG['OPC']['CANCEL']:

                    if verb: log("Cancel upload operation.")
                    res = stg.get_upload_by_id(lec_id)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_CANCEL_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) > 0:
                        tup = None
                        for r in res['data']:
                            if r['up_status'] in CFG['STC']['PROCESSING_CODES']:
                                tup = r
                                break
                        if tup != None:
                            if CFG['AUT']['ENABLED'] and user_id != tup['user_id'] and not(user_is_admin):
                                code = 3020
                                info = "Your are not allowed to cancel upload ID %s" % tup['id']
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_CANCEL_ERROR'], code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                continue                        
                            res = jbs.delete_jobs(tup['jobs'])
                            if res['code'] != 0:
                                print_excpt()
                                warn(res['info'])
                                #ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_CANCEL_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                #continue
                            if CFG['AUT']['ENABLED'] and tup['user_id'] != None:
                                 res = stg.restore_user_quota(tup['id'], tup['user_id'])
                                 if res['code'] != 0:
                                     warn("Cannot restore user quota (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                            info = "Upload cancelled due a cancel request (Upload ID '%s')." % up_id
                            ingest_update(stg, tup['id'], tup['external_id'], tup['language'], tup['user_id'], CFG['STC']['ST_CANCELLED'], info)
                            info = "Upload ID '%s' of External ID '%s' successfully canceled." % (tup['id'], lec_id)
                            ingest_update(stg, up_id, lec_id, None, user_id, CFG['STC']['ST_CANCEL_FINISHED'], info)
                        else:
                            info = "Upload ID '%s' was not being processed." % lec_id
                            ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_CANCEL_FINISHED'], info)
                    else:
                        info = "Upload ID '%s' was not being processed." % lec_id
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_CANCEL_FINISHED'], info)
                                    

                # Delete Operation
                elif manif['operation_code'] == CFG['OPC']['DELETE']:

                    if verb: log("Delete lecture operation.")
                    delete_mode = manif.get("delete_mode", DELETE_MODE_HARD)
                    res = stg.get_lecture_info(lec_hash)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) == 0:
                        info = "External ID '%s' does not exist in Database." % lec_id
                        error_code = 3000
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    owner_user_id = res['data'][0]['user_id']
                    if CFG['AUT']['ENABLED'] and user_id != owner_user_id and not(user_is_admin):
                        code = 3021
                        info = "Your are not allowed to delete lecture ID %s" % lec_id
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue                        
                    if delete_mode == DELETE_MODE_SOFT:
                        if res['data'][0]['removed']:
                            info = "Lecture ID '%s' already removed." % lec_id
                            error_code = 3001
                            nngest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        res = stg.delete_lecture(lec_id, lec_hash)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        info = "Lecture succesfully removed."
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DEL_FINISHED'], info)
                        #if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                    elif delete_mode == DELETE_MODE_HARD:
                        res = stg.hard_delete_lecture(lec_id, lec_hash)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        info = "Lecture succesfully removed (hard mode)."
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DEL_FINISHED'], info)
                        #if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                    else:
                        info = "Unknown delete mode '%s'." % delete_mode
                        error_code = 3050
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue

                # Update Operation
                elif manif['operation_code'] == CFG['OPC']['UPDATE']:

                    if verb: log("Update lecture operation.")
                    # Check if lecture's external_id already exists in DB
                    res = stg.get_lecture_info(lec_hash)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) == 0:
                        info = "External ID '%s' does not exist in Database." % lec_id
                        error_code = 3002
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    owner_user_id = res['data'][0]['user_id']
                    if CFG['AUT']['ENABLED'] and user_id != owner_user_id and not(user_is_admin):
                        code = 3022
                        info = "Your are not allowed to update lecture ID %s" % lec_id
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue                        
                    # Check whether this external_id lecture is already being processed
                    res = stg.get_upload_by_hash(lec_hash)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) > 0:
                        cont = False
                        for r in res['data']:
                            if r['up_status'] in CFG['STC']['PROCESSING_CODES']:
                                cont = True
                                info = "External ID '%s' is already being processed (Upload ID: '%s')." % (lec_id, r['id'])
                                error_code = 3004
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                break
                        if cont: 
                            continue
                    # Test Mode
                    if manif.get('test_mode', False):
                        dfxp_fn = "subs.%s.tx.dfxp" % lec_lang
                        dfxp_fp = stg.output_file_abspath(dfxp_fn, up_id)
                        res = create_empty_dfxp(manif['metadata']['duration'], dfxp_fp)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                            continue
                        res = add_file_output_manifest(stg, up_id, dfxp_fn, CFG['FTC']['captions'], 'dfxp', 'captions', lang=lec_lang, human=False, otype=CAPTION_TYPE_TX)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                            continue
                        if CFG['TEST_THUMB'] != None:
                            thumb_fn = os.path.basename(CFG['TEST_THUMB'])
                            thumb_fp = stg.output_file_abspath(thumb_fn, up_id)
                            thumb_ext = os.path.splitext(CFG['TEST_THUMB'])[1][1:]
                            res = stg.copy_file(CFG['TEST_THUMB'], thumb_fp)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue
                            res = add_file_output_manifest(stg, up_id, thumb_fn, CFG['FTC']['thumbnail'], thumb_ext, 'attachments')
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue
                        res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        scode = res['scode']
                        info = res['info']
                        jobs = res.get('jobs',[1])
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                        if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                    else:
                        if CFG['SES']['ENABLED']:
                            # Start edit session
                            res = stg.start_session(lec_hash, user_id, CFG['SES']['AID'], CFG['SES']['ACF'], author_name=CFG['SES']['ANM'])
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, commit_errors=commit_errors)
                                continue
                            session_id = res['data']
                            up['session_id'] = session_id
                        else:
                            session_id = None
    
                        # Process data
                        res = transLectures_processing(CFG, stg, jbs, manif, up_id, user_id, lec_hash, verb, mailer, email, session_id=session_id)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        scode = res['scode']
                        info = res['info']
                        jobs = res.get('jobs',[1])
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, session_id=session_id, jobs=jobs, eta=eta)
                        if scode in (CFG['STC']['ST_NEW_FINISHED'], CFG['STC']['ST_UP_FINISHED'], CFG['STC']['ST_DEL_FINISHED']):
                            if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                        else:
                            if mailer != None: send_mail_processing(mailer, manif, CFG, up_id, email, eta=etas)
                            if eta != None and verb: log("Estimated time to complete: %s." % (eta))        
    
                # New Operation
                elif manif['operation_code'] == CFG['OPC']['NEW']:

                    if verb: log("New lecture operation.")
                    # Check if lecture's external_id already exists in DB
                    res = stg.get_lecture_info(lec_hash)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) > 0:
                        info = "External ID '%s' already exists in Database." % lec_id
                        error_code = 3003
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    # Check whether this external_id lecture is already being processed
                    res = stg.get_upload_by_hash(lec_hash)
                    if res['code'] != 0:
                        ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                        continue
                    if len(res['data']) > 0:
                        cont = False
                        for r in res['data']:
                            if r['up_status'] in CFG['STC']['PROCESSING_CODES']:
                                cont = True
                                info = "External ID '%s' is already being processed (Upload ID: '%s')." % (lec_id, r['id'])
                                error_code = 3004
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], error_code, info, mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                                break
                        if cont: 
                            continue
                    # Empty DFXP if Test Mode or tL option disabled
                    if manif.get('test_mode', False) or not(tL):
                        dfxp_fn = "subs.%s.tx.dfxp" % lec_lang
                        dfxp_fp = stg.output_file_abspath(dfxp_fn, up_id)
                        res = create_empty_dfxp(manif['metadata']['duration'], dfxp_fp)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                            continue
                        res = add_file_output_manifest(stg, up_id, dfxp_fn, CFG['FTC']['captions'], 'dfxp', 'captions', lang=lec_lang, human=False, otype=CAPTION_TYPE_TX)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                            continue
                    # Test Mode
                    if manif.get('test_mode', False):
                        if CFG['TEST_THUMB'] != None:
                            thumb_fn = os.path.basename(CFG['TEST_THUMB'])
                            thumb_fp = stg.output_file_abspath(thumb_fn, up_id)
                            thumb_ext = os.path.splitext(CFG['TEST_THUMB'])[1][1:]
                            res = stg.copy_file(CFG['TEST_THUMB'], thumb_fp)
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue
                            res = add_file_output_manifest(stg, up_id, thumb_fn, CFG['FTC']['thumbnail'], thumb_ext, 'attachments')
                            if res['code'] != 0:
                                ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, inform_user=True, session_id=session_id, commit_errors=commit_errors)
                                continue
                        res = commit_changes_to_storage(CFG, stg, jbs, up_id, manif, lec_id, lec_hash, user_id, verb)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_DB_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        scode = res['scode']
                        info = res['info']
                        jobs = res.get('jobs',[1])
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs)
                        if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                    else:
                        res = transLectures_processing(CFG, stg, jbs, manif, up_id, user_id, lec_hash, verb, mailer, email)
                        if res['code'] != 0:
                            ingest_error(stg, manif, CFG, up_id, lec_id, lec_lang, user_id, CFG['STC']['ST_MP_ERROR'], res['code'], res['info'], mailer, email, debug, session_id=session_id, commit_errors=commit_errors)
                            continue
                        scode = res['scode']
                        info = res['info']
                        jobs = res.get('jobs',[1])
                        ingest_update(stg, up_id, lec_id, lec_lang, user_id, scode, info, jobs=jobs, eta=eta)
                        if scode in (CFG['STC']['ST_NEW_FINISHED'], CFG['STC']['ST_UP_FINISHED'], CFG['STC']['ST_DEL_FINISHED']):
                            if mailer != None: send_mail_finished(mailer, manif, CFG, up_id, email)
                        else:
                            if mailer != None: send_mail_processing(mailer, manif, CFG, up_id, email, eta=etas)
                            if eta != None and verb: log("Estimated time to complete: %s." % (eta))        
                        
        except:
            bug("Upload ID: %s" % up['id'])
            continue


#------------- Main Function ---------------#


def run():
    """ Main Function 

        Default configuration file: assuming this python script filename 
        but changing the extension .py -> .cfg

    """
    #-------- Initialize ---------#

    if opts.verb: log("----------------- INGEST INIT ------------------")
    # Check Configuration file
    if opts.debug: log("[INIT] Checking Configuration file...")
    cfg_file = DEFAULT_CFG_FP
    if opts.config_file != None: cfg_file = opts.config_file
    try:
        cfg_fd = open(cfg_file, 'r')
    except IOError:
        error_code = 5000
        error("Cannot open config file %s" % cfg_file, exit_code=error_code)
    # Parse Configuration File
    if opts.debug: log("[INIT] Parsing Configuration file...")
    CFG = parse_config(cfg_fd)
    cfg_fd.close()

    if opts.debug: log("[INIT] Creating Storage Object...")
    db_name = CFG['STO']['DB_NAME']
    if opts.db_name != None: 
        db_name = opts.db_name
        CFG['STO']['DB_NAME'] = db_name
    db_user = CFG['STO']['DB_USER']
    db_passwd = CFG['STO']['DB_PASS']
    db_host = CFG['STO']['DB_HOST']
    try:
        stg = Storage(Storage.OBJ_SPEECH, CFG['GEN']['HOSTNAME'], CFG['GEN']['TL_USER'], CFG['GEN']['TL_GROUP'], db_name, db_user, db_host, db_passwd)
    except:
        print_excpt()
        error_code = 5001
        error("Cannot create Storage Object.", exit_code=error_code)

    # Create Scheduler object
    if opts.debug: log("[INIT] Creating Scheduler Object...")
    try:
        jbs = Scheduler(CFG['SCH']['LOCALHOST'], CFG['SCH']['STATUS_CMD'], CFG['SCH']['SUBMIT_SCR'], CFG['SCH']['SUBMIT_OPTS'])
    except:
        print_excpt()
        error_code = 5002
        error("Cannot create Scheduler Object.", exit_code=error_code)

    # Create Mailer object
    if CFG['MAI']['ENABLED']:
        if CFG['MAI']['SRC_ADDRESS'] != "" and CFG['MAI']['SMTP_SERVER']  != "":
            if opts.debug: log("[INIT] Creating Mailer Object...")
            try:
                mailer = SpeechMailer(CFG['MAI']['SRC_ADDRESS'], CFG['MAI']['SMTP_SERVER'], CFG['MAI']['AAS'], CFG['MAI']['AAE'], CFG['MAI']['AAF'])
            except:
                print_excpt()
                warn("[INIT] Cannot create Mailer object.")
                mailer = None
    else:
        mailer = None
     
    # Get file types from DB
    res = stg.get_file_types()
    if res['code'] != 0:
        error(res['info'], exit_code=res['code'])
    CFG['FTC'] = res['data']

    # Get file formats from DB
    res = stg.get_file_formats()
    if res['code'] != 0:
        error(res['info'], exit_code=res['code'])
    FFT = res['data']

    # Create list of allowed pairs (file_type, file_format) for easy and fast MP checking
    # Also check whether file formats exist in DB.
    CFG['ATT'] = []
    def append_list_ATT(L, FFT, ft, lff):
        for ff in lff:
            if ff not in FFT:
                error_code = 5010
                error("Unrecognised format type '%s'. Please check configuration file ([file_formats] section), otherwise insert it into the DB table 'file_types'." % (ff), exit_code=error_code)
            t = (ft, ff)
            if t not in CFG['ATT']:
                L.append(t)
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['media'], CFG['FFT']['RVF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['media'], CFG['FFT']['AVF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['media'], CFG['FFT']['AAF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['slides'], CFG['FFT']['ASTF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['slides'], CFG['FFT']['ASVF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['docs'], CFG['FFT']['ADF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['captions'], CFG['FFT']['ACF'])
    append_list_ATT(CFG['ATT'], FFT, CFG['FTC']['thumbnail'], CFG['FFT']['ATF'])

    # Get Systems from DB
    res = stg.get_systems()
    if res['code'] != 0:
        error(res['info'], exit_code=res['code'])
    sys_list = res['data']

    # Load ASR Modules
    if opts.debug: log("[INIT] Loading ASR Modules...")
    CFG['ASR'] = {}
    CFG['ASR']['_ids'] = {}
    for s in sys_list:
        if not(s['enabled']): continue
        if s['type'] == 'asr':
            l = s['lang1']
            module_fp = "%s/%s" %(stg.systems_mp, s['file_path'])
            module_id = "asr_%s__sysid_%d" % (l, s['id'])
            try:
                mod = load_source(module_id, module_fp)
            except:
                print_excpt()
                error_code = 5003
                error("Cannot load '%s' ASR module at %s" % (l, module_fp), exit_code=error_code)
            # Check if module implements transcribe_lecture() function
            if "transcribe_lecture" not in dir(mod):
                error_code = 5004
                error("transcribe_lecture() function not found in '%s' ASR module at %s" % (l, module_fp), exit_code=error_code)
            dsys = { 'id': s['id'], 'public': s['public'], 'allowed_users': s['user_list'], 'module': mod, 'features_tadapt':s['features_tadapt'], 'rtf':s['rtf'], 'const_time':s['const_time']}
            if l not in CFG['ASR'].keys(): CFG['ASR'][l] = { 'default':None, 'dsys':{} }
            CFG['ASR'][l]['dsys'][s['id']] = dsys
            CFG['ASR']['_ids'][s['id']] = l
            if s['as_default']: 
                if CFG['ASR'][l]['default'] == None:
                    CFG['ASR'][l]['default'] = s['id']
                else:
                    warn("[INIT] Already exists a default '%s' ASR System. Ignoring." % (l))
            if opts.debug: log("[INIT]  --> '%s' ASR module succesfully loaded" % l)
    # Check defaults
    for l in CFG['ASR'].keys():
        if l != '_ids':
            if CFG['ASR'][l]['default'] == None:
                CFG['ASR'][l]['default'] = CFG['ASR'][l]['dsys'].keys()[0]
                warn("[INIT] Auto-defining default '%s' ASR Module: ID %d" % (l, CFG['ASR'][l]['default']))

    # Load MT Modules
    if opts.debug: log("[INIT] Loading MT Modules...")
    CFG['MT'] = {}
    CFG['MT']['_ids'] = {}
    for s in sys_list:
        if s['type'] == 'mt':
            sl = s['lang1']
            tl = s['lang2']
            pair = "%s-%s" % (sl, tl)
            module_fp = "%s/%s" %(stg.systems_mp, s['file_path'])
            module_id = "mt_%s__sysid_%d" % (pair, s['id'])
            try:
                mod = load_source(module_id, module_fp)
            except:
                print_excpt()
                error_code = 5005
                error("Cannot load '%s' MT module at %s" % (pair, module_fp), exit_code=error_code)
            # Check if module implements translate() function
            if "translate" not in dir(mod):
                error_code = 5006
                error("translate() function not found in '%s' MT module at %s" % (pair, module_fp), exit_code=error_code)
            if "ID" not in dir(mod):
                error_code = 5006
                error("ID attribute not found in '%s' MT module at %s" % (pair, module_fp), exit_code=error_code)
            dsys = { 'id': s['id'], 'public': s['public'], 'allowed_users': s['user_list'], 'module': mod, 'features_tadapt':s['features_tadapt'], 'rtf':s['rtf'], 'const_time':s['const_time']}
            if pair not in CFG['MT'].keys(): CFG['MT'][pair] = { 'default':None, 'dsys':{} }
            CFG['MT'][pair]['dsys'][s['id']] = dsys
            CFG['MT']['_ids'][s['id']] = pair
            if s['as_default']: 
                if CFG['MT'][pair]['default'] == None:
                    CFG['MT'][pair]['default'] = s['id']
                else:
                    warn("[INIT] Already exists a default '%s' MT System. Ignoring." % (pair))
            if opts.debug: log("[INIT]  --> '%s' MT module succesfully loaded" % pair)
    # Check defaults
    for p in CFG['MT'].keys():
        if p != '_ids':
            if CFG['MT'][p]['default'] == None:
                CFG['MT'][p]['default'] = CFG['MT'][p]['dsys'].keys()[0]
                warn("[INIT] Auto-defining default '%s' MT Module: ID %d" % (p, CFG['MT'][p]['default']))

    # Load TTS Modules
    if opts.debug: log("[INIT] Loading TTS Modules...")
    CFG['TTS'] = {}
    CFG['TTS']['_ids'] = {}
    for s in sys_list:
        if not(s['enabled']): continue
        if s['type'] == 'tts':
            l = s['lang1']
            module_fp = "%s/%s" %(stg.systems_mp, s['file_path'])
            module_id = "tts_%s__sysid_%d" % (l, s['id'])
            try:
                mod = load_source(module_id, module_fp)
            except:
                print_excpt()
                error_code = 5003
                error("Cannot load '%s' TTS module at %s" % (l, module_fp), exit_code=error_code)
            # Check if module implements transcribe_lecture() function
            if "synthesize_track" not in dir(mod):
                error_code = 5004
                error("synthesize_track() function not found in '%s' TTS module at %s" % (l, module_fp), exit_code=error_code)
            dsys = { 'id': s['id'], 'public': s['public'], 'allowed_users': s['user_list'], 'module': mod, 'voice_gender': s['tts_voice_gender'], 'features_tadapt':s['features_tadapt'], 'rtf':s['rtf'], 'const_time':s['const_time'] }
            if l not in CFG['TTS'].keys(): CFG['TTS'][l] = { 'default':None, 'dsys':{} }
            CFG['TTS'][l]['dsys'][s['id']] = dsys
            CFG['TTS']['_ids'][s['id']] = l
            if s['as_default']: 
                if CFG['TTS'][l]['default'] == None:
                    CFG['TTS'][l]['default'] = s['id']
                else:
                    warn("[INIT] Already exists a default '%s' TTS System. Ignoring." % (l))
            if opts.debug: log("[INIT]  --> '%s' TTS module succesfully loaded" % l)
    # Check defaults
    for l in CFG['TTS'].keys():
        if l != '_ids':
            if CFG['TTS'][l]['default'] == None:
                CFG['TTS'][l]['default'] = CFG['TTS'][l]['dsys'].keys()[0]
                warn("[INIT] Auto-defining default '%s' TTS Module: ID %d" % (l, CFG['TTS'][l]['default']))


    # Load Text Retrieval Module
    if CFG['TXT'] != None:
        if opts.debug: log("[INIT] Loading Text Retrieval Module...")
        try:
            mod = load_source('text_retrieval', CFG['TXT'])
        except:
            print_excpt()
            error_code = 5007
            error("Cannot load 'text_retrieval' module at %s" % CFG['TXT'], exit_code=error_code)
        else:
            # Check if module implements translate_lecture() function
            if "video_slides2text" not in dir(mod):
                error_code = 5008
                error("'video_slides2text()' function not found in 'text_retrieval' module at %s" % CFG['TXT'], exit_code=error_code)
            if "jpg_slides2text" not in dir(mod):
                error_code = 5008
                error("'jpg_slides2text()' function not found in 'text_retrieval' module at %s" % CFG['TXT'], exit_code=error_code)
            if "retrieve_external_resources" not in dir(mod):
                error_code = 5009
                error("'retrieve_external_resources()' function not found in 'text_retrieval' module at %s" % CFG['TXT'], exit_code=error_code)
            CFG['TXT'] = mod
            if opts.debug: log("[INIT]  --> 'text_retrieval' module succesfully loaded")

    # Load URL Decoders Modules
    if opts.debug: log("[INIT] Loading URL Decoders...")
    for domain in CFG['URLDC'].keys():
        try:
            mod = load_source(domain, CFG['URLDC'][domain])
        except:
            print_excpt()
            error_code = 5010
            error("Cannot load '%s' URL Decoder at %s" % (domain, CFG['URLDC'][domain]), exit_code=error_code)
        else:
            # Check if module implements decode() function
            if "decode" not in dir(mod):
                error_code = 5011
                error("decoder() function not found in '%s' URL Decoder at %s" % (domain, CFG['URLDC'][domain]), exit_code=error_code)
            CFG['URLDC'][domain] = mod
            if opts.debug: log("[INIT]  --> '%s' URL Decoder succesfully loaded" % domain)

    CFG['URLDW'] = UrlDownloader()
    CFG['URLDW'].url_handlers = CFG['URLDC']

    # Get Uploads
    if opts.debug: log("[INIT] Retrieving uploads list...")
    res = stg.get_uploads()
    if res['code'] != 0:
        error(res['info'], exit_code=res['code'])
    uploads = res['data']

    myuploads = []
    for up in uploads:
        if opts.debug and not(up['debug']):
            #log("[Debug Mode] Skipping regular upload ID '%s'" % up['id'])
            pass
        elif not(opts.debug) and up['debug']:
            #log("Skipping debug upload ID '%s'" % up['id'])
            pass
        else:
            myuploads.append(up)

    #-------- Main Stuff ---------#

    # Remove finished uploads
    if opts.verb: log("----------- Checking failed uploads ------------")
    process_failed_uploads(CFG, stg, myuploads, opts.verb, opts.debug, opts.errors)

    # Remove finished uploads
    if opts.verb: log("---------- Checking finished uploads -----------")
    process_finished_uploads(CFG, stg, myuploads, opts.verb, opts.debug, opts.errors)

    # Check running uploads
    if opts.verb: log("----------- Checking running uploads -----------")
    process_running_uploads(CFG, stg, jbs, myuploads, opts.verb, mailer, opts.debug, opts.errors)

    # Check new uploads
    if opts.verb: log("------------ Checking new uploads --------------")
    process_new_uploads(CFG, stg, jbs, myuploads, opts.verb, mailer, opts.debug, opts.errors)

    if opts.verb: log("------------------- EXIT 0 ---------------------")



if __name__ == "__main__":

    DEFAULT_CFG_FP = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))
    DEFAULT_CFG_FN = "config.ini"
    SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

    # Define Option Parser
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("-v", "--verbose", dest="verb", action="store_true", default=False, help="Verbose power on!")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="Debug mode")
    parser.add_option("-e", "--disable-errors", dest="errors", action="store_false", default=True, help="Disable ingest errors")
    parser.add_option("-C", "--config-file", dest="config_file", type="string", default=None, help="Configuration file. Default: %s" % DEFAULT_CFG_FN)
    parser.add_option("-D", "--database", dest="db_name", type="string", default=None, help="Database with which to work. Default: specified in config file")
    (opts, args) = parser.parse_args()

    try:
        run()
    except SystemExit:
        raise
    except:
        bug("Error on main function", exit=True)




