#!/bin/bash -e 

set -u

export LC_ALL="C.UTF-8"
CURRDIR=$PWD; SCRIPTDIR=$(cd ${0%/*} &> /dev/null; echo $PWD; cd $CURRDIR &> /dev/null)

# log
function log {
  [ $VERBOSE -eq 0 ] || echo -e $1 >&2
}

function error {
  echo -e "[EE] $1" >&2
}

PRGNAME=$(basename $0)
NARGS=2

# Show help
function print_help {
    TO=$1
    SHOWHELP=$2
    echo "$PRGNAME [options] <IN_DFXP_LIST> <OUT_TXT_LIST>" > $TO
    echo "" > $TO
    [ $SHOWHELP -eq 0 ] && return
    cat <<EOF > $TO

  Options:

    -h: Shows this help
    -v: Verbose power on!
    -r: Remove tmp directory at the end. 
    -f: File mode. All lists are considered as a single file.
    -t <str>: Specify temporal directory. By default it is created in /tmp.
    -S <str>: Scripts directory. Default: dirname of this script.
 
EOF
}

SCRIPTS="$SCRIPTDIR"
VERBOSE=0
TMP="NULL"
RMTMP=0
LIST=1

ARGS=`getopt -o "hvrt:S:fe:T:o:s" -n $PRGNAME -- "$@"`
eval set -- "$ARGS"
while true
do
    case $1 in
	-h) print_help /dev/stdout 1; exit 0;;
        -v) VERBOSE=1;;
        -t) TMP="$2";;
        -r) RMTMP=1;;
        -f) LIST=0;;
        -S) SCRIPTS="$(readlink -f $2)";;
	--) shift; break;;
    esac
    shift
done
[ $# != $NARGS ] && { print_help /dev/stderr 0; error "wrong number of arguments"; exit 1; }

INLIST="$(readlink -f $1)"
OUTLIST="$(readlink -f $2)"

export PATH=$SCRIPTS:$PATH

set -e

if [ "$TMP" = "NULL" ]; then 
  TMP=${TMPDIR:-/tmp}/$(basename $0 .sh).$$; mkdir $TMP 2>/dev/null;
  #trap "rm -rf $TMP 2>/dev/null" EXIT
else
  mkdir -p "$TMP"
fi

if [ $LIST -eq 1 ]; then
  [ $(cat "$INLIST" | wc -l) -ne $(cat "$OUTLIST" | wc -l) ] && error "Input and Output lists have different number of lines." && exit 1;
  [ $(cat "$INLIST" | wc -l) -eq 0 ] && error "Empty input list." && exit 1;
else
  echo "$INLIST" > "$TMP/in.lst"
  INLIST="$TMP/in.lst"
  echo "$OUTLIST" > "$TMP/out.lst"
  OUTLIST="$TMP/out.lst"
fi

set -x

mkdir -p "$TMP"

id=0
paste -d '|' $INLIST $OUTLIST | while IFS='|' read f trgf; do
  id=$[id+1]
  [ ! -s "$f" ] && error "File $f does not exist or is empty." && exit 1
  "$SCRIPTS/dfxp2txt.py" "$f" | awk '{if(NF==1) print $1,"__EMPTY__"; else print}' > "$TMP/$id.tlftxt"
  if [ $(cat "$TMP/$id.tlftxt" | wc -l) -eq 0 ]; then
    echo "[WW] No input sentences. Skipping." 1>&2
    #cp "$f" "$trgf"
    #continue
  fi
  awk '{print $1}' "$TMP/$id.tlftxt" > "$TMP/$id.segids"
  awk '{for(i=2;i<=NF;i++) printf("%s ",$i); printf("\n");}' "$TMP/$id.tlftxt" > "$TMP/$id.ftxt"
  cp "$TMP/$id.ftxt" $trgf
done

[ $RMTMP -eq 1 ] && rm -rf "$TMP"

set +x 

cd $CURRDIR

exit 0
