#!/usr/bin/python

import argparse
from libdfxp import *
import sys
from datetime import datetime

def processDfxp(pdfxp, tl, authorType = "automatic", authorId = "UPV", authorConf = 0.5, confMeasure = 0.5):
  timeStamp = datetime.now()
  pdfxp.timeStamp = timeStamp
  pdfxp.authorType = authorType
  pdfxp.authorId = authorId
  pdfxp.authorConf = authorConf
  pdfxp.confMeasure = confMeasure
  pdfxp.status = "fully_automatic"
  for i in xrange(len(pdfxp.data)):
    d1 = pdfxp.data[i]
    if isinstance(d1,TlSegment):
      processTlSegment(d1, tl)
    elif isinstance(d1,TlAlt):
      processTlAlt(d1, tl)
  pdfxp.oridata = pdfxp.data[:]
  

def processTlSegment(s, tl):
  try:
    s.data = [TlRawText(tl[int(s.segmentId)])]
  except KeyError:
    s.data = [TlRawText("__EMPTY__")]
  s.authorType = None
  s.authorId = None
  s.authorConf = None
  s.confMeasure = None
  s.timeStamp = None

def processTlAlt(a, tl):
  for d in a.data:
    if isinstance(d,TlSegment):
      processTlSegment(d, tl)

def read_txt(fn):
  fd = open(fn)
  D = {}
  for l in fd:
    segId = int(l.split('-#',1)[0].replace("#-","").replace("-#",""))
    try:
      txt = l.split('-#',1)[1].strip()
      D[segId] = txt.decode("UTF-8")
    except:
      txt = "__EMPTY__"
      D[segId] = txt
  fd.close()
  return D

parser = argparse.ArgumentParser()
parser.add_argument("tltxt", help="TLTXT file")
parser.add_argument("dfxp", help="DFXP file")
parser.add_argument("-t", "--trg-lang", default=None, help="Translation language")
parser.add_argument("-s", "--mt-system-label", help="MT system label", default="MLLP")

args = parser.parse_args()

txt_fn = args.tltxt
dfxp_fn = args.dfxp

translations = read_txt(txt_fn)
pdfxp = parseDfxp(dfxp_fn)
pdfxp.data = filterDfxp(pdfxp)
processDfxp(pdfxp,translations, authorId=args.mt_system_label)
pdfxp.lang = args.trg_lang
print pdfxp.toXml()

