#!/bin/bash -e 

set -u

export LC_ALL="C.UTF-8"
CURRDIR=$PWD; SCRIPTDIR=$(cd ${0%/*} &> /dev/null; echo $PWD; cd $CURRDIR &> /dev/null)

# log
function log {
  [ $VERBOSE -eq 0 ] || echo -e $1 >&2
}

function error {
  echo -e "[EE] $1" >&2
}

PRGNAME=$(basename $0)
NARGS=4

# Show help
function print_help {
    TO=$1
    SHOWHELP=$2
    echo "$PRGNAME [options] <DFXP2TXT_TMP_DIR> <IN_DFXP_LIST> <IN_TXT_LIST> <OUT_DFXP_LIST>" > $TO
    echo "" > $TO
    [ $SHOWHELP -eq 0 ] && return
    cat <<EOF > $TO

  Options:

    -h: Shows this help
    -v: Verbose power on!
    -r: Remove tmp directory at the end. 
    -f: File mode. All lists are considered as a single file.
    -t <str>: Specify temporal directory. By default it is created in /tmp.
    -S <str>: Scripts directory. Default: dirname of this script.
    -T <str>: Set target language (by default is read from DFXP files)
    -s <str>: MT system label
 
EOF
}

SCRIPTS="$SCRIPTDIR"
VERBOSE=0
TMP="NULL"
RMTMP=0
LIST=1
TRG="NULL"
SYS="NULL"

ARGS=`getopt -o "hvrt:S:fT:s:" -n $PRGNAME -- "$@"`
eval set -- "$ARGS"
while true
do
    case $1 in
	-h) print_help /dev/stdout 1; exit 0;;
        -v) VERBOSE=1;;
        -t) TMP="$2";;
        -r) RMTMP=1;;
        -f) LIST=0;;
        -S) SCRIPTS="$(readlink -f $2)";;
        -T) TRG="$2";; 
        -s) SYS="$2";; 
	--) shift; break;;
    esac
    shift
done
[ $# != $NARGS ] && { print_help /dev/stderr 0; error "wrong number of arguments"; exit 1; }
TMP_IN="$(readlink -f $1)"
IN_DFXP_LIST="$(readlink -f $2)"
IN_TXT_LIST="$(readlink -f $3)"
OUT_DFXP_LIST="$(readlink -f $4)"

export PATH=$SCRIPTS:$PATH

set -e

if [ "$TMP" = "NULL" ]; then 
  TMP=${TMPDIR:-/tmp}/$(basename $0 .sh).$$; mkdir $TMP 2>/dev/null;
  #trap "rm -rf $TMP 2>/dev/null" EXIT
else
  mkdir -p "$TMP"
fi

if [ $LIST -eq 1 ]; then
  [ $(cat "$IN_DFXP_LIST" | wc -l) -ne $(cat "$IN_TXT_LIST" | wc -l) ] && error "Input and Output lists have different number of lines." && exit 1;
  [ $(cat "$IN_TXT_LIST" | wc -l) -ne $(cat "$OUT_DFXP_LIST" | wc -l) ] && error "Input and Output lists have different number of lines." && exit 1;
  [ $(cat "$IN_TXT_LIST" | wc -l) -eq 0 ] && error "Empty input list." && exit 1;
else
  echo "$IN_DFXP_LIST" > "$TMP/in_dfxp.lst"
  IN_DFXP_LIST="$TMP/in_dfxp.lst"
  echo "$IN_TXT_LIST" > "$TMP/in_txt.lst"
  IN_TXT_LIST="$TMP/in_txt.lst"
  echo "$OUT_DFXP_LIST" > "$TMP/out_dfxp.lst"
  OUT_DFXP_LIST="$TMP/out_dfxp.lst"
fi

set -x

gopts=""
[ "$TRG" != "NULL" ] && gopts="$gopts -t $TRG"
[ "$SYS" != "NULL" ] && gopts="$gopts -s $SYS"
id=0
paste -d '|' $IN_DFXP_LIST $IN_TXT_LIST $OUT_DFXP_LIST | while IFS='|' read in_dfxp in_txt out_dfxp; do
  id=$[id+1]
  nL=$(cat "$TMP_IN/$id.tlftxt" | wc -l)
  [ $nL -eq 0 ] && continue
  paste "$TMP_IN/$id.segids" "$in_txt" > "$TMP/$id.tltxt"
  "$SCRIPTS/txt2dfxp.py" $gopts "$TMP/$id.tltxt" "$in_dfxp" > $TMP/$id.dfxp
  cp $TMP/$id.dfxp "$out_dfxp"
done

[ $RMTMP -eq 1 ] && rm -rf "$TMP"

set +x 

cd $CURRDIR

exit 0
