#!/usr/bin/python

import argparse
from libdtlx import *
import sys

def filter_text(text):
    return text.replace('\n', ' ').replace('\r', '').replace('\t', '')

def print_tltxt(doc):
    for s in doc.sentences:
        if len(s.translations) == 0:
            print '###-#%s#-### %s' % (s.sentId.encode('utf-8'), filter_text(s.origin.formattedText).encode("utf-8"))
        elif s.filteredTranslation.authorType != Document.AUTHOR_TYPE_MANUAL:
            print '###-#%s#-### %s' % (s.sentId.encode('utf-8'), filter_text(s.origin.formattedText).encode("utf-8"))

parser = argparse.ArgumentParser()
parser.add_argument("src_lang", help="Source language")
parser.add_argument("dtlx", help="Input dtlx file")
parser.add_argument("-e", "--existing-dtlx", dest="ex_dtlx", help="Provide an existing dtlx to preserve supervised translations")
parser.add_argument("-t", "--trg-lang", help="Overwrite target language")
parser.add_argument("-o", "--trg-dtlx", help="Write target DTLX file")
parser.add_argument("-s", "--allow-translation-as-source", dest="swap", default=False, action="store_true", help="Allow to use best translation of input DTLX as source text, if needed.")

args = parser.parse_args()

parser = Parser()
doc = parser.parse(args.dtlx)
if doc.srcLang != args.src_lang:
    if (doc.trgLang != args.src_lang) or (doc.trgLang == args.src_lang and not(args.swap)):
        sys.stderr.write("Error: Input language is %s, but provided DTLX file is a %s-%s translation\n" % (args.src_lang, doc.srcLang, doc.trgLang))
        sys.exit(1)
    else:
        sys.stderr.write("Warning: Input language is %s, but provided DTLX file is a %s-%s translation. Using %s translation as source.\n" % (args.src_lang, doc.srcLang, doc.trgLang, doc.trgLang))
        doc.translations_to_source()

if args.trg_lang != None:
    doc.trgLang = args.trg_lang

if args.ex_dtlx != None:
    ex_doc = parser.parse(args.ex_dtlx)
    if ex_doc.srcLang != doc.srcLang or ex_doc.trgLang != doc.trgLang:
        sys.stderr.write("Error: Input DTLX %s-%s, Existing DTLX %s-%s\n" % ( doc.srcLang, doc.trgLang, ex_doc.srcLang, ex_doc.trgLang))
        sys.exit(1)
    doc.copy_translations_given_origin(ex_doc, allowed_author_types=[Document.AUTHOR_TYPE_MANUAL])

doc.filter(Document.FILTER_GET_BEST)
print_tltxt(doc)

if args.trg_dtlx != None:
    with open(args.trg_dtlx, 'w') as fd:
        fd.write(doc.toXML())

