# -*- coding: utf-8 -*-
#
#  Copyright 2013, 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
from libdtlx import *

DIR = "%s" % os.path.dirname(os.path.abspath(__file__))
DTLX2TXT = "%s/dtlx2txt.sh" % DIR
TXT2DTLX = "%s/txt2dtlx.sh" % DIR


def dtlx2txt(jbs, src_lang, trg_lang, in_dtlx_lst, out_txt_lst, out_dtlx_lst, tmp_dir, ex_dtlx_lst=None, log_file=None, job_name=None, hold_job_ids=None):
    cmd = "%s -v -t %s -T %s -o %s -s %s %s %s" % (DTLX2TXT, tmp_dir, trg_lang, out_dtlx_lst, src_lang, in_dtlx_lst, out_txt_lst)
    if ex_dtlx_lst != None:
        cmd += " -e %s" % (ex_dtlx_lst)
    if log_file == None:
        log_file = "%s/dtlx2txt.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=2.0, time="00:30:00", output=log_file, job_name=job_name, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid], 'cmd':cmd, 'log':log_file}

def txt2dtlx(jbs, trg_lang, dtlx2txt_tmp_dir, in_txt_lst, dtlx_lst, tmp_dir, log_file=None, job_name=None, hold_job_ids=None):
    my_tmp_dir = "%s/txt2dtlx" % tmp_dir
    cmd = "%s -v -t %s -T %s %s %s %s %s" % (TXT2DTLX, tmp_dir, trg_lang, dtlx2txt_tmp_dir, dtlx_lst, in_txt_lst, dtlx_lst)
    if log_file == None:
        log_file = "%s/txt2dtlx.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=2.0, time="00:30:00", output=log_file, job_name=job_name, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid], 'cmd':cmd, 'log':log_file}

def create_empty_dtlx(in_dtlx_fp, out_dtlx_fp, obj_id, slang, tlang):
    parser = Parser()
    doc = parser.parse(in_dtlx_fp)
    doc.empty_translations()
    doc.docId = obj_id
    doc.srcLang = slang
    doc.trgLang = tlang
    data = doc.toXML()
    with open(out_dtlx_fp, 'w') as fd:
        fd.write(data)
    return {'code':0}

def compute_dtlx_src_stats(dtlx_fp):
    parser = Parser()
    doc = parser.parse(dtlx_fp)
    nsents = len(doc.sentences)
    nwords = 0
    nchars = 0
    for s in doc.sentences:
        nwords += len(s.origin.rawText.strip().split())
        nchars += len(filter(unicode.isalnum, s.origin.rawText))
    d = { 'sentences':nsents, 'words':nwords, 'characters':nchars }
    return {'code':0, 'data':d}

def compute_dtlx_tl_stats(dtlx_fp):
    parser = Parser()
    doc = parser.parse(dtlx_fp)
    doc.updateStatus()
    status = doc.status
    doc.filter(Document.FILTER_GET_LAST)
    nsents = len(doc.sentences)
    nwords = 0
    nchars = 0
    sup = 0
    if status != Document.STATUS_NOT_TRANSLATED:
        for s in doc.sentences:
            if s.filteredTranslation.authorType == Document.AUTHOR_TYPE_MANUAL:
                sup += 1
            nwords += len(s.filteredTranslation.rawText.strip().split())
            nchars += len(filter(unicode.isalnum, s.filteredTranslation.rawText))
    try:
        sup = sup/float(nsents)
    except:
        sup = 0
    d = { 'status':status, 'sup_ratio':sup, 'sentences':nsents, 'words':nwords, 'characters':nchars }
    return {'code':0, 'data':d}



