#!/usr/bin/python
import argparse
from libdtlx import *
import bs4
import sys


def read_tltxt(fn):
    fd = open(fn)
    D = {}
    for l in fd:
        segId = l.split('#-###',1)[0].replace("###-#","").replace("#-###","")
        txt = l.split('#-###',1)[1].strip()
        D[segId] = txt.decode("UTF-8")
    fd.close()
    return D

def add_translations(doc, translations, authorType="automatic", authorId="UPV", authorConf=0.5, confMeasure=0.5):
    doc.authorType = authorType
    doc.authorId = authorId
    doc.authorConf = authorConf
    doc.confMeasure = confMeasure
    for s in doc.sentences:
        if s.sentId in translations:
            txt = translations[s.sentId]
            t = Translation()
            t.formattedText = txt
            t.rawText = bs4.BeautifulSoup(txt).get_text().strip()
            t.authorType = authorType
            t.authorId = authorId
            t.authorConf = authorConf
            t.confMeasure = confMeasure
            s.translations.append(t)

parser = argparse.ArgumentParser()
parser.add_argument("tltxt", help="TLTXT file")
parser.add_argument("dtlx", help="DTLX file")
parser.add_argument("-t", "--trg-lang", help="Overwrite target language")
parser.add_argument("-s", "--mt-system-label", help="MT system label", default="MLLP")

args = parser.parse_args()

tltxt_fn = args.tltxt
dtlx_fn = args.dtlx

translations = read_tltxt(tltxt_fn)
parser = Parser()
doc = parser.parse(dtlx_fn)
add_translations(doc, translations, authorId=args.mt_system_label)
if args.trg_lang != None:
    doc.trgLang = args.trg_lang
doc.updateStatus()
print doc.toXML()
