# -*- coding: utf-8 -*-
#
#  Copyright 2013, 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os

def download_media(jbs, media_url, dest_file, tmp_dir, job_name=None, hold_job_ids=None):
    dest_dir = os.path.dirname(dest_file)
    dest_fn = os.path.basename(dest_file)
    cmd = "aria2c -k 5M -j 16 -x 16 -s 16 '%s' -d %s -o %s" % (media_url, dest_dir, dest_fn)
    log_file = "%s/download_media.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="00:30:00", output=log_file, job_name=job_name, local=False, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}

def media2mp4(jbs, src_file, trg_file, tmp_dir, job_name=None, hold_job_ids=None):
    cmd = "ffmpeg -y -i '%s' -c:a libfdk_aac -b:a 128k -c:v libx264 -pix_fmt yuv420p -preset medium '%s'" % (src_file, trg_file)
    log_file = "%s/media2mp4.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="24:00:00", output=log_file, job_name=job_name, local=True, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}


def media2ogv(jbs, src_file, trg_file, tmp_dir, job_name=None, hold_job_ids=None):
    cmd = "ffmpeg -i '%s' -acodec libvorbis -qscale:a 6 -vcodec libtheora -qscale:v 5 -y '%s'" % (src_file, trg_file)
    log_file = "%s/media2ogv.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="24:00:00", output=log_file, job_name=job_name, local=True, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}


def media2pcm(jbs, src_file, trg_file, tmp_dir, job_name=None, hold_job_ids=None):
    cmd = "ffmpeg -i '%s' -ac 1 -filter:a aresample=500 -map 0:a -c:a pcm_u8 -f data -y '%s'" % (src_file, trg_file)
    log_file = "%s/media2pcm.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="24:00:00", output=log_file, job_name=job_name, local=True, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}


def audio2mp4(jbs, audio_file, video_file, bg_image=None, tmp_dir=None, job_name=None, hold_job_ids=None):
    cmd = ""
    if bg_image == None:
        bg_image = "%s/black.png" % tmp_dir
        cmd = "convert -size 1280x720 xc:black '%s' &&" % bg_image
    cmd = """%s ffmpeg -y -loop 1 -i '%s' -i '%s' -c:a libfdk_aac -b:a 128k -shortest -c:v libx264 -pix_fmt yuv420p -preset medium -tune stillimage -vf scale=-1:720 -f mp4 -y '%s'"""\
             %(cmd, bg_image, audio_file, video_file)
    log_file = "%s/audio2mp4.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="24:00:00", output=log_file, job_name=job_name, local=True, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}


def make_thumbnail(jbs, video_file, thumb_file, tmp_dir, job_name=None, hold_job_ids=None, default_thumbnail=None):
    cmd = "ffmpeg -i '%s' -ss 5 -vframes 1 -an -f image2 -y '%s'" %(video_file, thumb_file)
    if default_thumbnail != None:
        cmd += " || cp '%s' '%s'" % (default_thumbnail, thumb_file)
    log_file = "%s/mkthumb.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=3.0, time="5:00:00", output=log_file, job_name=job_name, local=True, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid]}


def get_media_length(jbs, media_file, job_name=None, hold_job_ids=None):
    cmd = "ffprobe -show_format '%s' 2>&1 | grep 'duration=' | cut -d'=' -f2" % media_file
    res = jbs.execute_job_output(cmd)
    if res['code'] != 0:
        return res
    try:
        length = int(float(res['data'].strip()))
    except:
        return {'code':40000, 'info':'Cannot compute media file audio length (are you sure it is a media file?)'}
    return {'code':0, 'data':length}


def check_media_audio(jbs, media_file, job_name=None, hold_job_ids=None):
    cmd = "ffprobe -i '%s' -show_streams -select_streams a -loglevel error | wc -l" % media_file
    res = jbs.execute_job_output(cmd)
    if res['code'] != 0:
        return {'code':40001, 'info':'Invalid media file (are you sure it is a media file?)'}
    if int(res['data'].strip()) == 0:
        return {'code':40002, 'info':'Media file has no audio track'}
    return {'code':0, 'info':''}
