import urllib2
import urlparse
import mimetypes
import tldextract
import os

class UrlDownloader(object):
    
    CHUNK_SIZE = 10485760 # 10 MB

    def __init__(self):
        self.url_handlers = {}

    def __open_url(self, url):
        try:
            response = urllib2.urlopen(url) 
        except ValueError:
            return {'code':1050, 'info':'Invalid URL format (%s).' % url}
        except urllib2.URLError:
            return {'code':1051, 'info':'Unexisting URL or URL not publicy accessible (%s).' % url}
        except urllib2.HTTPError:
            return {'code':1052, 'info':'URL not accessible (%s).' % url}
        except:
            return {'code':1053, 'info':'Could not download URL (%s).' % url}
        return {'code':0, 'data':response }
    
    
    def download_url(self, url, dest_prefix, def_ext=None):
        res = self.__open_url(url)
        if res['code'] != 0:
            return res
        response = res['data']
        real_url = response.geturl()
        parsed_url = urlparse.urlparse(real_url)
        content_type = response.info().get('content-type', None)
        ext = None
        decoded = False
        # Look for url_handlers
        try:
            website = tldextract.extract(parsed_url.netloc).domain
        except:
            website = parsed_url.netloc
        if website in self.url_handlers.keys():
            res = self.url_handlers[website].decode(real_url)
            if res['code'] != 0:
                return res
            decoded_url, ext = res['data']
            decoded = True
            res = self.__open_url(decoded_url)
            if res['code'] != 0:
                return res
            response = res['data']
            real_url = response.geturl()
            parsed_url = urlparse.urlparse(real_url)
            content_type = response.info().get('content-type', None)
        # Guess extension
        if ext == None:
            if content_type != None:
                ext = mimetypes.guess_extension(content_type)
            else:
                path = parsed_url.path
                ext = os.path.splitext(path)[1]
                if ext == '': ext = None
        if ext != None:
           ext = ext.replace('.','')
        else:
            if def_ext != None:
                ext = def_ext
            else:
                return {'code':1054, 'info':'Could not guess file extension from URL (%s).' % real_url}
        # Download File
        dest_fp = "%s.%s" % (dest_prefix, ext)
        try:
            with open(dest_fp, 'wb') as fd:
                chunk = response.read(self.CHUNK_SIZE)
                while chunk != '':
                    fd.write(chunk)
                    chunk = response.read(self.CHUNK_SIZE)
        except:
            return {'code':1055, 'info':'Could not download URL (%s).' % real_url}
        return {'code':0, 'data':(ext, website, decoded)}
