import traceback
from lib.html2dtlx import html2dtlx_str
from lib.dtlx2html import regen_html_from_dtlx_files, clean_html


def parse_document(in_doc_fp, doc_id, src_lang, out_html_fp, out_dtlx_fp):
 
    try:
        with open(in_doc_fp, 'r') as fd:
            html = fd.read()
    except:
        return {'code':7000, 'info':'Cannot read input original document'}

    try:
        phtml, dtlx = html2dtlx_str(html, doc_id, src_lang, "xx")  
    except:
        traceback.print_exc()
        return {'code':7001, 'info':'Could not parse input original document'}
    
    try:
        with open(out_html_fp, 'w') as fd:
            fd.write(phtml)
    except:
        return {'code':7002, 'info':'Could not save parsed html document'}

    try:
        with open(out_dtlx_fp, 'w') as fd:
            fd.write(dtlx)
    except:
        return {'code':7003, 'info':'Could not save internal dtlx document'}

    return {'code':0}


def regenerate_document(in_dtlx_fp, in_phtml_fp, out_phtml_fp, out_html_fp):  
    try:
        phtml = regen_html_from_dtlx_files(in_dtlx_fp, in_phtml_fp)
    except:
        traceback.print_exc()
        return {'code':7004, 'info':'Could not regenerate parsed translated html document'}

    try:
        with open(out_phtml_fp, 'w') as fd:
            fd.write(phtml)
    except:
        return {'code':7005, 'info':'Could not save parsed translated html document'}

    try:
        cphtml = clean_html(phtml)
    except:
        traceback.print_exc()
        return {'code':7006, 'info':'Could regenerate original translated html document'}

    try:
        with open(out_html_fp, 'w') as fd:
            fd.write(cphtml)
    except:
        return {'code':7007, 'info':'Could not save original translated html document'}

    return {'code':0}
