#!/bin/bash 
#
#  Copyright 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

set -u
export LC_ALL=C

yesno() {
  echo -n " - $1 (y/n)? "
  while read -r answer; do
    if [[ $answer = [YyNn] ]]; then
      [[ $answer = [Yy] ]] && val=1
      [[ $answer = [Nn] ]] && val=0
      break
    else
      echo -en "   Please type Y (Yes) or N (No): "
    fi
  done
}

getvar() {
  # $1-> Question; $2-> Default value; $3-> varname
  var=""
  while [ -z $var ]; do
    read -e -p " - $1: " -i "$2" var
  done
  eval "${3}=\"$var\""
}

getpasswd() {
  # $1-> Question; $2-> varname
  var="0"; var2="1";
  while [ "$var" != "$var2" ]; do
    var=""
    while [ -z $var ]; do
      read -e -s -p " - $1: " var
      echo ""
    done
    var2=""
    while [ -z $var2 ]; do
      read -e -s -p "   Retype $1: " var2
      echo ""
    done
    [ "$var" != "$var2" ] && prompt "   ** Non-matching keys. Please try again. **"
  done
  eval "${2}=\"$var\""
}

prompt() {
    echo -e "$1" >&2
}

log() {
    echo -e "[II] $1" >&2
}

warn() {
    echo -e "[WW] $1" >&2
}

error() {
    echo -e "\n[EE] $1\n" >&2
    exit 1
}

cexit() {
    echo -e "\n[II] $1\n" >&2
    exit 0
}

#trap_error() {
#   error "Something (probably unexpected) failed." 
#}
#
#trap trap_error ERR

save_config() {

  { cat << EOF
  TLUSER="$TLUSER"
  TLGROUP="$TLGROUP"
  TLP_INSTALL_DIR="$TLP_INSTALL_DIR"
  TLP_REPOSITORIES_DIR="$TLP_REPOSITORIES_DIR"
  TLPADMINPASSWD="$TLPADMINPASSWD"
  TLPADMINAUTHKEY="$TLPADMINAUTHKEY"
  DBNAME="$DBNAME"
  DBPASSWD="$DBPASSWD"
  WSLOCATION="$WSLOCATION"
  WSKEYGENERATOR="$WSKEYGENERATOR"
  CRONTAB="$CRONTAB"
EOF
  } > "$TMP/INPUT_OPTIONS.bash"
 

}

#--- PACKAGE DEPENDENCES ---#

APT_PKG_LST="postgresql postgresql-server-dev-all apache2 apache2-utils libapache2-mod-wsgi php5 libapache2-mod-php5 php5-curl zip python-pip python-dev"
PIP_PKG_LST="tldextract beautifulsoup4 nltk Paste psycopg2"

#--- DEFAULT VALUES ---#

VERS="3.4.0"
DIRS="ingest-service lib misc player transedit docedit webtleditor web-service"

CDIR="$(dirname $(readlink -f $0))"

HOSTNAME=$(hostname)
IP=$(hostname -I | awk '{print $1}')
CANONICAL_NAME=$(nslookup $HOSTNAME | grep -A 1 "canonical name" | grep "^Name:" | awk '{print $2}')
if [ "$CANONICAL_NAME" = "" ]; then
  CANONICAL_NAME="localhost"
fi

TLUSER="tluser"
TLGROUP="tlgroup"
APACHE_USER="www-data"

DBNAME="tldb"
DBPASSWD=""
DBHOST="$CANONICAL_NAME"
DBIP="$IP"

HTTP_ROOT="http://$CANONICAL_NAME"
WSLOCATION="api/v${VERS:0:1}"
WSKEYGENERATOR=""

TLPADMIN="admin"
TLPADMINPASSWD=""
TLPADMINAUTHKEY=""

TXELOCATION="transedit"
PLLOCATION="player"
TELOCATION="docedit"
WTELOCATION="webtleditor"
PLKEY=""
PLDATADIRNAME="data"

if [ "$UID" -ne 0 ]; then 
  error "Please run this script as root."
fi

prompt "
****************************************************************
*                                                              *
*     transLectures-UPV Platform (TLP) Installation Script     *
*                        Version $VERS                         *
*                                                              *
****************************************************************\n"


prompt "\nThis script will try to automatically install and configure the following TLP components:\n"
prompt "- Database"
prompt "- Web Service"
prompt "- Ingest Service"
prompt "- Transcription Editor, Media Player, Translation Editor, and Web Translation Editor"
prompt "- Postprocessor module"
prompt "\nWARNING: this script was conceived to be executed on a fresh installation of the Ubuntu 14.04 LTS Server Operating System." 
prompt "         If this is not your case, please avoid to use this script and refer to the manual installation instructions, or execute it at your own risk.\n"
yesno "Proceed to install the transLectures-UPV Platform"
if [ $val -eq 0 ]; then
  cexit "Aborting Installation."
fi

#TMP=${TMPDIR:-/tmp}/tlp-install.$$; mkdir $TMP 2>/dev/null;
TMP="$CDIR/.tlp-installation"; 
if [ -e "$TMP" ]; then
  date=$(cat "$TMP/date.txt")
  prompt ""
  prompt "WARNING: There exist a previous, unfinished installation process (was started at: $date)"
  yesno "Do you want to resume it?"
  if [ $val -eq 0 ]; then
    rm -rf "$TMP"
    mkdir "$TMP"
    last_step=-1
    date > "$TMP/date.txt"
    prompt "\n\n********* Starting TLP Installation and Configuration **********\n"
  else
    last_step=$(cat "$TMP/last_step.txt")
    source "$TMP/INPUT_OPTIONS.bash"
    prompt "\n\n********* Resuming TLP Installation and Configuration **********\n"
  fi
else
  mkdir "$TMP"
  last_step=-1
  date > "$TMP/date.txt"
  prompt "\n\n********* Starting TLP Installation and Configuration **********\n"
fi

if [ 0 -gt $last_step ]; then
  
  echo "-1" > "$TMP/last_step.txt"
  
  prompt "\n0) TLP Configuration\n"
  
  prompt "\n0.1) UNIX user and group for TLP.\n     Note: this script will create the specified user (along with its home directory) and group if they do not exist.\n"
  
  getvar "transLectures user (default: $TLUSER)" "$TLUSER" TLUSER
  getvar "transLectures group (default: $TLGROUP)" "$TLGROUP" TLGROUP
  
  TLP_INSTALL_DIR="/home/$TLUSER/tlp/$VERS/app"
  TLP_REPOSITORIES_DIR="/home/$TLUSER/tlp/$VERS/data"
  
  prompt "\n0.2) Installation and data directories.\n"
  
  getvar "TLP installation directory (default: $TLP_INSTALL_DIR)" "$TLP_INSTALL_DIR" TLP_INSTALL_DIR
  getvar "Repository root directory (default: $TLP_REPOSITORIES_DIR)" "$TLP_REPOSITORIES_DIR" TLP_REPOSITORIES_DIR
  
  prompt "\n0.3) TLP Users.\n"
  
  getpasswd "TLP admin user password" TLPADMINPASSWD
  
  prompt "\n0.4) Database Configuration.\n"
  
  getvar "Database name (default: $DBNAME)" "$DBNAME" DBNAME
  getpasswd "Database user '$TLUSER' password" DBPASSWD
  
  prompt "\n0.5) Web Service Configuration.\n"
   
  getvar "Web Service relative location (URL) (default: $WSLOCATION, i.e. '$HTTP_ROOT/$WSLOCATION')" "$WSLOCATION" WSLOCATION
  getvar "Secret Key Generator" "$WSKEYGENERATOR" WSKEYGENERATOR
  
  
  prompt "\n0.6) Other options.\n"
  yesno "Do you want to install a new crontab to periodically run the Ingest Service and the Postprocessor module?"
  CRONTAB=$val
 
  #prompt "\n0.7) Reconfiguring locales.\n"
  #dpkg-reconfigure locales
  
  save_config
 
  echo "0" > "$TMP/last_step.txt"

fi
  
if [ 1 -gt $last_step ]; then
  
  prompt "\n1) Check software dependencies\n"
  
  dpkg -l | awk '{print $2}' &> "$TMP/dpkg.lst"
  
  NF_PKG_LST=""
  for p in $APT_PKG_LST; do
    echo -ne " - Checking if $p is installed... "
    if [ -z $(grep "^$p$" "$TMP/dpkg.lst") ]; then
      echo "no"
      NF_PKG_LST="$NF_PKG_LST $p"
    else
      echo "yes"
    fi
  done

  pip list | awk '{print $1}' &> "$TMP/pip.lst"
  
  NF_PIP_LST=""
  for p in $PIP_PKG_LST; do
    echo -ne " - Checking if $p is installed... "
    if [ -z $(grep "^$p$" "$TMP/pip.lst") ]; then
      echo "no"
      NF_PIP_LST="$NF_PIP_LST $p"
    else
      echo "yes"
    fi
  done
 
  if [ ! -z "$NF_PKG_LST" ]; then
    prompt "\nThe following packages will be automatically downloaded and installed via apt-get:\n"
    prompt "$NF_PKG_LST\n"
    yesno "Do you agree"
    if [ $val -eq 0 ]; then
      error "Missing software dependencies. Aborting installation."
    fi
  fi

  if [ ! -z "$NF_PIP_LST" ]; then
    prompt "\nThe following python packages will be automatically downloaded and installed via python-pip:\n"
    prompt "$NF_PIP_LST\n"
    yesno "Do you agree"
    if [ $val -eq 0 ]; then
      error "Missing software dependencies. Aborting installation."
    fi
  fi

  if [ ! -z "$NF_PKG_LST" ]; then
    prompt ""
    apt-get install $NF_PKG_LST -y
    [ $? -ne 0 ] && error "Could not install software dependencies. Aborting."
  fi
  
  if [ ! -z "$NF_PIP_LST" ]; then
    prompt ""
    pip install $NF_PIP_LST 
    [ $? -ne 0 ] && error "Could not install software dependencies. Aborting."
  fi
  
  # FIX postgresql problem with locales
  export LC_ALL=C
  [ $(pg_lsclusters | wc -l) -eq 1 ] && pg_createcluster 9.3 main --start &> /dev/null

  echo "1" > "$TMP/last_step.txt"
 
fi

if [ 2 -gt $last_step ]; then
  
  prompt "\n2) Prepare environment and copy files.\n"
  
  # Create UNIX user tluser
  if [ -z "$(cat /etc/passwd | grep "^${TLUSER}:")" ]; then
    log "Creating transLectures user '$TLUSER'..."
    useradd -d "/home/$TLUSER" -m -s /bin/bash $TLUSER
    [ $? -ne 0 ] && error "Could not create transLectures user '$TLUSER'. Aborting."
  else
    warn "transLectures user '$TLUSER' already exists."
  fi
  
  if [ -z "$(cat /etc/group | grep "^${TLGROUP}:")" ]; then
    log "Creating transLectures group '$TLGROUP'..."
    groupadd $TLGROUP
    [ $? -ne 0 ] && error "Could not create transLectures group '$TLGROUP'. Aborting."
  else
    warn "transLectures group '$TLGROUP' already exists."
  fi
  
  log "Adding user '$TLUSER' to group '$TLGROUP'"
  usermod -a -G "$TLGROUP" "$TLUSER"
  [ $? -ne 0 ] && error "Could not add user '$TLUSER' to group '$TLGROUP'. Aborting"
  
  # Copy TLP to install dir.
  log "Installing transLectures Platform (TLP) sources..."
  sudo -u $TLUSER mkdir -p "$TLP_INSTALL_DIR"
  [ $? -ne 0 ] && error "Could not create installation dir. Aborting."
  for d in $DIRS; do
    sudo -u $TLUSER cp -r "$CDIR/$d" "$TLP_INSTALL_DIR"
    [ $? -ne 0 ] && error "Could not copy directory $d into installation dir. Aborting."
  done
  
  echo "2" > "$TMP/last_step.txt"

fi

if [ 3 -gt $last_step ]; then
   
  prompt "\n3) Set up transLectures Database.\n"
 
  POSTGRES_CONF=/etc/postgresql/9.3/main/postgresql.conf
  log "Backing up postgresql config file..."
  cp "$POSTGRES_CONF" "$POSTGRES_CONF.bak"
  [ $? -ne 0 ] && error "Could not back up Postgresql config file. Aborting."
  cat "$POSTGRES_CONF.bak" | awk '$1 == "listen_addresses" {$3 = "\x27*\x27"} {print}' > "$POSTGRES_CONF"
  [ $? -ne 0 ] && error "Could not set up Postgresql config file. Aborting."
 
  #Install postgres, add user tluser
  log "Configuring PostgreSQL..."
  ret=$(sudo -u postgres psql postgres -tAc "SELECT count(*) FROM pg_roles WHERE rolname='$TLUSER'")
  if [ $ret -eq 0 ]; then
    log "Creating PostgreSQL user '$TLUSER'..."
    sudo -u postgres createuser -s $TLUSER &> "$TMP/psql_create_user.log"
    [ $? -ne 0 ] && error "Could not create postgres user $TLUSER. See '$TMP/psql_create_user.log' for clues. Aborting."
    sudo -u postgres psql -c "ALTER USER $TLUSER WITH PASSWORD '$DBPASSWD'" &> "$TMP/psql_alter_user.log"
    [ $? -ne 0 ] && error "Could not set password for postgres user $TLUSER. See '$TMP/psql_alter_user.log' for clues. Aborting."
  else
    log "PostgreSQL user '$TLUSER' already exists."
  fi
  
  log "Creating database $DBNAME ..."
  sudo -u $TLUSER createdb "$DBNAME" &> "$TMP/psql_createdb.log"
  [ $? -ne 0 ] && error "Could not create database $DBNAME. See '$TMP/psql_createdb.log' for clues. Aborting."
  sudo -u $TLUSER psql -f "$CDIR/db/sql/schema.sql" "$DBNAME" &> "$TMP/psql_schema.log"
  [ $? -ne 0 ] && error "Could not create database schema. See '$TMP/psql_schema.log' for clues. Aborting."
  sudo -u $TLUSER psql -f "$CDIR/db/sql/static_data.sql" "$DBNAME" &> "$TMP/psql_static_data.log"
  [ $? -ne 0 ] && error "Could not insert static data. See '$TMP/psql_static_data.log' for clues. Aborting."

  md5passwd=$(echo -n "$TLPADMINPASSWD" | md5sum | awk '{print $1}')
  [ $? -ne 0 ] && error "Could not compute md5sum of the admin's password. Aborting."
  sudo -u $TLUSER psql "$DBNAME" -c "UPDATE users SET password = '$md5passwd' WHERE id = 0;" &> "$TMP/psql_admin_passwd.log"
  [ $? -ne 0 ] && error "Could not update admin user password. See '$TMP/psql_admin_passwd.log' for clues. Aborting."
 
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/speech/media"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/speech/media' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/speech/trans"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/speech/trans' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/speech/uploads"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/speech/uploads' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/speech/align"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/speech/align' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/text/trans"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/text/trans' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/text/uploads"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/text/uploads' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/data/web/uploads"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/data/web/uploads' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/systems"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/systems' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/logs"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/logs' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/logs/player"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/logs/player' directory. Aborting."
  sudo mkdir -p "$TLP_REPOSITORIES_DIR/logs/docedit"
  [ $? -ne 0 ] && error "Could not create '$TLP_REPOSITORIES_DIR/logs/docedit' directory. Aborting."
  sudo chown $TLUSER:$TLGROUP "$TLP_REPOSITORIES_DIR"/*
  [ $? -ne 0 ] && error "Could not set dir permisions. Aborting."
  sudo chmod 775 "$TLP_REPOSITORIES_DIR"/*
  [ $? -ne 0 ] && error "Could not set dir permisions. Aborting."
  sudo chmod g+s "$TLP_REPOSITORIES_DIR"/*
  [ $? -ne 0 ] && error "Could not set dir permisions. Aborting."
  
  sudo -u $TLUSER psql "$DBNAME" -c "INSERT INTO machines (id, hostname, ip) VALUES (0, '$DBHOST', '$DBIP');" &> "$TMP/psql_machines.log"
  [ $? -ne 0 ] && error "Could not insert machines data. See '$TMP/psql_machines.log' for clues. Aborting."
  sudo -u $TLUSER psql "$DBNAME" -c "INSERT INTO mount_points (name, machine_id, path) VALUES
       ('systems', 0, '$TLP_REPOSITORIES_DIR/systems'),
       ('media', 0, '$TLP_REPOSITORIES_DIR/data/speech/media'),
       ('transcriptions', 0, '$TLP_REPOSITORIES_DIR/data/speech/trans'),
       ('valign_uploads', 0, '$TLP_REPOSITORIES_DIR/data/speech/align'),
       ('uploads', 0, '$TLP_REPOSITORIES_DIR/data/speech/uploads'),
       ('text_documents', 0, '$TLP_REPOSITORIES_DIR/data/text/trans'),
       ('text_uploads', 0, '$TLP_REPOSITORIES_DIR/data/text/uploads'),
       ('web_uploads', 0, '$TLP_REPOSITORIES_DIR/data/web/uploads');" &> "$TMP/psql_mount_points.log"
  [ $? -ne 0 ] && error "Could not insert mount points. See '$TMP/psql_mount_points.log' for clues. Aborting."
  
  echo "3" > "$TMP/last_step.txt"

fi
 
WSPATH_SP="$TLP_INSTALL_DIR/web-service/ws-speech.py"
WSPATH_TE="$TLP_INSTALL_DIR/web-service/ws-text.py"
WSPATH_WTL="$TLP_INSTALL_DIR/web-service/ws-web.py"
WSDIR="$TLP_INSTALL_DIR/web-service"

if [ 4 -gt $last_step ]; then
  
  prompt "\n4) Set up transLectures Web Service.\n"
  
  # Configure Web Service
  log "Setting up Web Service config file..."
  
  WSCFG="$TLP_INSTALL_DIR/web-service/config.ini"
  
  sets="authentication->key_generator->$WSKEYGENERATOR"
  sets="$sets; storage->db_name->$DBNAME"
  sets="$sets; storage->db_user->$TLUSER"
  sets="$sets; storage->db_passwd->$DBPASSWD"
  sets="$sets; storage->db_host->$DBHOST"
  sets="$sets; media_urls->use_urls->yes"
  sets="$sets; media_urls->base_url->$HTTP_ROOT/$PLLOCATION/$PLDATADIRNAME"
  sets="$sets; mailing->enabled->no"
  
  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modconf.py "$WSCFG" "$sets" > "$WSCFG".new
  [ $? -ne 0 ] && error "Could not modify Web Service config file. Aborting."
  sudo -u $TLUSER mv "$WSCFG".new "$WSCFG"
  [ $? -ne 0 ] && error "Could not modify Web Service config file. Aborting."
  
  GENTOKEN=$TLP_INSTALL_DIR/web-service/generate-authentication-token.py

  TLPADMINAUTHKEY=$(python $GENTOKEN admin)
  [ $? -ne 0 ] && error "Could not generate auth secret key for the admin user. Aborting."

  save_config

  log "Setting up Client-Tools config file..."
  
  WSCLICFG=$TLP_INSTALL_DIR/misc/client-tools/python/scr/config.ini

  sets="general->web_service_url->${HTTP_ROOT}/${WSLOCATION}"
  sets="$sets; general->transcription_editor_url->${HTTP_ROOT}/${TXELOCATION}"
  sets="$sets; general->player_url->${HTTP_ROOT}/${PLLOCATION}"
  sets="$sets; general->translation_editor_url->${HTTP_ROOT}/${TELOCATION}"
  sets="$sets; general->web_translation_editor_url->${HTTP_ROOT}/${WTELOCATION}"
  sets="$sets; http_auth->enabled->no"
  sets="$sets; api_client_auth->username->admin"
  sets="$sets; api_client_auth->secret_key->$TLPADMINAUTHKEY"
  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modconf.py "$WSCLICFG" "$sets" > "$WSCLICFG".new
  [ $? -ne 0 ] && error "Could not modify python client tools config file. Aborting."
  sudo -u $TLUSER mv "$WSCLICFG".new "$WSCLICFG"
  [ $? -ne 0 ] && error "Could not modify python client tools config file. Aborting."
  
  echo "4" > "$TMP/last_step.txt"
  
fi

TXEPATH="$TLP_INSTALL_DIR/transedit"
PLPATH="$TLP_INSTALL_DIR/player"
TEPATH="$TLP_INSTALL_DIR/docedit"
WTEPATH="$TLP_INSTALL_DIR/webtleditor"

if [ 5 -gt $last_step ]; then

  prompt "\n5) Set up Transcription Editor, Media Player, Translation Editor and Web Translation Editor.\n"

  log "Setting up Transcription Editor config file..."

  # Configure Transcription Editor
  TXECFG="$TXEPATH/lib/tlplib.php"
 
  cat $TXECFG | 
     sed -e "s|\(private \$API_URL =\)[^;]\{1,\};|\1 ${HTTP_ROOT}/${WSLOCATION}|g" > $TXECFG.new && 
  mv $TXECFG.new $TXECFG
 
  # Configure Player
  PLCFG="$PLPATH/translectures/config.json"
  
  log "Linking directories..."
  
  sudo -u $TLUSER ln -s "$TLP_REPOSITORIES_DIR/data/speech/media" $TLP_INSTALL_DIR/player/$PLDATADIRNAME
  [ $? -ne 0 ] && error "Could not symlink data directory. Aborting."
  
  sudo -u $TLUSER ln -s "$TLP_REPOSITORIES_DIR/logs/player" $TLP_INSTALL_DIR/player/translectures/stats
  [ $? -ne 0 ] && error "Could not symlink logs directory. Aborting."

  sudo -u $TLUSER ln -s "$TLP_REPOSITORIES_DIR/logs/docedit" $TLP_INSTALL_DIR/docedit/stats
  [ $? -ne 0 ] && error "Could not symlink logs directory. Aborting."

  log "Setting up Player JSON config file..."
  
  sets="ws_url->${HTTP_ROOT}/${WSLOCATION}"
  sets="$sets; data_url->$HTTP_ROOT/$PLLOCATION/$PLDATADIRNAME"
  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modjson.py "$PLCFG" "$sets" > "$PLCFG".new
  [ $? -ne 0 ] && error "Could not modify Player JSON config file. Aborting."
  sudo -u $TLUSER mv "$PLCFG".new "$PLCFG"
  [ $? -ne 0 ] && error "Could not modify Player JSON config file. Aborting."
  sudo chown $APACHE_USER:$TLGROUP "$PLCFG"
  sudo chmod 600 "$PLCFG"

  # Configure Editor
  TECFG="$TEPATH/config.json"
  
  log "Setting up Translation Editor JSON config file..."
  
  sets="ws_url->${HTTP_ROOT}/${WSLOCATION}"
  sets="$sets; data_url->$HTTP_ROOT/$PLLOCATION/$PLDATADIRNAME"
  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modjson.py "$TECFG" "$sets" > "$TECFG".new
  [ $? -ne 0 ] && error "Could not modify Translation Editor JSON config file. Aborting."
  sudo -u $TLUSER mv "$TECFG".new "$TECFG"
  [ $? -ne 0 ] && error "Could not modify Translation Editor JSON config file. Aborting."
  sudo chown $APACHE_USER:$TLGROUP "$TECFG"
  sudo chmod 600 "$TECFG"

  # Configure Web Editor
  WTECFG="$WTEPATH/js/main.js"
  
  log "Setting up Web Translation Editor..."
  
  sets="${HTTP_ROOT}/${WSLOCATION}/web/editor/"
  sudo -u $TLUSER sed -e "s|^var TLP_API_URL = '.*';|var TLP_API_URL = '$sets';|" "$WTECFG" > "$WTECFG".new
  [ $? -ne 0 ] && error "Could not modify Web Translation Editor JSON config file. Aborting."
  sudo -u $TLUSER mv "$WTECFG".new "$WTECFG"
  [ $? -ne 0 ] && error "Could not modify Web Translation Editor JSON config file. Aborting."
  sudo chown $APACHE_USER:$TLGROUP "$WTECFG"
  sudo chmod 600 "$WTECFG"


  echo "5" > "$TMP/last_step.txt"

fi
 
if [ 6 -gt $last_step ]; then
  
  prompt "\n6) Configure Apache for the transLectures Web Service and Player.\n"
  
  # Configure Apache
  usermod -a -G "$TLGROUP" $APACHE_USER
   
  APACHE_SITES_FILE="/etc/apache2/sites-enabled/000-default.conf"
  #if [ ! -s "$APACHE_SITES_FILE" ]; then
  #  warn "Could not find Apache Sites-Enabled file at $APACHE_SITES_FILE"
  #  getvar "Please, enter Apache Sites-Enabled file location" APACHE_SITES_FILE
  #  if [ ! -s "$APACHE_SITES_FILE" ]; then
  #    error "Apache Sites-Enabled file not found at the provided location ($APACHE_SITES_FILE). Aborting."
  #  fi
  #fi

  log "Backing up Apache sites-enabled file..."
  cp "$APACHE_SITES_FILE" "$APACHE_SITES_FILE.bak"
  [ $? -ne 0 ] && error "Could not back up Apache sites-enabled file. Aborting."
  log "Updating Apache sites file..."
  cat "$APACHE_SITES_FILE" | grep -v "</VirtualHost>" > "$APACHE_SITES_FILE.tmp"
  
  { cat << EOF

    # TLP Transcription Editor
    Alias /$TXELOCATION $TXEPATH
    <Directory $TXEPATH>
      Options Indexes FollowSymLinks MultiViews
      Require all granted
    </Directory>

    # TLP Player
    Alias /$PLLOCATION $PLPATH
    <Directory $PLPATH>
      Options Indexes FollowSymLinks MultiViews
      Require all granted
    </Directory>

    # TLP Translation Editor
    Alias /$TELOCATION $TEPATH
    <Directory $TEPATH>
      Options Indexes FollowSymLinks MultiViews
      Require all granted
    </Directory>

    # TLP Web Translation Editor
    Alias /$WTELOCATION $WTEPATH
    <Directory $WTEPATH>
      Options Indexes FollowSymLinks MultiViews
      Require all granted
    </Directory>
  
    # TLP Web Service
    WSGIScriptAlias /$WSLOCATION/speech $WSPATH_SP
    WSGIScriptAlias /$WSLOCATION/text $WSPATH_TE
    WSGIScriptAlias /$WSLOCATION/web $WSPATH_WTL
    <Directory $WSDIR>
      Require all granted
    </Directory>
  
  </VirtualHost>
EOF
  } >> "$APACHE_SITES_FILE.tmp"
   
  mv "$APACHE_SITES_FILE.tmp" "$APACHE_SITES_FILE"
  [ $? -ne 0 ] && error "Could not replace Apache sites-enabled file. Aborting."
  
  log "Restarting Apache server..."
  
  sudo service apache2 restart &> "$TMP/restart_apache.log"
  [ $? -ne 0 ] && { cp "$APACHE_SITES_FILE.bak" "$APACHE_SITES_FILE"; error "Apache server could not be restarted. See '$TMP/restart_apache.log' for clues. Restoring backup file and aborting."; }
  
  echo "6" > "$TMP/last_step.txt"
 
fi
 
if [ 7 -gt $last_step ]; then
  
  prompt "\n7) Set up transLectures Ingest Service and Postprocessor module.\n"
  
  # Configure Ingest Service
  log "Setting up Ingest Service config file..."
  
  ISCFG="$TLP_INSTALL_DIR/ingest-service/config.ini"
  
  sets="general->hostname->$CANONICAL_NAME"
  sets="$sets; general->tl_user->$TLUSER"
  sets="$sets; general->tl_group->$TLGROUP"
  sets="$sets; storage->db_name->$DBNAME"
  sets="$sets; storage->db_user->$TLUSER"
  sets="$sets; storage->db_passwd->$DBPASSWD"
  sets="$sets; storage->db_host->$DBHOST"
  sets="$sets; scheduler->localhost->$HOSTNAME"
  sets="$sets; sessions->enabled->no"
  sets="$sets; mailing->enabled->no"
  
  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modconf.py "$ISCFG" "$sets" > "$ISCFG".new
  [ $? -ne 0 ] && error "Could not modify Ingest Service config file. Aborting."
  sudo -u $TLUSER mv "$ISCFG".new "$ISCFG"
  [ $? -ne 0 ] && error "Could not modify Ingest Service config file. Aborting."

  # Configure Postprocessor
  log "Setting up Postprocessor config file..."
  
  POSTCFG="$TLP_INSTALL_DIR/misc/postprocessor/config.ini"
  
  sets="database->db_name->$DBNAME"
  sets="$sets; database->db_user->$TLUSER"
  sets="$sets; database->db_passwd->$DBPASSWD"
  sets="$sets; database->db_host->$DBHOST"
  sets="$sets; libtlp->web_service_url->${HTTP_ROOT}/${WSLOCATION}"
  sets="$sets; libtlp->api_username->admin"
  sets="$sets; libtlp->api_secret_key->$TLPADMINAUTHKEY"

  sudo -u $TLUSER $TLP_INSTALL_DIR/misc/other/modconf.py "$POSTCFG" "$sets" > "$POSTCFG".new
  [ $? -ne 0 ] && error "Could not modify Postprocessor config file. Aborting."
  sudo -u $TLUSER mv "$POSTCFG".new "$POSTCFG"
  [ $? -ne 0 ] && error "Could not modify Postprocessor config file. Aborting."

  if [ $CRONTAB -eq 1 ]; then

    log "Installing new crontab for the Ingest Service and the Postprocessor module..."

    { cat << EOF
      date;
      if [ ! -e "/home/$TLUSER/.cron-tlp.lock" ]; then
        touch "/home/$TLUSER/.cron-tlp.lock";
        source "/home/$TLUSER/.bashrc";
        python "$TLP_INSTALL_DIR/misc/postprocessor/postprocessor.py" >> "$TLP_REPOSITORIES_DIR/logs/postprocessor.log" 2>&1;
        python "$TLP_INSTALL_DIR/ingest-service/is-speech.py" -v >> "$TLP_REPOSITORIES_DIR/logs/is-speech.log" 2>&1;
        python "$TLP_INSTALL_DIR/ingest-service/is-text.py" -v >> "$TLP_REPOSITORIES_DIR/logs/is-text.log" 2>&1;
        python "$TLP_INSTALL_DIR/ingest-service/is-web.py" -v >> "$TLP_REPOSITORIES_DIR/logs/is-web.log" 2>&1;
        python "$TLP_INSTALL_DIR/ingest-service/is-align.py" -v >> "$TLP_REPOSITORIES_DIR/logs/is-align.log" 2>&1;
        python "$TLP_INSTALL_DIR/misc/postprocessor/postprocessor.py" >> "$TLP_REPOSITORIES_DIR/logs/postprocessor.log" 2>&1;
        rm "/home/$TLUSER/.cron-tlp.lock";
      fi
EOF
    } > "$TLP_INSTALL_DIR/misc/tlp-cron-jobs.sh"
    chmod +x "$TLP_INSTALL_DIR/misc/tlp-cron-jobs.sh"
    chown $TLUSER:$TLGROUP "$TLP_INSTALL_DIR/misc/tlp-cron-jobs.sh" 

    { cat << EOF
MAILTO=""
*/1 * * * * "$TLP_INSTALL_DIR/misc/tlp-cron-jobs.sh" &> "$TLP_REPOSITORIES_DIR/logs/cron.log"
EOF
    } > "$TMP/crontab"
  
    sudo -u $TLUSER crontab "$TMP/crontab"
    [ $? -ne 0 ] && error "Could not install new crontab file. Aborting."

  fi
  
  echo "7" > "$TMP/last_step.txt"
  
fi


prompt "\n********* INSTALLATION COMPLETED! *********\n"

rm -rf "$TMP"

exit 0

