#!/usr/bin/python

import sys
import bs4
import nltk.data
import re
import libdtlx
import cgi
import string

LANGS = {"en":"english", "es":"spanish", "it":"italian", "nl":"dutch", "pt":"portuguese", "ca":"spanish", "fr":"french", "et":"estonian" }

SKIP_TAGS = [ "script", "noscript", "applet", "embed", "object", "param", "head", "meta", "base", "basefont", "style", "audio", "source", "track", "video", "frame", "frameset", "noframes", "iframe", "map", "area", "canvas" ]
FORM_TAGS = [ "b", "i", "strong", "center", "em", "font", "acronym", "abbr", "bdi", "bdo", "big", "del", "dfn", "ins", "kbd", "mark", "meter", "progress", "q", "rp", "rt", "ruby", "s", "small", "strike", "sub", "sup", "time", "tt", "u", "var", "wbr" ]
ATTR_TAGS = { "a":(["title"], 'form'), "img":(["alt", "title"], 'skip') }

def mycomment(st):
  return u"""<!--%s-->""" % st

def mywrap(st):
  htmlstr = u"""<span data-dtlx-mywrap="1">%s</span>""" % st
  return bs4.BeautifulSoup(htmlstr)
  
def mycopy(n, comment=False):
  if comment:
    ss = mywrap(mycomment(unicode(n)))
  else:
    ss = mywrap(unicode(n))
  return ss.span.contents[0]

def span(soup, node, ltxt, gid, dtlx, sd, tokenize=True):
  sp = soup.new_tag("span")
  for e in ltxt:
    e1 = mycopy(e)
    sp.append(e1)
  aux = "".join([ unicode(x) for x in sp.contents ])
  aux = re.sub(r">([^\s<]+)", r"><myauxmark/> \1", aux)
  aux = re.sub(r"([^\s>]+)<", r"\1 <myauxmark/><", aux)
  if tokenize:
    sents = sd.tokenize(aux)
  else:
    sents = [aux]
  for s in sents:
    s = re.sub(r"><myauxmark/> *", r">", s)
    s = re.sub(r" *<myauxmark/><", r"<", s)
    ss =  mywrap(s)
    txt = ss.get_text().strip()
    any_attr_tag =  any( any( t.has_attr(a) for a in ATTR_TAGS.get(t.name, ([],''))[0] ) for t in ss.span.contents )
    #if (len(txt) > 0 and any(c.isalpha() for c in txt)) or any_attr_tag:
    if (len(txt) > 0 and any(c.isalnum() for c in txt)) or any_attr_tag:
      sId = "%d" % gid[0]
      dtlx_sent = libdtlx.Sentence(sId)
      dtlx_sent.origin = libdtlx.Origin()
      dtlx_sent.origin.rawText = txt
      dtlx_sent.origin.formattedText = s
      dtlx.sentences.append(dtlx_sent)
      sp = soup.new_tag("span")
      sp['data-dtlx-id'] = gid[0]
      sp.contents = ss.span.contents
      node.append(sp)
      gid[0] += 1
    else:
      try:
        dtlx.sentences[-1].origin.formattedText += s
      except IndexError:
        continue
      if len(ss.contents) == 0:
        try:
          node.contents[-1].string += s
        except IndexError:
          node.contents = [ soup.new_string(s) ]
      else:
        try:
          node.contents[-1].contents += ss.span.contents
        except IndexError:
          node.contents = ss.span.contents

def lsents(nsoup, soup, nnode, node, gid, dtlx, sd, text):
  ltxt = []
  for n in node.children:
    if isinstance(n, bs4.element.Comment):
      if text[0] == True:
        text[0] = False
        res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
        if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
        ltxt = []
      n2 = mycopy(n, comment=True)
      nnode.append(n2)
    elif isinstance(n, bs4.element.Tag):
      if n.name == "img" and str(n.attrs.get('src', None)).find("data:") == 0: # Skip binary images
        if text[0] == True:
          text[0] = False
          res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
          if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
          ltxt = []
        n2 = mycopy(n)
        nnode.append(n2)
      elif n.name in ATTR_TAGS.keys():
        if any(n.has_attr(a[0]) for a in ATTR_TAGS[n.name]):
          if text[0] == True:
            text[0] = False
            res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
            if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
            ltxt = []
          ltxt.append(n)
          span(nsoup, nnode, ltxt, gid, dtlx, sd, tokenize=False)
          ltxt = []
        else:
          if ATTR_TAGS[n.name][1] == 'skip':
            if text[0] == True:
              text[0] = False
              res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
              if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
              ltxt = []
            n2 = mycopy(n)
            nnode.append(n2)
          else:
            if any(t not in FORM_TAGS for t in n.find_all(True)):
              if text[0] == True:
                text[0] = False
                res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
                if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
                ltxt = []
              nt = nsoup.new_tag(n.name)
              nnode.append(nt)
              nt.attrs = n.attrs
              lsents(nsoup, soup, nt, n, gid, dtlx, sd, text)
            else:
              text[0] = True
              ltxt.append(n)
      #elif len(n.get_text().strip()) == 0 or n.name in SKIP_TAGS:
      elif n.name in SKIP_TAGS:
        if text[0] == True:
          text[0] = False
          res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
          if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
          ltxt = []
        n2 = mycopy(n)
        nnode.append(n2)
      else:
        if n.name in FORM_TAGS:
          if any(t not in FORM_TAGS for t in n.find_all(True)):
            if text[0] == True:
              text[0] = False
              res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
              if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
              ltxt = []
            nt = nsoup.new_tag(n.name)
            nnode.append(nt)
            nt.attrs = n.attrs
            lsents(nsoup, soup, nt, n, gid, dtlx, sd, text)
          else:
            text[0] = True
            ltxt.append(n)
        else:
          if text[0] == True:
            text[0] = False
            res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
            if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
            ltxt = []
          nt = nsoup.new_tag(n.name)
          nnode.append(nt)
          nt.attrs = n.attrs
          lsents(nsoup, soup, nt, n, gid, dtlx, sd, text)
    elif isinstance(n, bs4.element.NavigableString):
      if unicode(n).strip() != "":
        text[0] = True
        ltxt.append(bs4.element.NavigableString(cgi.escape(n)))
  if text[0] == True:
    text[0] = False
    res = pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text)
    if res: span(nsoup, nnode, ltxt, gid, dtlx, sd)
    ltxt = []

def pre_span(nsoup, soup, nnode, ltxt, gid, dtlx, sd, text):
  any_str = any(isinstance(t, bs4.element.NavigableString) for t in ltxt)
  any_attr_tag = any( any( t.has_attr(a) for a in ATTR_TAGS.get(t.name, ([],''))[0] ) for t in ltxt)
  if not(any_str) and not(any_attr_tag):
    for t in ltxt:
      nt = nsoup.new_tag(t.name)
      nnode.append(nt)
      nt.attrs = t.attrs
      lsents(nsoup, soup, nt, t, gid, dtlx, sd, text)
    return False
  else:
    return True



def shall_merge_texts(txt):
  if txt[0].islower():
    return True
  if txt[0] in string.punctuation:
    return True
  return False

def shall_remove(txt):
  if len(txt) == 0:
    return True
  if txt.find("<<") == 0: # Shit from extracting PDF text
    return True
  return False

def preprocess_soup(s):
  n = s.body
  for tag in n.find_all("p"):
    if isinstance(tag, bs4.element.Tag):
      txt = tag.text.encode('utf-8').strip()
      if shall_remove(txt):
        tag.extract()
        continue
      for e in tag.find_next_siblings("p"):
        txt = e.text.encode('utf-8').strip()
        if len(txt) > 0:
          if shall_merge_texts(txt):
            ret = e.extract()
            for c in ret.contents:
              tag.append(c)
          else:
            break
        else:
          e.extract()



def html2dtlx(html, doc_id, src_lang_code, trg_lang_code, title=None, empty_tl=False):

  # EMMA FIX
  html = html.replace('\\t', '').replace('\\n', '')
  
  # Sentence tokenizer from nltk
  lang = LANGS.get(src_lang_code, None)
  if lang == None:
    lang = "english" 
  sent_detector = nltk.data.load('tokenizers/punkt/%s.pickle' % lang)
  
  bs = bs4.BeautifulSoup(html, from_encoding='utf-8')

  #preprocess_soup(bs)

  nbs = bs4.BeautifulSoup("<body></body>")

  try:
    nbs.html.attrs = bs.html.attrs.copy()
    nbs.body.attrs = bs.body.attrs.copy()
  except:
    pass


  head = nbs.new_tag('head')
  nbs.html.insert(0, head)
  if bs.head != None:
    nbs.head.contents = bs.head.contents[:]
 
  found = False 
  for h in nbs.head.contents:
    if h.name == "meta" and h.has_attr('charset'):
      found = True
      break

  if not(found):
    meta = nbs.new_tag('meta')
    meta.attrs['charset'] = 'UTF-8'
    nbs.head.contents.append(meta)

  dtlx = libdtlx.Document()
  dtlx.docId = doc_id
  dtlx.srcLang = src_lang_code
  dtlx.trgLang = trg_lang_code
  if title != None:
     dtlx.title = title

  gid = [1]# Must be a list instead of a integer, because ints cannot be modified by reference
  text = [False]
  if bs.body != None:
    lsents(nbs, bs, nbs.body, bs.body, gid, dtlx, sent_detector, text)

  if empty_tl:
    for s in dtlx.sentences:
      t = libdtlx.Translation()
      t.authorType = "automatic"
      t.rawText = ""
      t.formattedText = ""
      s.translations.append(t)

  dtlx.updateStatus()
  return nbs, dtlx

  
def html2dtlx_str(html, doc_id, src_lang_code, trg_lang_code, title=None, empty_tl=False):
  bs, dtlx = html2dtlx(html, doc_id, src_lang_code, trg_lang_code, title=title, empty_tl=empty_tl)
  return bs.prettify().encode('utf-8'), dtlx.toXML()


if __name__ == "__main__":
  
  if len(sys.argv) != 6:
    sys.stderr.write("Usage: %s <HTML> <DOC_ID> <SRC_LANG> <TRG_LANG> <OUT_FILE_PREFIX>\n" % sys.argv[0])
    sys.exit(1)
  
  html_fp = sys.argv[1]
  doc_id = sys.argv[2]
  src_lang_code = sys.argv[3]
  trg_lang_code = sys.argv[4]
  out_file_prefix = sys.argv[5]
  
  with open(html_fp, 'r') as fd:
    html = fd.read()


  bs, dtlx = html2dtlx(html, doc_id, src_lang_code, trg_lang_code)  
 
  dtlx_fp = "%s.dtlx" % out_file_prefix 
  html_fp = "%s.html" % out_file_prefix 
    
  with open(dtlx_fp, 'w') as fd:
    fd.write(dtlx.toXML())
  
  with open(html_fp, 'w') as fd:
    fd.write(bs.prettify().encode('utf-8'))
  
 
