# -*- coding: utf-8 -
#
#  Copyright 2015 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import hashlib

class TLPAuth(object):

    def __init__(self, key_generator):
        """ Constructor """
        self.key_generator = key_generator
        self.f_skey_hash = hashlib.sha256
        self.f_rkey_hash = hashlib.sha1
        self.request_subkey_length = int(self.f_rkey_hash().digest_size)*2
   
 
    def __generate_secret_key(self, user, passwd):
        s = self.f_skey_hash()
        s.update(self.key_generator + user + passwd)
        return s.hexdigest()


    def get_secret_key(self, user, passwd):
        return self.__generate_secret_key(user, passwd)

   
    def get_request_key(self, user, passwd, oid, expire, author=None, author_conf=None):
        secret_key = self.__generate_secret_key(user, passwd)
        s1 = self.f_rkey_hash()
        s1.update(oid + str(expire) + user + secret_key)
        rk = s1.hexdigest()
        if author != None:
          s2 = self.f_rkey_hash()
          s2.update(author + str(author_conf) + str(expire) + user + secret_key)
          rk += s2.hexdigest()
        return rk

