# -*- coding: utf-8 -
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import math

MAX_NGRAM_ORDER = 4

def get_ngram_counts(sl):
  d = {}
  for i in xrange(len(sl)):
    for k in xrange( i+1, min(i+MAX_NGRAM_ORDER+1 , len(sl)+1) ):
      ngk = tuple(sl[i:k])
      try:
        d[ngk] += 1
      except:
        d[ngk] = 1
  return d

def get_ngram_overlapping(hyp, ref):
  ov = [ 0 for i in xrange(MAX_NGRAM_ORDER) ]
  for k in hyp:
    if len(k) <= MAX_NGRAM_ORDER:
      cnh = hyp[k]
      cnr = ref.get(k,0)
      ov[len(k)-1] += min(cnh, cnr)
  return ov

def my_log(x):
  if x == 0:
    return -9999999999
  else:
    return math.log(x)

def bleu_score(hyp, ref):
  hs = [ l.strip().split() for l in hyp ]
  rs = [ l.strip().split() for l in ref ]
  hn = [ get_ngram_counts(s) for s in hs ]
  rn = [ get_ngram_counts(s) for s in rs ]
  nol = [ get_ngram_overlapping(hn[i],rn[i]) for i in xrange(len(hn)) ]
  htn = [ [ max((len(s) - i), 0) for i in xrange(MAX_NGRAM_ORDER) ] for s in hs ]
  cnol  = [ 0 for i in xrange(MAX_NGRAM_ORDER) ]
  ctn = [ 0 for i in xrange(MAX_NGRAM_ORDER) ]
  haclen = 0
  raclen = 0
  for i in xrange(len(hs)):
    haclen += len(hs[i])
    raclen += len(rs[i])
    for n in xrange(4):
      cnol[n] += nol[i][n]
      ctn[n] += htn[i][n]
  ngprec = [ float(cnol[i]) / float(ctn[i]) if ctn[i] > 0 else 0 for  i in xrange(MAX_NGRAM_ORDER) ]
  aux = sum( [ my_log(x) for x in ngprec ] ) / MAX_NGRAM_ORDER
  bp = min( 1, math.exp( 1.0 - ( raclen / float(haclen) ) ) )
  aux += math.log(bp)
  bleu = math.exp(aux) * 100
  return bleu

if __name__ == "__main__":

  import sys

  with open(sys.argv[1], "r") as fh:
    hyp = [ l.decode('utf-8') for l in fh ]

  with open(sys.argv[2], "r") as fr:
    ref = [ l.decode('utf-8') for l in fr ]

  print bleu_score(hyp, ref)
