$(document).ready(function(){
  var TLP_VERSION_TAG = "TLP3.1";
  /* Fix console logs to IE and Opera browsers. */
  if (!window.console) {
    window.console = {};
    window.console.log = function(text) { var emptyvar = text; };
  }
  // ttime = total time (stops counting on mouseleave)
  // etime = editing time (sum of segment editing times)
  // itext = initial text
  // ftext = final text
  // sup_session = percentaje of supervised segments in this session
  // sup_total = percentaje of total supervised segments
  var debug = false;
  var stats = {version: TLP_VERSION_TAG, src: {lang: "", ttime: 0, etime: 0, itext: "", ftext: "", sup_session: 0.0, sup_total: 0.0}, tar: {lang: "", ttime: 0, etime: 0, itext: "", ftext: "", sup_session: 0.0, sup_total: 0.0}};
  var session_timestamp = new Date().getTime();
  var ttime_src_last_init = new Date().getTime();
  var ttime_tar_last_init = new Date().getTime();
  var etime_src_last_init = null;
  var etime_tar_last_init = null;
  var wasEditingTar = false;
  var wasEditingSrc = false;

  /* Event handlers to retrieve the user actions. */
  $(document).on("mouseenter",function(e) {
    if (debug) console.log("** STATS: \"mouseenter\" event captured");
    stats_activesystem();
  });
  $(document).on("mouseleave",function(e) {
    if (debug) console.log("** STATS: \"mouseleave\" event captured");
    stats_disablesystem();
  });
  $(document).on("tLsave",function(e) {
    save_stats(e.container);
    if (debug) console.log("** STATS: \"tLsave\" event captured");
  });
  $(document).on("tLload",function(e) {
    // e.container, e.lang
    stats[e.container]["lang"] = e.lang;
    stats[e.container]["ttime"] = 0.0;
    stats[e.container]["etime"] = 0.0;
    stats[e.container]["sup_session"] = 0.0;
    stats[e.container]["sup_total"] = 0.0;
    stats[e.container]["itext"] = "";
    stats[e.container]["ftext"] = "";
    if (e.container == "tar") {
      if (dfxptar != null && "s" in dfxptar) { 
        stats[e.container]["idfxp"] = JSON.parse(JSON.stringify(dfxptar));
        for (var i=0; i<dfxptar['s'].length; i++) stats[e.container]["itext"] += dfxptar['s'][i]['text']+" ";
      }
      ttime_tar_last_init = new Date().getTime();
      etime_tar_last_init = null;
    }
    else if (e.container == "src") {
      if (dfxpsrc != null && "s" in dfxpsrc) {
        stats[e.container]["idfxp"] = JSON.parse(JSON.stringify(dfxpsrc));
        for (var i=0; i<dfxpsrc['s'].length; i++) stats[e.container]["itext"] += dfxpsrc['s'][i]['text']+" ";
      }
      ttime_src_last_init = new Date().getTime();
      etime_src_last_init = null;
    }
    if (debug) console.log("** STATS: \"tLload\" event captured");
  });
  $(document).on("edit",function(e) {
    if (debug) console.log("** STATS: \"edit\" event captured ("+e.container+")");
    if (e.container == "tar") {
      etime_tar_last_init = new Date().getTime();
    }
    else if (e.container == "src") {
      etime_src_last_init = new Date().getTime();
    }
  });
  $(document).on("confirm",function(e) {
    if (debug && typeof e.container != 'undefined') console.log("** STATS: \"confirm\" event captured ("+e.container+")");
    if (e.container == "tar") {
      if (etime_tar_last_init != null) {
        var editDuration = (new Date().getTime()-etime_tar_last_init)/1000.0;
        if (editDuration < 300) stats[e.container]["etime"] += editDuration;
      }
      etime_tar_last_init = null;
    }
    else if (e.container == "src") {
      if (etime_src_last_init != null) {
        var editDuration = (new Date().getTime()-etime_src_last_init)/1000.0;
        if (editDuration < 300) stats[e.container]["etime"] += editDuration;
      }
      etime_src_last_init = null;
    }
  });

  function save_stats(lst) {
    if (debug) console.log("** STATS: Saving user stats");
    var lst_lang = stats[lst]["lang"];
    if (typeof mediainfo != "undefined" && "language" in mediainfo && mediainfo["language"] != stats[lst]["lang"]) stats[lst]["lang"] = mediainfo["language"]+"-"+stats[lst]["lang"];
    stats[lst]["ftext"] = "";
    stats[lst]["duration"] = $("video").get(0).duration;
    if (jQuery('#subtitle-waveform-container').is(':visible')) stats[lst]["advanced_mode"] = true;
    else stats[lst]["advanced_mode"] = false;
    var seg_sup = 0;
    var seg_man = 0;
    if (lst == "tar") {
      stats[lst]["fdfxp"] = JSON.parse(JSON.stringify(dfxptar));
      for (var i=0; i<dfxptar['s'].length; i++) {
        stats[lst]["ftext"] += dfxptar['s'][i]['text']+" ";
        if ("edited" in dfxptar['s'][i] && dfxptar['s'][i]["edited"] != "none") seg_sup++;
        if ("aT" in dfxptar['s'][i] && dfxptar['s'][i]["aT"] == "human") seg_man++;
      }
      stats[lst]["sup_session"] += (parseFloat(seg_sup)/parseFloat(dfxptar['s'].length))*100;
      stats[lst]["sup_total"] = (parseFloat(seg_man)/parseFloat(dfxptar['s'].length))*100;
      if (ttime_tar_last_init != null) {
        stats[lst]["ttime"] += ((new Date().getTime())-ttime_tar_last_init)/1000;
        ttime_tar_last_init = new Date().getTime();
      }
    }
    else if (lst == "src") {
      stats[lst]["fdfxp"] = JSON.parse(JSON.stringify(dfxpsrc));
      for (var i=0; i<dfxpsrc['s'].length; i++) {
        stats[lst]["ftext"] += dfxpsrc['s'][i]['text']+" ";
        if ("edited" in dfxpsrc['s'][i] && dfxpsrc['s'][i]["edited"] != "none") seg_sup++;
        if ("aT" in dfxpsrc['s'][i] && dfxpsrc['s'][i]["aT"] == "human") seg_man++;
      }
      stats[lst]["sup_session"] += (parseFloat(seg_sup)/parseFloat(dfxpsrc['s'].length))*100;
      stats[lst]["sup_total"] = (parseFloat(seg_man)/parseFloat(dfxpsrc['s'].length))*100;
      if (ttime_src_last_init != null) {
        stats[lst]["ttime"] += ((new Date().getTime())-ttime_src_last_init)/1000;
        ttime_src_last_init = new Date().getTime();
      }
    }
    if (debug) console.log(stats);
    $.post("ttp-player/translectures/lib/stats.php", {lang: stats[lst]["lang"], data: JSON.stringify(stats[lst]), video: request_data["id"], user:request_data["author_id"], conf:request_data["author_conf"], timestamp: session_timestamp}).done(function(data) { console.log("** STATS: Stats saved"); });
    stats[lst]["lang"] = lst_lang;
  }

  var stats_activesystem = function() {
    if (ttime_src_last_init == null) ttime_src_last_init = new Date().getTime();
    else {
      stats["src"]["ttime"] += (new Date().getTime()-ttime_src_last_init)/1000.0;
      ttime_src_last_init = new Date().getTime();
    }
    if (ttime_tar_last_init == null) ttime_tar_last_init = new Date().getTime();
    else { 
      stats["tar"]["ttime"] += (new Date().getTime()-ttime_tar_last_init)/1000.0;
      ttime_tar_last_init = new Date().getTime();
    }
    if (wasEditingTar) {
      etime_tar_last_init = new Date().getTime();
      wasEditingTar = false;
    }
    if (wasEditingSrc) {
      etime_src_last_init = new Date().getTime();
      wasEditingSrc = false;
    }
  };

  var stats_disablesystem = function() {
    if (ttime_src_last_init != null) stats["src"]["ttime"] += (new Date().getTime()-ttime_src_last_init)/1000.0;
    ttime_src_last_init = null;
    if (ttime_tar_last_init != null) stats["tar"]["ttime"] += (new Date().getTime()-ttime_tar_last_init)/1000.0;
    ttime_tar_last_init = null;
    if (etime_tar_last_init != null) {
      stats["tar"]["etime"] += (new Date().getTime()-etime_tar_last_init)/1000.0;
      wasEditingTar = true;
    }
    etime_tar_last_init = null;
    if (etime_src_last_init != null) {
      stats["src"]["etime"] += (new Date().getTime()-etime_src_last_init)/1000.0;
      wasEditingSrc = true;
    }
    etime_src_last_init = null;
  };
});
