/***

Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

***/

/* GLOBAL VARS */

/*
 request_data:
   id = video id
   lang = user desired target transcription language
   author_id = author user name
   author_name = author name (optional)
   author_conf = author confidence
   sel_data_policy = dfxp alternative filter policy method (optional)
   api_user = web service user name
   expire = expire UNIX timestamp
   request_key = request key
   
  userSettings:
    player:
      vmode
      tlalignment
      autoconfirm

  pcmWaveformFile;
  ttsAvailTracks;
*/


/*
 SETTINGS 
*/
var shortcutDefinitions = {};
shortcutDefinitions["replay"] = "shift tab";
shortcutDefinitions["playpause"] = "tab";
shortcutDefinitions["enterconfirm"] = "enter";
shortcutDefinitions["newsegment"] = "ctrl enter";
shortcutDefinitions["splitsegment"] = "ctrl s";
shortcutDefinitions["joinprev"] = "ctrl backspace";
shortcutDefinitions["joinnext"] = "ctrl shift backspace";
shortcutDefinitions["jumpback"] = "alt left";
shortcutDefinitions["jumpforward"] = "alt right";
shortcutDefinitions["frameback"] = "ctrl num_4";
shortcutDefinitions["frameforward"] = "ctrl num_6";
shortcutDefinitions["prevsegment"] = "up";
shortcutDefinitions["nextsegment"] = "down";
shortcutDefinitions["undo"] = "ctrl z";
var userSettings = { 
  player: { 
    vmode: 1,
    tlalignment: true,
    autoconfirm: true,
    tlautoedit: false,
    advanced: false,
    volume: 1.0, autoplay: true,
    aspect_ratio: 1.78,
    cps: 19,
    cps_allowance: 0.3,
    smpte: false,
    max_cpl: 0 
  }
};
userSettings['shortcuts'] = shortcutDefinitions;

var debugmode = false;
var editing = false; // false if no editing, the focused editing object focused otherwise
var emode = 0; // 0=normal/advanced, 1=forced-align, 2=intelligent-interaction
var vmode = 1; // 0=over, 1=horizontal, 2=vertical
var currents = 0; // current segment (target)
var currents_src = 0; // current segment (source)
var dfxpsrc; // contains all dfxp info and changes (source selected language)
var dfxptar; // contains all dfxp info and changes (target selected language)
var langs = {}; // available languages, example: ({"media_lang": "es", "langs": [{"code": "es", "type": 2, "value": "Spanish"}, {"code": "en", "type": 1, "value": "English"}, {"code": "ca", "type": 1, "value": "Catalan"}], "rcode": 0, "rcode_description": "Language list available"})
var edit_history_data = false; // revisions data from web service
var sremovedtar = [];
var smaxtar = 0;
var sremovedsrc = [];
var smaxsrc = 0;
var showTranslationAlignments = userSettings["player"]["tlalignment"];
var volume = userSettings["player"]["volume"];
var autoEdit = userSettings["player"]["tlautoedit"];
var autoconfirm = userSettings["player"]["autoconfirm"];
var mod_session = false;
var reviewing_session_id = false;
var savedTranscriptionInSession = false;
var frameRate = 24.0;
var keyListener = new window.keypress.Listener(); // Keypress plugin http://dmauro.github.io/Keypress/
var registeredCombos = null;
var max_ops_heap = 10;
var undo_heap = [];
var redo_heap = [];

$(document).ready(function() {
var tlplayer = $("video").get(0);
var helpvisible = false; // true if help layout is currently visible
var confThreshold = 0.25; // word confidence measure threshold for printing words/segments in red color [0.0, 1.0]
var showLowConfWords = false; // print word colors depending on transcription quality (manual, low confidence, etc.)
var loadFlag = 0; // loadFlag is used for removing "loading" div once loading of transcriptions is completed
var firstLoad = true;
var saveFlag = 0; // saveFlag is used for removing "saving" div once loading of transcriptions is completed
var saveMessage = "";

/*
 PLAYER BUTTONS CODE 
*/

$('.control-bars').click(function(e) {
  // hide any visible panel
  if ($(e.target).parents('.tlpanel').length == 0 && !$(e.target).hasClass("tlpanel")) $('.tlpanel').hide();
  // stop click propagation outside control-bars
  if ($(e.target).parents('.tlpanel').length == 0) return false;
});

$('video').click(function(e) {
  if (editing === false) {
    if (tlplayer.paused) tlplayer.play();
    else tlplayer.pause();
  }
});

$('button.play').click(function(e) {
  if (tlplayer.paused) tlplayer.play();
  else tlplayer.pause();
});

$('button.forward').click(function(e) {
  moveForward();
});

$('button.backward').click(function(e) {
  moveBackward();
});

$('button.stepforward').click(function(e) {
  nextSegment();
});

$('button.stepbackward').click(function(e) {
  previousSegment();
});

$('button.options').click(function(e) {
  $('.tlpanel').not('.options-panel').hide();
  $('.options-panel').toggle();
  return false;
});

$('button.layout').click(function(e) {
  // if (editing != false) $(editing).trigger("confirm");
  $('.tlpanel').not('.layouts-panel').hide();
  $('.layouts-panel').toggle();
  return false;
});

$('button.fullscreen').click(function(e) {
  var playerdiv = $('#tl-player-container').get(0);
  if (window.screenTop || window.screenY) {
    // enter fullscreen
    $(document).on('webkitfullscreenchange mozfullscreenchange fullscreenchange MSFullscreenChange', function() {
      if (!(window.screenTop || window.screenY) && /firefox/.test(navigator.userAgent.toLowerCase())) {
        var mtop = parseInt(($('#tl-player-container').height()-$(tlplayer).outerHeight(true))/2)-($('.control-bars').height());
        if (mtop < 0) mtop = 0;
        jQuery('#subtitle-bar-container').css('bottom', (mtop+50)+'px');
      }
    });
    if (playerdiv.requestFullscreen) {
      playerdiv.requestFullscreen();
    } else if (playerdiv.mozRequestFullScreen) {
      playerdiv.mozRequestFullScreen();
    } else if (playerdiv.webkitRequestFullscreen) {
      playerdiv.webkitRequestFullscreen();
    } else if (playerdiv.msRequestFullscreen) {
      playerdiv.msRequestFullscreen();
    }
    console.log("Entering fullscreen");
    jQuery('#tl-player-container').css('width', '100%');
    $(tlplayer).css('height', '');
    jQuery('.control-bars').css({'width': '100%', 'position': 'fixed', 'bottom': '0px', 'left': '0px'});
    var fontSize = ((Math.round($(window).width()/66) < 11) ? "11pt" : (Math.round($(window).width()/66) + 'pt'));
    $('.subtitle-bar').css('fontSize', fontSize);
  }
  else {
    // exit fullscreen
    if (document.exitFullscreen) {
      document.exitFullscreen();
    } else if (document.mozCancelFullScreen) {
      document.mozCancelFullScreen();
    } else if (document.webkitExitFullscreen) {
      document.webkitExitFullscreen();
    } else if (document.msExitFullscreen) {
      document.msExitFullscreen();
    }
    // not fullscreen
    console.log("Exiting fullscreen");
    jQuery('#tl-player-container').css('width', '');
    jQuery('.control-bars').css({'width': '', 'position': '', 'bottom': '', 'left': ''});
    jQuery(tlplayer).css('margin-top', '');
    jQuery('#subtitle-bar-container').css('z-index', '').css('bottom', '50px');
    jQuery('.subtitle-bar').css('font-size', (Math.round($('#tl-player').width()/66) < 11) ? "11pt" : Math.round($('#tl-player').width()/54) + 'pt');
    $(window).trigger("resize");
  }
  return false;
});

$(document).on('webkitfullscreenchange mozfullscreenchange fullscreenchange', function(e) {
  if (window.screenTop || window.screenY) {
    console.log("Exiting fullscreen");
    jQuery('#tl-player-container').css('width', '');
    jQuery('.control-bars').css({'width': '', 'position': '', 'bottom': '', 'left': ''});
    jQuery(tlplayer).css('margin-top', '');
    jQuery('#subtitle-bar-container').css('z-index', '').css('bottom', '50px');
    jQuery('.subtitle-bar').css('font-size', (Math.round($('#tl-player').width()/66) < 11) ? "11pt" : Math.round($('#tl-player').width()/54) + 'pt');
    $(window).trigger("resize");
  }
});

// requires jQuery Tools plugin
$('input[id="volume-bar"]').rangeinput({value: volume});

var vtimeout;

$('button.volume').click(function(e){
  tlplayer.muted = !tlplayer.muted;
  if (tlplayer.muted) {
    $('button.volume').css('background-position','-196px 0px');
    $('button.volume').attr('title','Activate sound');
    $('input[id="volume-bar"]').val(0);
    $('.slider .handle').css({"left": "0px"});
  }
  else {
    tlplayer.volume = 1;
    $('input[id="volume-bar"]').val(1);
    $('.slider .handle').css({"left": "82px"});
    if (volume <= 0.4) $('button.volume').css('background-position','-168px 0px');
    else $('button.volume').css('background-position','-224px 0px');
    $('button.volume').attr('title','Mute');
  }
});

$('button.volume').on("mouseover", function(e) {
  clearTimeout(vtimeout);
  if ($('#controls .slider').css("width") != "94px") $('#controls .slider').animate({"width": "94px"}, "fast", function() { $('#controls .slider .handle').show(); });
});

$('button.volume').on("mouseleave", function(e) {
  vtimeout = setTimeout(function() {
    $('#controls .slider .handle').hide();
    $('#controls .slider').animate({"width": "0px"}, "fast");
  }, 500);
});

$('#controls .slider').on("mouseover", function(e) {
  clearTimeout(vtimeout);
  if ($('#controls .slider').css("width") != "94px") $('#controls .slider').animate({"width": "94px"}, "fast", function() { $('#controls .slider .handle').show(); });
});

$('#controls .slider').on("mouseleave", function(e) {
  vtimeout = setTimeout(function() {
    $('#controls .slider .handle').hide();
    $('#controls .slider').animate({"width": "0px"}, "fast");
  }, 500);
});

var aud_comp_audioCtx = null;
var aud_comp_source = null;
var aud_comp_compressor = null;
var aud_comp_enabled = false;
function enableAudioCompression() {
  var AudioContext = window.AudioContext || window.webkitAudioContext;
  if (aud_comp_audioCtx == null) {
    aud_comp_audioCtx = new AudioContext();
    aud_comp_source = aud_comp_audioCtx.createMediaElementSource(tlplayer);
    aud_comp_compressor = aud_comp_audioCtx.createDynamicsCompressor();
    aud_comp_compressor.threshold.value = -50;
    aud_comp_compressor.knee.value = 40;
    aud_comp_compressor.ratio.value = 12;
    aud_comp_compressor.reduction.value = -20;
    aud_comp_compressor.attack.value = 0;
    aud_comp_compressor.release.value = 0.25;
    aud_comp_source.connect(aud_comp_audioCtx.destination);
  }
  aud_comp_source.disconnect(aud_comp_audioCtx.destination);
  aud_comp_source.connect(aud_comp_compressor);
  aud_comp_compressor.connect(aud_comp_audioCtx.destination);
  aud_comp_enabled = true;
}

function disableAudioCompression() {
  if (aud_comp_audioCtx != null && aud_comp_source != null && aud_comp_compressor != null) {
    aud_comp_source.disconnect(aud_comp_compressor);
    aud_comp_compressor.disconnect(aud_comp_audioCtx.destination);
    aud_comp_source.connect(aud_comp_audioCtx.destination);
    aud_comp_enabled = false;
  }
}

/*
$('button.audiocompress').click(function(e){
  if (aud_comp_enabled) {
    disableAudioCompression();
  }
  else {
    enableAudioCompression();
  }
});
*/

$('input[id="volume-bar"]').change(function(e) {
  tlplayer.volume = $(this).val();
  volume = $(this).val();
  if ($(this).val() > 0) {
    $('button.volume').attr('title','Mute');
    if (volume <= 0.4) $('button.volume').css('background-position','-168px 0px');
    else $('button.volume').css('background-position','-224px 0px');
  }
  else {
    $('button.volume').css('background-position','-196px 0px');
    $('button.volume').attr('title','Activate sound');
  }
  userSettings["player"]["volume"] = volume;
  onSettingChange();
});

$('#progress-bar').click(function(e) {
  if (emode < 2 && editing == false) {
    var time = ((e.pageX-$(this).offset().left)/$(this).width()*tlplayer.duration).toFixed(2);
    var tr = tlplayer.seekable;
    if (tr != null) {
      for (i=0; i<tr.length; i++) {
        if (tr.end(i) > time) {
          tlplayer.currentTime = time;
          break;
        }
      }
    }
  }
});

$('#progress-bar').bind('mouseover mousemove', function(e) {
  if (tlplayer.readyState >= 1) { // metadata loaded
    var per = (e.pageX-$(this).offset().left)/$(this).width();
    var totaltime = (tlplayer.duration*per).toFixed(2);
    $('#progress-bar').attr('title', parsetime(totaltime));
  }
});

$('button.save').click(function(e) {
  $(".player-body").prepend('<div class="processing"><div id="logo" style="margin: 0 auto;"><img src="translectures/images/tllogo.png"/></div><div style="margin: 0 auto; background-color: #999; padding: 10px; margin-top: 20px;">saving changes, please wait<br/><br/><img src="translectures/images/loading-bars.gif"></div></div>');
  if (editing != false) $(editing).trigger("confirm");
  if (!tlplayer.paused) tlplayer.pause();
  setTimeout(function() {
    saveTranscriptionChanges();
  }, 250);
});

$('button.edit_history').click(function(e) {
  show_edit_history();
});

$('button.toggle-advanced-mode').click(function(e) {
  toggle_advanced_mode();
  if (wa_enabled) { 
    $(this).html('<i class="fa fa-dot-circle-o"></i> Basic mode');
  }
  else {
    $(this).html('<i class="fa fa-crop"></i> Advanced mode');
  }
});

$('button.export').click(function(e) {
  if (!tlplayer.paused) tlplayer.pause();
  bootbox.dialog({
    title: 'Export subtitles',
    message: '<div class="col-md-6" style="padding-top: 5px;">Select format:</div><div class="col-md-6"><select id="export_subs_format" class="form-control"><option value="srt">SubRip (srt)</option><option value="txt">Text (txt)</option></select></div>',
    buttons: {
      danger: {
        label: "Cancel",
        className: "btn-danger"
      },
      success: {
        label: "Download",
        className: "btn-primary",
        callback: function() {
          if ($('#export_subs_format').val() == "srt") {
            var srtText = "";
            for (var i=0; i<dfxptar['s'].length; i++) {
              srtText = srtText + (i+1) + "#@#";
              srtText = srtText + secondsToTime(dfxptar['s'][i]['b']) + " --> " + secondsToTime(dfxptar['s'][i]['e']) + "#@#";
              srtText = srtText + getSubtitleLines(dfxptar['s'][i]['text'].replace(/"/g, "&quot;").replace(/<br>/g, "#@#").replace(/\/([^\/]*)\/([^\/]*)\//gi, "$2").replace(/\s\[hesitation\]/gi, "").replace(/\[hesitation\]\s/gi, "").replace(/\[hesitation\]/gi, ""), "#@#") + "#@##@#";
            }
            var url = 'translectures/lib/dlsrt.php?id=' + request_data["id"] + "&lang=" + $('.languages-panel #tarlang-list').data("lcode");
            var dlform = $('<form id="srtdownload" style="display:none;height:0;" action="' + url + '" method="POST" target="_blank"><input type="text" name="content" value="' + srtText + '" /></form>');
            $('body').append(dlform);
            $(dlform).submit();
            $(dlform).remove();
          }
          else if ($('#export_subs_format').val() == "txt") {
            var text = "";
            for (var i=0; i<dfxptar['s'].length; i++) {
              text = text + getSubtitleLines(dfxptar['s'][i]['text'].replace(/\/([^\/]*)\/([^\/]*)\//gi, "$2").replace(/\s\[hesitation\]/gi, "").replace(/\[hesitation\]\s/gi, "").replace(/\[hesitation\]/gi, ""), " ")+'\n';
            }
            var a = window.document.createElement('a');
            a.href = window.URL.createObjectURL(new Blob([text], {type: 'text/plain'}));
            a.download = request_data["id"]+'_'+$('.languages-panel #tarlang-list').data("lcode");
            document.body.appendChild(a);
            a.click();
            document.body.removeChild(a);
          }
        }
      }
    }
  });
});

$('button.import').click(function(e) {
  $('form#import_form').show();
  $('form#import_form input#import_file').trigger("click");
});

$('input#import_file').on("change", function(e) {
  var ext = $(this).val().split('.').pop().toLowerCase();
  console.log("Importing ."+ext+" file...");
  $('form#import_form').submit();
});

$('form#import_form').on("submit", function() {
  $.ajax({
    url: $(this).attr('action'), // Url to which the request is send
    type: "POST",             // Type of request to be send, called as method
    data: new FormData(this), // Data sent to server, a set of key/value pairs (i.e. form fields and values)
    dataType: "json",
    contentType: false,       // The content type used when sending data to the server.
    cache: false,             // To unable request pages to be cached
    processData:false,        // To send DOMDocument or non processed data file it is set to false
    success: function(data) {
      if ("rcode" in data && data["rcode"] == 0) {
        if (dfxptar != null) for (var i=0; i<dfxptar['s'].length; i++) sremovedtar.push(parseInt(dfxptar['s'][i]['sI']));
        dfxptar = data;
        fillTranscriptionContainer(false);
        if (wa_enabled) genWaveformSubtitles();
        editing = false;
        wa_movingElement = false;
        wa_resizingElement = false;
      }
      else {
        if ("description" in data) bootbox.alert("The subtitles file you uploaded is not supported: "+data["description"]);

        else bootbox.alert("The subtitles file you uploaded is not supported.");
      }
    },
    error: function() {
      bootbox.alert("Sorry, there was an error importing your subtitles file.");
    }
  });
  return false;
});

$('button.help').click(function(e) {
  helpvisible = true;
  if (!tlplayer.paused) tlplayer.pause();
  $('.inhibit').css({"width": "100%", "height": "100%"});
  $('button.play').tooltipster({
    content: '<div class="shortcut-bg" style="background-position: 0px -360px; width: 60px; height: 40px;"></div> or '+keypressCodeNicer(userSettings["shortcuts"]["playpause"])+': Play/pause video',
    position: 'top-right',
    trigger:'custom'
  }).addClass("tooltipster");
  $('button.save').tooltipster({
    content: 'Save changes',
    position: 'left',
    trigger:'custom'
  }).addClass("tooltipster");
  $('button.languages').tooltipster({
    content: 'Select caption language(s)',
    position: 'bottom',
    trigger:'custom'
  }).addClass("tooltipster");
  $('button.layout').tooltipster({
    content: 'Change view mode',
    position: 'top',
    trigger:'custom'
  }).addClass("tooltipster");
  if (vmode > 0) {
    if ($('#transcription-tar li.current').length == 0) {
      $('#transcription-tar li:eq('+currents+')').addClass("current");
      if ($('#transcription-src').is(':visible')) {
        $('#transcription-src li:eq('+currents+')').addClass("current");
      }
    }
    updateScrollPos(false);
    $('#replay-button').remove();
    $('#transcription-tar li:eq('+currents+')').prepend('<div id="replay-button"></div>');
    $('#transcription-tar li.current').tooltipster({
      content: '<table class="table help-shortcuts"><tr><td><div class="shortcut-bg" style="background-position: 0px -360px; width: 60px; height: 40px;"></div> or '+keypressCodeNicer(userSettings["shortcuts"]["enterconfirm"])+'</td><td>Edit/Confirm</td></tr><tr><td>'+keypressCodeNicer(userSettings["shortcuts"]["prevsegment"])+' / '+keypressCodeNicer(userSettings["shortcuts"]["nextsegment"])+'</td><td>Previous/Next segment</td></tr><tr><td>'+keypressCodeNicer(userSettings["shortcuts"]["jumpback"])+' / '+keypressCodeNicer(userSettings["shortcuts"]["jumpforward"])+'</td><td>Jump -/+ 1,5 sec.</td></tr></table>',
      position:'bottom',
      trigger:'custom'
    }).addClass("tooltipster");
    $('#replay-button').tooltipster({
      content: '<div class="shortcut-bg" style="background-position: 0px -360px; width: 60px; height: 40px;"></div> or '+keypressCodeNicer(userSettings["shortcuts"]["replay"])+': Replay current segment',
      position:'left',
      functionAfter: function(origin) {
        if (editing == false) {
          $('#replay-button').remove();
        }
      },
      trigger:'custom'
    }).addClass("tooltipster");
  }
  else {
    $('#sbar-tar').tooltipster({
      content: '<table class="table help-shortcuts"><tr><td><div class="shortcut-bg" style="background-position: 0px -360px; width: 60px; height: 40px;"></div> or '+keypressCodeNicer(userSettings["shortcuts"]["enterconfirm"])+'</td><td>Edit/Confirm</td></tr><tr><td>'+keypressCodeNicer(userSettings["shortcuts"]["prevsegment"])+' / '+keypressCodeNicer(userSettings["shortcuts"]["nextsegment"])+'</td><td>Previous/Next segment</td></tr><tr><td>'+keypressCodeNicer(userSettings["shortcuts"]["jumpback"])+' / '+keypressCodeNicer(userSettings["shortcuts"]["jumpforward"])+'</td><td>Jump -/+ 1,5 sec.</td></tr></table>',
      position:'top',
      trigger:'custom'
    }).addClass("tooltipster");
  }
  
  $('.tooltipster').tooltipster('show');
});

$('.inhibit').click(function(e) {
  if (helpvisible) {
    $('.tooltipster').tooltipster('hide');
    $('.tooltipster').removeClass("tooltipster");
    // set button titles
    $('button.play').prop("title", "Play/Pause (Tab)");
    $('button.save').prop("title", "Save changes");
    $('button.languages').prop("title", "Select editing and reference languages");
    $('button.layout').prop("title", "Select editor layout");
    helpvisible = false;
  }
  $('.inhibit').css({"width": "0%", "height": "0%"});
});

$('button.languages').click(function(e) {
  if (editing != false) {
    $(editing).trigger("confirm");
  }
  if (!tlplayer.paused) tlplayer.pause();
  $('.tlpanel').not('.languages-panel').hide();
  $('.languages-panel').toggle();
  return false;
});

$('.languages-panel #srclang-list').change(function(e) {
  if (!tlplayer.paused) tlplayer.pause();

  var loadFunction = function() {
    var slang = $('.languages-panel #srclang-list option:selected').val();
    if (slang == "none") {
      autoEdit = false;
      dfxpsrc = {};
      $('#transcription-src').hide();
      $('#transcription-tar').css("width", "100%");
      $('#transcription-tar li').css("height", "");
      // throw tLload event
      $.event.trigger({
        type: "tLload",
        container: "src",
        lang: "none"
      });
      $('.languages-panel #srclang-list').data("lname", $('.languages-panel #srclang-list option:selected').text());
      $('.languages-panel #srclang-list').data("lcode", $('.languages-panel #srclang-list option:selected').val());
      if (typeof editing == 'undefined' || editing == null || (editing != false && $(editing).is(':hidden'))) editing = false;
    }
    else if (slang != $('.languages-panel #tarlang-list option:selected').val()) {
      // source language != target language
      loadFlag = 1;
      $('#transcription-src').css("display", "inline-block");
      $('#transcription-tar').css("width", "49%");
      loadTranscription(slang, true, (mod_session != false && "session_id" in mod_session ? mod_session["session_id"] : null));
      //autoEdit = true;
    }
    else {
      // source language == target language
      bootbox.alert("If you want to have " + $('.languages-panel #srclang-list option:selected').text() + " as reference language you must first select a different editing language.");
      $('.languages-panel #srclang-list option[value="'+$('.languages-panel #srclang-list').data("lcode")+'"]').prop("selected", true);
    }
    updateScrollPos(false, true);
  }
  if (e.originalEvent) loadFunction();
});

function setLanguageSelect(lcode, src) {
  if (src) {
    $('.languages-panel #srclang-list').val(lcode);
    $('.languages-panel #srclang-list').data("lname", $('.languages-panel #srclang-list option:selected').text());
    $('.languages-panel #srclang-list').data("lcode", $('.languages-panel #srclang-list option:selected').val());
  }
  else {
    $('.languages-panel #tarlang-list').val(lcode);
    $('.languages-panel #tarlang-list').data("lname", $('.languages-panel #tarlang-list option:selected').text());
    $('.languages-panel #tarlang-list').data("lcode", $('.languages-panel #tarlang-list option:selected').val());
  }
}

$('.languages-panel #tarlang-list').change(function(e) {
  if (!tlplayer.paused) tlplayer.pause();

  var loadFunction = function() {
    var slang = $('.languages-panel #tarlang-list option:selected').val();
    if (slang != $('.languages-panel #srclang-list option:selected').val()) {
      if ($('.languages-panel #srclang-list option:selected').val() != "none") {
        // source language != target language
        $('#transcription-src').css("display", "inline-block");
        $('#transcription-tar').css("width", "49%");
      }
      loadFlag = 1;
      loadTranscription(slang, false, (mod_session != false && "session_id" in mod_session ? mod_session["session_id"] : null));
    }
    else {
      // source language == target language
      dfxptar = dfxpsrc;
      dfxpsrc = null;
      $('#transcription-tar').empty();
      $('#transcription-src li').clone().appendTo($('#transcription-tar'));
      $('#transcription-src').empty().hide();
      $('.languages-panel #srclang-list option[value="none"]').prop("selected", true);
      $('#transcription-tar').css("width", "100%");
      $('#transcription-tar li').css("height", "");
      // throw tLload event twice
      $.event.trigger({
        type: "tLload",
        container: "src",
        lang: "none"
      });
      $.event.trigger({
        type: "tLload",
        container: "tar",
        lang: slang
      });
      $('.languages-panel #tarlang-list').data("lname", $('.languages-panel #tarlang-list option:selected').text());
      $('.languages-panel #tarlang-list').data("lcode", $('.languages-panel #tarlang-list option:selected').val());
      if (typeof editing == 'undefined' || editing == null || (editing != false && $(editing).is(':hidden'))) editing = false;
    }
    updateScrollPos(false, false);
  }
  if (e.originalEvent) {
    loadFunction();
  }
});

function check_unsaved_changes(src) {
  var changes = false;
  if (typeof src == 'undefined' || src) {
    if (sremovedsrc.length > 0) return true;
    else if (typeof dfxpsrc != 'undefined' && dfxpsrc != null && "s" in dfxpsrc) {
      $.each(dfxpsrc['s'], function(key, val) {
        if ("edited" in val && val['edited'] != "none") {
          changes = true;
          return false;
        }
      });
    }
  }
  if (typeof src == 'undefined' || !src) {
    if (sremovedtar.length > 0) return true;
    else if (typeof dfxptar != 'undefined' && dfxptar != null && "s" in dfxptar) {
      var changes = false;
      $.each(dfxptar['s'], function(key, val) {
        if ("edited" in val && val['edited'] != "none") {
          changes = true;
          return false;
        }
      });
    }
  }
  return changes;
}

function changeViewMode(vsel) {
  if (vmode != vsel) {
    vmode = vsel;
    userSettings["player"]["vmode"] = vmode;
    onSettingChange();
    switch(vmode) {
      case 0: // over-layout
        if (typeof dfxpsrc != 'undefined' && dfxpsrc != null && "s" in dfxpsrc) $('#sbar-src').html(getSubtitleLines(dfxpsrc['s'][currents]['text'], "<br/>"));
        if (typeof dfxptar != 'undefined' && dfxptar != null && "s" in dfxptar) $('#sbar-tar').html(getSubtitleLines(dfxptar['s'][currents]['text'], "<br/>"));
        $('.subtitle-bar').show();
        $('#tl-editor').hide();
        $('video').show();
        $('#tl-player').css({"float": "none", "width": "66%"});
        break;
      case 1: // horizontal-layout
        $('#tl-player').after($('#tl-editor'));
        $('#tl-editor').show();
        $('video').show();
        $('#sbar-tar').show();
        $('#sbar-src').hide();
        $('#tl-player').css({"float": "left", "width": "52%"});
        $('#tl-editor').css({"float": "right", "width": "46%"});
        if (emode == 1) $('.subtitle-bar').hide();
        break;
      case 2: // vertical-layout
        $('#tl-player').after($('#tl-editor'));
        $('#tl-editor').show();
        $('video').show();
        $('#sbar-tar').show();
        $('#sbar-src').hide();
        $('#tl-player').css({"float": "none", "width": "52%"});
        $('#tl-editor').css({"float": "none"});
        if (emode == 1) $('.subtitle-bar').hide();
        break;
      case 3: // text layout
        $('video').hide();
        $('.subtitle-bar').hide();
        $('#tl-editor').insertBefore($('#tl-player'));
        $('#tl-player').css({"float": "none", "width": "90%"});
        $('#tl-editor').css({"float": "none", "width": "90%"});
        $('.transcription-container').css("height", $('video').css('height'));
        if ($('.transcription-container').height() > $(window).height()/1.5) $('.transcription-container').height($(window).height()/1.5);
        $('#tl-editor').show();
      default:
        break;
    }
    // unset editing object if hidden
    if (editing != false && $(editing).is(':hidden')) editing = false;
    onResize();
    updateScrollPos();
  }
}

$('.layouts-list li').click(function(e) {
  var vsel = $('.layouts-list li').index($(this));
  changeViewMode(vsel);
  $('.layouts-panel').hide();
});

/* 
  EDITOR CODE 
*/

function isLanguageAvailable(lcode) {
  var available = false;
  $.each(langs['langs'], function(key, value) {
    if (value["lang_code"] == lcode) {
      available = true;
      return false;
    }
  });
  return available;
}

function getAvailableLanguages() {
  var langs_url = "translectures/lib/langs.php?id="+request_data["id"];
  if ("hash" in request_data) langs_url = langs_url+"&hash="+request_data["hash"];
  if ("api_user" in request_data) langs_url = langs_url+"&user="+request_data["api_user"];
  if ("expire" in request_data) langs_url = langs_url+"&expire="+request_data["expire"];
  if ("request_key" in request_data) langs_url = langs_url+"&auth_token="+request_data["request_key"];
  $.ajax({
    url: langs_url,
    dataType: "json",
    cache: false,
    async: false,
    success: function(rdata) {
      if (rdata != null && "rcode" in rdata && rdata["rcode"] == 0) {
        langs = rdata;
        lavailable = false;
        if ("subs_locked" in rdata && rdata["subs_locked"] && request_data["author_conf"] != 100) {
          $('button.save').hide();
          bootbox.alert("The author has disabled user revisions for this media. You will not be able to submit any modifications.");
        }
        else if (request_data["request_key"].length == 80) {
          start_mod_session();
        }
        if ("lang" in request_data && request_data["lang"] != null && request_data["lang"] != "") {
          lavailable = isLanguageAvailable(request_data["lang"]);
        }
        $.each(langs['langs'], function(key, value) {
          if (lavailable == false) {
            // if current language == media language
            if (value["lang_code"] == langs["media_lang"]) {
              $('.languages-panel #srclang-list option[value="none"]').prop("selected", true);
              $('.languages-panel #srclang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
              $('.languages-panel #tarlang-list').append('<option value="'+value['lang_code']+'" selected>'+value['lang_name']+'</option>');
            }
            else {
              $('.languages-panel #srclang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
              $('.languages-panel #tarlang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
            }
          }
          else {
            // if current language == media language
            if (value["lang_code"] == langs["media_lang"]) {
              // if current language == user selected language
              if (value["lang_code"] == request_data["lang"]) {
                $('.languages-panel #srclang-list option[value="none"]').prop("selected", true);
                $('.languages-panel #srclang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
                $('.languages-panel #tarlang-list').append('<option value="'+value['lang_code']+'" selected>'+value['lang_name']+'</option>');
              }
              else {
                $('.languages-panel #srclang-list').append('<option value="'+value['lang_code']+'" selected>'+value['lang_name']+'</option>');
                $('.languages-panel #tarlang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
              }
            }
            // if current language != media language
            else {
              if (value["lang_code"] == request_data["lang"]) {
                $('.languages-panel #srclang-list').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
                $('.languages-panel #tarlang-list').append('<option value="'+value['lang_code']+'" selected>'+value['lang_name']+'</option>');
              }
              else $('.languages-panel select').append('<option value="'+value['lang_code']+'">'+value['lang_name']+'</option>');
            }
          }
        });
        if (!lavailable) {
          if ("lang" in request_data && request_data["lang"] != "") {
            bootbox.alert("Transcriptions with language code \""+request_data["lang"]+"\" are not available for this media.");
          }
          $('.languages-panel #srclang-list option[value="none"]').prop("selected", true);
          // setting media language as default
          request_data["lang"] = langs["media_lang"];
        }
        $('#srclang-list').data("lname", $('#srclang-list option:selected').text());
        $('#srclang-list').data("lcode", $('#srclang-list option:selected').val());
        $('#tarlang-list').data("lname", $('#tarlang-list option:selected').text());
        $('#tarlang-list').data("lcode", $('#tarlang-list option:selected').val());
        // show language code(s) on languages button
        // if ($('.languages-panel #srclang-list option:selected').val() != "none") $('button.languages').html($('.languages-panel #srclang-list option:selected').val() + "<br/>" + $('.languages-panel #tarlang-list option:selected').val());
        // else $('button.languages').html($('.languages-panel #tarlang-list option:selected').val());
        addTtsLanguages();
      }
      else {
        console.log("Error retrieving available languages or invalid format ("+langs_url+")");
        $('.processing').remove();
        bootbox.alert("There are no transcriptions available for this media.");
        request_data["lang"] = null;
	// hide editor buttons
	$('button.layout').hide();
        $('button.save').hide();
        $('button.config').hide();
        $('button.ii').hide();
        $('button.languages').hide();
        $('button.help').hide();
        $('button.options').hide();
        // hide transcription editor
        $('#tl-editor').hide();
        $('#tl-player').css('float', 'none');
      }
    },
    error: function(e) {
      $('.processing').remove();
      console.log("Error checking available languages ("+langs_url+")");
      bootbox.alert("Couldn't retrieve available transcriptions for this media.");
      request_data["lang"] = null;
      $('button.layout').hide();
      $('button.save').hide();
      $('button.config').hide();
      $('button.ii').hide();
      $('button.languages').hide();
      $('button.help').hide();
      $('button.options').hide();
      // hide transcription editor
      $('#tl-editor').hide();
      $('#tl-player').css('float', 'none');
    }
  });
}

function markRevised(session_id) {
  var markurl = "translectures/lib/mark_revised.php?id="+request_data["id"]+"&session_id="+session_id+"&author_id="+request_data["author_id"]+"&author_conf="+request_data["author_conf"]+"&revision_session_id="+mod_session["session_id"];
  if ("hash" in request_data) markurl = markurl+"&hash="+request_data["hash"];
  if ("author_name" in request_data) markurl = markurl+"&author_name="+request_data["author_name"];
  if ("api_user" in request_data) markurl = markurl+"&user="+request_data["api_user"];
  if ("expire" in request_data) markurl = markurl+"&expire="+request_data["expire"];
  if ("request_key" in request_data) markurl = markurl+"&auth_token="+request_data["request_key"];
  $.ajax({
    url: markurl,
    dataType: "json",
    cache: false,
    async: true,
    success: function(data) {
    },
    error: function() {
      console.log("Could not mark session id "+session_id+" as revised");
    }
  });
}

function loadTranscription(lang, src, session_id, sel_data_policy, is_revision) {
  src = typeof src !== 'undefined' ? src : true;
  is_revision = typeof is_revision !== 'undefined' ? is_revision : false;
  session_id = typeof session_id !== 'undefined' ? session_id : null;
  if (check_unsaved_changes(src) == true) {
    var langName = (src ? $('#srclang-list').data("lname") : $('#tarlang-list').data("lname"));
    bootbox.confirm("You have not saved your <strong>"+langName.toLowerCase()+"</strong> transcription changes. If you continue, all your changes will be discarded. Are you sure?", function(resp) {
      if (resp) {
        setLanguageSelect(lang, src);
        loadTranscriptionOK(lang, src, session_id, sel_data_policy, is_revision);
      }
      else {
        if (src) setLanguageSelect($('#srclang-list').data("lcode"), src);
        else setLanguageSelect($('#tarlang-list').data("lcode"), src);
      }
    });
  }
  else {
    setLanguageSelect(lang, src);
    loadTranscriptionOK(lang, src, session_id, sel_data_policy, is_revision);
  }
}

function loadTranscriptionOK(lang, src, session_id, sel_data_policy, is_revision) {
  src = typeof src !== 'undefined' ? src : true;
  is_revision = typeof is_revision !== 'undefined' ? is_revision : false;
  session_id = typeof session_id !== 'undefined' ? session_id : null;
  console.log("Loading \"" + lang + "\" transcription for video id \"" + request_data["id"] + "\"");
  loadurl = "translectures/lib/subs.php?id="+request_data["id"]+"&lang="+lang;
  if ("hash" in request_data) loadurl = loadurl+"&hash="+request_data["hash"];
  if (session_id != null) {
    loadurl = loadurl+"&session_id="+session_id;
    if (is_revision) {
      markRevised(session_id);
      reviewing_session_id = session_id;
    }
  }
  else if ("session_id" in request_data) {
    loadurl = loadurl+"&session_id="+request_data["session_id"];
    markRevised(request_data["session_id"]);
    reviewing_session_id = request_data["session_id"];
  }
  else if (mod_session != false) {
    loadurl = loadurl+"&session_id="+mod_session["session_id"];
    reviewing_session_id = false;
  }
  else reviewing_session_id = false;
  if ("api_user" in request_data) loadurl = loadurl+"&user="+request_data["api_user"];
  if ("expire" in request_data) loadurl = loadurl+"&expire="+request_data["expire"];
  if ("request_key" in request_data) loadurl = loadurl+"&auth_token="+request_data["request_key"];
  if (typeof sel_data_policy !== 'undefined' && sel_data_policy != null) loadurl = loadurl+"&sel_data_policy="+sel_data_policy;
  else if ("sel_data_policy" in request_data) loadurl = loadurl+"&sel_data_policy="+request_data["sel_data_policy"];
  if ("seg_filt_policy" in request_data) loadurl = loadurl+"&seg_filt_policy="+request_data["seg_filt_policy"];
  if (typeof editing == 'undefined' || editing == null || (editing != false && $(editing).is(':hidden'))) editing = false;
  setLanguageSelect(lang, src);
  $(".player-body").prepend('<div class="processing"><div id="logo" style="margin: 0 auto;"><img src="translectures/images/tllogo.png"/></div><div style="margin: 0 auto; background-color: #999; padding: 10px; margin-top: 20px;">loading, please wait<br/><br/><img src="translectures/images/loading-bars.gif"></div></div>');
  $.ajax({
    url: loadurl,
    dataType: "json",
    cache: false,
    async: true,
    success: function(rdata) {
      var emsg = false;
      if ("rcode" in rdata) {
        if (rdata["rcode"] == 0) {
          if (rdata['s'].length > 0) {
            if (src) dfxpsrc = rdata;
            else dfxptar = rdata;
            fillTranscriptionContainer(src);
            if (!src) genWaveformSubtitles();
            // throw tLload event
            $.event.trigger({
              type: "tLload",
              container: (src == true) ? "src" : "tar",
              lang: lang
            });
          }
          else {
            emsg = "Error loading "+getLanguageName(lang)+" transcription: transcription invalid or empty";
            console.log("Error loading \""+lang+"\" transcription: transcription is invalid or empty"); 
          }
        }
        else {
          emsg = "There was an error loading "+getLanguageName(lang)+" transcription: " + rdata["rcode_description"];
          console.log("Error loading \"" + lang + "\" transcription: " + rdata["rcode_description"]);
        }
      }
      else {
        emsg = "There was an error loading "+getLanguageName(lang)+" transcription: Web Service unavailable";
        console.log("Error loading \"" + lang + "\" transcription: unexpected response from Web Service ("+JSON.stringify(rdata)+")");
      }
      if (emsg != false) {
        $('.processing').remove();
        bootbox.alert(emsg);
        if (src) {
          $('#transcription-src').empty();
        }
        else {
          $('#transcription-tar').empty();
        }
      }
    },
    error: function(e) {
      console.log("Couldn't retrieve \"" + lang + "\" transcription for video \"" + request_data["id"] + "\"");
      bootbox.alert("There was an error loading "+getLanguageName(lang)+" transcription for this media.");
      $('.processing').remove();
      if (src) {
        $('#transcription-src').empty();
      }
      else {
        $('#transcription-tar').empty();
      }
    }
  });
}

function fillTranscriptionContainer(src) {
  if (editing != false) $(editing).trigger("confirm");
  if (src) {
    var trans = dfxpsrc;
    var tcontainer = $('#transcription-src');
    $('#transcription-src').empty();
    smaxsrc = 0;
  }
  else {
    var trans = dfxptar;
    var tcontainer = $('#transcription-tar');
    $('#transcription-tar').empty();
    smaxtar = 0;
  }
  var author_name = false;
  if (reviewing_session_id != false) {
    var author_name = get_revision_author_name(reviewing_session_id);
    if (author_name.length == 0) author_name = false;
  }
  var global_text = "";
  $.each(trans['s'], function(skey, sval) {
    if (src && smaxsrc < parseInt(sval["sI"])) smaxsrc = parseInt(sval["sI"]);
    else if (!src && smaxtar < parseInt(sval["sI"])) smaxtar = parseInt(sval["sI"]);
    if ("d" in sval && parseInt(sval["d"]) == 1) {
      if (src) sremovedsrc.push(parseInt(sval["sI"]));
      else sremovedtar.push(parseInt(sval["sI"]));
      return true;
    }
    if (emode != 1) tcontainer.append('<li data-id="'+sval["sI"]+'"><div class="segment-text" contenteditable="false"></div></li>');
    // if no word(s) label(s) inside segment
    var stext = "";
    if (!("w" in sval) || sval['w'] == null || sval['w'].length == 0) stext = sval["text"];
    else {
      $.each(sval['w'], function(wkey, wval) {
        stext += wval["text"];
      });
    }
    if (emode == 1) {
      global_text += stext+' ';
    }
    else if ("aT" in sval && sval['aT'] != null && sval['aT'] == "human") {
      if ("h" in sval) {
        tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').append('<span id="w0" class="word-text manual-conf hl">'+sval["text"]+'</span>');
        if (author_name != false) tcontainer.children('li[data-id="'+sval["sI"]+'"]').append('<span class="rev_author"><i class=\"fa fa-user\"></i> '+author_name+'</span>');
        trans['s'][skey]["edited"] = "full";
      }
      else tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').append('<span id="w0" class="word-text manual-conf">'+sval["text"]+'</span>');
    }
    else if ("cM" in sval && sval['cM'] != null && sval['cM'] < confThreshold) tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').append('<span id="w0" class="word-text low-conf">'+sval["text"]+'</span>');
    else tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').append('<span id="w0" class="word-text">'+sval["text"]+'</span>');
    /*
    if (!("w" in sval) || sval['w'] == null || sval['w'].length == 0) {
      stext = sval["text"];
      if ("aT" in sval && sval['aT'] != null && sval['aT'] == "human") tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w0" class="word-text manual-conf">'+sval["text"]+'</span>');
      else if ("cM" in sval && sval['cM'] != null && sval['cM'] < confThreshold) tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w0" class="word-text low-conf">'+sval["text"]+'</span>');
      else tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w0" class="word-text">'+sval["text"]+'</span>');
    }
    else {
      $.each(sval['w'], function(wkey, wval) {
        stext += wval["text"];
        if ("g" in wval && wval['g'] != null && wval['g'] == "1") tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w'+wkey+'" class="word-text manual-conf">'+wval["text"]+'</span> ');
        if ("cM" in wval && wval['cM'] != null && wval['cM'] < confThreshold) tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w'+wkey+'" class="word-text low-conf">'+wval["text"]+'</span> ');
        else tcontainer.children('li:eq('+skey+')').children('.segment-text').append('<span id="w'+wkey+'" class="word-text">'+wval["text"]+'</span> ');
      });
    }
    */
    if (emode != 1 && wa_enabled) {
      var scps = stext.replace(/\s+/g, '').length/(sval["e"]-sval["b"]);
      if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) tcontainer.children('li[data-id="'+sval["sI"]+'"]').append('<span class="cpsmeter cpsred">'+scps.toFixed(1)+' cps</span>');
      else tcontainer.children('li[data-id="'+sval["sI"]+'"]').append('<span class="cpsmeter cpsgreen">'+scps.toFixed(1)+' cps</span>');
    }
  });
  smaxtar++;
  smaxsrc++;
  if (emode == 1) {
    tcontainer.append('<textarea class="fa-text form-control" placeholder="Write here what is said in the video">'+global_text+'</textarea>');
  }
  if (showLowConfWords == false) $('.low-conf').css('color','#000000');
}

function emptyTranscriptionContainer(src) {
  if (src) {
    var trans = dfxpsrc;
    var tcontainer = $('#transcription-src');
  }
  else {
    var trans = dfxptar;
    var tcontainer = $('#transcription-tar');
  }
  $.each(trans['s'], function(skey, sval) {
    if ("aT" in sval && sval['aT'] != null && sval['aT'] != "human") {
      tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').empty();
      tcontainer.children('li[data-id="'+sval["sI"]+'"]').children('.segment-text').append('<span id="w0" class="word-text"></span>');
    }
  });
}

function saveTranscriptionChanges() {
  if (mod_session == false || !("rcode" in mod_session) || mod_session["rcode"] != 0) {
    bootbox.alert("No editing session started. You will not be able to save any changes.");
    return;
  }
  var mods = {};
  stats_activesystem();

  if (!($('#srclang-list').data("lcode") == "none") && (dfxpsrc != null) && ("s" in dfxpsrc) && (dfxpsrc['s'].length > 0)) {
    var srcmod = {"txt": [], "del": sremovedsrc, "rev_time": parseInt(stats_src_total_time)};
    $.each(dfxpsrc['s'], function(key, val) {
      smod = {};
      smod['sI'] = parseInt(val['sI']);
      smod['b'] = parseFloat(val['b']);
      smod['e'] = parseFloat(val['e']);
      if ("edited" in val && val['edited'] != "none") {
        smod['t'] = val['text'];
        srcmod['txt'].push(smod);
      }
    });
    if (srcmod['txt'].length > 0 || srcmod['del'].length > 0) {
      mods[$('#srclang-list').data("lcode")] = srcmod;
    }
  }
  if (dfxptar != null && "s" in dfxptar && dfxptar['s'].length > 0) {
    var tarmod = {"txt": [], "del": sremovedtar, "rev_time": parseInt(stats_tar_total_time)};
    $.each(dfxptar['s'], function(key, val) {
      smod = {};
      smod['sI'] = parseInt(val['sI']);
      smod['b'] = parseFloat(val['b']);
      smod['e'] = parseFloat(val['e']);
      if ("edited" in val && val['edited'] != "none") {
        smod['t'] = val['text'];
        tarmod['txt'].push(smod);
      }
    });
    if (tarmod['txt'].length > 0 || tarmod['del'].length > 0) {
      mods[$('#tarlang-list').data("lcode")] = tarmod;
    }
  }

  if (jQuery.isEmptyObject(mods)) {
    bootbox.alert("There are no changes to save.");
    $('.processing').remove();
  }
  else {
    var mod_url = "translectures/lib/mod.php?id="+request_data["id"]+"&session_id="+mod_session["session_id"]+"&author_id="+request_data["author_id"]+"&author_conf="+parseInt(request_data["author_conf"]);
    if ("author_name" in request_data) mod_url = mod_url+"&author_name="+request_data["author_name"];
    if ("hash" in request_data) mod_url = mod_url+"&hash="+request_data["hash"];
    if ("api_user" in request_data) mod_url = mod_url+"&user="+request_data["api_user"];
    if ("expire" in request_data) mod_url = mod_url+"&expire="+request_data["expire"];
    if ("request_key" in request_data) mod_url = mod_url+"&auth_token="+request_data["request_key"];
    if ("sel_data_policy" in request_data) mod_url = mod_url+"&sel_data_policy="+request_data["sel_data_policy"];
    $.ajax({
      type: 'POST',
      url: mod_url,
      data: { data: JSON.stringify(mods) },
      dataType: 'json',
      async: true,
      success: function(response) {
        var langsString = "";
        var index = 0;
        $.each(mods, function(key, value) {
          if (index == 0) langsString += "<span style=\"font-weight: 600;\">"+getLanguageName(key)+"</span>";
          else if (index != mods.length-1) langsString += ", <span style=\"font-weight: 600;\">"+getLanguageName(key)+"</span>";
          else langsString += " and <span style=\"font-weight: 600;\">"+getLanguageName(key)+"</span>";
          index++;
        });
        $('.processing').remove();
        if (response != false && "rcode" in response && response['rcode'] == 0) {
          if ($('#srclang-list').data("lcode") in mods) resetTranscriptionChanges(dfxpsrc, true);
          if ($('#tarlang-list').data("lcode") in mods) resetTranscriptionChanges(dfxptar, false);
          if (request_data["author_conf"] == 100) bootbox.alert(langsString+" changes successfully submitted.");
          else bootbox.alert(langsString+" changes successfully submitted. The author of this media must now approve your changes before they become available to everyone.");
          if ($('#tarlang-list').data("lcode") == langs["media_lang"] && $('#tarlang-list').data("lcode") in mods) savedTranscriptionInSession = true;
        }
        else {
          if (response != false && "rcode_description" in response) bootbox.alert("There was an error saving "+langsString+" subtitle changes: "+response['rcode_description']);
          else bootbox.alert("There was an error saving "+langsString+" subtitle changes. If the problem persists, you can download the modified subtitles in SRT format and import them later in a future session.");
        }
        if ($('#srclang-list').data("lcode") in mods) {
          $.event.trigger({
            type: "tLsave",
            container: "src"
          });
        }
        if ($('#tarlang-list').data("lcode") in mods) {
          $.event.trigger({
            type: "tLsave",
            container: "tar"
          });
        }
      },
      error: function() {
        var langsString = "";
        var index = 0;
        $.each(mods, function(key, value) {
          if (index == 0) langsString += getLanguageName(key);
          else if (index != mods.length-1) langsString += ", "+getLanguageName(key);
          else langsString += " and "+getLanguageName(key);
          index++;
        });
        $('.processing').remove();
        bootbox.alert("There was an error saving "+langsString+" subtitle changes. If the problem persists, you can download the modified subtitles in SRT format and import them later in a future session.");
        $.event.trigger({
          type: "tLsave"
        });
      }
    });
  }
}

function resetTranscriptionChanges(dfxpvar, src) {
  $.each(dfxpvar['s'], function(key, val) {
    val['edited'] = "none";
  });
  if (src) sremovedsrc = [];
  else sremovedtar = [];
}

function toggleWordColors() {
  if ($('.low-conf').css("color") == "#333") $('.low-conf').css("color", "#C80000");
  else $('.low-conf').css("color", "#333");
}

function updateScrollPos(animate, src) {
  if (emode == 0) {
    if (typeof animate == 'undefined') animate = true;
    if (typeof src == 'undefined') src = false;
    if ($('.transcription-container').length > 0 && loadFlag == 0) {
      if (src && dfxpsrc != null && "s" in dfxpsrc && dfxpsrc["s"].length > currents_src) {
        $('#transcription-src').stop();
        var yoffset = ($('#transcription-src').scrollTop() + $('#transcription-src li:eq('+currents_src+')').position().top - ($('#transcription-src').height()/2.5));
        if (animate) {
          if (Math.abs(yoffset - ($('#transcription-src').scrollTop())) < 150) $('#transcription-src').animate({ scrollTop: yoffset }, "slow");
          else $('#transcription-src').scrollTop(yoffset);
        }
        else $('#transcription-src').scrollTop(yoffset);
      }
      else if (dfxptar != null && "s" in dfxptar) {
        $('#transcription-tar').stop();
        var yoffset = ($('#transcription-tar').scrollTop() + $('#transcription-tar li:eq('+currents+')').position().top - ($('#transcription-tar').height()/2.5));
        $('#transcription-tar').stop();
        if (animate) {
          if (Math.abs(yoffset - ($('#transcription-tar').scrollTop())) < 150) $('#transcription-tar').animate({ scrollTop: yoffset }, "slow");
          else $('#transcription-tar').scrollTop(yoffset);
        }
        else $('#transcription-tar').scrollTop(yoffset);
      }
    }
  }
}

function autoConfirmPreviousSegment() {
  // auto-confirm
  if (((currents == (dfxptar['s'].length - 1)) || (tlplayer.currentTime < parseFloat(dfxptar['s'][currents+1]['e']))) && (dfxptar['s'][currents]['aT'] == "automatic" && (!("edited" in dfxptar['s'][currents]) || dfxptar['s'][currents]['edited'] == "none"))) {
    dfxptar['s'][currents]['edited'] = "full";
    dfxptar['s'][currents]['aT'] = "human";
    $('#transcription-tar li:eq('+currents+') .segment-text').empty().append('<span id="w0" class="word-text manual-conf">'+dfxptar['s'][currents]['text']+'</span>');
    // throw tLconfirm event
    $.event.trigger({
      type: "tLconfirm",
      container: "tar",
      segment: currents,
    });
  }
}

var min_chars_per_line = 24;
var max_chars_per_subtitle = 100;
function getSubtitleLines(text, lineBreak) {
  if (typeof lineBreak === undefined) lineBreak = "<br/>";
  // if (lineBreak == "<br/>") return text; // trick to disable this function
  if (text.length <= min_chars_per_line || text.length > max_chars_per_subtitle) return text;
  var line1 = "";
  var line2 = "";
  var mindiff = text.length;
  words = text.split(" ");
  for (var i=0; i<words.length; i++) {
    diff = Math.abs(words.slice(0,i).join(" ").length-words.slice(i,words.length).join(" ").length);
    if (diff < mindiff) {
      mindiff = diff;
      line1 = words.slice(0,i).join(" ") + lineBreak;
      line2 = words.slice(i,words.length).join(" ");
    }
  }
  return line1+line2;
}

var translationAlignmentOffset = 0.2; // seconds
function updateTranslationAlignment() {
  if (showTranslationAlignments && dfxpsrc != null && $('#transcription-src').is(':visible') && $('#transcription-tar').is(':visible')) {
    /* from source to target */
    $('#transcription-tar li.aligned').removeClass("aligned");
    var srcb = parseFloat(dfxpsrc['s'][currents_src]['b']);
    var srce = parseFloat(dfxpsrc['s'][currents_src]['e']);
    for (var i=0; i<dfxptar['s'].length; i++) {
      if (parseFloat(dfxptar['s'][i]['e']) >= (srcb+translationAlignmentOffset) && parseFloat(dfxptar['s'][i]['b']) < (srce-translationAlignmentOffset)) {
        $('#transcription-tar li:eq('+i+')').addClass("aligned");
      }
    }
    /* from target to source */
    $('#transcription-src li.aligned').removeClass("aligned");
    var tarb = parseFloat(dfxptar['s'][currents]['b']);
    var tare = parseFloat(dfxptar['s'][currents]['e']);
    for (var i=0; i<dfxpsrc['s'].length; i++) {
      if (parseFloat(dfxpsrc['s'][i]['e']) >= (tarb+translationAlignmentOffset) && parseFloat(dfxpsrc['s'][i]['b']) < (tare-translationAlignmentOffset)) {
        $('#transcription-src li:eq('+i+')').addClass("aligned");
      }
    }
  }
}

function updateCurrentSubtitle() {
  var change = false;
  // search for subtitle corresponding to video time (target lang)
  if (emode != 1 && dfxptar != null && "s" in dfxptar && dfxptar['s'].length > 0) {
    if (editing == false) {
      if (tlplayer.currentTime >= parseFloat(dfxptar['s'][currents]['b']) && tlplayer.currentTime < parseFloat(dfxptar['s'][currents]['e']) && !$('#transcription-tar li:eq('+currents+')').hasClass('current')) {
        change = true;
      }
      else if (tlplayer.currentTime > parseFloat(dfxptar['s'][currents]['e'])) {
        $('#transcription-tar li.current').removeClass("current");
        $('#transcription-src li.aligned').removeClass("aligned");
        $('#sbar-tar').html("");
        if (wa_enabled) $('.tl-editor-controls-advanced input').val('');
        if (autoconfirm) autoConfirmPreviousSegment();
        for (var i=currents+1; i<dfxptar['s'].length; i++) {
          if (tlplayer.currentTime > parseFloat(dfxptar['s'][i]['e'])) continue;
          // move video to the beggining of next segment if distance > 5s
          if (false && tlplayer.currentTime < parseFloat(dfxptar['s'][i]['b'])-5) {
            change = true;
            tlplayer.currentTime = parseFloat(dfxptar['s'][i]['b'])+(1/frameRate);
            currents = i;
          }
          else if (tlplayer.currentTime >= parseFloat(dfxptar['s'][i]['b'])) {
            change = true;
            currents = i;
          }
          else if (i > currents+1) {
            currents = i-1;
            updateScrollPos(true, false);
          }
          break;
        }
      }
      else if (tlplayer.currentTime < parseFloat(dfxptar['s'][currents]['b'])) {
        $('#transcription-tar li.current').removeClass("current");
        $('#transcription-src li.aligned').removeClass("aligned");
        $('#sbar-tar').html("");
        if (wa_enabled) $('.tl-editor-controls-advanced input').val('');
        for (var i=currents-1; i>=0; i--) {
          if (tlplayer.currentTime < parseFloat(dfxptar['s'][i]['b'])) continue;
          // move video to the beggining of next segment if distance > 5s
          if (false && tlplayer.currentTime < parseFloat(dfxptar['s'][i]['b'])-5) {
            change = true;
            tlplayer.currentTime = parseFloat(dfxptar['s'][i]['b'])+(1/frameRate);
            currents = i;
          }
          else if (tlplayer.currentTime >= parseFloat(dfxptar['s'][i]['b'])) {
            change = true;
            currents = i;
          }
          else if (i < currents-1) {
            if (i > 0) currents = i-1;
            else currents = 0;
            updateScrollPos(true, false);
          }
          break;
        }
      }
      if (change) {
        update_session_status();
        updateScrollPos(true, false);
        updateTranslationAlignment();
        $('#transcription-tar li:eq('+currents+')').addClass("current");
	      // start editing next segment if it is from an automatic transcription
        if (autoEdit && ("aT" in dfxptar["s"][currents] && dfxptar["s"][currents]['aT'] != null && dfxptar["s"][currents]['aT'] != "human")) $('#transcription-tar li:eq('+currents+') .segment-text').trigger("click", false);
        $('#sbar-tar').html(getSubtitleLines(dfxptar['s'][currents]['text'], "<br/>"));
        if (wa_enabled) {
          updateSegmentTimeControl();
        }
      }
    }
    else if ($(editing).parents('#transcription-tar').length > 0) {
      if (tlplayer.currentTime > parseFloat(dfxptar['s'][currents]['e'])) tlplayer.pause();
    }
  }
  change = false;
  // search for subtitle corresponding to video time (source lang)
  if (emode != 1 && dfxpsrc != null && "s" in dfxpsrc && dfxpsrc['s'].length > 0) {
    if (editing == false) {
      if (tlplayer.currentTime >= parseFloat(dfxpsrc['s'][currents_src]['b']) && tlplayer.currentTime < parseFloat(dfxpsrc['s'][currents_src]['e']) && !$('#transcription-src li:eq('+currents_src+')').hasClass('current')) {
        change = true;
      }
      else if (tlplayer.currentTime > parseFloat(dfxpsrc['s'][currents_src]['e'])) {
        $('#transcription-src li.current').removeClass("current");
        $('#transcription-tar li.aligned').removeClass("aligned");
        $('#sbar-src').html("");
        for (var i=currents_src+1; i<dfxpsrc['s'].length; i++) {
          if (tlplayer.currentTime > parseFloat(dfxpsrc['s'][i]['e'])) continue;
          else if (tlplayer.currentTime >= parseFloat(dfxpsrc['s'][i]['b'])) {
            change = true;
            currents_src = i;
          }
          else if (i > currents_src+1) {
            currents_src = i-1;
            updateScrollPos(true, true);
          }
          break;
        }
      }
      else if (tlplayer.currentTime < parseFloat(dfxpsrc['s'][currents_src]['b'])) {
        $('#transcription-src li.current').removeClass("current");
        $('#transcription-tar li.aligned').removeClass("aligned");
        $('#sbar-src').html("");
        for (var i=currents_src-1; i>=0; i--) {
          if (tlplayer.currentTime < parseFloat(dfxpsrc['s'][i]['b'])) continue;
          else if (tlplayer.currentTime >= parseFloat(dfxpsrc['s'][i]['b'])) {
            change = true;
            currents_src = i;
          }
          else if (i < currents_src-1) {
            if (i > 0) currents_src = i-1;
            else currents_src = 0;
            updateScrollPos(true, true);
          }
          break;
        }
      }
      if (change) {
        updateScrollPos(true, true);
        updateTranslationAlignment();
        $('#transcription-src li:eq('+currents_src+')').addClass("current");
        $('#sbar-src').html(getSubtitleLines(dfxpsrc['s'][currents_src]['text'], "<br/>"));
      }
    }
    else if ($(editing).parents('#transcription-src').length > 0) {
      if (tlplayer.currentTime > parseFloat(dfxpsrc['s'][currents_src]['e'])) tlplayer.pause();
    }
  }
}

function toTimecode(seconds) {
  var secs = Math.floor(seconds)%60;
  var mins = Math.floor(seconds/60)%60;
  var hours = Math.floor(seconds/3600)%60;
  hours = parseInt(hours)+10;
  frate = 1/frameRate;
  var frame = Math.floor((seconds-Math.floor(seconds))/frate);
  if (frameRate == 29.97) {
    
  }
  str = (pad(hours)+':'+pad(mins)+':'+pad(secs)+':'+pad(frame));
  return str;
}

function timecodeToSeconds(timecode) {
  var times = timecode.split(':');
  if (times.length != 4) return false;
  for (var i=0; i<4; i++) {
    if (times[i].length != 2) return false;
    if (isNaN(times[i])) return false;
  }
  if (parseInt(times[3]) > frameRate) return false;
  return (((parseInt(times[0])-10)*3600)+(parseInt(times[1])*60)+parseInt(times[2])+(Math.round((parseInt(times[3])/frameRate)*100)/100));
}

function updateSegmentTimeControl() {
  if (wa_enabled && currents >= 0 && currents < dfxptar["s"].length) {
    $('#s_begin_control').css('backgroundColor', '');
    $('#s_end_control').css('backgroundColor', '');
    if (dfxptar['s'][currents]['b'] <= tlplayer.currentTime && dfxptar['s'][currents]['e'] >= tlplayer.currentTime) {
      if (userSettings["player"]["smpte"]) {
        jQuery('#s_begin_control').val(toTimecode(dfxptar["s"][currents]["b"]));
        jQuery('#s_duration_control').val((toTimecode(dfxptar["s"][currents]["e"]-dfxptar["s"][currents]["b"])).substring(6));
        jQuery('#s_end_control').val(toTimecode(dfxptar["s"][currents]["e"]));
      }
      else {
        jQuery('#s_begin_control').val(dfxptar["s"][currents]["b"].toFixed(2));
        jQuery('#s_duration_control').val((dfxptar["s"][currents]["e"]-dfxptar["s"][currents]["b"]).toFixed(2));
        jQuery('#s_end_control').val(dfxptar["s"][currents]["e"].toFixed(2));
      }
    }
    else {
      jQuery('#s_begin_control').val('');
      jQuery('#s_duration_control').val('');
      jQuery('#s_end_control').val('');
    }
  }
}

/*
  EDITOR EVENT HANDLING CODE
*/

$('#transcription-tar').on("scroll wheel mousedown DOMMouseScroll mousewheel", function(e) {
  if (emode != 1 && e.which > 0 || e.type == "mousewheel" || e.type == "mousedown" || e.type == "DOMMouseScroll") {
    if (!tlplayer.paused) tlplayer.pause();
  }
});

$('#transcription-src').on("scroll wheel DOMMouseScroll mousewheel", function(e) {
  if (emode != 1 && e.which > 0 || e.type == "mousewheel" || e.type == "mousedown" || e.type == "DOMMouseScroll") {
    if (!tlplayer.paused) tlplayer.pause();
  }
});

$('.transcription-container').on("mouseenter", "li", function(e) {
  if (emode == 0 && editing == false) {
    if ($('#replay-button').length == 0) $(this).prepend('<div id="replay-button" title="Replay subtitle (Shift+Tab)"></div>');
  }
  /* show delete and add segment buttons */
  if (wa_enabled) {
    $(this).prepend('<div class="segment-del-btn" title="Remove subtitle"><i class="fa fa-remove fa-fw"></i></div>');
    if ($(this).parents('#transcription-tar').length > 0) $(this).prepend('<div class="segment-add-btn" title="Add new subtitle"><i class="fa fa-plus fa-fw"></i></div>');
  }
});

$('.transcription-container').on("mouseleave", "li", function(e) {
  if (editing == false) {
    $(this).children('#replay-button').remove();
  }
  if (wa_enabled) {
    $(this).children('.segment-del-btn').remove();
    $(this).children('.segment-add-btn').remove();
  }
});

function register_operation(op) {
  undo_heap.push(op);
  if (undo_heap.length > max_ops_heap) undo_heap.shift();
  redo_heap = [];
}

function undo() {
  if (undo_heap.length > 0) {
    var op = undo_heap.pop();
    redo_heap.push(op);
    if (op.op == "add") {
      removeSegment(op.index, op.src, false);
    }
    else if (op.op == "remove") {
      if (op.src) {
        var index = sremovedsrc.indexOf(op.s.sI);
        if (index > -1) sremovedtar.splice(index, 1);
      }
      else {
        var index = sremovedtar.indexOf(op.s.sI);
        if (index > -1) sremovedtar.splice(index, 1);
      }
      addSegment(op.s.b, op.src, op.s.e, op.s.text, false, op.s.sI);
    }
    else if (op.op == "join") {
      splitSegment(op.index, op.src, op.fe, op.lb, op.ftext, op.ltext, false);
    }
    else if (op.op == "split") {
      joinSegments(op.a, op.b, false);
    }
    else if (op.op == "move" || op.op == "resize") {
      positionSegment(op.index, op.src, op.b, op.e, op.left, op.width);
    }
  }
  else console.log("No operations to undo");
}

function redo() {
  if (redo_heap.length > 0) {
    var op = redo_heap.pop();
    register_operation(op);
    console.log("Redoing "+op.op);
    if (op.op == "add") {
      removeSegment(op.index, op.src);
    }
    else if (op.op == "remove") {
      if (op.src) {
        var index = sremovedsrc.indexOf(op.s.sI);
        if (index > -1) sremovedtar.splice(index, 1);
      }
      else {
        var index = sremovedtar.indexOf(op.s.sI);
        if (index > -1) sremovedtar.splice(index, 1);
      }
      addSegment(op.s.b, op.src, op.s.e, op.s.text, false, op.s.sI);
    }
    else if (op.op == "join") {
      splitSegment(op.index, op.src, op.fe, op.lb, op.ftext, op.ltext);
    }
    else if (op.op == "split") {
      joinSegments(op.a, op.b);
    }
  }
  else console.log("No operations to undo");
}

function positionSegment(index, src, begin, end, left, width) {
  if (!src) {
    dfxptar['s'][index]['b'] = begin;
    dfxptar['s'][index]['e'] = end;
    $('#subtitle-waveform-container .swf-sub:eq('+index+')').css('left', left+'px');
    $('#subtitle-waveform-container .swf-sub:eq('+index+')').width(width+'px');
    updateSegmentTimeControl();
    // update segment cps meter
    var sduration = dfxptar['s'][index]['e']-dfxptar['s'][index]['b'];
    var scps = 0.0;
    if (sduration > 0) scps = dfxptar['s'][index]['text'].replace(/\s+/g, '').length/sduration;
    if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+index+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
    else $('#transcription-tar li:eq('+index+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
  }
}

var newSegmentInitialDuration = 1.0; // seconds
function addSegment(begin, src, end, txt, reg_op, sI) {
  if (src && typeof sI == "undefined") sI = smaxsrc;
  if (!src && typeof sI == "undefined") sI = smaxtar;
  if (typeof reg_op == "undefined") reg_op = true;
  if (parseFloat(begin) == 'NaN') return;
  if (begin < 0) begin = 0;
  if (typeof end == "undefined") end = begin+newSegmentInitialDuration;
  if (typeof txt == "undefined") txt = "";
  var prevIndex = -2;
  var dfxpvar;
  if (src) dfxpvar = dfxpsrc;
  else dfxpvar = dfxptar;
  for (var i=0; i<dfxpvar['s'].length; i++) {
    if (parseFloat(dfxpvar['s'][i]['e']) > begin) {
      prevIndex = i-1;
      break;
    }
  }
  if (prevIndex == -2) prevIndex = dfxpvar['s'].length-1;
  /* Check if there's enough space between segments */
  if (prevIndex >= 0 && prevIndex < dfxpvar['s'].length-1 && dfxpvar['s'].length > 0 && (begin < dfxpvar['s'][prevIndex]['e'] || end > dfxpvar['s'][i]['b'])) {
    bootbox.alert("There's not enough space to insert a new segment, it should be at least "+newSegmentInitialDuration+" seconds between segments.");
  }
  else if (prevIndex == -1 && dfxpvar['s'].length > 0 && end > dfxpvar['s'][0]['b']) {
    bootbox.alert("There's not enough space to insert a new segment, it should be at least "+newSegmentInitialDuration+" seconds between segments.");
  }
  else if (prevIndex == dfxpvar['s'].length-1 && dfxpvar['s'].length > 0 && begin < dfxpvar['s'][prevIndex]['e']) {
    bootbox.alert("There's not enough space to insert a new segment, it should be at least "+newSegmentInitialDuration+" seconds between segments.");
  }
  else if (end > tlplayer.duration) {
    bootbox.alert("There's not enough space to insert a new segment, it should be at least "+newSegmentInitialDuration+" seconds between segments.");
  }
  else { // No collision, insert segment
    var cpscolor = "cpsgreen";
    var cpstxt = txt.replace(/\s+/g, '').length/(end-begin);
    if (cpstxt > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || cpstxt < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) cpscolor = "cpsred";
    if (src) {
      if (prevIndex >= 0) $('<li data-id="'+smaxsrc+'"><div class="segment-text" contenteditable="false"><span id="w0" class="word-text manual-conf">'+txt+'</span></div><span class="cpsmeter '+cpscolor+'">'+(txt.length/(end-begin)).toFixed(1)+' cps</span></li>').insertAfter($('#transcription-src li:eq('+prevIndex+')'));
      else $('<li data-id="'+smaxsrc+'"><div class="segment-text" contenteditable="false"><span id="w0" class="word-text manual-conf">'+txt+'</span></div><span class="cpsmeter '+cpscolor+'">'+(txt.length/(end-begin)).toFixed(1)+' cps</span></li>').insertBefore($('#transcription-src li:eq(0)'));
      if (prevIndex == dfxpsrc["s"].length-1) dfxpsrc["s"].push({"sI": smaxsrc, "aT": "human", "cM": 1.0, "b": begin, "e": end, "text": txt, "edited": "full"});
      else dfxpsrc["s"].splice((prevIndex+1), 0, {"sI": smaxsrc, "aT": "human", "cM": 1.0, "b": begin, "e": end, "text": txt, "edited": "full"});
      if (reg_op) register_operation({ op: "add", index: (prevIndex+1), src: true });
      smaxsrc++;
    }
    else {
      if (prevIndex >= 0) {
        $('<li data-id="'+sI+'"><div class="segment-text" contenteditable="false"><span id="w0" class="word-text manual-conf">'+txt+'</span></div><span class="cpsmeter '+cpscolor+'">'+(txt.length/(end-begin)).toFixed(1)+' cps</span></li>').insertAfter($('#transcription-tar li:eq('+prevIndex+')'));
        $('<div id="swf-s'+sI+'" class="swf-sub" style="left: '+Math.round(begin*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px; width: '+Math.round((end-begin)*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px"><div class="swf-sub-rs swf-sub-leftrs">&#171;</div><div class="swf-sub-text">'+txt+'</div><div class="swf-sub-rs swf-sub-rightrs">&#187;</div><div class="swf-sub-remove"><i class="fa fa-remove"></i></div></div>').insertAfter('#subtitle-waveform-container .swf-sub:eq('+prevIndex+')');
      }
      else {
        $('<li data-id="'+sI+'"><div class="segment-text" contenteditable="false"><span id="w0" class="word-text manual-conf">'+txt+'</span></div><span class="cpsmeter '+cpscolor+'">'+(txt.length/(end-begin)).toFixed(1)+' cps</span></li>').insertBefore($('#transcription-tar li:eq(0)'));
        $('<div id="swf-s'+sI+'" class="swf-sub" style="left: '+Math.round(begin*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px; width: '+Math.round((end-begin)*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px"><div class="swf-sub-rs swf-sub-leftrs">&#171;</div><div class="swf-sub-text">'+txt+'</div><div class="swf-sub-rs swf-sub-rightrs">&#187;</div><div class="swf-sub-remove"><i class="fa fa-remove"></i></div></div>').insertBefore('#subtitle-waveform-container .swf-sub:eq(0)');
      }
      $('#swf-s'+sI).contextmenu({
        target: '#swf-sub-context-menu',
        before: function (e) {
          e.preventDefault();
          return true;
        },
        onItem: function(context, e) {
          if ($(e.target).parent().attr("id") == "wa_swf_contextMenu_remove") {
            removeSegment($(context).index('.swf-sub'), false);
          }
          else if ($(e.target).parent().attr("id") == "wa_swf_contextMenu_split") {
            splitSegment($(context).index('.swf-sub'), false);
          }
        }
      });
      if (prevIndex == dfxptar["s"].length-1) dfxptar["s"].push({"sI": sI, "aT": "human", "cM": 1.0, "b": begin, "e": end, "text": txt, "edited": "full"});
      else dfxptar["s"].splice((prevIndex+1), 0, {"sI": sI, "aT": "human", "cM": 1.0, "b": begin, "e": end, "text": txt, "edited": "full"});
      if (reg_op) register_operation({ op: "add", index: (prevIndex+1), src: false });
      smaxtar++;
    }
    $(tlplayer).trigger("timeupdate");
    updateSegmentTimeControl();
  }
}

function removeSegment(index, src, reg_op) {
  if (typeof reg_op == "undefined") reg_op = true;
  if (!src) {
    sremovedtar.push(parseInt($('#transcription-tar li:eq('+index+')').attr('data-id')));
    $('#transcription-tar li:eq('+index+')').remove();
    $('#subtitle-waveform-container .swf-sub:eq('+index+')').remove();
    if (editing != false && !$(editing).is(':visible')) editing = false;
    if (reg_op) register_operation({ op: "remove", src: false, id: index, s: dfxptar["s"][index] });
    dfxptar["s"].splice(index, 1);
    if (currents == index) {
      if ($('#sbar-tar').is(':visible')) $('#sbar-tar').html("");
    }
    if (currents >= dfxptar["s"].length) currents = dfxptar["s"].length-1;
  }
  else {
    sremovedsrc.push(parseInt($('#transcription-src li:eq('+index+')').attr('data-id')));
    $('#transcription-src li:eq('+index+')').remove();
    if (reg_op) register_operation({ op: "remove", src: true, id: index, s: dfxpsrc["s"][index] });
    dfxpsrc["s"].splice(index, 1);
    if ($('#sbar-src').is(':visible') && currents_src == index) $('#sbar-src').html("");
    if (currents_src >= dfxpsrc["s"].length) currents_src = dfxpsrc["s"].length-1;
  }
  $(tlplayer).trigger("timeupdate");
  updateSegmentTimeControl();
}

function splitSegment(index, src, fend, lbegin, ftext, ltext, reg_op) {
  if (typeof reg_op == 'undefined') reg_op = true;
  if (!src) {
    var sbegin = parseFloat(dfxptar['s'][index]['b']);
    var send = parseFloat(dfxptar['s'][index]['e']);
    if (editing != false) {
      $(editing).trigger("confirm");
      tlplayer.pause();
    }
    var stext = dfxptar['s'][index]['text'];
    removeSegment(index, src, false);
    var stmid = Math.round(((sbegin+send)/2)*100)/100;
    if (typeof fend == "undefined") {
      if (tlplayer.currentTime >= dfxptar['s'][index]['b'] && tlplayer.currentTime <= dfxptar['s'][index]['e']) fend = tlplayer.currentTime;
      else fend = stmid;
    }
    if (typeof lbegin == "undefined") lbegin = fend+0.01;
    if (typeof ftext == "undefined" || typeof ltext == "undefined") { 
      addSegment(sbegin+0.01, src, fend, stext, false);
      addSegment(lbegin, src, send-0.01, "", false);
    }
    else {
      addSegment(sbegin+0.01, src, fend, ftext, false);
      addSegment(lbegin, src, send-0.01, ltext, false);
    }
    if (reg_op) register_operation({ op: "split", a: index, b: (index+1) });
    $(tlplayer).trigger("timeupdate");
  }
  else {
    console.log("Not implemented");
  }
  updateSegmentTimeControl();
}

$('.transcription-container').on("click", "li .segment-del-btn", function(e) {
  var sid = $(this).parent().index();
  var container = $(this).parent().parent().attr("id").substr($('.transcription-container').attr("id").length-3);
  if (container == "tar") {
    removeSegment(sid, false);
  }
  else if (container == "src") {
    removeSegment(sid, true);
  }
});

$('.transcription-container').on("click", "li .segment-add-btn", function(e) {
  var sid = $(this).parent().index();
  var container = $(this).parent().parent().attr("id").substr($('.transcription-container').attr("id").length-3);
  if (container == "tar") addSegment(parseFloat(dfxptar["s"][sid]["e"])+(1/frameRate), false);
  else if (container == "src") addSegment(parseFloat(dfxpsrc["s"][sid]["e"])+(1/frameRate), true);
});

$('.transcription-container').on("click", "li #replay-button", function(e) {
  e.stopPropagation();
  var sid = $(this).parent().index();
  var container = $(this).parent().parent().attr("id").substr($('.transcription-container').attr("id").length-3);
  /*
  if ($('li.current').length == 0) {
    $('#transcription-tar li:eq('+sid+')').addClass("current");
    if ($('#transcription-src').is(':visible')) {
      $('#transcription-src li:eq('+sid+')').addClass("current");
    }
  }
  */
  if (container == "tar") {
    if (emode == 0) tlplayer.currentTime = parseFloat(dfxptar['s'][sid]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
    else tlplayer.currentTime = parseFloat(dfxptar['s'][sid]['w'][currentw['f']]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
  }
  else if (container == "src") {
    if (emode == 0) tlplayer.currentTime = parseFloat(dfxpsrc['s'][sid]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
    else tlplayer.currentTime = parseFloat(dfxpsrc['s'][sid]['w'][currentw['f']]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
  }
  if (tlplayer.paused) tlplayer.play();
  if (editing != false) $(editing).trigger("focus");
  return false;
});

$('.transcription-container').on("click", ".segment-text", function(e, muststop) {
  if (typeof muststop == 'undefined') muststop = true;
  if (e.which != 2 && e.which != 3 && editing == false) {
    hasEditedSegmentText = false;
    $(this).prop("contenteditable", true);
    // pause if playing
    if (muststop && !tlplayer.paused) tlplayer.pause();
    var sid = $(this).parent().index();
    /* disable step-forward/backward buttons */
    $('button.stepforward').css('opacity','0.3');
    $('button.stepbackward').css('opacity','0.3');
    // no auto-confirm if clicked on next segment workaround
    if ($(this).parents('#transcription-tar').length > 0) {
      currents = sid;
      $('#transcription-tar li.current').removeClass('current');
      $('#transcription-tar li:eq('+currents+')').addClass("current");
      $('#sbar-tar').html($('#transcription-tar li:eq('+currents+') .segment-text').text());
      updateScrollPos(true, false);
    }
    else if ($(this).parents('#transcription-src').length > 0) {
      currents_src = sid;
      $('#transcription-src li.current').removeClass('current');
      $('#transcription-src li:eq('+currents_src+')').addClass("current");
      $('#sbar-src').html($('#transcription-src li:eq('+currents_src+') .segment-text').text());
      updateScrollPos(true, true);
    }
    // end workaround
    // set time on focused segment begin if it wasn't the current segment
    if (tlplayer.currentTime < parseFloat(dfxptar['s'][sid]['b']) || tlplayer.currentTime > parseFloat(dfxptar['s'][sid]['e'])) {
      if ($(this).parents('#transcription-tar').length > 0) tlplayer.currentTime = parseFloat(dfxptar['s'][sid]['b'])+(1/frameRate); // maximum frame accuracy 1s/24fps = 0.041
      else if ($(this).parents('#transcription-src').length > 0) tlplayer.currentTime = parseFloat(dfxpsrc['s'][sid]['b'])+(1/frameRate); // maximum frame accuracy 1s/24fps = 0.041
    }
    updateCurrentSubtitle(false);
    // set editing object
    editing = this;
    /* throw editing event */
    $.event.trigger({
      type: "edit",
      container: (($(this).parents('#transcription-tar').length > 0) ? "tar" : "src")
    });
    // enable replay button
    $('#replay-button').remove();
    $(this).parent().prepend('<div id="replay-button"></div>');
    $(this).focus();
    // hide any visible panel
    $('.tlpanel').hide();
    // remove segment color if no word-level info
    if ($(this).children('span').length == 1) {
      $(this).children('span').removeClass("low-conf");
      $(this).children('span').removeClass("manual-conf");
    }
    updateSegmentTimeControl();
    return false;
  }
});

$('.transcription-container').on("click", ".segment-text .word-text", function(e) {
  if (!editing || editing == $(this).parents(".segment-text").get(0)) {
    $(this).removeClass("low-conf");
    $(this).removeClass("manual-conf");
  }
});

$('.transcription-container').on("blur", ".segment-text", function(e) {
  setTimeout(function() { 
    if (editing != false && isEditorControlsAdvancedFocused == false) {
      $(editing).trigger("focus");
    }
  }, 250);
});

$('.transcription-container').on("confirm", ".segment-text", function(e) {
  hasEditedSegmentText = false;
  var container = $(this).parent().parent().attr("id").substr($('.transcription-container').attr("id").length-3);
  var sid = $(this).parent().index();
  var stext = $(this).children('.word-text').html();
  if (stext.length == 0) stext = $(this).text();
  while (stext.charAt(stext.length-1) == ' ') stext = stext.substr(0, stext.length-1);
  $(this).empty().append('<span id="w0" class="word-text manual-conf">'+stext+'</span>');
  if (container == "tar") {
    dfxptar['s'][sid]['text'] = stext;
    dfxptar['s'][sid]['edited'] = "full";
    dfxptar['s'][sid]['aT'] = "human";
    dfxptar['s'][sid]['w'] = [];
  }
  else {
    dfxpsrc['s'][sid]['text'] = stext;
    dfxpsrc['s'][sid]['edited'] = "full";
    dfxpsrc['s'][sid]['aT'] = "human";
    dfxpsrc['s'][sid]['w'] = [];
  }
  $(this).parent().children('#replay-button').remove();
  editing = false;
  $.event.trigger({
    type: "confirm",
    container: container
  });
  /* enable step-forward/backward buttons */
  $('button.stepforward').css('opacity','1.0');
  $('button.stepbackward').css('opacity','1.0');
  if (tlplayer.paused && vmode != 3 && userSettings["player"]["autoplay"] == true) tlplayer.play();
  $(this).trigger("blur");
  $(this).prop("contenteditable", false);
  // equalize segment heights (translation only)
  if ($('#transcription-src').is(':visible')) {
    var hsrc = $('#transcription-src li:eq('+sid+')').height();
    var htar = $('#transcription-tar li:eq('+sid+')').height();
    if (hsrc > htar) $('#transcription-tar li:eq('+sid+')').height(hsrc);
    else if (hsrc < htar) $('#transcription-src li:eq('+sid+')').height(htar); 
  }
});

$('.subtitle-bar').on("click", function(e) {
  if (editing == false) {
    hasEditedSegmentText = false;
    $(this).prop("contenteditable", true);
    $(this).trigger("focus");
    editing = this;
    $.event.trigger({
      type: "edit",
      container: $(this).attr("dfxp")
    });
    if (!tlplayer.paused) tlplayer.pause();
    return false;
  }
});

$('.subtitle-bar').on("blur", function(e) {
  if (this === editing) setTimeout(function() { if (this === editing) $(this).trigger("focus"); }, 0);
});

$('.subtitle-bar').on("confirm", function(e) {
  hasEditedSegmentText = false;
  if ($(this).attr("dfxp") == "tar") {
    dfxptar['s'][currents]['text'] = $(this).html();
    dfxptar['s'][currents]['edited'] = "full";
    dfxptar['s'][currents]['aT'] = "human";
    dfxptar['s'][currents]['w'] = [];
    $('#transcription-tar li:eq('+currents+') .segment-text').empty().append('<span id="w0" class="word-text manual-conf">'+$(this).html()+'</span>'); 
  }
  else if ($(this).attr("dfxp") == "src") {
    dfxpsrc['s'][currents]['text'] = $(this).html();
    dfxpsrc['s'][currents]['edited'] = "full";
    dfxpsrc['s'][currents]['aT'] = "human";
    dfxpsrc['s'][currents]['w'] = [];
    $('#transcription-src li:eq('+currents+') .segment-text').empty().append('<span id="w0" class="word-text manual-conf">'+$(this).html()+'</span>');
  }
  editing = false;
  $.event.trigger({
    type: "confirm",
    container: $(this).attr("dfxp")
  });
  $(this).trigger("blur");
  $(this).prop("contenteditable", false);
  if (tlplayer.paused && userSettings["player"]["autoplay"] == true) tlplayer.play();
});

$('#tl-main').on("click", function(e) {
  // clicking outside panel -> close panel
  if ($(e.target).parents('.tlpanel').length == 0) $('.tlpanel:visible').hide();
  // clicking outside editing element -> confirm edition
  //if (editing != false && !(e.target === editing) && !($(e.target).parent().get(0) === editing) && !(e.target === editing)) {
    //$(editing).trigger("confirm");
  //}
});

$('body').on("click", function(e) {
  // clicking outside editing element -> confirm edition
  if (editing != false && !(e.target === editing) && !($(e.target).parent().get(0) === editing) && !(e.target === editing) && $(e.target).parents('#subtitle-waveform-container').length == 0 && $(e.target).parents('.tl-editor-controls-advanced').length == 0) {
    $(editing).trigger("confirm");
  }
  if ($(e.target).parents('.tl-editor-controls-advanced').length == 0) isEditorControlsAdvancedFocused = false;
});

var isEditorControlsAdvancedFocused = false;
$('.tl-editor-controls-advanced').on("click", function(e) {
  isEditorControlsAdvancedFocused = true;
});

$('#s_begin_control').on("keyup", function(e) {
  $(this).css('backgroundColor', '');
  var secs = 0;
  if (userSettings["player"]["smpte"]) secs = timecodeToSeconds($(this).val());
  else secs = parseFloat($(this).val());
  if (secs != false && !isNaN(secs) && (currents == 0 || dfxptar["s"][currents-1]["e"] < secs) && (dfxptar["s"][currents]["e"]-secs > 0)) {
    dfxptar["s"][currents]["b"] = secs;
    dfxptar["s"][currents]["edited"] = "full";
    $('#swf-s'+currents).css('left', Math.round(secs*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px');
    $('#swf-s'+currents).width(Math.round((dfxptar["s"][currents]["e"]-secs)*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px');
    if (userSettings["player"]["smpte"]) $('#s_duration_control').val((toTimecode(dfxptar["s"][currents]["e"]-dfxptar["s"][currents]["b"])).substring(6));
    else $('#s_duration_control').val((dfxptar["s"][currents]["e"]-dfxptar["s"][currents]["b"]).toFixed(2));
    // update segment cps meter
    var sduration = dfxptar["s"][currents]["e"]-secs;
    var scps = 0.0;
    if (sduration > 0) scps = dfxptar['s'][currents]['text'].replace(/\s+/g, '').length/sduration;
    if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
    else $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
  }
  else if (secs == false || isNaN(secs)) {
    $(this).css('backgroundColor', '#FF6666');
  }
  else {
    $(this).css('backgroundColor', '#FFCE4A');
  }
});

$('#s_end_control').on("keyup", function(e) {
  $(this).css('backgroundColor', '');
  var secs = timecodeToSeconds($(this).val());
  if (secs != false && (currents == dfxptar["s"].length-1 || dfxptar["s"][currents+1]["b"] > secs) && (secs-dfxptar["s"][currents]["b"] > 0)) {
    dfxptar["s"][currents]["e"] = secs;
    dfxptar["s"][currents]["edited"] = "full";
    $('#swf-s'+currents).width(Math.round((secs-dfxptar["s"][currents]["b"])*wa_zoomLevelsPps[wa_pixelsPerSecond]));
    $('#s_duration_control').val((toTimecode(dfxptar["s"][currents]["e"]-dfxptar["s"][currents]["b"])).substring(6));
    // update segment cps meter
    var sduration = secs-dfxptar["s"][currents]["b"];
    var scps = 0.0;
    if (sduration > 0) scps = dfxptar['s'][currents]['text'].replace(/\s+/g, '').length/sduration;
    if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
    else $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
  }
  else if (secs == false) {
    $(this).css('backgroundColor', '#FF6666');
  }
  else {
    $(this).css('backgroundColor', '#FFCE4A');
  }
});

function moveBackward() {
  if (emode < 2) {
    if (editing != false) {
      if (tlplayer.currentTime >= parseFloat(dfxptar['s'][currents]['b']) + 1.5) {
        tlplayer.currentTime = tlplayer.currentTime - 1.5;
      }
      else tlplayer.currentTime = parseFloat(dfxptar['s'][currents]['b']);
      if (tlplayer.paused) tlplayer.play();
    }
    else {
      tlplayer.currentTime = tlplayer.currentTime - 1.5;
      if (tlplayer.paused) tlplayer.play();
    }
  }
}

function moveForward() {
  if (emode < 2) {
    if (editing != false) {
      if (tlplayer.currentTime < (parseFloat(dfxptar['s'][currents]['e']) - 1.5)) {
        tlplayer.currentTime = tlplayer.currentTime + 1.5;
        if (tlplayer.paused) tlplayer.play();
      }
    }
    else {
      tlplayer.currentTime = tlplayer.currentTime + 1.5;
      if (tlplayer.paused) tlplayer.play();
    }
  }
}

function previousSegment() {
  if (editing == false) {
    var tos = (currents > 0) ? currents-1 : 0;
    if (tos > 0 && vmode > 0 && $('li.current').length == 0) {
      tos++;
      $('#transcription-tar li:eq('+tos+')').addClass("current");
      if ($('#transcription-src').is(':visible')) {
        $('#transcription-src li:eq('+tos+')').addClass("current");
      }
    }
    tlplayer.currentTime = parseFloat(dfxptar['s'][tos]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
  }
}

function nextSegment() {
  if (editing == false) {
    var tos = (currents < dfxptar['s'].length-1) ? currents+1 : dfxptar['s'].length-1;
    if (currents == 0 && vmode > 0 && $('li.current').length == 0) tos = 0;
    tlplayer.currentTime = parseFloat(dfxptar['s'][tos]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
    updateCurrentSubtitle(false);
  }
}

// keyboard shortcuts
$(document).on("keyup", function(e) {
  if (editing != false) {
    var cstext = $(editing).text().replace(/\s+/g, '');
    var cscps = (cstext.length/(dfxptar['s'][currents]['e']-dfxptar['s'][currents]['b']));
    var text = "";
    if ($(editing).hasClass("subtitle-bar")) {
      text = $(editing).html();
      if ($(editing).is('#sbar-tar')) {
        $('#transcription-tar li:eq('+currents+') .segment-text .word-text').html(text);
        $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+" cps");
        if (cscps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || cscps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
        else $('#transcription-tar li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
      }
      else if ($(editing).is('#sbar-src')) {
        $('#transcription-src li:eq('+currents+') .segment-text .word-text').html(text);
        $('#transcription-src li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+" cps");
        if (cscps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || cscps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-src li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
        else $('#transcription-src li:eq('+currents+') span.cpsmeter').html(cscps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
      }
    }
    else if ($(editing).hasClass("segment-text")) {
      if ($(editing).children('.word-text').length == 0) $(editing).append('<span id="w0" class="word-text"></span>');
      text = $(editing).children('.word-text').html();
      if (text.length == 0) text = $(editing).text();
      if ($(editing).parents('#transcription-tar').length > 0) {
        $('#sbar-tar').html(getSubtitleLines(text, "<br/>")); 
      }
      else if ($(editing).parents('#transcription-src').length > 0) {
        $('#sbar-src').html(getSubtitleLines(text, "<br/>"));
      }
      $(editing).parent().children('span.cpsmeter').html(cscps.toFixed(1)+" cps");

      if (cscps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || cscps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $(editing).parent().children('span.cpsmeter').removeClass('cpsgreen').addClass('cpsred');
      else $(editing).parent().children('span.cpsmeter').removeClass('cpsred').addClass('cpsgreen');
    }
    /* autobreak lines if longer than max_cpl */
    if (wa_enabled && userSettings["player"]["max_cpl"] > 0) {
      if (text.length == userSettings["player"]["max_cpl"]) {
        console.log("Line too long, trying to break...");
        /*
        $(editing).children('.word-text').html($(editing).children('.word-text').html()+"<br/>");
        var el = $(editing).children('.word-text').get(0);
        var range = document.createRange();
        var sel = window.getSelection();
        range.setStart(el.childNodes[1], 0);
        range.collapse(true);
        sel.removeAllRanges();
        sel.addRange(range);
        el.focus();
        */
      }
    }
  }
});

function toggle_advanced_mode() {
  if (wa_enabled) {
    disableWa();
    fillTranscriptionContainer(false);
    if (dfxpsrc != null) fillTranscriptionContainer(true);
  }
  else {
    enableWa();
    fillTranscriptionContainer(false);
    if (dfxpsrc != null) fillTranscriptionContainer(true);
  }
}

function unregisterKeypressCombos() {
  keyListener.reset();
  registeredCombos = null;
}

function registerKeypressCombos() {
  registeredCombos = keyListener.register_many([
    {
        "keys"          : userSettings["shortcuts"]["replay"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if ($('li.current').length == 0) {
            $('#transcription-tar li:eq('+currents+')').addClass("current");
            if ($('#transcription-src').is(':visible')) {
              $('#transcription-src li:eq('+currents+')').addClass("current");
            }
          }
          if (emode == 0) {
            $('#replay-button').css('background-position', '-28px');
            setTimeout(function() { $('#replay-button').css('background-position', ''); }, 150);
            tlplayer.currentTime = parseFloat(dfxptar['s'][currents]['b'])+(1/frameRate); // maximum frame accuracy: 1s/24fps = 0.041
            if (tlplayer.paused) tlplayer.play();
          }
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["playpause"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if (emode == 0 && $('li.current').length == 0) {
            $('#transcription-tar li:eq('+currents+')').addClass("current");
            if ($('#transcription-src').is(':visible')) {
              $('#transcription-src li:eq('+currents+')').addClass("current");
            }
          }
          if (emode < 2) {
            if (tlplayer.paused) tlplayer.play();
            else tlplayer.pause();
          }
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["enterconfirm"],
        "is_exclusive"  : true,
        "is_solitary"   : true,
        "on_keydown"    : function(e) {
          if (emode == 0) {
            if (editing != false) $(editing).trigger("confirm");
            else {
              if (vmode > 0) $('#transcription-tar li:eq('+currents+') .segment-text').trigger("click");
              else $('#sbar-tar').trigger("click");
            }
            return false;
          }
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["newsegment"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if (wa_enabled) {
            addSegment(tlplayer.currentTime, false);
          }
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["joinprev"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if (wa_enabled && currents > 0) joinSegments(currents, currents-1);
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["joinnext"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if (wa_enabled && currents < dfxptar["s"].length-1) joinSegments(currents, currents+1);
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["splitsegment"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          if (wa_enabled) {
            splitSegment(currents, false);
          }
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["jumpback"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          moveBackward();
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["jumpforward"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          moveForward();
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["frameback"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          tlplayer.currentTime = tlplayer.currentTime-(1.0/frameRate);
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["frameforward"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          tlplayer.currentTime = tlplayer.currentTime+(1.0/frameRate);
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["prevsegment"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          previousSegment();
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["nextsegment"],
        "is_exclusive"  : true,
        "on_keydown"    : function() {
          nextSegment();
          return false;
        }
    },
    {
        "keys"          : userSettings["shortcuts"]["undo"],
        "is_exclusive"  : true,
        "on_keydown"    : function(e) {
          if (wa_enabled && ($(editing) == false || hasEditedSegmentText == false)) {
            e.preventDefault();
            undo();
            return false;
          }
          else {
            hasEditedSegmentText = false;
            return true;
          }
        }
    }
  ]);
}

$(document).on("keydown", function(e) {
  if (emode == 0 && e.keyCode == 8 && editing == false) e.preventDefault();
  if (isSettingShortcut) e.preventDefault();
  // Non user-defined (fixed) key controls 
  if ("rcode" in langs && langs["rcode"] == 0) {
    if (helpvisible) {
      $('.inhibit').trigger("click");
      return false;
    }
    if (e.ctrlKey) { // CTRL +
      switch (e.keyCode) {
        case 190: // .
          if (aud_comp_enabled) {
            disableAudioCompression();
          }
          else {
            enableAudioCompression();
          }
          break;
      }
    }
    else {
      switch (e.keyCode) {
        case 27: // ESC
          if (emode == 0) {
            if (editing != false) $(editing).trigger("confirm");
            return false;
          }
          break;
        case 32: // SPACE
          if (emode == 0 && editing == false) {
            (tlplayer.paused) ? tlplayer.play() : tlplayer.pause();
            return false;
          }
          break;
        case 35: // END
          // END span bugfix
          if (editing != false) {
            if ($(editing).children('span').length > 1) {
              var range = document.createRange();
              var sel = window.getSelection();
              range.setStart(editing.childNodes[editing.childNodes.length-1], editing.childNodes[editing.childNodes.length-1].length);
              range.collapse(true);
              sel.removeAllRanges();
              sel.addRange(range);
              return false;
            }
          }
          break;
        case 36: // HOME
          // HOME span bugfix
          if (editing != false) {
            if ($(editing).children('span').length > 1) {
              var range = document.createRange();
              var sel = window.getSelection();
              range.setStart(editing.childNodes[0], 0);
              range.collapse(true);
              sel.removeAllRanges();
              sel.addRange(range);
              return false;
            }
          }
          break;
      }
    }
  }
});

var hasEditedSegmentText = false;
$(document).on("keyup", function(e) {
  if (editing != false) {
    if (wa_enabled) {
      var text = "";
      if ($(editing).hasClass("subtitle-bar")) {
        text = $(editing).html();
        $('#subtitle-waveform-container .swf-sub:eq('+currents+')').children('.swf-sub-text').html(text);
        if (hasEditedSegmentText == false && text != dfxptar['s'][currents]['text']) hasEditedSegmentText = true;
      }
      else {
        text = $(editing).children('.word-text').html();
        if (text.length == 0) text = $(editing).text();
        $('#subtitle-waveform-container .swf-sub:eq('+$(editing).parent().index()+')').children('.swf-sub-text').html(text);
        if (hasEditedSegmentText == false && text != dfxptar['s'][currents]['text']) hasEditedSegmentText = true;
      }
    }
  }
});

/*
  TTS - Text-to-speech code
*/

var ttsAudioTrack = null;
var ttsEnabled = false;

function addTtsLanguages() {
  for (var lang in ttsAvailTracks) {
    if ("gender" in ttsAvailTracks[lang]) $('#ttslang-list').append('<option value="'+lang+'">'+getLanguageName(lang)+' ('+ttsAvailTracks[lang]["gender"]+')</option>');
    else $('#ttslang-list').append('<option value="'+lang+'">'+getLanguageName(lang)+'</option>');
  };
}

function enableTts(lang) {
  if (lang in ttsAvailTracks) {
    ttsAudioTrack = document.createElement('audio');
    ttsAudioTrack.setAttribute('src', ttsAvailTracks[lang]["url"]);
    ttsAudioTrack.volume = tlplayer.volume;
    $("button.tts").css("background-image", "url('translectures/images/player-controls-blue.png')");
    tlplayer.volume = 0;
    ttsEnabled = true;
  }
}

function disableTts() {
  tlplayer.volume = ttsAudioTrack.volume;
  ttsAudioTrack.volume = 0;
  ttsEnabled = false;
  ttsAudioTrack.remove();
  ttsAudioTrack = null;
  $("button.tts").css("background-image", "");
}

$('button.tts').click(function() {
  if (editing != false) {
    $(editing).trigger("confirm");
  }
  $('.tlpanel').not('.tts-panel').hide();
  $('.tts-panel').toggle();
  return false;
});

$('#ttslang-list').change(function(e) {
  var slang = $('#ttslang-list option:selected').val();
  if (!tlplayer.paused) tlplayer.pause();
  if (slang == "none") {
    disableTts();
  }
  else {
    enableTts(slang);
  }
  $('.tts-panel').hide();
});

$('button.volume').click(function(e) {
  if (ttsEnabled) {
    ttsAudioTrack.muted = !ttsAudioTrack.muted;
    tlplayer.volume = 0;
  }
});

$('input[id="volume-bar"]').change(function(e) {
  if (ttsEnabled) {
    ttsAudioTrack.volume = $(this).val();
    tlplayer.volume = 0;
  }
});

tlplayer.addEventListener('play', function() {
  if (ttsEnabled && ttsAudioTrack.paused) ttsAudioTrack.play();
}, false);

tlplayer.addEventListener('pause', function() {
  if (ttsEnabled && !ttsAudioTrack.paused) ttsAudioTrack.pause();
}, false);

tlplayer.addEventListener('ended', function() {
  tlplayer.pause();
  if (ttsEnabled) ttsAudioTrack.pause();
}, false);

tlplayer.addEventListener('timeupdate', function() {
  if (ttsEnabled && ttsAudioTrack.readyState >= 4) {
    if (ttsAudioTrack.paused && !tlplayer.paused) ttsAudioTrack.play();
    if (Math.abs(ttsAudioTrack.currentTime-tlplayer.currentTime) > 0.5) {
      ttsAudioTrack.currentTime = tlplayer.currentTime;
    }
  }
}, false);

/* FORCED ALIGN MODE */

// function enableForcedAlignMode() {
//   editing = false;
//   if (wa_enabled) disableWa();
//   emode = 1;
//   if (vmode == 0) changeViewMode(1);
//   $('.subtitle-bar').hide();
//   $('#olayout').hide();
//   $('button.stepforward').hide();
//   $('button.stepbackward').hide();
//   $('.tl-editor-controls').show();
//   if (dfxpsrc != null) {
//     $('#transcription-src').hide();
//     $('#transcription-tar').css('width', '100%');
//   }
//   fillTranscriptionContainer(false);
// }

// function disableForcedAlignMode() {
//   emode = 0;
//   $('.subtitle-bar').show();
//   $('#olayout').show();
//   $('button.stepforward').show();
//   $('button.stepbackward').show();
//   $('.tl-editor-controls').hide();
//   if (dfxpsrc != null) {
//     $('#transcription-src').show();
//     $('#transcription-tar').css('width', '49%');
//   }
//   fillTranscriptionContainer(false);
//   if (dfxpsrc != null) fillTranscriptionContainer(true);
// }

// $('button#force-align').click(function() {
//   synchronizeTranscription();
// });

// function synchronizeTranscription() {
//   bootbox.dialog({
//     title: "Automatic synchronization",
//     message: "<div>Please select the time range you have transcribed:</div><br/><div><label id=\"fa-time-range-begin\" style=\"margin-right: 1em;\">00:00</label><input id=\"fa-time-range\" type=\"number\" value=\"\" data-slider-min=\"0\" data-slider-id=\"fa-time-range-slider\" data-slider-tooltip=\"hide\" data-slider-max=\""+(Math.floor(tlplayer.duration))+"\" data-slider-step=\"1\" data-slider-value=\"[0,"+(Math.round(tlplayer.currentTime))+"]\"/><label id=\"fa-time-range-end\" style=\"margin-left: 1em;\">"+parsetime(Math.round(tlplayer.currentTime))+"</label></div>",
//     buttons: {
//       danger: {
//         label: "Cancel",
//         className: "btn-danger"
//       },
//       success: {
//         label: "Synchronize",
//         className: "btn-primary",
//         callback: function() {
//           var begin = $('#fa-time-range').slider('getValue')[0];
//           var end = $('#fa-time-range').slider('getValue')[1];
//           /* Launch synchronization process */
//         }
//       }
//     }
//   });
//   $('#fa-time-range').slider();
//   $('#fa-time-range').on("slide", function(slideEvt) {
//     $('#fa-time-range-begin').html(parsetime(slideEvt.value[0]));
//     $('#fa-time-range-end').html(parsetime(slideEvt.value[1]));
//   });
// }

/* 
  PLAYER WAVEFORM ALIGNMENT
*/

var wa_waveformDataHz = 500;
var wa_zoomLevelsPps = [parseInt(wa_waveformDataHz/50), parseInt(wa_waveformDataHz/25), parseInt(wa_waveformDataHz/10), parseInt(wa_waveformDataHz/5), parseInt(wa_waveformDataHz/2), wa_waveformDataHz];
var wa_pixelsPerSecond = 2; // wa_zoomLevelsPps index
var wa_maxPeakHeight = 64;
var wa_enabled = false;
var wa_animationFrameId = null;
var wa_waveformData = null;
var wa_smoothingWindowSize = 5;

function getWaveformData() {
  if (pcmWaveformFile != null) {
    console.log("Getting waveform data...");
    var xhr = new XMLHttpRequest();
    xhr.open('GET', pcmWaveformFile, true);
    xhr.responseType = 'arraybuffer';
    xhr.onreadystatechange = function (e) {
      if (xhr.readyState == 4) {
        if (xhr.status == 200) {
          wa_waveformData = new Uint8Array(this.response);
          for (var i=0; i<wa_waveformData.length; i++) wa_waveformData[i] = Math.abs(wa_waveformData[i]-128);
          smoothWaveform();
          /* Increase waveform size if too low volume */
          var mean = 0;
          var n = 0;
          for (var i=0; i<wa_waveformData.length; i++) {
            if (wa_waveformData[i] > 8) {
              mean+=wa_waveformData[i];
              n++;
            }
          }
          if (n == 0 || (mean/n) < 16) for (var i=0; i<wa_waveformData.length; i++) {
            if (n == 0) {
              if (wa_waveformData[i]*3 > 128) wa_waveformData[i] = 128;
              else wa_waveformData[i]=wa_waveformData[i]*3;
            }
            else {
              if (wa_waveformData[i]*2 > 128) wa_waveformData[i] = 128;
              else wa_waveformData[i]=wa_waveformData[i]*2;
            }
          }
          printWaveform();
          printTimebar();
        }
        else {
          wa_waveformData = null;
          console.log("Error loading waveform data. It won't be displayed.");
        }
      }
    };
    xhr.send();
  }
  else {
    console.log("Waveform data file not available.");
  }
}

function smoothWaveform() {
  if (wa_smoothingWindowSize == 1) return;
  if (wa_waveformData.length <= wa_smoothingWindowSize) return;
  if (wa_waveformData != null) {
    var wa_waveformDataSmoothed = new Uint8Array(wa_waveformData.length);
    var wMean = 0;
    for (var j=0; j<wa_smoothingWindowSize; j++) wMean += wa_waveformData[j];
    wa_waveformDataSmoothed[0] = Math.round(wMean/wa_smoothingWindowSize);
    wMean = wMean-wa_waveformData[0];
    for (var i=1; i<wa_waveformData.length; i++) {
      wMean += wa_waveformData[(i+wa_smoothingWindowSize-1)];
      wa_waveformDataSmoothed[i] = Math.round(wMean/wa_smoothingWindowSize);
      wMean = wMean-wa_waveformData[i];
    }
    wa_waveformData = wa_waveformDataSmoothed;
  }
}

function enableWa() {
  emode = 0;
  if (wa_waveformData == null) getWaveformData();
  else printWaveform();
  printTimebar();
  genWaveformSubtitles();
  wa_enabled = true;
  $('#subtitle-waveform-container').show();
  $('.tl-editor-controls-advanced').show();
  $('.tl-editor-controls').hide();
  if (userSettings["player"]["smpte"]) {
    $('span.current-time').html(toTimecode(tlplayer.currentTime));
    $('span.duration-time').html(toTimecode(tlplayer.duration));
  }
  onResize();
  updateWaveformPosition();
  wa_animationFrameId = window.requestAnimationFrame(updateWaveformPosition);
  userSettings["player"]["advanced"] = true;
  onSettingChange();
}

function disableWa() {
  emode = 0;
  window.cancelAnimationFrame(wa_animationFrameId);
  wa_enabled = false;
  $('#subtitle-waveform-container').hide();
  $('.tl-editor-controls-advanced').hide();
  $('span.current-time').html(parsetime(tlplayer.currentTime));
  $('span.duration-time').html(parsetime(tlplayer.duration));
  onResize();
  userSettings["player"]["advanced"] = false;
  onSettingChange();
}

function updateWaveformPosition(animate) {
  if (typeof(animate) == 'undefined') animate = true;
  var ppos = Math.round(tlplayer.currentTime*wa_zoomLevelsPps[wa_pixelsPerSecond]-($('#subtitle-waveform-container').width()/2));
  $('#subtitle-waveform-container').scrollLeft(ppos);
  $('#subtitle-waveform-container').children('.swf-mark').css('left', (ppos+($('#subtitle-waveform-container').width()/2))+'px');
  printWaveform();
  printTimebar();
  if (!tlplayer.paused) {
    wa_animationFrameId = window.requestAnimationFrame(updateWaveformPosition);
  }
}

function genWaveformSubtitles() {
  $('#subtitle-waveform-void').width(Math.round(wa_zoomLevelsPps[wa_pixelsPerSecond]*tlplayer.duration));
  $('#subtitle-waveform-container').children('.swf-sub').remove();
  $.each(dfxptar['s'], function(key, val) {
    $('#subtitle-waveform-container').append('<div id="swf-s'+key+'" class="swf-sub" style="left: '+Math.round(val["b"]*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px; width: '+Math.round((val['e']-val['b'])*wa_zoomLevelsPps[wa_pixelsPerSecond])+'px;"><div class="swf-sub-rs swf-sub-leftrs">&#171;</div><div class="swf-sub-text">'+val["text"]+'</div><div class="swf-sub-rs swf-sub-rightrs">&#187;</div><div class="swf-sub-remove"><i class="fa fa-remove"></i></div></div>');
  });
  $('.swf-sub').contextmenu({
    target: '#swf-sub-context-menu',
    before: function (e) {
      e.preventDefault();
      return true;
    },
    onItem: function(context, e) {
      if ($(e.target).parent().attr("id") == "wa_swf_contextMenu_remove") {
        removeSegment($(context).index('.swf-sub'), false);
      }
      else if ($(e.target).parent().attr("id") == "wa_swf_contextMenu_split") {
        splitSegment($(context).index('.swf-sub'), false);
      }
    }
  });
  $('#subtitle-waveform-void').contextmenu({
    target: '#subtitle-waveform-void-context-menu',
    before: function (e) {
      e.preventDefault();
      return true;
    },
    onItem: function(context, e) {
      if ($(e.target).parent().attr("id") == "wa_contextMenu_addSegment") {
        addSegment(tlplayer.currentTime, false);
      }
      else if ($(e.target).parent().attr("id") == "wa_contextMenu_zoomIn") {
        wa_zoomIn();
      }
      else if ($(e.target).parent().attr("id") == "wa_contextMenu_zoomOut") {
        wa_zoomOut();
      }
    }
  });
}

function printTimebar() {
  printTimebarAlt();
  /*
  var c = $('#subtitle-waveform-canvas')[0];
  var w = $('#subtitle-waveform-container').width();
  var h = $('#subtitle-waveform-container').height();
  if (wa_waveformData == null) {
    c.width = w;
    c.height = h;
  }
  var tmark = 5*wa_zoomLevelsPps[wa_pixelsPerSecond];
  var tsubmark = wa_zoomLevelsPps[wa_pixelsPerSecond];
  var ctx = c.getContext('2d');
  ctx.strokeStyle = '#333333';
  ctx.fillStyle = '#333333';
  ctx.lineWidth = 2;
  ctx.font = "11px Arial";
  var tbegin = 0;
  if ((tlplayer.currentTime-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) <= 0) tbegin = 0;
  else if ((tlplayer.currentTime+((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) > tlplayer.duration) tbegin = Math.round((tlplayer.duration*wa_zoomLevelsPps[wa_pixelsPerSecond])-w);
  else tbegin = Math.round((tlplayer.currentTime*wa_zoomLevelsPps[wa_pixelsPerSecond])-(w/2));
  ctx.beginPath();
  ctx.moveTo(0, h);
  ctx.lineTo(w, h);
  for (var j=Math.round((Math.ceil(tbegin/tmark)*tmark)-tbegin); j<w; j+=tmark) {
    var time = ((tbegin+j)/wa_zoomLevelsPps[wa_pixelsPerSecond]);
    if (time != 0) {
      if (parsetime(time).length > 5) ctx.fillText(parsetime(time), j-20, h-20);
      else ctx.fillText(parsetime(time), j-14, h-20);
    }
    ctx.moveTo(j, h);
    ctx.lineTo(j, h-15);
  }
  for (var j=Math.round((Math.ceil(tbegin/tsubmark)*tsubmark)-tbegin); j<w; j+=tsubmark) {
    ctx.moveTo(j, h);
    ctx.lineTo(j, h-5);
  }
  ctx.stroke();
  ctx.closePath();
  $('#subtitle-waveform-canvas').css('left', $('#subtitle-waveform-container').scrollLeft());
  */
}

function printTimebarAlt() {
  var c = $('#subtitle-waveform-canvas')[0];
  var w = $('#subtitle-waveform-container').width();
  var h = $('#subtitle-waveform-container').height();
  var p = 0.5 / window.devicePixelRatio;
  if (wa_waveformData == null) {
    c.width = w;
    c.height = h;
  }
  var tmark = 5*wa_zoomLevelsPps[wa_pixelsPerSecond];
  var tsubmark = wa_zoomLevelsPps[wa_pixelsPerSecond];
  var ctx = c.getContext('2d');
  ctx.strokeStyle = '#333333';
  ctx.fillStyle = '#333333';
  ctx.lineWidth = 1;
  ctx.font = "11px Arial";
  var tbegin = 0;
  if ((tlplayer.currentTime-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) <= 0) {
    tbegin = 0;
  }
  else if ((tlplayer.currentTime+((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) > tlplayer.duration) {
    tbegin = Math.round((tlplayer.duration*wa_zoomLevelsPps[wa_pixelsPerSecond])-w);
    if (tbegin < 0) tbegin = 0;
  }
  else {
    tbegin = Math.round((tlplayer.currentTime*wa_zoomLevelsPps[wa_pixelsPerSecond])-(w/2));
  }
  ctx.beginPath();
  ctx.moveTo(0+p, h-25+p);
  ctx.lineTo(w+p, h-25+p);
  for (var j=Math.round((Math.ceil(tbegin/tmark)*tmark)-tbegin); j<w; j+=tmark) {
    var time = ((tbegin+j)/wa_zoomLevelsPps[wa_pixelsPerSecond]);
    if (time != 0) {
      if (userSettings["player"]["smpte"]) {
        ctx.fillText(toTimecode(time), j-28, h-5);
      }
      else {
        if (parsetime(time).length > 5) ctx.fillText(parsetime(time), j-20, h-5);
        else ctx.fillText(parsetime(time), j-14, h-5);
      }
    }
    ctx.moveTo(j+p, h-15+p);
    ctx.lineTo(j+p, h-25+p);
  }
  for (var j=Math.round((Math.ceil(tbegin/tsubmark)*tsubmark)-tbegin); j<w; j+=tsubmark) {
    ctx.moveTo(j+p, h-20+p);
    ctx.lineTo(j+p, h-25+p);
  }
  ctx.stroke();
  ctx.closePath();
  $('#subtitle-waveform-canvas').css('left', $('#subtitle-waveform-container').scrollLeft()+'px');
}


function printWaveform() {
  printWaveformAlt();
  /*
  if (wa_waveformData != null) {
    var valuesperpixel = parseInt(wa_waveformDataHz/wa_zoomLevelsPps[wa_pixelsPerSecond]);
    var p = 0.5 / window.devicePixelRatio;
    var w = $('#subtitle-waveform-container').width();
    var h = $('#subtitle-waveform-container').height();
    var c = $('#subtitle-waveform-canvas')[0]; 
    c.width = w;
    c.height = h;
    h = h-25;
    var ctx = c.getContext('2d');
    ctx.fillStyle = 'rgb(128,128,128)';
    ctx.beginPath();
    ctx.moveTo(p, h);
    var firstValIndex;
    if ((tlplayer.currentTime*wa_zoomLevelsPps[wa_pixelsPerSecond]) < w/2) firstValIndex = 0;
    else if (tlplayer.currentTime > tlplayer.duration-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) firstValIndex = Math.round((Math.round(Math.round((tlplayer.duration-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond]))*wa_waveformDataHz)/valuesperpixel)*valuesperpixel)-((w/2)*valuesperpixel));
    else firstValIndex = Math.round((Math.round((Math.round(tlplayer.currentTime*wa_waveformDataHz)/valuesperpixel))*valuesperpixel)-((w/2)*valuesperpixel));
    for (var i=firstValIndex, j=0; j<w && i<wa_waveformData.length; i+=valuesperpixel, j++) {
      ctx.lineTo((j%w)+p, h-wa_waveformData[i]+p);
    }
    ctx.lineTo(c.width + p, h);
    ctx.fill();
    ctx.closePath();
  }
  */
}

function printWaveformAlt() {
  if (wa_waveformData != null) {
    var valuesperpixel = parseInt(wa_waveformDataHz/wa_zoomLevelsPps[wa_pixelsPerSecond]);
    var p = 0.5 / window.devicePixelRatio;
    var w = $('#subtitle-waveform-container').width();
    var h = $('#subtitle-waveform-container').height();
    var c = $('#subtitle-waveform-canvas')[0];
    c.width = w;
    c.height = h;
    h = h-25;
    var ctx = c.getContext('2d');
    ctx.fillStyle = 'rgb(128,128,128)';
    ctx.beginPath();
    ctx.moveTo(p, h);
    var firstValIndex;
    if ((tlplayer.currentTime*wa_zoomLevelsPps[wa_pixelsPerSecond]) < w/2) firstValIndex = 0;
    else if (tlplayer.currentTime > tlplayer.duration-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond])) {
      firstValIndex = Math.round((Math.round(Math.round((tlplayer.duration-((w/2)/wa_zoomLevelsPps[wa_pixelsPerSecond]))*wa_waveformDataHz)/valuesperpixel)*valuesperpixel)-((w/2)*valuesperpixel));
      if (firstValIndex < 0) firstValIndex = 0;
    }
    else firstValIndex = Math.round((Math.round((Math.round(tlplayer.currentTime*wa_waveformDataHz)/valuesperpixel))*valuesperpixel)-((w/2)*valuesperpixel));
    for (var i=firstValIndex, j=0; j<w && i<wa_waveformData.length; i+=valuesperpixel, j++) {
      ctx.lineTo((j%w)+p, h-wa_waveformData[i]+p);
    }
    ctx.lineTo(c.width + p, h);
    ctx.fill();
    ctx.closePath();
    ctx.fillStyle = 'rgba(128,128,128,0.33)';
    ctx.beginPath();
    ctx.moveTo(p, h);
    for (var i=firstValIndex, j=0; j<w && i<wa_waveformData.length; i+=valuesperpixel, j++) {
      ctx.lineTo((j%w)+p, h+wa_waveformData[i]+p);
    }
    ctx.lineTo(c.width + p, h);
    ctx.fill();
    ctx.closePath();
  }
}

$('#subtitle-waveform-container').mousedown(function(e) {
  e.preventDefault();
  if (wa_resizingElement == null && wa_movingElement == null) {
    if ($(e.target).is("#subtitle-waveform-void")) {
      if (e.which == 3 && !tlplayer.paused) tlplayer.pause(); 
      var offX = (e.offsetX || e.clientX - $(e.target).offset().left);
      tlplayer.currentTime = offX/wa_zoomLevelsPps[wa_pixelsPerSecond];
    }
  }
});

$('#subtitle-waveform-container').on("mouseup", function(e) {
  $('#subtitle-waveform-container').unbind("mousemove");
  if (wa_resizingElement != null && wa_movingElement == null) {
    var sid = $(wa_resizingElement).index('.swf-sub');
    $(wa_resizingElement).children('.swf-sub-leftrs').show();
    $(wa_resizingElement).children('.swf-sub-rightrs').show();
    $(wa_resizingElement).css('border-left', '');
    $(wa_resizingElement).css('border-right', '');
  }
  else if (wa_movingElement != null && wa_resizingElement == null) {
    var sid = $(wa_movingElement).index('.swf-sub');
    $(wa_movingElement).children('.swf-sub-leftrs').show();
    $(wa_movingElement).children('.swf-sub-rightrs').show();
    $(wa_movingElement).css('border-left', '');
    $(wa_movingElement).css('border-right', '');
  }
  wa_resizingElement = null;
  wa_movingElement = null;
  if (wa_joinSegments == 1) joinSegments(sid, sid-1);
  else if (wa_joinSegments == 2) joinSegments(sid, sid+1);
  wa_joinSegments = 0;
  e.preventDefault();
  e.stopPropagation();
  $(tlplayer).trigger("timeupdate");
  return false;
});

function wa_zoomOut() {
  $('#wa_contextMenu_zoomIn').removeClass("disabled");
  if (wa_pixelsPerSecond > 0) wa_pixelsPerSecond--;
  if (wa_pixelsPerSecond == 0) $('#wa_contextMenu_zoomOut').addClass("disabled");
  wa_minWidth = wa_zoomLevelsPps[wa_pixelsPerSecond]; // 1s
  genWaveformSubtitles();
  printWaveform();
  printTimebar();
  updateWaveformPosition(false);
}

function wa_zoomIn() {
  $('#wa_contextMenu_zoomOut').removeClass("disabled");
  if (wa_pixelsPerSecond < wa_zoomLevelsPps.length-1) wa_pixelsPerSecond++;
  if (wa_pixelsPerSecond == wa_zoomLevelsPps.length-1) $('#wa_contextMenu_zoomIn').addClass("disabled");
  wa_minWidth = wa_zoomLevelsPps[wa_pixelsPerSecond]; // 1s
  genWaveformSubtitles();
  printWaveform();
  printTimebar();
  updateWaveformPosition(false);
}

$('#subtitle-waveform-container').bind('mousewheel DOMMouseScroll', function(e) {
  e.preventDefault();
  if (e.originalEvent.wheelDelta > 0 || e.originalEvent.detail < 0) {
    // scroll up
    wa_zoomIn();
  }
  else {
    // scroll down
    wa_zoomOut();
  }
  return false;
});

var wa_movingElement = null;
var wa_resizingElement = null;
var wa_initialLeft = 0;
var wa_initialRight = 0;
var wa_initialClick = 0;
var wa_resizingDirection = 0; // 0: left, 1: right
var wa_joinSegments = 0; // 0 = no, 1 = prev (left); 2 = next (right);
var wa_minWidth = wa_zoomLevelsPps[wa_pixelsPerSecond]; // min 1s

/* Join segment indexes a and b */
function joinSegments(a, b, reg_op) {
  if (typeof reg_op == 'undefined') reg_op = true;
  if (editing != false) $(editing).trigger("confirm");
  tlplayer.pause();
  if (a >= 0 && b >= 0 && a < dfxptar['s'].length && b < dfxptar['s'].length && Math.abs(a-b) == 1) {
    dfxptar['s'][a]['edited'] = "full";
    dfxptar['s'][a]['aT'] = "human";
    dfxptar['s'][a]['w'] = [];
    if (a < b) {
      if (reg_op) register_operation({ op: "join", index: a, src: false, fe: dfxptar['s'][a]['e'], ftext: dfxptar['s'][a]['text'], lb: dfxptar['s'][b]['b']+0.01, ltext: dfxptar['s'][b]['text'] });
      dfxptar['s'][a]['e'] = dfxptar['s'][b]['e'];
      dfxptar['s'][a]['text'] = $('#transcription-tar li:eq('+a+') .segment-text .word-text').text()+" "+$('#transcription-tar li:eq('+b+') .segment-text .word-text').text();
    }
    else {
      if (reg_op) {
        if (dfxptar['s'][b]['e'] > dfxptar['s'][a]['b']) register_operation({ op: "join", index: b, src: false, fe: dfxptar['s'][b]['e'], ftext: dfxptar['s'][b]['text'], lb: dfxptar['s'][b]['e']+0.01, ltext: dfxptar['s'][a]['text'] });
        else register_operation({ op: "join", index: b, src: false, fe: dfxptar['s'][b]['e'], ftext: dfxptar['s'][b]['text'], lb: dfxptar['s'][a]['b']+0.01, ltext: dfxptar['s'][a]['text'] });
      }
      dfxptar['s'][a]['b'] = dfxptar['s'][b]['b'];
      dfxptar['s'][a]['text'] = $('#transcription-tar li:eq('+b+') .segment-text .word-text').text()+" "+$('#transcription-tar li:eq('+a+') .segment-text .word-text').text();
    }
    $('#transcription-tar li:eq('+a+') .segment-text .word-text').text(dfxptar['s'][a]['text']);
    if (tlplayer.currentTime >= dfxptar['s'][a]['b'] && tlplayer.currentTime < dfxptar['s'][a]['e']) {
      $('#transcription-tar li.current').removeClass("current");
      $('#transcription-tar li:eq('+a+')').addClass("current");
      $('#sbar-tar').html(dfxptar['s'][a]['text']);
    }
    var scps = dfxptar['s'][a]['text'].replace(/\s+/g, '').length/(dfxptar['s'][a]['e']-dfxptar['s'][a]['b']);
    if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+a+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
    else $('#transcription-tar li:eq('+a+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
    $('#subtitle-waveform-container .swf-sub:eq('+a+')').css('left', Math.round(dfxptar['s'][a]['b']*wa_zoomLevelsPps[wa_pixelsPerSecond])+"px");
    $('#subtitle-waveform-container .swf-sub:eq('+a+')').width(Math.round((dfxptar['s'][a]['e']-dfxptar['s'][a]['b'])*wa_zoomLevelsPps[wa_pixelsPerSecond]));
    $('#subtitle-waveform-container .swf-sub:eq('+a+') .swf-sub-text').text(dfxptar['s'][a]['text']);
    $('#subtitle-waveform-container .swf-sub:eq('+b+')').remove();
    $('#transcription-tar li:eq('+b+')').remove();
    sremovedtar.push(parseInt(dfxptar['s'][b]['sI']));
    dfxptar['s'].splice(b, 1);
    if (!$(editing).is(':visible')) editing = false;
    if (currents == b) updateCurrentSubtitle(false);
    updateSegmentTimeControl();
  }
}

$('#subtitle-waveform-container').on("mouseenter", ".swf-sub", function(e) {
  $(this).children('.swf-sub-remove').show();
});

$('#subtitle-waveform-container').on("mouseleave", ".swf-sub", function(e) {
  $(this).children('.swf-sub-remove').hide();
});

$('#subtitle-waveform-container').on("mousedown", function(e) {
  registerMousemoveEvent();
});

var firstMoveOrResize = true;
function registerMousemoveEvent() {
  $('#subtitle-waveform-container').on("mousemove", function(e) {
    if (e.which == 1 && ($(e.target).hasClass("swf-sub") || $(e.target).hasClass("swf-sub-text")) && wa_movingElement == null && wa_resizingElement == null) {
      if ($(e.target).hasClass("swf-sub")) wa_movingElement = $(e.target);
      else if ($(e.target).hasClass("swf-sub-text")) wa_movingElement = $(e.target).parent();
      wa_initialLeft = parseInt($(wa_movingElement).css('left'));
      wa_initialRight = parseInt($(wa_movingElement).css('left'))+$(wa_movingElement).width();
      wa_initialClick = e.pageX;
      firstMoveOrResize = true;
    }
    else if (wa_movingElement != null && wa_resizingElement == null) {
      var sid = $(wa_movingElement).index('.swf-sub');
      if (firstMoveOrResize) {
        firstMoveOrResize = false;
        register_operation({ op: "move", src: false, index: sid, b: dfxptar['s'][sid]['b'], e: dfxptar['s'][sid]['e'], left: wa_initialLeft, width: $(wa_movingElement).width() });
      }
      /* moving segment: detect collisions ;) */
      var cleft = wa_initialLeft-(wa_initialClick-e.pageX);
      var fleft = 0;
      if (cleft < 0) cleft = 0;
      if (sid > 0 && cleft <= (parseInt($('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')').css('left'))+$('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')').width())) {
        var prevs = $('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')');
        wa_joinSegments = 1;
        $(wa_movingElement).children('.swf-sub-leftrs').hide();
        $(prevs).children('.swf-sub-rightrs').hide();
        $(wa_movingElement).css('border-left', '1px dashed #428bca');
        $(prevs).css('border-right', '1px dashed #428bca');
        fleft = parseInt($(prevs).css('left'))+$(prevs).width();
      }
      else if ((sid+1) < dfxptar['s'].length && (cleft+$(wa_movingElement).width()) >= parseInt($('#subtitle-waveform-container .swf-sub:eq('+(sid+1)+')').css('left'))) {
        var nexts = $('#subtitle-waveform-container .swf-sub:eq('+(sid+1)+')');
        wa_joinSegments = 2;
        $(wa_movingElement).children('.swf-sub-rightrs').hide();
        $(nexts).children('.swf-sub-leftrs').hide();
        $(wa_movingElement).css('border-right', '1px dashed #428bca');
        $(nexts).css('border-left', '1px dashed #428bca');
        fleft = parseInt($(nexts).css('left'))-$(wa_movingElement).width();
      }
      else {
        wa_joinSegments = 0;
        $(wa_movingElement).children('.swf-sub-leftrs').show();
        $(wa_movingElement).children('.swf-sub-rightrs').show();
        $('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')').children('.swf-sub-rightrs').show();
        $('#subtitle-waveform-container .swf-sub:eq('+(sid+1)+')').children('.swf-sub-leftrs').show();
        $(wa_movingElement).css('border-left', '');
        $(wa_movingElement).css('border-right', '');
        $('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')').css('border-right', '');
        $('#subtitle-waveform-container .swf-sub:eq('+(sid+1)+')').css('border-left', '');
        fleft = cleft;
      }
      if (fleft < 0) fleft = 0;
      $(wa_movingElement).css('left', fleft+"px");
      dfxptar['s'][sid]['b'] = fleft/wa_zoomLevelsPps[wa_pixelsPerSecond];
      dfxptar['s'][sid]['e'] = (fleft+$(wa_movingElement).width())/wa_zoomLevelsPps[wa_pixelsPerSecond];
      dfxptar['s'][sid]['edited'] = "full";
      updateSegmentTimeControl();
    }
    else if (wa_resizingElement != null && wa_movingElement == null) {
      /* resizing segment: detect collisions ;) */
      var sid = $(wa_resizingElement).index('.swf-sub');
      if (firstMoveOrResize) {
        firstMoveOrResize = false;
        register_operation({ op: "resize", src: false, index: sid, b: dfxptar['s'][sid]['b'], e: dfxptar['s'][sid]['e'], left: wa_initialLeft, width: $(wa_resizingElement).width() });
      }
      dfxptar['s'][sid]['edited'] = "full";
      var sw = 0;
      var cleft = 0;
      var fleft = 0;
      if (wa_resizingDirection == 0) {
        /* resizing direction: left */
        cleft = (wa_initialLeft-(wa_initialClick-e.pageX));
        sw = wa_initialRight-cleft;
        if (sw >= wa_minWidth) {
          fleft = cleft;
          if (sid > 0) {
            var prevs = $('#subtitle-waveform-container .swf-sub:eq('+(sid-1)+')');
            if (cleft <= (parseInt($(prevs).css('left'))+$(prevs).width())) {
              wa_joinSegments = 1;
              $(wa_resizingElement).children('.swf-sub-leftrs').hide();
              $(prevs).children('.swf-sub-rightrs').hide();
              $(wa_resizingElement).css('border-left', '1px dashed #428bca');
              $(prevs).css('border-right', '1px dashed #428bca');
              cleft = parseInt($(prevs).css('left'))+$(prevs).width();
              fleft = parseInt($(prevs).css('left'));
              sw = wa_initialRight-cleft;
              $(wa_resizingElement).css('left', cleft+"px");
              $(wa_resizingElement).width(sw+"px");
              if (fleft < 0) fleft = 0;
              dfxptar['s'][sid]['b'] = fleft/wa_zoomLevelsPps[wa_pixelsPerSecond];
            }
            else {
              wa_joinSegments = 0;
              $(wa_resizingElement).children('.swf-sub-leftrs').show();
              $(prevs).children('.swf-sub-rightrs').show();
              $(wa_resizingElement).css('border-left', '');
              $(prevs).css('border-right', '');
              if (fleft < 0) fleft = 0;
              $(wa_resizingElement).css('left', fleft+"px");
              $(wa_resizingElement).width(sw+"px");
              dfxptar['s'][sid]['b'] = fleft/wa_zoomLevelsPps[wa_pixelsPerSecond];
              dfxptar['s'][sid]['e'] = (fleft+sw)/wa_zoomLevelsPps[wa_pixelsPerSecond];
            }
          }
          else {
            if (cleft < 0) cleft = 0;
            if (fleft < 0) fleft = 0;
            $(wa_resizingElement).css('left', fleft+"px");
            $(wa_resizingElement).width(sw+"px");
            dfxptar['s'][sid]['b'] = fleft/wa_zoomLevelsPps[wa_pixelsPerSecond];
            dfxptar['s'][sid]['e'] = (fleft+sw)/wa_zoomLevelsPps[wa_pixelsPerSecond];
          }
        }
        else {
          sw = wa_minWidth;
          fleft = wa_initialRight-sw;
          if (fleft < 0) fleft = 0;
          $(wa_resizingElement).css('left', fleft+"px");
          $(wa_resizingElement).width(sw+"px");
          dfxptar['s'][sid]['b'] = fleft/wa_zoomLevelsPps[wa_pixelsPerSecond];
          dfxptar['s'][sid]['e'] = (fleft+sw)/wa_zoomLevelsPps[wa_pixelsPerSecond];
        }
      }
      else {
        /* resizing direction: right */
        sw = wa_initialRight-wa_initialLeft-(wa_initialClick-e.pageX);
        if (sw >= wa_minWidth) {
          if ((sid+1) < dfxptar['s'].length) {
            var nexts = $('#subtitle-waveform-container .swf-sub:eq('+(sid+1)+')');
            if ((wa_initialLeft+sw) >= parseInt($(nexts).css('left'))) {
              wa_joinSegments = 2;
              $(wa_resizingElement).children('.swf-sub-rightrs').hide();
              $(nexts).children('.swf-sub-leftrs').hide();
              $(wa_resizingElement).css('border-right', '1px dashed #428bca');
              $(nexts).css('border-left', '1px dashed #428bca');
              sw = parseInt($(nexts).css('left'))-wa_initialLeft;
            }
            else {
              wa_joinSegments = 0;
              $(wa_resizingElement).children('.swf-sub-rightrs').show();
              $(nexts).children('.swf-sub-leftrs').show();
              $(wa_resizingElement).css('border-right', '');
              $(nexts).css('border-left', '');
            }
          }
        }
        else sw = wa_minWidth;
        $(wa_resizingElement).width(sw+"px");
        dfxptar['s'][sid]['e'] = (wa_initialLeft+sw)/wa_zoomLevelsPps[wa_pixelsPerSecond];
      }
      var sduration = $(wa_resizingElement).width()/wa_zoomLevelsPps[wa_pixelsPerSecond];
      var scps = 0.0;
      if (sduration > 0) scps = dfxptar['s'][sid]['text'].replace(/\s+/g, '').length/sduration;
      if (scps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || scps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) $('#transcription-tar li:eq('+sid+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsgreen").addClass("cpsred");
      else $('#transcription-tar li:eq('+sid+') span.cpsmeter').html(scps.toFixed(1)+' cps').removeClass("cpsred").addClass("cpsgreen");
      updateSegmentTimeControl();
    }
  });
}

$('#subtitle-waveform-container').on("mouseup", ".swf-sub", function(e) {
  $('#subtitle-waveform-container').unbind("mousemove");
  firstMoveOrResize = true;
  if (wa_resizingElement == null && wa_movingElement == null && e.which == 1) {
    var sid = $(this).index('.swf-sub');
    if ($(e.target).hasClass('swf-sub-remove') || $(e.target).parent().hasClass('swf-sub-remove')) {
      removeSegment(sid, false);
    }
    else {
      if (editing == false) {
        if ($('#transcription-tar').is(':visible')) $('#transcription-tar li:eq('+sid+') .segment-text').trigger("click");
        else if ($('#sbar-tar').is(':visible')) {
          tlplayer.currentTime = dfxptar['s'][sid]['b'];
          updateCurrentSubtitle(false);
          $('#sbar-tar').trigger("click");
        }
      }
    }
  }
});

$('#subtitle-waveform-container').on("mousedown", ".swf-sub-leftrs", function(e) {
  e.preventDefault();
  e.stopPropagation();
  firstMoveOrResize = true;
  wa_resizingElement = $(this).parent();
  wa_initialLeft = parseInt($(this).parent().css('left'));
  wa_initialRight = parseInt($(this).parent().css('left'))+$(this).parent().width();
  wa_initialClick = e.pageX;
  wa_resizingDirection = 0;
  wa_joinSegments = 0;
  registerMousemoveEvent();
  return false;
});

$('#subtitle-waveform-container').on("mousedown", ".swf-sub-rightrs", function(e) {
  e.preventDefault();
  e.stopPropagation();
  firstMoveOrResize = true;
  wa_resizingElement = $(this).parent();
  wa_initialLeft = parseInt($(this).parent().css('left'));
  wa_initialRight = parseInt($(this).parent().css('left'))+$(this).parent().width();
  wa_initialClick = e.pageX;
  wa_resizingDirection = 1;
  wa_joinSegments = 0;
  registerMousemoveEvent();
  return false;
});

// begin auxiliar functions

function getLanguageName(lcode) {
  var lname = lcode;
  $.each(langs['langs'], function(key, value) {
    if (value["lang_code"] == lcode) {
      lname = value["lang_name"];
      return false;
    }
  });
  return lname;
}

function isMobile() {
  try { document.createEvent("TouchEvent"); return true; }
  catch(e) { return false; }
}

$.fn.selectRange = function(start, end) {
  if(!end) end = start; 
  return this.each(function() {
    if (this.setSelectionRange) {
      this.focus();
      this.setSelectionRange(start, end);
    } else if (this.createTextRange) {
      var range = this.createTextRange();
      range.collapse(true);
      range.moveEnd('character', end);
      range.moveStart('character', start);
      range.select();
    }
  });
};

function pad(num){
  return ((num < 10 ? '0' : '') + num);
}

function parsetime(seconds) {
  var secs = pad(Math.floor(seconds)%60);
  var mins = pad(Math.floor(seconds/60)%60);
  var hours = pad(Math.floor(seconds/3600)%60);
  str = (hours > 0 ? hours+':'+mins+':'+secs : mins+':'+secs);
  return str;
}
// end auxiliar functions

$('video').on("timeupdate", function(e) {
  // update current time span and video progress bar
  if (wa_enabled && userSettings["player"]["smpte"]) $('span.current-time').html(toTimecode(this.currentTime));
  else $('span.current-time').html(parsetime(this.currentTime));
  $('.progress-left').css('width', (($(this).get(0).currentTime/$(this).get(0).duration)*100).toFixed(2)+'%');
  updateCurrentSubtitle();
  if (this.paused && wa_enabled) updateWaveformPosition();
});

$('video').on("progress", function(e) {
  // update progress loaded bar
  if ($(this).get(0).buffered.length > 0) {
    $('.progress-loaded').css('width', (($(this).get(0).buffered.end(0)/$(this).get(0).duration)*100).toFixed(2)+'%');
  }
});

$('video').on("play", function(e) {
  $('button.play').css({backgroundPosition: '-252px -28px'});
  // if editing enabled and current time > segment end time, start playing from the beggining of the current segment
  if (editing != false && emode == 0 && $(this).get(0).currentTime >= dfxptar['s'][currents]['e']) $(this).get(0).currentTime = parseFloat(dfxptar['s'][currents]['b'])+(1/frameRate);
  if (wa_enabled) {
    wa_animationFrameId = window.requestAnimationFrame(updateWaveformPosition);
  }
  update_session_status();
});

$('video').on("pause", function(e) {
  $('button.play').css({backgroundPosition: '-84px -28px'});
  if (wa_enabled) {
    window.cancelAnimationFrame(wa_animationFrameId);
  }
});

$('video').on("loadedmetadata", function(e) {
  if (wa_enabled && userSettings["player"]["smpte"]) $('span.duration-time').html(toTimecode(this.duration));
  else $('span.duration-time').html(parsetime(this.duration));
  $('video').css('height', $('#tl-player').width()*(1/userSettings["player"]["aspect_ratio"]));
  if (vmode == 1) $('.transcription-container').height($('#tl-player').height());
});

/* UTILITY FUNCTIONS */

function getCaretCharacterOffsetWithin(element) {
  var caretOffset = 0;
  var doc = element.ownerDocument || element.document;
  var win = doc.defaultView || doc.parentWindow;
  var sel;
  if (typeof win.getSelection != "undefined") {
    var range = win.getSelection().getRangeAt(0);
    var preCaretRange = range.cloneRange();
    preCaretRange.selectNodeContents(element);
    preCaretRange.setEnd(range.endContainer, range.endOffset);
    caretOffset = preCaretRange.toString().length;
  } 
  else if ((sel = doc.selection) && sel.type != "Control") {
    var textRange = sel.createRange();
    var preCaretTextRange = doc.body.createTextRange();
    preCaretTextRange.moveToElementText(element);
    preCaretTextRange.setEndPoint("EndToEnd", textRange);
    caretOffset = preCaretTextRange.text.length;
  }
  return caretOffset;
}

function secondsToTime(secs) {
  var secs_split = secs.toString().split(".");
  var sec_numb = parseInt(secs_split[0]);
  var sec_ms = "0";
  if (secs_split.length > 1) sec_ms = secs_split[0].substr(0,3);
  var hours = Math.floor(sec_numb / 3600);
  var minutes = Math.floor((sec_numb - (hours * 3600)) / 60);
  var seconds = sec_numb - (hours * 3600) - (minutes * 60);
  while (sec_ms.length < 3) sec_ms = sec_ms + "0";
  if (hours   < 10) { hours = "0"+hours;}
  if (minutes < 10) { minutes = "0"+minutes;}
  if (seconds < 10) { seconds = "0"+seconds;}
  return hours+':'+minutes+':'+seconds+','+sec_ms;
}

function onResize() {
  if (wa_enabled) {
    if (userSettings["player"]["aspect_ratio"] != 0) $('video').css('height', ($('#tl-player').width()*(1/userSettings["player"]["aspect_ratio"]))-$('#subtitle-waveform-container').height()/1.5);
    else $('video').css('height', '');
    printWaveform();
    printTimebar();
  }
  else {
    if (userSettings["player"]["aspect_ratio"] != 0) $('video').css('height', $('#tl-player').width()*(1/userSettings["player"]["aspect_ratio"]));
    else $('video').css('height', '');
  }
  // dynamic font-size for closed caption (vmode=0)
  $('.subtitle-bar').css('font-size', (Math.round($('#tl-player').width()/66) < 11) ? "11pt" : Math.round($('#tl-player').width()/54) + 'pt');
  if (vmode == 1) $('.transcription-container').height($('#tl-player').height());
  if (vmode == 2) {
    $('#tl-editor').width($('#tl-player').width());
    var th = ($(window).height()*0.9)-$('#tl-player').height();
    th = (th > $('#tl-player').height()) ? $('#tl-player').height() : th;
    $('.transcription-container').height(th);
  }
  if (vmode == 3) {
    $('.transcription-container').height($(window).height()*0.6);
  }
  setTimeout(function() {
    if (dfxpsrc != null && dfxptar != null && $('#transcription-src').is(':visible') && $('#transcription-tar').is(':visible') && (dfxpsrc['s'].length == dfxptar['s'].length)) {
      $.each(dfxptar['s'], function(key, val) {
        // set equal height for both transcription and translation
        $('#transcription-tar li:eq('+key+')').css("height", "");
        $('#transcription-src li:eq('+key+')').css("height", "");
        var hsrc = Math.round($('#transcription-src li:eq('+key+')').height());
        var htar = Math.round($('#transcription-tar li:eq('+key+')').height());
        if (hsrc != htar) {
          if (htar < hsrc) $('#transcription-tar li:eq('+key+')').height(hsrc);
          else $('#transcription-src li:eq('+key+')').height(htar);
        }
      });
    }
    if ($('#tl-editor').is(':visible') && vmode > 0) {
      updateScrollPos(false, false);
      if (dfxpsrc != null && $('#transcription-src').is(':visible')) updateScrollPos(false, true);
    }
    if (helpvisible) $('.tooltipster').tooltipster('reposition');
  }, 0);
}

/* 
  EXTERNAL EVENTS HANDLING 
*/

var resizeTimeout;

$(window).resize(function() {
  if (window.screenTop || window.screenY) {
    // not fullscreen
    if (helpvisible) $('.tooltipster').tooltipster('reposition');
    if ($('#transcription-src').is(':visible') && $('#transcription-tar').is(':visible') && (dfxpsrc['s'].length == dfxptar['s'].length)) {
      clearTimeout(resizeTimeout);
      resizeTimeout = setTimeout(function() {
        onResize();
      }, 500);
    }
    else onResize();
  }
});

$(window).on("beforeunload", function() {
  if (check_unsaved_changes()) {
    if (mod_session != false && "rcode" in mod_session && mod_session["rcode"] == 0) {
      end_mod_session();
    }
    setTimeout(function() {
      setTimeout(function() {
        start_mod_session();
      }, 1000);
    }, 1);
    return "Your last changes have not been saved. If you close this window, your unsaved changes will be discarded. To avoid losing your changes, stay in the page and press the save button on the player button bar.";
  }
  else {
    if (mod_session != false && "rcode" in mod_session && mod_session["rcode"] == 0) {
      end_mod_session();
    }
    return;
  }
});

$(document).on("tLload", function(e) {
  if (loadFlag <= 1) {
    loadFlag = 0;
    if (firstLoad) {
      firstLoad = false;
      var checkLoadState = setInterval(function() {
        if (tlplayer.readyState == 4) {
          setTimeout(function() {
            if (userSettings["player"]["advanced"] == true) {
              enableWa();
              fillTranscriptionContainer(false);
              if (dfxpsrc != null) fillTranscriptionContainer(true);
              $('button.toggle-advanced-mode').html('<i class="fa fa-dot-circle-o"></i> Basic mode');
            }
            $('.processing').remove();
            onResize();
            // autoplay on load
            setTimeout(function() {
              if ($('.apprise').is(':hidden')) {
                tlplayer.play();
              }
            }, 1000);
          }, 0);
          clearInterval(checkLoadState);
        }
        else if (tlplayer.networkState == 0 || tlplayer.networkState == 3) {
          console.log("Couldn't load video source (check if mp4/ogv/webm video URLs are accessible by users)");
          $('.processing').remove();
          $('#tl-player').hide();
          $('#tl-editor').hide();
          bootbox.alert("It was not possible to load the video sources: mp4/ogv/webm video URLs are not accessible");
          clearInterval(checkLoadState);
        }
      }, 250);
    }
    else { // firstLoad = false;
      onResize();
      updateCurrentSubtitle(false);
      $('.processing').remove();
    }
  }
  else if (loadFlag > 1) {
    loadFlag = loadFlag-1;
    if (request_data["lang"] != langs["media_lang"]) {
      loadTranscriptionOK(langs["media_lang"], true, (mod_session != false && "session_id" in mod_session ? mod_session["session_id"] : null));
    }
  }
});

/* USER SETTINGS */
function onSettingChange() {
  /* Saving settings not available in standalone version */
  return false;
}

$('#mllp_user_settings').click(function() {
  showSettingsPanel();
});

function keypressCodeNicer(kpc) {
  if (typeof kpc === "undefined") return " Not set";
  var text = kpc.replace(/left/g, "&#8592;").replace(/up/g, "&#8593;").replace(/right/g, "&#8594;").replace(/down/g, "&#8595;").split(' ');
  var nicertext = "";
  for (var i=0; i<text.length; i++) {
    if (i > 0) nicertext += "+" + "<div class=\"keyboard-key\">" + text[i].charAt(0).toUpperCase() + text[i].slice(1) + "</div>";
    else nicertext += "<div class=\"keyboard-key\">" + text[i].charAt(0).toUpperCase() + text[i].slice(1) + "</div>";
  }
  return nicertext;
}

var isSettingShortcut = false;
function showSettingsPanel() {
  bootbox.dialog({
    title: "User settings",
    message: '<div class="row"><div class="col-md-12"><form class="form-horizontal">'+'<div class="panel panel-default"><div class="panel-heading" style="border-bottom: none;"><i class="fa fa-cog"></i> General settings</div><div class="panel-body"><div class="form-group"><label class="col-md-8 control-label" for="autoconfirm">Confirm segment automatically after being played</label><div class="col-md-4"> <button id="btn_autoconfirm" type="button" class="btn btn-default'+((userSettings["player"]["autoconfirm"]) ? ' active">Enabled' : '">Disabled')+'</button></div></div>'+'<div class="form-group"><label class="col-md-8 control-label" for="tlalignment">Highlight segment alignment if two languages present</label><div class="col-md-4"><button id="btn_tlalignment" type="button" class="btn btn-default'+((userSettings["player"]["tlalignment"]) ? ' active">Enabled' : '">Disabled')+'</button></div></div>'+'<div class="form-group"><label class="col-md-8 control-label" for="autoplay">Start playing automatically after confirm segment</label><div class="col-md-4"><button id="btn_autoplay" type="button" class="btn btn-default'+((userSettings["player"]["autoplay"]) ? ' active">Enabled' : '">Disabled')+'</button></div></div>'+'<div class="panel panel-default"><div class="panel-heading"><i class="fa fa-object-ungroup"></i> Advanced options</div><div class="panel-body"><div class="form-group"><label class="col-md-8 control-label" for="chars_per_sec" title="Optimum number of characters per second">Optimum character rate (cps)</label><div class="col-md-4"><input id="chars_per_sec" type="number" class="form-control" value="'+userSettings["player"]["cps"]+'"/></div></div></div></div>'+'<div class="panel panel-default"><div class="panel-heading" style="border-bottom: none;"><i class="fa fa-keyboard-o"></i> Keyboard shortcuts</div><div class="panel-body" style="padding: 0;"><table id="settings_key_shortcuts_table" class="table table-striped"><tbody><tr><td>Play/Pause</td><td data-scid="playpause">'+keypressCodeNicer(userSettings["shortcuts"]["playpause"])+'</td></tr><tr><td>Replay segment</td><td data-scid="replay">'+keypressCodeNicer(userSettings["shortcuts"]["replay"])+'</td></tr><tr><td>Jump -1.5s</td><td data-scid="jumpback">'+keypressCodeNicer(userSettings["shortcuts"]["jumpback"])+'</td></tr><tr><td>Jump +1.5s</td><td data-scid="jumpforward">'+keypressCodeNicer(userSettings["shortcuts"]["jumpforward"])+'</td></tr><tr><td>Jump -1 frame</td><td data-scid="frameback">'+keypressCodeNicer(userSettings["shortcuts"]["frameback"])+'</td></tr><tr><td>Jump +1 frame</td><td data-scid="frameforward">'+keypressCodeNicer(userSettings["shortcuts"]["frameforward"])+'</td></tr><tr><td>Previous segment</td><td data-scid="prevsegment">'+keypressCodeNicer(userSettings["shortcuts"]["prevsegment"])+'</td></tr><tr><td>Next segment</td><td data-scid="nextsegment">'+keypressCodeNicer(userSettings["shortcuts"]["nextsegment"])+'</td></tr><tr><td>Edit/Confirm</td><td data-scid="enterconfirm">'+keypressCodeNicer(userSettings["shortcuts"]["enterconfirm"])+'</td></tr><tr><td>Add new segment</td><td data-scid="newsegment">'+keypressCodeNicer(userSettings["shortcuts"]["newsegment"])+'</td></tr><tr><td>Split segment</td><td data-scid="splitsegment">'+keypressCodeNicer(userSettings["shortcuts"]["splitsegment"])+'</td></tr><tr><td>Join current and previous segments</td><td data-scid="joinprev">'+keypressCodeNicer(userSettings["shortcuts"]["joinprev"])+'</td></tr><tr><td>Join current and next segments</td><td data-scid="joinnext">'+keypressCodeNicer(userSettings["shortcuts"]["joinnext"])+'</td></tr><tr><td>Undo last operation</td><td data-scid="undo">'+keypressCodeNicer(userSettings["shortcuts"]["undo"])+'</td></tr></tbody></table></div></div>'+'</div></form></div></div>',
    buttons: {
      success: {
        label: "Save",
        className: "btn-primary",
        callback: function () {
          jQuery("#chars_per_sec").trigger("change");
          onSettingChange();
          if (registeredCombos == null) registerKeypressCombos();
        }
      }
    }
  });

  jQuery("#btn_autoconfirm").click(function() { 
    if ($(this).hasClass("active")) {
      autoconfirm = false;
      userSettings["player"]["autoconfirm"] = false; 
      $(this).removeClass("active");
      $(this).html("Disabled");
    } 
    else {
      autoconfirm = true;
      userSettings["player"]["autoconfirm"] = true;
      jQuery(this).addClass("active");
      $(this).html("Enabled");
    }
  });

  jQuery("#btn_tlalignment").click(function() {
    if ($(this).hasClass("active")) {
      showTranslationAlignments = false;
      userSettings["player"]["tlalignment"] = false;
      $(this).removeClass("active");
      $(this).html("Disabled");
    }
    else {
      showTranslationAlignments = true;
      userSettings["player"]["tlalignment"] = true;
      $(this).addClass("active");
      $(this).html("Enabled");
    }
  });

  jQuery("#btn_autoplay").click(function() {
    if ($(this).hasClass("active")) {
      userSettings["player"]["autoplay"] = false;
      $(this).removeClass("active");
      $(this).html("Disabled");
    }
    else {
      userSettings["player"]["autoplay"] = true;
      $(this).addClass("active");
      $(this).html("Enabled");
    }
  });

  jQuery("#chars_per_sec").on("change", function() {
    userSettings["player"]["cps"] = parseFloat(jQuery(this).val());
    jQuery("span.cpsmeter").each(function() {
      jQuery(this).removeClass("cpsgreen").removeClass("cpsyellow").removeClass("cpsred");
      var tcps = jQuery(this).text().split(" cps");
      tcps = parseFloat(tcps[0]);
      if (tcps > (parseFloat(userSettings["player"]["cps"])+(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"])) || tcps < (parseFloat(userSettings["player"]["cps"])-(userSettings["player"]["cps"]*userSettings["player"]["cps_allowance"]))) jQuery(this).addClass("cpsred");
      else jQuery(this).addClass("cpsgreen");
    });
  });
  jQuery("#chars_per_sec").val(userSettings["player"]["cps"]);

  jQuery("#settings_key_shortcuts_table tr td:nth-child(2)").click(function() {
    if (isSettingShortcut == false) {
      unregisterKeypressCombos();
      isSettingShortcut = true;
      jQuery(this).html("<span style=\"color: #428bca\">Press keys...</span>");
      var thisClass = jQuery(this);
      jQuery(document).on("keydown.shortcuts", function(e) {
        e.preventDefault();
        if (isSettingShortcut) {
          var shortcut = "";
          if (e.ctrlKey) shortcut = "ctrl";
          if (e.shiftKey) {
            if (shortcut.length > 0) shortcut += " shift";
            else shortcut = "shift";
          }
          if (e.altKey) {
            if (shortcut.length > 0) shortcut += " alt";
            else shortcut = "alt";
          }
          if (e.which != 16 && e.which != 17 && e.which != 18) {
            if (shortcut.length > 0) shortcut += " " + window.keypress._keycode_dictionary[e.which];
            else shortcut = window.keypress._keycode_dictionary[e.which];
            userSettings["shortcuts"][thisClass.data("scid")] = shortcut;
            thisClass.html(keypressCodeNicer(shortcut));
            jQuery(document).unbind("keydown.shortcuts");
            isSettingShortcut = false;
            onSettingChange();
            registerKeypressCombos();
          }
          return false;
        }
      });
    }
  });
}

/* User stats */
var stats_src_total_time = 0.0;
var stats_src_last_init = null;
var stats_tar_total_time = 0.0;
var stats_tar_last_init = null;

jQuery(document).on("mouseenter",function(e) {
  stats_activesystem();
});
jQuery(document).on("mouseleave",function(e) {
  stats_disablesystem();
});

var stats_activesystem = function() {
  if (stats_src_last_init == null) stats_src_last_init = new Date().getTime();
  else {
    stats_src_total_time += (new Date().getTime()-stats_src_last_init)/1000.0;
    stats_src_last_init = new Date().getTime();
  }
  if (stats_tar_last_init == null) stats_tar_last_init = new Date().getTime();
  else { 
    stats_tar_total_time += (new Date().getTime()-stats_tar_last_init)/1000.0;
    stats_tar_last_init = new Date().getTime();
  }
};

var stats_disablesystem = function() {
  if (stats_src_last_init != null) stats_src_total_time += (new Date().getTime()-stats_src_last_init)/1000.0;
  stats_src_last_init = null;
  if (stats_tar_last_init != null) stats_tar_total_time += (new Date().getTime()-stats_tar_last_init)/1000.0;
  stats_tar_last_init = null;
};

jQuery(document).on("tLload tLsave", function(e) {
  if (e.container == "src") {
    stats_src_total_time = 0.0;
    stats_src_last_init = null;
  }
  else {
    stats_tar_total_time = 0.0;
    stats_tar_last_init = null;
  }
  stats_activesystem();
});

/*
  MAIN EXECUTION 
*/

function start_mod_session(showalert) {
  var smodurl = "translectures/lib/start_mod_session.php?id="+request_data["id"]+"&author_id="+request_data["author_id"]+"&author_conf="+request_data["author_conf"];
  if ("author_name" in request_data) smodurl = smodurl+"&author_name="+request_data["author_name"];
  if ("hash" in request_data) smodurl = smodurl+"&hash="+request_data["hash"];
  if ("api_user" in request_data) smodurl = smodurl+"&user="+request_data["api_user"];
  if ("expire" in request_data) smodurl = smodurl+"&expire="+request_data["expire"];
  if ("request_key" in request_data) smodurl = smodurl+"&auth_token="+request_data["request_key"];
  jQuery.ajax({
    url: smodurl,
    dataType: "json",
    cache: false,
    async: false,
    success: function(rdata) {
      if (rdata != false && "rcode" in rdata) {
        if (rdata["rcode"] != 0) {
          if (request_data["author_conf"] == 100) {
            if (rdata["author_type"] == "computer") {
              $("button.save").hide();
              $("button.edit_history").hide();
              bootbox.alert("The transcription and translation systems are currently improving the accuracy of the captions for this media. You will not be able to save any changes until this process is finished. Please come back in a few minutes.");
            }
            else {
              bootbox.dialog({
                title: "Cannot open editing session",
                message: "User <strong>"+(rdata["author_name"] != null ? rdata["author_name"] : rdata["author_id"])+"</strong> is currently editing the captions for this media. He started the editing session on "+rdata["started_at"].substring(0, rdata["started_at"].length-10)+"h. You will not be able to save any changes unless you force close the opened session. If you do that, <strong>"+(rdata["author_name"] != null ? rdata["author_name"] : rdata["author_id"])+"</strong> will not be able to save any further changes. Close <strong>"+(rdata["author_name"] != null ? rdata["author_name"] : rdata["author_id"])+"</strong>'s session?",
                buttons: {
                  success: {
                    label: "Force closing",
                    className: "btn-danger",
                    callback: function() {
                      end_mod_session(0, rdata["session_id"], 1);
                    }
                  },
                  danger: {
                    label: "Continue without saving",
                    className: "btn-primary",
                    callback: function() {
                      $("button.save").hide();
                      $("button.edit_history").hide();
                    }
                  }
                }
              });
            }
          }
          else {
            $("button.save").remove();
            bootbox.alert(rdata["rcode_description"]+"<br/><br/>You will not be able to save any changes, but you can still modify the captions and download the subtitles in a local file.");
            console.log("Mod session start failed: "+JSON.stringify(rdata));
          }
        }
        else {
          mod_session = rdata;
          console.log("Mod session started (Session ID: "+mod_session["session_id"]+")");
          if (typeof showalert !== 'undefined' && showalert == true) bootbox.alert("Your session has been opened correctly."); 
          if (request_data["author_conf"] == 100) get_edit_history();
        }
      }
    },
    error: function() {
      $("button.save").remove();
      bootbox.alert("It was not possible to start an editing session. You will not be able to save any changes, but you can still modify the captions and download the subtitles in a local file.");
    }
  });
}

window.end_mod_session = end_mod_session;
function end_mod_session(regenerate, session_id, force) {
  if (typeof session_id == 'undefined' && mod_session == false) return;
  var emodurl = "translectures/lib/end_mod_session.php?id="+request_data["id"]+"&author_id="+request_data["author_id"]+"&author_conf="+request_data["author_conf"];;
  if ("hash" in request_data) emodurl = emodurl+"&hash="+request_data["hash"];
  if ("api_user" in request_data) emodurl = emodurl+"&user="+request_data["api_user"];
  if ("expire" in request_data) emodurl = emodurl+"&expire="+request_data["expire"];
  if ("request_key" in request_data) emodurl = emodurl+"&auth_token="+request_data["request_key"];
  if (typeof session_id !== 'undefined' && typeof session_id !== 'object') emodurl = emodurl+"&session_id="+session_id;
  else if (mod_session != false && "session_id" in mod_session) emodurl = emodurl+"&session_id="+mod_session["session_id"];
  if (typeof force !== 'undefined') {
    emodurl = emodurl+"&force="+force;
    if (force == 1) emodurl = emodurl+"&regenerate=0";
  }
  if (typeof regenerate !== 'undefined') emodurl = emodurl+"&regenerate="+(regenerate ? "1" : "0");
  $.ajax({
    url: emodurl,
    dataType: "json",
    cache: false,
    async: true,
    success: function(data) {
      if (typeof force !== 'undefined' && force == 1) {
        if ("rcode" in data && data["rcode"] == 0) start_mod_session();
        else {
          bootbox.alert("Unable to close the session. You will not be able to save any changes.");
          $("button.save").hide();
        }
      }
      else if (typeof session_id !== 'undefined' && typeof session_id !== 'object') mod_session = false;
    },
    error: function() {
      if (typeof force !== 'undefined' && force == 1) {
        bootbox.alert("Unable to close the session. You will not be able to save any changes.");
        $("button.save").hide();
      }
    }
  });
}

var sessionStatusTimerEnabled = false;
function update_session_status() {
  if (sessionStatusTimerEnabled == false) {
    sessionStatusTimerEnabled = true;
    session_status();
    setTimeout(function() {
      sessionStatusTimerEnabled = false;
    }, 60000);
  }
}

setInterval(function() {
  session_status(1);
}, 60000);

function session_status(alive) {
  if (mod_session != false) {
    var sstatusurl = "translectures/lib/session_status.php?id="+request_data["id"]+"&author_id="+request_data["author_id"]+"&author_conf="+request_data["author_conf"]+"&session_id="+mod_session["session_id"];
    if (typeof alive !== 'undefined') sstatusurl = sstatusurl+"&alive="+(alive ? "1" : "0");
    if ("hash" in request_data) sstatusurl = sstatusurl+"&hash="+request_data["hash"];
    if ("api_user" in request_data) sstatusurl = sstatusurl+"&user="+request_data["api_user"];
    if ("expire" in request_data) sstatusurl = sstatusurl+"&expire="+request_data["expire"];
    if ("request_key" in request_data) sstatusurl = sstatusurl+"&auth_token="+request_data["request_key"];
    $.ajax({
      url: sstatusurl,
      dataType: "json",
      cache: false,
      async: true,
      success: function(rdata) {
        if (rdata != false && "rcode" in rdata) {
          if (rdata["rcode"] != 0) {
            mod_session = false;
            var msg = "";
            if (rdata["rcode"] == 4 && "ended_by_id" in rdata && rdata["ended_by_id"] != null) {
              msg = "Your session has been closed by <strong>"+(rdata["ended_by_name"] != null ? rdata["ended_by_name"] : rdata["ended_by_id"])+"</strong>. ";
            }
            else msg = "Your session has expired. ";
            bootbox.confirm(msg+"Do you want to try re-opening the session?", function(resp) {
              if (resp) {
                start_mod_session(true);
              }
              else {
                bootbox.alert("You will not be able to save any changes, but you can still download the subtitles with your changes and import them in a future session.");
                $('button.save').hide();
              }
            });
          }
        }
        else {
          console.log("ERROR: Could not retrieve nor update session status. Web Service unavailable?");
        }
      },
      error: function() {
        console.log("ERROR: Could not retrieve nor update session status (/lib/session_status.php failure?)");
      }
    });
  }
}

function get_edit_history() {
  var revurl = "translectures/lib/edit_history.php?id="+request_data["id"];
  if ("hash" in request_data) revurl = revurl+"&hash="+request_data["hash"];
  if ("api_user" in request_data) revurl = revurl+"&user="+request_data["api_user"];
  if ("request_key" in request_data) revurl = revurl+"&auth_token="+request_data["request_key"];
  if ("expire" in request_data) revurl = revurl+"&expire="+request_data["expire"];
  $.ajax({
    url: revurl,
    dataType: "json",
    cache: false,
    async: true,
    success: function(rdata) {
      if (rdata != false && "rcode" in rdata && rdata["rcode"] == 0) {
        edit_history_data = rdata;
        if (!("session_id" in request_data)) show_edit_history(true);
      }
      else {
        edit_history_data = false;
        if ("rcode_description" in rdata) console.log("Unable to retrieve edit_history: "+rdata["rcode_description"]);
        else console.log("Unable to retrieve edit_history: call to edit_history.php failed.");
      }
    },
    error: function() {
      console.log("Unable to retrieve edit_history: call to edit_history.php failed.");
      edit_history_data = false;
    }
  });
}

function get_revision_author_name(session_id) {
  if (edit_history_data != false) {
    for (var i=0; i<edit_history_data["edit_history"].length; i++) {
      if (edit_history_data["edit_history"][i]["session_id"] == session_id) {
        if (edit_history_data["edit_history"][i]["author_name"] != null) return edit_history_data["edit_history"][i]["author_name"];
        else return edit_history_data["edit_history"][i]["author_id"];
      }
    }
  }
  return "";
}

function lock_subs(lock) {
  var lockurl = "translectures/lib/lock_subs.php?id="+request_data["id"]+"&author_id="+request_data["author_id"]+"&author_conf="+parseInt(request_data["author_conf"])+"&lock="+(lock == true ? "1" : "0");
  if ("author_name" in request_data) lockurl = lockurl+"&author_name="+request_data["author_name"];
  if ("hash" in request_data) lockurl = lockurl+"&hash="+request_data["hash"];
  if ("api_user" in request_data) lockurl = lockurl+"&user="+request_data["api_user"];
  if ("expire" in request_data) lockurl = lockurl+"&expire="+request_data["expire"];
  if ("request_key" in request_data) lockurl = lockurl+"&auth_token="+request_data["request_key"];
  $.ajax({
    type: 'GET',
    url: lockurl,
    dataType: 'json',
    async: true,
    success: function(response) {
      if (response != false && "rcode" in response && response["rcode"] == 0) {
        if (lock) bootbox.alert("Subtitles locked. Other users will not be able to submit additional modifications for the captions of this media.");
        else bootbox.alert("Subtitles unlocked. Users are now able to submit additional modifications for the captions of this media.");
        langs["subs_locked"] = lock;
      }
      else {
        if (response != false && "rcode_description" in response) bootbox.alert("Subtitles could not be "+(lock ? "" : "un")+"locked: "+response["rcode_description"]);
        else bootbox.alert("Error: Subtitles could not be "+(lock ? "" : "un")+"locked.");
      }
    },
    error: function() {
      bootbox.alert("Error: Subtitles could not be "+(lock ? "" : "un")+"locked.");
    }
  });
}

window.loadTranscription = loadTranscription;
function show_edit_history(only_pending) {
  if (typeof only_pending == 'undefined') only_pending = false;
  var msg = "";
  var some = false;
  msg = "<div class=\"edit-history-table\"><table class=\"table table-hover\" style=\"font-size: 1em;\"><tbody>";
  if (edit_history_data != false && "rcode" in edit_history_data && edit_history_data["rcode"] == 0 && "edit_history" in edit_history_data && edit_history_data["edit_history"].length > 0) {
    for (var i=0; i<edit_history_data["edit_history"].length; i++) {
      if (!only_pending || (edit_history_data["edit_history"][i]["requires_revision"] == true && edit_history_data["edit_history"][i]["revised"] == false)) {
        var rev_langs = "";
        var first = true;
        for (var lang in edit_history_data["edit_history"][i]["edit_stats"]) {
          if (!first) {
            rev_langs += ", ";
            first = false;
          }
          if (edit_history_data["edit_history"][i]["edit_stats"][lang]["edit_time_percent"] < 100) rev_langs += "<a href=\"#\" onclick=\"bootbox.hideAll(); window.loadTranscription('"+lang+"', false, "+edit_history_data["edit_history"][i]["session_id"]+", null, true); return false;\"><span class=\"v_lang v_cap_type_1\" title=\""+Math.round(edit_history_data["edit_history"][i]["edit_stats"][lang]["edit_time_percent"])+"% revised\">"+lang+"</span></a>";
          else rev_langs += "<a href=\"#\" onclick=\"bootbox.hideAll(); window.loadTranscription('"+lang+"', false, "+edit_history_data["edit_history"][i]["session_id"]+", null, true); return false;\"><span class=\"v_lang v_cap_type_2\" title=\"100% revised\">"+lang+"</span></a>";
        }
        msg += "<tr><td><i class=\"fa fa-user\" style=\"color: #4BA6AF; margin-right: 0.5em;\"></i> <strong>"+(edit_history_data["edit_history"][i]["author_name"] != null ? edit_history_data["edit_history"][i]["author_name"] : edit_history_data["edit_history"][i]["author_id"])+"</strong></td><td>"+edit_history_data["edit_history"][i]["timestamp"].slice(0, -10)+"</td><td>"+rev_langs+"</td><td>"+(edit_history_data["edit_history"][i]["revised"] == true ? "<i class=\"fa fa-check\" style=\"cursor: help; color: #4BA6AF; margin-top: 4px;\" title=\"Revised by "+(edit_history_data["edit_history"][i]["revised_by_name"] != null ? edit_history_data["edit_history"][i]["revised_by_name"] : edit_history_data["edit_history"][i]["revised_by_id"])+" ("+edit_history_data["edit_history"][i]["revised_at"].substring(0, edit_history_data["edit_history"][i]["revised_at"].length-10)+")\"></i>" : "")+"</td></tr>";
        some = true;
      }
    }
  }
  if (!only_pending) {
    msg += "<tr><td><i class=\"fa fa-cogs\" style=\"color: #4BA6AF; margin-right: 0.5em;\"></i> Ingest Service</td><td>-</td><td>";
    if ("langs" in langs) {
      $.each(langs['langs'], function(key, value) {
        msg += "<a href=\"#\" onclick=\"bootbox.hideAll(); window.loadTranscription('"+value["lang_code"]+"', false, null, 1); return false;\"><span class=\"v_lang v_cap_type_0\" title=\"Automatic\">"+value["lang_code"]+"</span></a>"
      });
    }
    msg += "</td><td></td></tr>";
  }
  msg += "</tbody></table></div>";
  if (only_pending) msg = "<p style=\"padding-top: 15px; text-align: center;\">There are some pending revisions for this media:</p>"+msg;
  var dialog_btns = {};
  if (request_data["author_conf"] == 100 && !jQuery.isEmptyObject(langs)) {
    if ("subs_locked" in langs && langs["subs_locked"] == true) {
      dialog_btns["success"] = {
        label: "Enable user reviews",
        className: "btn-success",
        callback: function() {
          bootbox.confirm("This will allow users to submit reviews for this media. Are you sure?", function(resp) {
            if (resp) lock_subs(false);
          });
        }
      };
    }
    else {
      dialog_btns["danger"] = {
        label: "Disable user reviews",
        className: "btn-danger",
        callback: function() {
          bootbox.confirm("This will prevent users from sending further reviews for this media. Are you sure?", function(resp) {
            if (resp) lock_subs(true);
          });
        }
      };
    }
  }
  dialog_btns["main"] = {
    label: "Close",
    className: "btn-primary",
  };
  if (!only_pending || (only_pending && some)) {
    bootbox.dialog({
      message: msg,
      title: "Revisions history",
      buttons: dialog_btns
    });
    jQuery('.modal-body').css('padding', '0');
  }
}

/* check of expire date */
var expireAlert5min = false;
var expiredAlert = false;
if ("expire" in request_data) {
  setInterval(function() {
    var now = new Date();
    if (!expiredAlert && now.getTime() > parseInt(request_data["expire"])*1000) {
      bootbox.alert("Your session has expired. You will not be able to save any changes, but you can still download the subtitles with your changes and import them in a future session.");
      expiredAlert = true;
    }
    else if (!expireAlert5min && (now.getTime()+(5*60*1000)) > parseInt(request_data["expire"])*1000) {
      var remString = "";
      if (Math.round((parseInt(request_data["expire"])-(now.getTime()/1000))) < 60) remString = Math.round((parseInt(request_data["expire"])-(now.getTime()/1000)))+" seconds";
      else remString = Math.round((parseInt(request_data["expire"])-(now.getTime()/1000))/60)+" minutes";
      bootbox.alert("Your session expires in "+remString+". Once expired you will not be able to save any changes.");
      expireAlert5min = true;
    }
  }, 15000);
}

$(".player-body").prepend($('.processing'));
registerKeypressCombos();
changeViewMode(userSettings["player"]["vmode"]);
function initLoadingTranscriptions() {
  if (typeof request_data != 'undefined' && "api_user" in request_data && request_data["api_user"] != "" && "id" in request_data && request_data["id"] != "" && "request_key" in request_data && "author_id" in request_data && "author_conf" in request_data) {
    //if (isMobile()) changeViewMode(0);
    getAvailableLanguages();
    if ("rcode" in langs && langs["rcode"] == 0) {
      if (request_data["lang"] != langs["media_lang"]) {
        // if selected language != media original language
        // load original media language as source (translation default mode)
        loadFlag = 2; // 2 transcriptions remaining
        $('#transcription-tar').css({"width": "49%"});
        $('#transcription-src').css({"display": "inline-block"});
        loadTranscriptionOK(request_data["lang"], false, (mod_session != false && "session_id" in mod_session ? mod_session["session_id"] : null));
      }
      else {
        loadFlag = 1;
        loadTranscriptionOK(request_data["lang"], false, (mod_session != false && "session_id" in mod_session ? mod_session["session_id"] : null));
      }
    }
  }
  else {
    bootbox.alert("Missing input parameters / Unable to decode URL parameters", function() { history.back(); });
    // hide editor buttons
    $('button.layout').hide();
    $('button.save').hide();
    $('button.config').hide();
    $('button.ii').hide();
    $('button.languages').hide();
    $('button.help').hide();
    $('button.options').hide();
    // hide transcription editor
    $('#tl-editor').hide();
    $('#tl-player').css('float', 'none');
    $('#tl-main').remove();
  }
}
initLoadingTranscriptions();
});
