var video_data = {
  "subs": {}
};

var options = {
  "editor": {
    "create_paragraphs": true,
    "create_paragraphs_sil_threshold": 2.00,
    "highlight_low_conf_words": false,
    "highlight_low_conf_words_conf_threshold": 0.75,
    "pause_on_keypress": true
  }
};

var ted_flags = {
  "closest_edit_time": Number.MAX_VALUE,
  "farthest_edit_time": 0.0,
  "closest_edit_segment_id": null,
  "farthest_edit_segment_id": null,
};

jQuery(document).ready(function() {
  var init_load = false;
  
  /* UI */

  // Auxiliar functions
  function getCurrentCaretElement() {
    var node = document.getSelection().anchorNode;
    return (node.nodeType == 3 ? node.parentNode : node);
  }

  function getSelectionCharacterOffsetWithin(element) {
    if (typeof element === 'undefined') element = getCurrentCaretElement();
    var start = 0;
    var end = 0;
    var doc = element.ownerDocument || element.document;
    var win = document.defaultView || document.parentWindow;
    var sel;
    if (typeof window.getSelection != "undefined") {
      sel = window.getSelection();
      if (sel.rangeCount > 0) {
        var range = window.getSelection().getRangeAt(0);
        var preCaretRange = range.cloneRange();
        preCaretRange.selectNodeContents(element);
        preCaretRange.setEnd(range.startContainer, range.startOffset);
        start = preCaretRange.toString().length;
        preCaretRange.setEnd(range.endContainer, range.endOffset);
        end = preCaretRange.toString().length;
      }
    } else if ( (sel = doc.selection) && sel.type != "Control") {
      var textRange = sel.createRange();
      var preCaretTextRange = document.body.createTextRange();
      preCaretTextRange.moveToElementText(element);
      preCaretTextRange.setEndPoint("EndToStart", textRange);
      end = preCaretTextRange.text.length;
      preCaretTextRange.setEndPoint("EndToEnd", textRange);
      end = preCaretTextRange.text.length;
    }
    return { "start": start, "end": end };
  }


  // Resize
  jQuery(window).on("resize", function() {
    wf_print();
    update_closest_farthest_edits();
  });

  // Resize sidebar
  var dragging_sidebar = false;
  jQuery('.ted-sidebar-draggable').on("mousedown", function(e) {
    dragging_sidebar = true;
    jQuery(document).on("mousemove", function(e) {
      if (dragging_sidebar) {
        jQuery('.ted-sidebar-right').css("width", jQuery("body").width()-e.pageX+2);
        jQuery('.ted-container').css("width", "calc(100% - "+jQuery('.ted-sidebar-right').width()+"px)");
      }
    });
  });
  jQuery(document).on("mouseup", function(e) {
    if (dragging_sidebar) {
      dragging_sidebar = false;
      jQuery(document).unbind("mousemove");
    }
  });

  // Media buttons
  jQuery('#ted-play-button').on("click", function() {
    var video = jQuery('.ted-media-player video')[0];
    if (video.paused) video.play();
    else video.pause();
  });
  jQuery('#ted-replay5-button').on("click", function() {
    var video = jQuery('.ted-media-player video')[0];
    video.currentTime = (video.currentTime-2 > 0 ? video.currentTime-2 : 0);
    if (video.paused) video.play();
  });

  jQuery('.ted-media-player video').on("play", function() {
    jQuery('#ted-play-button').html('<i class="material-icons">pause</i> Pause<br/><span class="small">(Tab)</span>');
    if (wf_animationFrameId == null) {
      wf_animationFrameId = window.requestAnimationFrame(wf_update_timemark);
    }
  });

  jQuery('.ted-media-player video').on("pause", function() {
    jQuery('#ted-play-button').html('<i class="material-icons">play_arrow</i> Play<br/><span class="small">(Tab)</span>');
    if (wf_animationFrameId != null) {
      window.cancelAnimationFrame(wf_animationFrameId);
      wf_animationFrameId = null;
    }
  });

  jQuery('.ted-media-player video').on("timeupdate", function(e) {
    // Highlight words
    jQuery('.ted-editor-word').each(function() {
      var begin = 0.0;
      var end = jQuery('.ted-media-player video')[0].duration;
      if (typeof jQuery(this).attr('data-begin') !== 'undefined') begin = parseFloat(jQuery(this).attr('data-begin'));
      else return true;
      if (typeof jQuery(this).attr('data-end') !== 'undefined') end = parseFloat(jQuery(this).attr('data-end'));
      else return true;
      if (begin <= jQuery('.ted-media-player video')[0].currentTime && end >= jQuery('.ted-media-player video')[0].currentTime) {
        jQuery('.ted-current-paragraph').removeClass('ted-current-paragraph');
        jQuery('.ted-current-word').removeClass('ted-current-word');
        jQuery(this).addClass('ted-current-word');
        jQuery(this).closest('.ted-editor-paragraph').addClass('ted-current-paragraph');
        return false;
      }
    });
    if (jQuery('.ted-waveform-container').is(":visible")) wf_animationFrameId = window.requestAnimationFrame(wf_update_timemark);
  });

  jQuery('.ted-editor-content').on("click", function(e) {
    var begin = null;
    if (typeof jQuery(e.target).attr('data-begin') !== 'undefined') begin = parseFloat(jQuery(e.target).attr('data-begin'));
    else {
      var caretElement = getCurrentCaretElement();
      if (typeof jQuery(caretElement).attr('data-begin') !== 'undefined') begin = parseFloat(jQuery(caretElement).attr('data-begin'));
    }
    if (begin != null) {
      jQuery('.ted-media-player video')[0].currentTime = begin+0.05;
      jQuery('.ted-media-player video').triggerHandler("timeupdate");
    }
  });

  // On edit content
  jQuery('.ted-editor-content').on("DOMCharacterDataModified", function(e) {
    if (jQuery(e.target).hasClass('ted-editor-word')) {
      jQuery(e.target).attr('data-edited', 'true').attr('data-conf', '1.0000');
      // Fix begin and end times: set begin and end to end and begin of previous and next words
      if (jQuery(e.target).prev().length) {
        if (jQuery(e.target).prev().hasClass('ted-editor-word')) jQuery(e.target).attr('data-begin', (parseFloat(jQuery(e.target).prev().attr('data-end'))+0.01).toFixed(2));
      }
      else jQuery(e.target).attr('data-begin', '0.00');
      if (jQuery(e.target).next().length) {
        if (jQuery(e.target).next().hasClass('ted-editor-word')) jQuery(e.target).attr('data-end', (parseFloat(jQuery(e.target).next().attr('data-begin'))-0.01).toFixed(2));
      }
      else jQuery(e.target).attr('data-end', jQuery('.ted-media-player video')[0].duration);
    }
    update_closest_farthest_edits();
  });

  // Aux function: update start and end edits
  function update_closest_farthest_edits() {
    return false;
    // Waveform canvas
    wf_print();
    var w = jQuery('.ted-waveform-container').width();
    var h = jQuery('.ted-waveform-container').height();
    var c = jQuery('#ted-waveform-canvas')[0];
    var ctx = c.getContext('2d');
    ctx.fillStyle = 'rgba(0,200,0,0.5)';
    ctx.fillRect(w*(parseFloat(ted_flags["closest_edit_time"]/jQuery('.ted-media-player video')[0].duration)), h-2, w*(parseFloat(ted_flags["farthest_edit_time"]-ted_flags["closest_edit_time"])/jQuery('.ted-media-player video')[0].duration), 2);
    // Mark edited text (debug)
    var is_edited_text = true;
    if (ted_flags["closest_edit_segment_id"] != null && jQuery('.ted-editor-segment[data-id="'+ted_flags["closest_edit_segment_id"]+'"]').length) is_edited_text = false;
    jQuery('.ted-editor-content .ted-editor-segment').each(function() {
      if (ted_flags["closest_edit_segment_id"] != null && jQuery(this).attr('data-id') == ted_flags["closest_edit_segment_id"]) is_edited_text = true;
      if (is_edited_text) jQuery(this).css('background-color', 'rgba(0,200,0,0.15)');
      else jQuery(this).css('background-color', '');
      if (ted_flags["farthest_edit_segment_id"] != null && jQuery('.ted-editor-segment[data-id="'+ted_flags["farthest_edit_segment_id"]+'"]').length && jQuery(this).attr('data-id') == ted_flags["farthest_edit_segment_id"]) is_edited_text = false;
    });
  }

  // Fix structure
  jQuery('.ted-editor-content').on("click blur", function(e) {
    jQuery(this).contents().each(function() {
      if (!jQuery(this).hasClass('ted-editor-paragraph')) {
        // Transcription from scratch? Create paragraph and put text contents in
        var paragraphId = Math.floor(Math.random()*(2147483648-10000000+1)+10000000);
        var begin = 0.0;
        var end = jQuery('.ted-media-player video')[0].duration;
        if (jQuery(this).prev('.ted-editor-paragraph').length) {
          if (jQuery(this).prev('.ted-editor-paragraph').find('.ted-editor-word:last-child').length) {
            begin = parseFloat(jQuery(this).prev('.ted-editor-paragraph').find('.ted-editor-word:last-child').attr('data-end'))+0.01;
          }
        }
        if (jQuery(this).next('.ted-editor-paragraph').length) {
          if (jQuery(this).next('.ted-editor-paragraph').find('.ted-editor-word:first-child').length) {
            end = parseFloat(jQuery(this).next('.ted-editor-paragraph').find('.ted-editor-word:first-child').attr('data-begin'))-0.01;
          }
        }
        var currentText = jQuery(this).text();
        jQuery('<div class="ted-editor-paragraph" data-id="'+paragraphId+'"><span class="ted-editor-word" data-begin="'+begin.toFixed(2)+'" data-end="'+end.toFixed(2)+'" data-edited="true">'+currentText+'</span></div>').insertAfter(jQuery(this));
        jQuery(this).remove();
      }
    });
    jQuery('.ted-editor-paragraph').children().each(function() {
      if (!(jQuery(this).hasClass('ted-editor-word'))) {
        var text = jQuery(this).text();
        var b = 0.0;
        var e = jQuery('.ted-media-player video')[0].duration;
        // Fix begin and end times: set begin and end to end and begin of previous and next words
        if (jQuery(e.target).prev().length) {
          if (jQuery(this).prev().hasClass('ted-editor-word')) b = parseFloat(jQuery(this).prev().attr('data-end'))+0.01;
        }
        else if (jQuery(this).closest('.ted-editor-paragraph').prev().length && jQuery(this).closest('.ted-editor-paragraph').prev().find('.ted-editor-word:last').length) {
          b = parseFloat(jQuery(this).closest('.ted-editor-paragraph').prev().find('.ted-editor-word:last').attr('data-end'));
        }
        if (jQuery(this).next().length) {
          if (jQuery(this).next().hasClass('ted-editor-word')) e = parseFloat(jQuery(this).next().attr('data-begin'))+0.01;
        }
        else if (jQuery(this).closest('.ted-editor-paragraph').next().length && jQuery(this).closest('.ted-editor-paragraph').next().find('.ted-editor-word:first').length) {
          e = parseFloat(jQuery(this).closest('.ted-editor-paragraph').next().find('.ted-editor-word:first').attr('data-begin'));
        }
        jQuery('<span class="ted-editor-word" data-begin="'+b.toFixed(2)+'" data-end="'+e.toFixed(2)+'" data-edited="true">'+text+'</span>').insertAfter(jQuery(this));
        jQuery(this).remove();
      }
    });
    // Fix splitted words
    jQuery('.ted-editor-word[data-edited="true"]').each(function() {
      if ((jQuery(this).text().length == 0 || /^\s+$/.test(jQuery(this).text())) && jQuery(this).prev().length) {
        jQuery(this).prev().text(jQuery(this).prev().text()+" ").attr('data-end', jQuery(this).attr('data-end'));
        jQuery(this).remove();
      }
      if (!/\s+$/g.test(jQuery(this).text()) && jQuery(this).next().length) {
        jQuery(this).attr('data-end', jQuery(this).next().attr('data-end')).text(jQuery(this).text()+jQuery(this).next().text());
        jQuery(this).next().remove();
      }
    });
  });

  jQuery('.ted-editor-content').on("paste", function(e) {
    e.preventDefault();
    var text = (event.originalEvent || event).clipboardData.getData("text/plain");
    document.execCommand("insertHTML", false, text);
  });

  jQuery('.ted-editor-content').on("keypress", function(e) {
    if (e.keyCode == 13) { // Enter
      e.preventDefault();
      e.stopPropagation();
      return false;
    }
    else if ((e.keyCode == 37 || e.keyCode == 39) && e.ctrlKey) { // Ctrl + Left/Right arrow
      var currentWord = getCurrentCaretElement();
      if (jQuery(currentWord).hasClass('ted-editor-word') && jQuery(currentWord).attr('data-begin') !== 'undefined') {
        jQuery('.ted-media-player video')[0].currentTime = parseFloat(jQuery(currentWord).attr('data-begin'));
      }
    }
    else {
      if (options["editor"]["pause_on_keypress"] && !jQuery('.ted-media-player video')[0].paused) jQuery('.ted-media-player video')[0].pause();
    }
  });

  // Shortcuts
  jQuery(document).on("keydown", function(e) {
    if (e.keyCode == 8) { // Backspace
      // Prevent page back
      if (!jQuery(e.target).is('input, textarea, [contenteditable="true"]')) {
        e.preventDefault();
      }
    }
    if (e.keyCode == 9) { // Tab
      if (e.shiftKey) {
        // Play from the begining of the word/segment
        var currentWord = getCurrentCaretElement();
        if (jQuery(currentWord).hasClass('ted-editor-word')) {
          if (typeof jQuery(currentWord).attr('data-begin') !== 'undefined' && typeof jQuery(currentWord).attr('data-begin') !== false) {
            jQuery('.ted-media-player video')[0].currentTime = parseFloat(jQuery(currentWord).attr('data-begin'));
            if (jQuery('.ted-media-player video')[0].paused) jQuery('.ted-media-player video')[0].play();
          }
        }
      }
      else {
        e.preventDefault();
        e.stopPropagation();
        if (jQuery('.ted-media-player video')[0].paused) jQuery('.ted-media-player video')[0].play();
        else jQuery('.ted-media-player video')[0].pause();
        return false;
      }
    }
    if (e.keyCode == 37) { // Left arrow (Alt)
      if (e.altKey) {
        e.preventDefault();
        e.stopPropagation();
        jQuery('.ted-media-player video')[0].currentTime = (jQuery('.ted-media-player video')[0].currentTime-2 > 0 ? jQuery('.ted-media-player video')[0].currentTime-2 : 0);
      }
    }
  });

  jQuery('.ted-waveform-container').on("click", function(e) {
    jQuery('.ted-media-player video')[0].currentTime = (e.pageX/jQuery('.ted-waveform-container').width())*jQuery('.ted-media-player video')[0].duration;
    jQuery('.ted-media-player video').triggerHandler("timeupdate");
    jQuery('html, body').animate({
      scrollTop: jQuery('.ted-current-word').offset().top-200
    }, 400);
  });

  /* Waveform code */
  var wf_waveformDataHz = 500;
  var wf_waveformData = null;
  var wf_animationFrameId = null;

  function wf_load_pcm_data() {
    jQuery(video_data["metadata"]["media"]).each(function() {
      if (this["type_code"] == 6) {
        var xhr = new XMLHttpRequest();
        xhr.open('GET', this["location"], true);
        //xhr.open('GET', "9407fbe400a33a1cebbd168cd1f75032.pcm", true);
        xhr.responseType = 'arraybuffer';
        xhr.onreadystatechange = function (e) {
          if (xhr.readyState == 4) {
            if (xhr.status == 200) {
              wf_waveformData = new Uint8Array(this.response);
              for (var i=0; i<wf_waveformData.length; i++) wf_waveformData[i] = Math.abs(wf_waveformData[i]-128);
              var mean = 0;
              var n = 0;
              for (var i=0; i<wf_waveformData.length; i++) {
                if (wf_waveformData[i] > 8) {
                  mean+=wf_waveformData[i];
                  n++;
                }
              }
              if (n == 0 || (mean/n) < 16) for (var i=0; i<wf_waveformData.length; i++) {
                if (n == 0) {
                  if (wf_waveformData[i]*3 > 128) wf_waveformData[i] = 128;
                  else wf_waveformData[i]=wf_waveformData[i]*3;
                }
                else {
                  if (wf_waveformData[i]*2 > 128) wf_waveformData[i] = 128;
                  else wf_waveformData[i]=wf_waveformData[i]*2;
                }
              }
              wf_print();
            }
            else {
              wf_waveformData = null;
              console.log("Error loading waveform data. It won't be displayed.");
              jQuery('.ted-waveform-container').hide();
            }
          }
        };
        xhr.send();
        return false;
      }
    });
  }

  function wf_print() {
    if (wf_waveformData != null) {
      jQuery('.ted-waveform-container').show();
      var p = 0.5 / window.devicePixelRatio;
      var w = jQuery('.ted-waveform-container').width();
      var h = jQuery('.ted-waveform-container').height();
      var c = jQuery('#ted-waveform-canvas')[0];
      var valuesPerPixel = parseFloat(wf_waveformData.length/w);
      c.width = w;
      c.height = h;
      h = h/2;
      var ctx = c.getContext('2d');
      ctx.fillStyle = '#2D3144';
      ctx.beginPath();
      ctx.moveTo(p, h);
      for (var i=0, j=0; j<w && i<wf_waveformData.length; i+=valuesPerPixel, j++) {
        ctx.lineTo((j%w)+p, h-((wf_waveformData[Math.floor(i)]/128)*(h*2))+p);
      }
      ctx.lineTo(c.width + p, h);
      ctx.fill();
      ctx.closePath();
      ctx.fillStyle = '#2D3144';
      ctx.beginPath();
      ctx.moveTo(p, h);
      for (var i=0, j=0; j<w && i<wf_waveformData.length; i+=valuesPerPixel, j++) {
        ctx.lineTo((j%w)+p, h+((wf_waveformData[Math.floor(i)]/128)*(h*2))+p);
      }
      ctx.lineTo(c.width + p, h);
      ctx.fill();
      ctx.closePath();
      if (wf_animationFrameId != null) wf_animationFrameId = window.requestAnimationFrame(wf_update_timemark);
    }
    else jQuery('.ted-waveform-container').hide();
  }

  function wf_update_timemark() {
    var x = (jQuery('.ted-media-player video')[0].currentTime/jQuery('.ted-media-player video')[0].duration)*jQuery('.ted-waveform-container').width();
    jQuery('.ted-waveform-timemark-canvas').css('left', parseInt(x)+'px');
    if (wf_animationFrameId != null) wf_animationFrameId = window.requestAnimationFrame(wf_update_timemark);
  }
  
  /* API event listeners */
  jQuery(document).on("speech/metadata", function(e) {
    var response = e.detail;
    if (response["error"] == false) {
      if (response["response"]["rcode"] == 0) {
        video_data["metadata"] = response["response"];

        jQuery(video_data["metadata"]["media"]).each(function() {
          if (this["type_code"] == 0) {
            jQuery('.ted-media-player video').append('<source src="'+this["location"]+'" type="video/'+this["media_format"]+'">');
          }
        });
        // Load waveform if available
        wf_load_pcm_data();
        if (init_load) {
          jQuery('#ted-init-modal .modal-body').html('<p>Retrieving available languages...</p>');
          tlpApi.speech.langs();
        }
      }
      else {
        if (init_load) {
          jQuery('#ted-init-modal .modal-title').html('Error');
          jQuery('#ted-init-modal .modal-body').html('<p>Error loading metadata.</p>');
        }
        else {
          bootbox.alert("Error loading metadata.");
        }
      }
    }
    else {
      jQuery('#ted-init-modal .modal-title').html('Error');
      jQuery('#ted-init-modal .modal-body').html('<p>Error loading metadata.</p>');
    }
  });

  jQuery(document).on("speech/langs", function(e) {
    var response = e.detail;
    if (response["error"] == false) {
      if (response["response"]["rcode"] == 0) {
        video_data["langs"] = response["response"];
        if ("langs" in video_data["langs"]) jQuery(video_data["langs"]["langs"]).each(function() {
          jQuery('.ted-language-list').append('<li><a href="#" onclick="load_language(\''+this["lang_code"]+'\'); return false";>'+this["lang_name"]+'</a></li>');
        });
        if (init_load) {
          jQuery('#ted-init-modal .modal-body').html('<p>Starting editing session...</p>');
          if ("author_name" in request_data) tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"], request_data["author_name"]);
          else tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"]);
        }
      }
      else {
        if (init_load) {
          jQuery('#ted-init-modal .modal-title').html('Error');
          jQuery('#ted-init-modal .modal-body').html('<p>Error retrieving available languages.</p>');
        }
        else bootbox.alert("Error retrieving available languages");
      }
    }
    else {
      jQuery('#ted-init-modal .modal-title').html('Error');
      jQuery('#ted-init-modal .modal-body').html('<p>Error retrieving available languages.</p>');
    }
  });

  jQuery(document).on("speech/start_session", function(e) {
    var response = e.detail;
    if (response["error"] == false) {
      video_data["session"] = response["response"];
      if (video_data["session"]["rcode"] == 0) {
        if (init_load) {
          jQuery('#ted-init-modal .modal-body').html('<p>Loading transcription...</p>');
          tlpApi.speech.get(("lang" in request_data ? request_data["lang"] : video_data["langs"]["media_lang"]), 5, video_data["session"]["session_id"]);
          /* Session status interval */
          setInterval(function() {
            tlpApi.speech.session_status(video_data["session"]["session_id"], video_data["session"]["author_id"], video_data["session"]["author_conf"], 1);
          }, 10000);
        }
      }
      else {
        bootbox.confirm(video_data["session"]["rcode_description"]+"<br/><br/>Do you want to try re-opening the session?", function(resp) {
          if (resp) {
            if ("author_name" in request_data) tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"], request_data["author_name"]);
            else tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"]);
          }
          else if (init_load) {
            jQuery('#ted-init-modal .modal-body').html('<p>Could not start editing session.</p>');
          }
        });
      }
    }
    else {
      jQuery('#ted-init-modal .modal-title').html('Error');
      jQuery('#ted-init-modal .modal-body').html('<p>Error starting an editing session.</p>');
    }
  });

  jQuery(document).on("speech/get", function(e) {
    bootbox.hideAll();
    var response = e.detail;
    if (init_load) {
      jQuery('#ted-init-modal').modal('hide');
      init_load = false;
    }
    if (response["error"] == false) {
      video_data["subs"]["meta"] = response["response"]["meta"];
      video_data["subs"]["meta"]["lang"] = response["lang"];
      video_data["subs"]["data"] = {}; 
      for (var i=0; i<response["response"]["data"].length; i++) {
        video_data["subs"]["data"][response["response"]["data"][i]["id"]] = response["response"]["data"][i];
      }
      jQuery('#ted-editor .ted-editor-content').empty();
      var paragraph_index = 0;
      jQuery('#ted-editor .ted-editor-content').append('<div class="ted-editor-paragraph" data-id="'+paragraph_index+'"></div>');
      for (var i=0; i<response["response"]["data"].length; i++) {
        var segment = response["response"]["data"][i];
        if (options["editor"]["create_paragraphs"]) {
          if (i>0 && (parseFloat(segment["b"])-parseFloat(response["response"]["data"][i-1]["e"])) > options["editor"]["create_paragraphs_sil_threshold"]) {
            paragraph_index++;
            jQuery('#ted-editor .ted-editor-content').append('<div class="ted-editor-paragraph" data-id="'+paragraph_index+'"></div>');
          }
        }
        for (var j=0; j<segment["wl"].length; j++) {
          jQuery('#ted-editor .ted-editor-content .ted-editor-paragraph[data-id="'+paragraph_index+'"]').append('<span class="ted-editor-word" data-segment-id="'+segment["id"]+'" data-at="'+segment["aT"]+'" data-begin="'+("b" in segment["wl"][j] ? parseFloat(segment["wl"][j]["b"]).toFixed(2) : segment["b"].toFixed(2))+'" data-end="'+("e" in segment["wl"][j] ? parseFloat(segment["wl"][j]["e"]).toFixed(2) : segment["e"].toFixed(2))+'" data-conf="'+("c" in segment["wl"][j] ? parseFloat(segment["wl"][j]["c"]).toFixed(4) : segment["c"].toFixed(4))+'">'+segment["wl"][j]["w"]+' </span>');
        }
      }
      if (options["editor"]["highlight_low_conf_words"]) highlight_low_confidence_words(true);
    }
    else {
      bootbox.alert("Error loading "+response["lang"]+" transcription.");
    }
  });

  jQuery(document).on("speech/mod", function(e) {
    var response = e.detail;
    if (response["error"] == false) {
      var mod_response = response["response"];
      if ("rcode" in mod_response && mod_response["rcode"] == 0) {
        jQuery('.ted-editor-word').removeAttr('data-edited');
        bootbox.alert("Transcription changes successfully saved.");
      }
      else {
        bootbox.alert("Error saving transcription changes: "+mod_response["rcode_description"]);
      }
    }
    else {
      bootbox.alert("Error saving transcription changes. We suggest you to export the transcription into a text file to prevent losing your changes.");
    }
  });

  jQuery(document).on("speech/session_status", function(e) {
    var response = e.detail;
    if (response["error"] == false) {
      if (video_data["session"]["rcode"] != 0) {
        bootbox.hideAll();
        bootbox.confirm(video_data["session"]["rcode_description"]+"<br/><br/>Do you want to try re-opening the session?", function(resp) {
          if (resp) {
            if ("author_name" in request_data) tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"], request_data["author_name"]);
            else tlpApi.speech.start_session(request_data["author_id"], request_data["author_conf"]);
          }
        });
      }
    }
  });

  window.onbeforeunload = function() {
    if (video_data["session"] != false && "rcode" in video_data["session"] && video_data["session"]["rcode"] == 0) {
      tlpApi.speech.end_session(video_data["session"]["session_id"], video_data["session"]["author_id"], video_data["session"]["author_conf"], video_data["session"]["author_name"], 0, 1);
    }
  };

  /* Main execution */
  if (typeof request_data === 'undefined' || request_data === null) {
    bootbox.alert("Couldn't read request data");
  }
  else {
    init_load = true;
    jQuery('#ted-init-modal').modal({"backdrop": "static", "keyboard": false});
    tlpApi.init(request_data["id"], request_data["api_user"], request_data["request_key"], request_data["expire"], ("hash" in request_data ? request_data["hash"] : null));
    tlpApi.speech.metadata();
  }
});

/* Global functions */
function highlight_low_confidence_words() {
  if (options["editor"]["highlight_low_conf_words"]) {
    jQuery('.ted-editor-word').each(function() {
      if (typeof jQuery(this).attr('data-conf') !== 'undefined') {
        if (parseFloat(jQuery(this).attr('data-conf')) < options["editor"]["highlight_low_conf_words_conf_threshold"]) jQuery(this).addClass('low-conf');
        else jQuery(this).removeClass('low-conf');
      }
    });
  }
  else {
    jQuery('.ted-editor-word.low-conf').removeClass('low-conf');
  }
}

function load_language(lang_code) {
  // Check for changes
  if (jQuery('.ted-editor-word[data-edited="true"]').length > 0) {
    bootbox.confirm("You have changes pending to saved. Do you want to continue and discard your current changes?", function(r) {
      if (r) {
        bootbox.alert({title: "Loading transcription", message: "Please wait..."});
        tlpApi.speech.get(lang_code, 5, video_data['session']['session_id']);
      }
    });
  }
  else tlpApi.speech.get(lang_code, 5, video_data['session']['session_id']);
}

function save_changes() {
  // NAIVE IMPLEMENTATION: Send one segment, from first edited to last edited word/segment
  if (jQuery('.ted-editor-content .ted-editor-word[data-edited="true"]').length > 0) {
    var first_edited_word = jQuery('.ted-editor-content .ted-editor-word[data-edited="true"]:first');
    if (typeof jQuery(first_edited_word).attr('data-segment-id') !== 'undefined') first_edited_word = jQuery('.ted-editor-content .ted-editor-word[data-segment-id="'+jQuery(first_edited_word).attr('data-segment-id')+'"]:first');
    var last_edited_word = jQuery('.ted-editor-content .ted-editor-word[data-edited="true"]:last');
    if (typeof jQuery(last_edited_word).attr('data-segment-id') !== 'undefined') last_edited_word = jQuery('.ted-editor-content .ted-editor-word[data-segment-id="'+jQuery(last_edited_word).attr('data-segment-id')+'"]:last');
    var edited_text = jQuery(first_edited_word).text()+" ";
    if (last_edited_word !== first_edited_word) {
      jQuery(first_edited_word).nextAll().each(function() {
        edited_text += jQuery(this).text()+" ";
        if (jQuery(this).is(last_edited_word)) return false;
      });
    }
    mods = {};
    mods[video_data["subs"]["meta"]["lang"]] = {
      "txt": [
        {
          "sI": Math.floor(Math.random()*(2147483648-10000000+1)+10000000),
          "b": parseFloat(jQuery(first_edited_word).attr('data-begin')),
          "e": parseFloat(jQuery(last_edited_word).attr('data-end')),
          "t": edited_text
        }
      ]
    };
    if ("author_name" in request_data) tlpApi.speech.mod(video_data["session"]["session_id"], request_data["author_id"], parseInt(request_data["author_conf"]), mods, request_data["author_name"]);
    else tlpApi.speech.mod(video_data["session"]["session_id"], request_data["author_id"], parseInt(request_data["author_conf"]), mods);
  }
  else {
    bootbox.alert("There are no changes to save.");
  }
}

function export_to() {
  var textFile = null,
  makeTextFile = function (text) {
    var data = new Blob([text], {type: 'text/plain'});
    if (textFile !== null) {
      window.URL.revokeObjectURL(textFile);
    }

    textFile = window.URL.createObjectURL(data);
    return textFile;
  };

  var text = "";
  jQuery('.ted-editor-content .ted-editor-paragraph').each(function() {
    text += jQuery(this).text();
    text += "\n\n";
  });

  bootbox.alert({
    title: "Export to",
    message: '<select class="form-control"><option value="txt">Text file</option></select><br/><br/><a download="transcription.txt" id="export_url_link" href="'+makeTextFile(text)+'"><button class="btn btn-primary">Download file</button></a>'
  });
}
