<?php
/**
 * TLP 3.0 PHP5 library
 *
*/

/**
 * TlpApi class
 *
*/
class TlpApi {
  private $API_URL = 'https://ttp.mllp.upv.es/api/v3';
  private $API_USER = '';
  private $API_SECRET_KEY = '';
  private $API_EXPIRE = NULL;
  private $SPEECH_EDITOR_URL = 'http://ttp.mllp.upv.es/index.php?page=player';
  private $TEXT_EDITOR_URL = 'http://ttp.mllp.upv.es/index.php?page=doc';
  private $TEMP_DIR = ''; // If empty, defaults to sys_get_temp_dir()

  /**
  * TlpApi Class constructor method
  *
  */
  function TlpApi($apiUser = NULL, $apiKey = NULL, $apiExpire = NULL) {
    if (strlen($this->TEMP_DIR) == 0) $this->TEMP_DIR = sys_get_temp_dir();
    if (!is_null($apiUser) && !is_null($apiKey)) {
      $this->API_USER = $apiUser;
      $this->API_SECRET_KEY = $apiKey;
    }
    if (!is_null($apiExpire)) $this->API_EXPIRE = intval($apiExpire);
    if (substr($this->API_URL, -1) != '/') $this->API_URL .= '/';
  }

  /**
  * Returns the status of an uploaded media in JSON format.
  *
  * @param string $upload_id the ID of the uploaded media for which the API should
  *  return media status.
  */
  public function speech_status($upload_id) {
    $data_json = array();
    $data_json["id"] = $upload_id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/status', false, $context);
    return $response;
  }

  /**
  * Returns metadata information of a given media ID in JSON format.
  *
  * @param string $id the ID of the media for which the API should
  *  return metadata information.
  */
  public function speech_metadata($id, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/metadata', false, $context);
    return $response;
  }

  /**
  * Returns a list of subtitles and languages available for a given media ID
  * in JSON format.
  *
  * @param string $id the ID of the media for which the API should
  *  return a list of available captions.
  */
  public function speech_langs($id, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/langs', false, $context);
    return $response;
  }

  /**
  * Returns a list with all available ASR, MT and TTS systems in JSON format.
  *
  */
  public function speech_systems() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/systems', false, $context);
    return $response;
  }

  /**
  * Returns a list with information of all uploaded media in JSON format.
  *
  */
  public function speech_uploadslist() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/uploadslist', false, $context);
    return $response;
  }

  /**
  * Returns a list with information of all existing media in JSON format.
  *
  */
  public function speech_list() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/list', false, $context);
    return $response;
  }

  /**
  * Returns a subtitles file for a given media ID and language 
  * in different formats.
  *
  * @param string $id the ID of the media for which the API should
  *  return the captions.
  * @param string $lang the language code of the captions.
  * @param int $format the format ID of the captions (check
  *  documentation)
  * @param int $session_id subtitle modifications from the given
  *  session ID (check documentation)
  * @param int $seg_filt_policy segment text filtering policy (check
  *  documentation)
  * @param int $sel_data_policy subtitle contents to be returned (check
  *  documentation)
  */
  public function speech_get($id, $lang, $format = NULL, $session_id = NULL, $seg_filt_policy = NULL, $sel_data_policy = NULL, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["lang"] = $lang;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    if (!is_null($format)) $data_json["format"] = $format;
    if (!is_null($session_id)) $data_json["session_id"] = $session_id;
    if (!is_null($seg_filt_policy)) $data_json["seg_filt_policy"] = $seg_filt_policy;
    if (!is_null($sel_data_policy)) $data_json["sel_data_policy"] = $sel_data_policy;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/get', false, $context);
    return $response;
  }

  /**
  * Modifies the captions of a given media ID and language with the
  * changes provided.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId the editing session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param array $mods list of changes (check documentation)
  */
  public function speech_mod($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $mods, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["mods"] = $mods;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/mod', false, $context);
    return $response;
  }

  /**
  * Returns a valid URL link for calling TLP Player given
  * input parameters.
  *
  * @param string $id the ID of the media.
  * @param string $lang the language code of the captions.
  * @param string $authorId user name of user doing the changes.
  * @param string $authorName name of user doing the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $expire expire date of this URL in UNIX timestamp
  *  format.
  */
  public function gen_speech_editor_url($id, $lang, $authorId, $authorName, $authorConf, $expire) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["lang"] = $lang;
    $data_json["author_id"] = $authorId;
    $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["expire"] = $expire;
    $data_json["api_user"] = $this->API_USER;
    $data_json["request_key"] = sha1($id.$expire.$this->API_USER.$this->API_SECRET_KEY).sha1($authorId.$authorConf.$expire.$this->API_USER.$this->API_SECRET_KEY);
    if (strpos($this->SPEECH_EDITOR_URL, '?') !== false) return $this->SPEECH_EDITOR_URL.'&request='.base64_encode(json_encode($data_json));
    else return $this->SPEECH_EDITOR_URL.'?request='.base64_encode(json_encode($data_json));
  }

  /**
  * Creates a Media Package in a temporal directory
  * and sends it to /ingest/new API interface.
  *
  * Returns a string if an error ocurred, or a JSON
  * string (/ingest/new response text) if the ingest
  * process was successful.
  *
  * @param string $title title of the media being ingested.
  * @param string $id ID media will have once ingested.
  * @param string $lang language code of the media audio.
  * @param array $requested_langs list of languages for which
  *  captions will be generated.
  * @param mixed $mediaFile DOLLAR_FILES PHP variable with
  *  uploaded media file or media file URL string.
  * @param array $slidesFile DOLLAR_FILES PHP variable with uploaded
  *  slides file (optional).
  * @param array $subtitlesFile DOLLAR_FILES PHP variable with uploaded
  *  subtitles file (optional).
  * @param array $attachmentFiles DOLLAR_FILES PHP variable with multiple
  *  attachment files (optional).
  * @param date $date ingestion date in d/m/y format.
  */
  public function speech_ingest_new($title, $id, $lang, $requested_langs, $mediaFile, $slidesFile = NULL, $subtitlesFile = NULL, $attachmentFiles = NULL, $from_webui = false) {
    $manifest=array();
    if ($from_webui) $manifest['from_webui'] = true;
    $manifest['metadata']=array();
    $manifest['metadata']['title'] = $title;
    $manifest['metadata']['language'] = $lang;
    $manifest['metadata']['external_id'] = $id;
    $manifest['requested_langs'] = $requested_langs;
    $manifest['media']=array();
    $tempDir = tempnam($this->TEMP_DIR, 'up-');
    if (file_exists($tempDir)) unlink($tempDir);
    if (!mkdir($tempDir, 0777)) return "/ingest error: unable to create temp directory ".$tempDir;
    /* media file */
    if (is_array($mediaFile)) { /* file */
      $manifest['media']['fileformat'] = pathinfo($mediaFile['name'], PATHINFO_EXTENSION);
      $manifest['media']['filename'] = $this->rand_str().'.'.pathinfo($mediaFile['name'], PATHINFO_EXTENSION);
      if (!move_uploaded_file($mediaFile['tmp_name'], $tempDir.'/'.$manifest['media']['filename'])) return "/ingest error: cannot move media uploaded file to temp directory";
      $manifest['media']['md5'] = md5_file($tempDir.'/'.$manifest['media']['filename']);
    }
    else { /* url */
      $manifest['media']['url'] = $mediaFile;
    }
    if (!is_null($slidesFile) || !is_null($subtitlesFile) || !is_null($attachmentFiles)) $manifest['attachments'] = array();
    /* slides file */
    if (!is_null($slidesFile)) {
      $slidesArray = array();
      $slidesArray['fileformat'] = pathinfo($slidesFile['name'], PATHINFO_EXTENSION);
      $slidesArray['filename'] = rand_str().'.'.pathinfo($slidesFile['name'], PATHINFO_EXTENSION);
      $slidesArray['type_code'] = 1;
      if (!move_uploaded_file($slidesFile['tmp_name'], $tempDir.'/'.$slidesArray['filename'])) return "/ingest error: cannot move slides uploaded file to temp directory";
      $slidesArray['md5'] = md5_file($tempDir.'/'.$slidesArray['filename']);
      $manifest['attachments'][] = $slidesArray;
    }
    /* subtitles (transcription) file */
    if (!is_null($subtitlesFile)) {
      $subtitlesArray = array();
      $subtitlesArray['fileformat'] = pathinfo($subtitlesFile['name'], PATHINFO_EXTENSION);
      $subtitlesArray['filename'] = rand_str().'.'.pathinfo($subtitlesFile['name'], PATHINFO_EXTENSION);
      $subtitlesArray['language'] = $lang;
      $subtitlesArray['type_code'] = 4;
      if (!move_uploaded_file($subtitlesFile['tmp_name'], $tempDir.'/'.$subtitlesArray['filename'])) return "/ingest error: cannot move subtitles uploaded file to temp directory";
      $subtitlesArray['md5'] = md5_file($tempDir.'/'.$subtitlesArray['filename']);
      $manifest['attachments'][] = $subtitlesArray;
    }
    /* attachment files */
    if (!is_null($attachmentFiles) && array_key_exists('name', $attachmentFiles) && is_array($attachmentFiles['name'])) {
      for ($i = 0; $i < count($attachmentFiles['name']); $i++) {
        $attArray = array();
        $attArray['fileformat'] = pathinfo($attachmentFiles['name'][$i], PATHINFO_EXTENSION);
        $attArray['filename'] = rand_str().'.'.pathinfo($attachmentFiles['name'][$i], PATHINFO_EXTENSION);
        $attArray['type_code'] = 2;
        if (!move_uploaded_file($attachmentFiles['tmp_name'][$i], $tempDir.'/'.$attArray['filename'])) return "/ingest error: cannot move attachments uploaded files to temp directory";
        $attArray['md5'] = md5_file($tempDir.'/'.$attArray['filename']);
        $manifest['attachments'][] = $attArray;
      }
    }
    /* mediapackage.zip */
    $mpzip = new ZipArchive;
    if ($mpzip->open($tempDir.'/'.preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).'.zip', ZipArchive::CREATE) !== true) return "/ingest error: cannot create ".$tempDir."/".preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).".zip file";
    if (is_array($mediaFile)) $mpzip->addFile($tempDir.'/'.$manifest['media']['filename'], $manifest['media']['filename']);
    if (array_key_exists("attachments", $manifest) && is_array($manifest['attachments'])) {
      for ($i = 0; $i < count($manifest['attachments']); $i++) $mpzip->addFile($tempDir.'/'.$manifest['attachments'][$i]['filename'], $manifest['attachments'][$i]['filename']);
    }
    $numZipFiles = $mpzip->numFiles;
    $mpzip->close();
    /* create manifest.json file */
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $data_json["manifest"] = $manifest;
    $mf=fopen($tempDir.'/'.preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).'.json', 'w');
    fwrite($mf, json_encode($data_json));
    fclose($mf);
    /* ingest files */
    $curl = curl_init($this->API_URL.'speech/ingest/new');
    curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($curl, CURLOPT_POST, true);
    curl_setopt($curl, CURLOPT_HTTPHEADER, array('Content-Type: multipart/form-data')); 
    if ($numZipFiles > 0) curl_setopt($curl, CURLOPT_POSTFIELDS, array('json' => '@'.$tempDir.'/'.preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).'.json'.';type=application/json', 'pkg' => '@'.$tempDir.'/'.preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).'.zip'.';type=application/zip'));
    else curl_setopt($curl, CURLOPT_POSTFIELDS, array('json' => '@'.$tempDir.'/'.preg_replace('/[^A-Za-z0-9_\-]/', '_', $id).'.json'.';type=application/json'));
    $response = curl_exec($curl);
    $status = curl_getinfo($curl, CURLINFO_HTTP_CODE);
    /* remove temp files and folder */
    $tmp_files = glob($tempDir.'/*');
    foreach ($tmp_files as $tf) if (is_file($tf)) unlink($tf);
    rmdir($tempDir);
    if ($status != 200) return "/ingest/new error: ".($status == 0 ? curl_error($curl) : "Web Service returned HTTP status ".$status);
    return $response;
  }

  /**
  * Cancel an upload file being processed
  *
  * @param string $id the Upload ID returned by the /ingest interface
  *  of the media to be canceled.
  */
  public function speech_ingest_cancel($id, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/ingest/cancel', false, $context);
    return $response;
  }

  /**
  * Delete existing media
  *
  * @param string $id the media ID of the media to be deleted.
  */
  public function speech_ingest_delete($id, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/ingest/delete', false, $context);
    return $response;
  }

  /**
  * Update existing media
  *   - Re-generate trancriptions/translations
  *   - Request new translations
  *
  * @param string $id the media ID of the media to be updated.
  */
  public function speech_ingest_update($manifest_json) {
    $data_json = array();
    $data_json["manifest"] = $manifest_json;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/ingest/update', false, $context);
    return $response;
  }

  /**
  * Returns a list of all edit sessions for all API user's
  * media files that are pending to be revised.
  *
  */
  public function speech_revisions() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/revisions', false, $context);
    return $response;
  }

  /**
  * Returns a list of all edit sessions carried out over an specific Media ID. 
  *
  * @param string $id the media ID of the media.
  */
  public function speech_edit_history($id, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/edit_history', false, $context);
    return $response;
  }

  /**
  * Starts an editing session.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  */
  public function speech_start_session($id, $authorId, $authorName = NULL, $authorConf, $sessionType = "player", $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["type"] = $sessionType;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/start_session', false, $context);
    return $response;
  }

  /**
  * Gets the status of an editing session.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $alive 0 for only retrieving, 1 to also update session status
  */
  public function speech_session_status($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $alive = 1, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["alive"] = $alive;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/session_status', false, $context);
    return $response;
  }

  /**
  * Ends an editing session.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $force force end session when user and/or author_id are not the owners.
  * @param int $regenerate request regeneration of subtitles and/or 
  *  synthesized audiotracks immediately after closing the session.
  */
  public function speech_end_session($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $force = 0, $regenerate = 0, $hash = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    if (!is_null($hash)) $data_json["hash"] = $hash;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["force"] = $force;
    $data_json["regenerate"] = $regenerate;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($this->API_EXPIRE)) $data_json["expire"] = $this->API_EXPIRE;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'speech/end_session', false, $context);
    return $response;
  }

  /******* TEXT API *******/

  /**
  * Returns the status of an uploaded document in JSON format.
  *
  * @param string $upload_id the ID of the uploaded document for which the API should
  *  return current status.
  */
  public function text_status($upload_id) {
    $data_json = array();
    $data_json["id"] = $upload_id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/status', false, $context);
    return $response;
  }

  /**
  * Returns metadata information of a given document ID in JSON format.
  *
  * @param string $id the ID of the document for which the API should
  *  return metadata information.
  */
  public function text_metadata($id) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/metadata', false, $context);
    return $response;
  }

  /**
  * Returns a list of translations available for a given document ID
  * in JSON format.
  *
  * @param string $id the ID of the document for which the API should
  *  return a list of available translations.
  */
  public function text_langs($id) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/langs', false, $context);
    return $response;
  }

  /**
  * Returns a list with all available MT systems in JSON format.
  *
  */
  public function text_systems() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/systems', false, $context);
    return $response;
  }

  /**
  * Returns a list with information of all uploaded documents in JSON format.
  *
  */
  public function text_uploadslist() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/uploadslist', false, $context);
    return $response;
  }

  /**
  * Returns a list with information of all existing documents in JSON format.
  *
  */
  public function text_list() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/list', false, $context);
    return $response;
  }

  /**
  * Returns a translation file for a given document ID and language 
  * in different formats.
  *
  * @param string $id the document ID
  * @param string $lang the language code of the translation.
  * @param int $format the format ID of the document (check
  *  documentation)
  * @param int $session_id translation modifications from the given
  *  session ID (check documentation)
  * @param int $sel_data_policy translation contents to be returned (check
  *  documentation)
  * @param string $expire expire date in UNIX timestamp format.
  */
  public function text_get($id, $lang, $format = NULL, $session_id = NULL, $sel_data_policy = NULL, $expire = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["lang"] = $lang;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($expire)) $data_json["expire"] = $expire;
    if (!is_null($format)) $data_json["format"] = $format;
    if (!is_null($session_id)) $data_json["session_id"] = $session_id;
    if (!is_null($sel_data_policy)) $data_json["sel_data_policy"] = $sel_data_policy;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/get', false, $context);
    return $response;
  }

  /**
  * Starts an editing session to send and commit modifications of a
  * translated document file.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param string $expire expiration date UNIX timestamp of the request.
  */
  public function text_start_session($id, $authorId, $authorName = NULL, $authorConf, $expire = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($expire)) $data_json["expire"] = $expire;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/start_session', false, $context);
    return $response;
  }

  /**
  * Send and commit modifications of a translation file made by a 
  * user under a session ID returned by /start_session interface.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId the editing session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param array $mods list of changes (check documentation)
  * @param string $expire expire date of the request in UNIX timestamp.
  */
  public function text_mod($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $mods, $expire = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["mods"] = $mods;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($expire)) $data_json["expire"] = $expire;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/mod', false, $context);
    return $response;
  }

  /**
  * Returns the current status of the given session ID. If it is alive,
  * it updates the last alive timestamp (last_update output key).
  * This interface is commonly used to avoid the automatic end of
  * session due to user inactivity.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $alive alive message type.
  * @param string $expire expiration date UNIX timestamp of the request.
  */
  public function text_session_status($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $alive = 0, $expire = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["alive"] = $alive;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($expire)) $data_json["expire"] = $expire;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/session_status', false, $context);
    return $response;
  }

  /**
  * Ends an opened editing session.
  *
  * @param string $id the ID of the media for which the API should
  *  modify the captions.
  * @param string $sessionId session ID.
  * @param string $authorId user name of user submitting the changes.
  * @param string $authorName name of user submitting the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $force force end session when user and/or author_id
  *  are not the owners.
  * @param string $expire expiration date UNIX timestamp of the request.
  */
  public function text_end_session($id, $sessionId, $authorId, $authorName = NULL, $authorConf, $force = 0, $expire = NULL) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["session_id"] = $sessionId;
    $data_json["author_id"] = $authorId;
    if (!is_null($authorName)) $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["force"] = $force;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    if (!is_null($expire)) $data_json["expire"] = $expire;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/end_session', false, $context);
    return $response;
  }

  /**
  * Returns a valid URL link for calling TLP Player given
  * input parameters.
  *
  * @param string $id the ID of the media.
  * @param string $lang the language code of the captions.
  * @param string $authorId user name of user doing the changes.
  * @param string $authorName name of user doing the changes.
  * @param int $authorConf integer from 0 to 100 indicating the
  *  user confidence.
  * @param int $expire expire date of this URL in UNIX timestamp
  *  format.
  */
  public function gen_text_editor_url($id, $lang, $authorId, $authorName, $authorConf, $expire) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["lang"] = $lang;
    $data_json["author_id"] = $authorId;
    $data_json["author_name"] = $authorName;
    $data_json["author_conf"] = $authorConf;
    $data_json["expire"] = $expire;
    $data_json["api_user"] = $this->API_USER;
    $data_json["request_key"] = sha1($id.$expire.$this->API_USER.$this->API_SECRET_KEY).sha1($authorId.$authorConf.$expire.$this->API_USER.$this->API_SECRET_KEY);
    if (strpos($this->TEXT_EDITOR_URL, '?') !== false) return $this->TEXT_EDITOR_URL.'&request='.base64_encode(json_encode($data_json));
    else return $this->TEXT_EDITOR_URL.'?request='.base64_encode(json_encode($data_json));
  }

  /**
  * Creates a Document Package in a temporal directory
  * and sends it to /ingest/new API interface.
  *
  * Returns a string if an error ocurred, or a JSON
  * string (/ingest/new response text) if the ingest
  * process was successful.
  * 
  * @param array $titles titles of the documents being ingested.
  * @param array $ids IDs documents will have once ingested.
  * @param string $lang language code of ingested documents.
  * @param array $requested_langs list of languages for which
  *  translations will be generated.
  * @param array $docFiles array of DOLLAR_FILES PHP variable
  *  with uploaded document files or file URL strings.
  */
  public function text_ingest_new($titles, $ids, $lang, $requested_langs, $docFiles, $from_webui = false) {
    if (!is_array($ids) || !is_array($titles) || !is_array($docFiles)) {
      return "/ingest error: wrong parameters";
    }
    else if (count($ids) != count($titles) || count($titles) != count($docFiles)) {
      return "/ingest error: element missmatch";
    }
    $manifest=array();
    if ($from_webui) $manifest['from_webui'] = true;
    $manifest['language']=$lang;
    $manifest['documents']=array();
    $tempDir = tempnam($this->TEMP_DIR, 'up-');
    if (file_exists($tempDir)) unlink($tempDir);
    if (!mkdir($tempDir, 0777)) return "/ingest error: unable to create temp directory ".$tempDir;
    for ($i=0; $i<count($ids); $i++) {
      $doc = array(
        'external_id' => $ids[$i],
        'title' => $titles[$i],
        'filename' => $this->rand_str().'.'.pathinfo($docFiles[$i]['name'], PATHINFO_EXTENSION),
        'fileformat' => pathinfo($docFiles[$i]['name'], PATHINFO_EXTENSION)
      );
      if (!move_uploaded_file($docFiles[$i]['tmp_name'], $tempDir.'/'.$doc['filename'])) return "/ingest error: cannot move uploaded file to temp directory";
      $doc['md5'] = md5_file($tempDir.'/'.$doc['filename']);
      $manifest['documents'][]=$doc;
    }
    $manifest['requested_langs'] = $requested_langs;
    /* docpackage.zip */
    $dpzip = new ZipArchive;
    if ($dpzip->open($tempDir.'/docpackage.zip', ZipArchive::CREATE) !== true) return "/ingest error: cannot create ".$tempDir."/docpackage.zip file";
    foreach ($manifest['documents'] as $doc) {
      $dpzip->addFile($tempDir.'/'.$doc['filename'], $doc['filename']);
    }
    $numZipFiles = $dpzip->numFiles;
    $dpzip->close();
    /* create manifest.json file */
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $data_json["manifest"] = $manifest;
    $mf=fopen($tempDir.'/manifest.json', 'w');
    fwrite($mf, json_encode($data_json));
    fclose($mf);
    /* ingest files */
    $curl = curl_init($this->API_URL.'text/ingest/new');
    curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
    curl_setopt($curl, CURLOPT_POST, true);
    curl_setopt($curl, CURLOPT_HTTPHEADER, array('Content-Type: multipart/form-data')); 
    if ($numZipFiles > 0) curl_setopt($curl, CURLOPT_POSTFIELDS, array('json' => '@'.$tempDir.'/manifest.json'.';type=application/json', 'pkg' => '@'.$tempDir.'/docpackage.zip'.';type=application/zip'));
    else curl_setopt($curl, CURLOPT_POSTFIELDS, array('json' => '@'.$tempDir.'/manifest.json'.';type=application/json'));
    $response = curl_exec($curl);
    $status = curl_getinfo($curl, CURLINFO_HTTP_CODE);
    /* remove temp files and folder */
    $tmp_files = glob($tempDir.'/*');
    foreach ($tmp_files as $tf) if (is_file($tf)) unlink($tf);
    rmdir($tempDir);
    if ($status != 200) return "/ingest/new error: ".($status == 0 ? curl_error($curl) : "Web Service returned HTTP status ".$status);
    return $response;
  }

  /**
  * Cancel an upload file being processed
  *
  * @param string $id the Upload ID returned by the /ingest interface
  *  of the document to be canceled.
  */
  public function text_ingest_cancel($id) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/ingest/cancel', false, $context);
    return $response;
  }

  /**
  * Delete existing media
  *
  * @param string $id the document ID.
  */
  public function text_ingest_delete($id) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/ingest/delete', false, $context);
    return $response;
  }

  /**
  * Returns a list of all edit sessions for all API user's
  * document files that are pending to be revised.
  *
  */
  public function text_revisions() {
    $data_json = array();
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/revisions', false, $context);
    return $response;
  }

  /**
  * Returns a list of all edit sessions carried out over an specific Document ID. 
  *
  * @param string $id the document ID.
  */
  public function text_edit_history($id) {
    $data_json = array();
    $data_json["id"] = $id;
    $data_json["user"] = $this->API_USER;
    $data_json["auth_token"] = $this->API_SECRET_KEY;
    $context = stream_context_create(array(
      'http' => array(
        'header'  => "Content-type: application/json\r\n",
        'method' => 'POST',
        'content' => json_encode($data_json)
      )
    ));
    $response = @file_get_contents($this->API_URL.'text/edit_history', false, $context);
    return $response;
  }

  /******* PRIVATE FUNCTIONS *******/

  private function rand_str($length = 20) {
    $chars = '0123456789abcdef';
    $rstring = '';
    for ($i = 0; $i < $length; $i++) {
      $rstring .= $chars[rand(0, strlen($chars)-1)];
    }
    return $rstring;
  }
}
?>
