#!/usr/bin/python
# -*- coding: utf-8 -
#
#  Copyright 2015 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
import sys
import os
from ConfigParser import ConfigParser
from lib.libtlpauth import TLPAuth
from lib.libwsstorage import tLWSSLayer


def parse_config(cfg_fd): 

    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}

    CFG['AUT'] = {}
    CFG['AUT']['KEY'] = config.get('authentication', 'key_generator')

    CFG['STG'] = {}
    CFG['STG']['DB_NAME'] = config.get('storage', 'db_name')
    CFG['STG']['DB_USER'] = config.get('storage', 'db_user')
    CFG['STG']['DB_HOST'] = config.get('storage', 'db_host')
    if CFG['STG']['DB_HOST'] != None: 
      if CFG['STG']['DB_HOST'].strip() == "": CFG['STG']['DB_HOST'] = None
    CFG['STG']['DB_PASS'] = config.get('storage', 'db_passwd')
    if CFG['STG']['DB_PASS'] != None: 
      if CFG['STG']['DB_PASS'].strip() == "": CFG['STG']['DB_PASS'] = None

    return CFG


path = os.path.dirname(__file__)
if path == "": path = "./"
CFG_FILE = "%s/config.ini" % path # tL-WS.cfg


if __name__ == "__main__":

    if len(sys.argv) != 2:
        sys.stderr.write("Usage: %s <username>\n" % sys.argv[0])
        sys.exit(1)

    user_name = sys.argv[1]

    with open(CFG_FILE) as cfg_fd:
        CFG = parse_config(cfg_fd)

    tlpauth = TLPAuth(CFG['AUT']['KEY'])
    tL = tLWSSLayer(CFG['STG']['DB_NAME'], CFG['STG']['DB_USER'], CFG['STG']['DB_HOST'], tLWSSLayer.OBJ_SPEECH, passwd=CFG['STG']['DB_PASS'])

    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        sys.stderr.write("[EE] Cannot get user information from DB.\n" % user_name)
        sys.exit(1)
    if len(WSSdata.data) == 0:
        sys.stderr.write("[EE] Username %s does not exist.\n" % user_name)
        sys.exit(1)

    user_id = WSSdata.data[0]['id']
    user_pass = WSSdata.data[0]['password']
    is_admin = WSSdata.data[0]['is_admin']
    
    secret_key = tlpauth.get_secret_key(user_name.encode('utf-8'), user_pass)

    sys.stdout.write("%s\n"% (secret_key))
    


