#!/usr/bin/python

import sys
import bs4
import cStringIO
import libdtlx

def regen_html_from_dtlx_files(dtlx_fp, html_fp):
  dtlx = libdtlx.Parser().parse(dtlx_fp)
  with open(html_fp) as html_fd:
    html = html_fd.read()
  return regen_html_from_dtlx(dtlx, html)

def regen_html_from_dtlx(dtlx, html):
  bs = bs4.BeautifulSoup(html)
  dtlx.filter(dtlx.FILTER_GET_LAST) # Best alts
  for s in dtlx.sentences:
    d_id = s.sentId.split("->")[0]
    n = bs.find(attrs={"data-dtlx-id": d_id})
    if not(n.get('data-dtlx-cleared', False)):
      n.clear()
      n['data-dtlx-cleared'] = 1
    if s.filteredTranslation.formattedText != None:
        d = s.filteredTranslation.formattedText
    else:
        d = s.filteredTranslation.rawText
    n.append(d)
  for i in bs.descendants:
    try:
      del i['data-dtlx-cleared']
    except:
      pass 
  return bs.prettify(formatter=None).encode('utf-8')

def regen_source_html_from_dtlx(dtlx_fp, html_fp):
  dtlx = libdtlx.Parser().parse(dtlx_fp)
  with open(html_fp) as html_fd:
    bs = bs4.BeautifulSoup(html_fd)
  for s in dtlx.sentences:
    d_id = s.sentId.split("->")[0]
    n = bs.find(attrs={"data-dtlx-id": d_id})
    if not(n.get('data-dtlx-cleared', False)):
      n.clear()
      n['data-dtlx-cleared'] = 1
    if s.origin.formattedText != None:
        d = s.origin.formattedText
    else:
        d = s.origin.rawText
    n.append(d)
  for i in bs.descendants:
    try:
      del i['data-dtlx-cleared']
    except:
      pass 
  return bs.prettify(formatter=None).encode('utf-8')


def has_span_dtlx_attr(tag):
  return tag.name == "span" and tag.has_attr('data-dtlx-id')

def has_dtlx_attr(tag): # Compatibility for old parsings
  return tag.has_attr('data-dtlx-id')

def clean_html(data):
  s = bs4.BeautifulSoup(data)
  for i in s.find_all(has_span_dtlx_attr):
    i.unwrap()
  for i in s.find_all(has_dtlx_attr): # Compatibility for old parsings
    del i['data-dtlx-id']
  return s.prettify().encode('utf-8')

def remove_html(data):
  s = bs4.BeautifulSoup(data)
  out = cStringIO.StringIO()
  for i in s.stripped_strings:
      out.write("%s\n" % i.encode('utf-8'))
  return out.getvalue()

if __name__ == "__main__":
  
  if len(sys.argv) != 3:
    sys.stderr.write("Usage: %s <DTLX> <HTML>\n" % sys.argv[0])
    sys.exit(1)

  dtlx_fp = sys.argv[1]
  html_fp = sys.argv[2]
  
  html = regen_html_from_dtlx_files(dtlx_fp, html_fp)

  print html

