#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from sys import argv, exit, stdin, stdout, stderr
import sys
import codecs
import os
import io
import subprocess
import shlex
import argparse
import tempfile
import libdfxp

reload(sys)
sys.setdefaultencoding('utf-8')

#awk style dictionary
class D(dict):
  def __getitem__(self, i):
    if i not in self: self[i] = 0
    return dict.__getitem__(self, i)


def escaped_word(word):
  if word == "<unk>" or word == "[hesitation]":
    return True
  else:
    return False

##############################################################################
#########   Gets the confidence measure at the X percentage ##################
##############################################################################
def get_cm_at_percentage(input_dfxp, percentage):
  #order words per confidence
  all_words = []
  for i in range(len(input_dfxp.data)):
    if input_dfxp.data[i].authorType == "human":
      continue
    words = input_dfxp.data[i].data
    for j in xrange(len(words)):
      if isinstance (words[j], libdfxp.TlGroup) or not escaped_word(words[j].txt):
        all_words.append(words[j].confMeasure)
  all_words.sort()
  #select the threshold under all words are supervised
  index = int(len(all_words)*percentage+0.5)
  if index > len(all_words) - 1:
    index = len(all_words) - 1
  thr_cm = all_words[index]

  return thr_cm

####################################################################################
#########   gets a vector with the index of words to be supervised         #########
#########   words below thr_cm are chosen along with the context marked by #########
#########   word_index - prev_context , word_index + aftr_context          #########
####################################################################################
def get_words_to_supervise_fusion(dfxp_segment, thr_cm, prev_context, aftr_context, 
    min_length=3, make_chunks=True):

  words = dfxp_segment.data
  sup_words = []

  #obtain words and their context to be supervised
  for i in xrange(len(words)):
    if isinstance(words[i], libdfxp.TlGroup):
      continue
    if not escaped_word(words[i].txt) and words[i].confMeasure <= thr_cm:
      #get the context and ignore escaped_words
      #previous context
      beg_word = i - 1
      j = 0
      while beg_word > 0 and j < prev_context:
        if isinstance(words[beg_word], libdfxp.TlGroup): #manual segment cannot be presented
          break
        if not escaped_word(words[beg_word].txt):
          j += 1
        beg_word-=1
      beg_word += 1 #add a position to solve the last increment

      #posterior context
      end_word = i + 1
      j = 0
      while end_word < len(words) and j < aftr_context:
        if isinstance(words[end_word], libdfxp.TlGroup):
          break
        if not escaped_word(words[end_word].txt):
          j += 1
        end_word+=1
      end_word -= 1

      phrase = range(beg_word, end_word+1)

      if len(phrase) < min_length:
        continue

      #check if the phrase is overlapped with the latter and fusion them if so
      if len(sup_words) > 0 and beg_word in sup_words[-1]:
        sup_words[-1] = sup_words[-1][0:sup_words[-1].index(beg_word)] + phrase #add those that not intersect
      else:
        sup_words.append(phrase)

      # split fused segment in chunks
      if make_chunks:
        phrase_len = int(prev_context+aftr_context+1)
        chunks = int(len(sup_words[-1])/float(phrase_len))
        if chunks > 1:
          sup_words_split = []
          ini = 0
          #insert chuks of fixed size
          for i in range(0,chunks-1):
            end = ((i+1)*phrase_len)
            if end > len(sup_words[-1]):
              end = len(sup_words[-1])
            sup_words_split.append(sup_words[-1][ini:end])
            ini = end
          #insert the last one with the remainder
          sup_words_split.append(sup_words[-1][ini:len(sup_words[-1])])
          sup_words = sup_words_split

  return (sup_words if len(sup_words) > 0 else None)

####################################################################################
####### reads a dfxp (with only automatic transcription) and order it according to their
####### confidence measure
####################################################################################
def order_dfxp_per_confidence(dfxp):
  v = []

  #Order segments per confidence
  for segment in dfxp.data:
    v.append((segment.data[-1].confMeasure, segment))
  v.sort()

  #reorder data
  data = []
  for i, segment in v:
    data.append(segment)
  dfxp.data = data

####################################################################################
####### reads a dfxp and selects for supervision the first percentage of words
####### as ordered per confidence, and along a context for each
####################################################################################
def filter_dfxp_over_percentage(input_dfxp, percentage, prev_context, aftr_context):
  out_dfxp = libdfxp.TlDfxp()

  #copy header
  out_dfxp.authorType = input_dfxp.authorType
  out_dfxp.authorId = input_dfxp.authorId
  out_dfxp.timeStamp = input_dfxp.timeStamp
  out_dfxp.lang = input_dfxp.lang

  thr_cm = get_cm_at_percentage(input_dfxp, percentage)

  #count sub_ids
  out_dfxp_segment_id = D()

  for segment in input_dfxp.data:

    if segment.authorType == "human":
      continue

    sup_words = get_words_to_supervise_fusion(segment, thr_cm, prev_context, aftr_context)

    #put this words in the new dfxp
    if sup_words != None:
      for i in sup_words:
        #create new segment
        new_data = []

        last = None
        for j in i:
          #get the word to push
          new_data.append(segment.data[j])
          last = j

        new_segment = libdfxp.TlSegment(out_dfxp,
            "%s.%d" %(segment.segmentId, out_dfxp_segment_id[segment.segmentId]),
            segment.data[i[0]].begin, segment.data[i[-1]].end)
        #increment sub.id to avoid collisions on the resulting dfxp
        out_dfxp_segment_id[segment.segmentId]+=1
        new_segment.data = new_data
        new_segment.confMeasure = new_data[-1].confMeasure #put the confidence of the word 
                                                           #to be supervised
        out_dfxp.data.append(new_segment)

  order_dfxp_per_confidence(out_dfxp)

  return out_dfxp

def convert_to_segmented_dfxp(dfxp, time_delay=1.00):
  i=0.0
  for segment in dfxp.data:
    segment.begin = round_25fps(segment.begin)
    segment.end = round_25fps(segment.end)
    duration = segment.end - segment.begin
    segment.end = (duration) + i
    segment.begin = i
    i = segment.end + time_delay

def round_25fps(number):
  base25 = int(number/ 0.04 +0.5)
  return base25*0.04

def cut_video(dfxp, video, video_out):
  tmpdir = tempfile.mkdtemp(dir="/home/tmp/")

  stdout.write("avconv -i %s -q 1 -c:v mjpeg -c:a adpcm_ima_wav -ar 48k %s/tmp.avi\n" %(video, tmpdir))
  clips=[]
  for segment in dfxp.data:
    begin = round_25fps(segment.begin)
    end = round_25fps(segment.end)

    stdout.write("avconv -i %s/tmp.avi -ss %f -t %f -c:v mjpeg -c:a adpcm_ima_wav %s/clip_%s.avi\n" \
        %(tmpdir, begin, end-begin, tmpdir, segment.segmentId))
    stdout.write("avconv -i %s/tmp.avi -r 1 -ss %f -t 0.04 -q 1 -f image2 %s/clip_%s_thumbnail.png\n" \
        %(tmpdir, begin + (end-begin), tmpdir, segment.segmentId))
    stdout.write("for i in `seq 0 24`; do ln -s %s/clip_%s_thumbnail.png %s/clip_%s_thumbnail_`printf \"%%02d\" $i`.png; done\n" \
        %(tmpdir, segment.segmentId, tmpdir, segment.segmentId))
    stdout.write("avconv -f image2 -i %s/clip_%s_thumbnail_%%02d.png -i silence.wav -r 25 -s 1280x720 \
-c:v mjpeg -c:a adpcm_ima_wav -q 1 %s/clip_%s_silence.avi\n" \
        %(tmpdir, segment.segmentId, tmpdir, segment.segmentId))

    clips.append("%s/clip_%s.avi|%s/clip_%s_silence.avi" %(tmpdir, segment.segmentId, tmpdir, segment.segmentId))
  stdout.write("avconv -i 'concat:%s' -c:v libx264 -c:a libvo_aacenc %s-cut.mp4\n" %("|".join(clips), video_out))
  stdout.write("rm -rf %s\n" %(tmpdir))

#####################################################################
#####################################################################
#####################################################################
def main():

  parser = argparse.ArgumentParser(description=
      'Cuts a dfxp into another one according to it confidence measures')

  parser.add_argument("input_dfxp", help='Input .dfxp file to be processed')
  parser.add_argument("percentage_of_words",
      help='Percentage of recognised words that will be supervised, ordered by confidence', type=float)

  parser.add_argument("-b", "--before", help='Number of words considered before the low confident one. Default (3)',
      type=int, default=3)
  parser.add_argument("-a", "--after", help='Number of words considered after the low confident one. Default (0)',
      type=int, default=0)

  parser.add_argument("-n", "--new-video", help='Make a new video with only the extracted chunks',
    action='store_true')
  parser.add_argument("--video", help='Path of the video to extract segmetns to create a new video')
  parser.add_argument("--video-out", help='Path of the output video')

  parser.add_argument('-v', help='Outputs additional verbose information',
    action='store_true')

  args = parser.parse_args(argv[1:])

  input_dfxp = libdfxp.parseDfxp(args.input_dfxp)

  output_dfxp = filter_dfxp_over_percentage(input_dfxp, args.percentage_of_words/100, args.before, args.after)

  if args.new_video:
    if args.video:
      cut_video(output_dfxp, args.video, args.video_out)

    convert_to_segmented_dfxp(output_dfxp)

  print output_dfxp.toXml()

####################################################################################################
if __name__ == "__main__":
  main()


