# -*- coding: utf-8 -
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

####---- Standard Module imports ----####

import sys
import os
import json
import hashlib
import base64
from cgi import parse_qs, FieldStorage
from datetime import datetime, timedelta
from imp import load_source
from ConfigParser import ConfigParser
from collections import OrderedDict


import smtplib
from email.mime.text import MIMEText
from datetime import datetime
import traceback
from cStringIO import StringIO
import pprint
import urllib

####---- tL libraries + Config file ----####

WS_PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, WS_PATH)

from lib.libdfxp import *
from lib.libtlpauth import TLPAuth
from lib.libwsstorage import tLWSSLayer, tLWSSData

CFG_FILE = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))

####---- Cache control ----####

CACHE_CONTROL="no-cache, no-store, private, must-revalidate, max-stale=0"
PRAGMA="no-cache"
EXPIRES="Tue, 01 Jan 1980 12:00:00 GMT"
CACHE_RESP_HEADERS = [('Cache-Control', CACHE_CONTROL), ('Pragma', PRAGMA), ('Expires', EXPIRES)]


#################################################
#--------------- Webservice API ----------------#
#################################################

def my_application(env, start_response, tL, tlpauth):


 
    #---------------------------------------------#
    #----------------- /systems ------------------#
    #---------------------------------------------#
  
    if env['PATH_INFO'] == '/systems':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Dresp = {}

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/systems] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/systems] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_systems_getUserSystems(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/systems] tL.tLint_systems_getUserSystems() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Dresp['asr'] = [ { 'id':s['id'], 'lang':s['lang1'], 'name':s['name'], 'description':s['description'], 'rtf':s['rtf'], 'topic_adaptation':s['features_tadapt'], 'topic_adaptation_time':s['const_time'] } for s in WSSdata.data if s['type'] == 'asr' ]
        Dresp['mt']  = [ { 'id':s['id'], 'source_lang':s['lang1'], 'target_lang':s['lang2'], 'name':s['name'], 'description':s['description'], 'topic_adaptation':s['features_tadapt'], 'processing_time':s['const_time'] } for s in WSSdata.data if s['type'] == 'mt' ]
        Dresp['tts'] = [ { 'id':s['id'], 'lang':s['lang1'], 'voice_gender':s['tts_voice_gender'], 'name':s['name'], 'description':s['description'] } for s in WSSdata.data if s['type'] == 'tts' ]
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #-------------------------------------------------#
    #----------------- /uploadslist ------------------#
    #-------------------------------------------------#
  
    if env['PATH_INFO'] == '/uploadslist':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        object_id = body.get("object_id", None)
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Lresp = []

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/uploadslist] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/uploadslist] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_uploadslist_getUploads(user_id, object_id)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/uploadslist] tL.tLint_uploadslists_getUploads() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Lresp = [ { 'id':s['id'], 'status_code':s['status_id'], 'object_id':s['object_id'], 'last_update':str(s['last_update']), 'uploaded_on':str(s['uploaded_on']) } for s in WSSdata.data ]
        data = json.dumps(Lresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #---------------------------------------------#
    #----------------- /list ------------------#
    #---------------------------------------------#
  
    if env['PATH_INFO'] == '/list':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/list] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/list] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_list_getList(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/list] tL.tLint_list_getList() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Lresp = [ { 'id':s['id'], 'title':s['title'], 'language':s['lang'], 'duration':s['duration'] } for s in WSSdata.data ]
        data = json.dumps(Lresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #--------------------------------------------#
    #------------------ /align ------------------#
    #--------------------------------------------#

    elif env['PATH_INFO'].find('/align') == 0: 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/align] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/align] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/align] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        if env['PATH_INFO'] == '/align/new':
            try:
                segs = body["segments"]
            except: 
                return error_400(start_response, "Missing mandatory parameters/keys." )
           
            if not(encoded_params):
                try:
                    segs = json.loads(base64.b64decode(segs), encoding="UTF-8")
                except:
                    error_400(start_response, "Could not decode 'segments' GET parameter (should be base64-encoded).")
 
            if not(isinstance(segs, list)):
                return error_400(start_response, "'segments' key must be a list." )
    
            for seg in segs:
                if not(isinstance(seg, dict)):
                    return error_400(start_response, "'segments' elements must be dicts." )
                try:
                    st = seg["b"]
                    et = seg["e"]
                    txt = seg["t"]
                except:
                    error_400(start_response, "'segments' elements must be dicts with 'st', 'et' and 'txt' keys." )
                try:
                    et = float(et)
                except:
                    return error_400(start_response, "'et' key value must be a float." )
                try:
                    st = float(st)
                except:
                    return error_400(start_response, "'st' key value must be a float." )

            align_id = body.get("align_id", None)
            debug_mode = body.get("debug", False)
            session_id = body.get("session_id", None)
            if session_id != None:
                try:
                    session_id = int(session_id)
                except:  
                    return error_400(start_response, "'session_id' value must be an integer." )
 
            if align_id != None:
                ret = tL.tLint_align_cancel_align(align_id)
                if ret.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/align] tL.tLint_align_new() returned %s" % (ret.scode))
                    return error_500(start_response)

            ret = tL.tLint_align_new(lec_hash, segs, session_id, debug_mode)
            if ret.scode != 0:
                report_error(CFG, env, "[TLP][WS][/align] tL.tLint_align_new() returned %s" % (ret.scode))
                return error_500(start_response)
            this_align_id = ret.data

            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Alignment request successfully stored.' 
            Dresp['align_id'] = this_align_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        elif env['PATH_INFO'] == '/align/get':
            try:
                align_id = body["align_id"]
            except: 
                return error_400(start_response, "Missing mandatory parameters/keys." )

            ret = tL.tLint_align_get(align_id, lec_hash)
            if ret.scode != 0:
                report_error(CFG, env, "[TLP][WS][/align] tL.tLint_align_status() returned %s" % (ret.scode))
                return error_500(start_response)
            if ret.results == 0:
                Dresp['rcode'] = 2
                Dresp['rcode_description'] = "Align ID [ %s ] for media ID [ %s ] does not exist." % (align_id, extid)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
            elif ret.results == 1:
                d = ret.data[0]
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = "Returning align info."
                Dresp['status_code'] = d['status_id']
                Dresp['status_info'] = d['info']
                Dresp['error_code'] = d['error_code']
                Dresp['sent_on'] = str(d['timestamp'])
                Dresp['estimated_time_complete'] = str(d['eta']) if d['eta'] != None else None
                Dresp['completed_on'] = str(d['completed_on']) if d['completed_on'] != None else None
                Dresp['align_data'] = d['output_data']
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
            else:
                report_error(CFG, env, "[TLP][WS][/align/status] tL.tLint_align_status() returned more than one row")
                return error_500(start_response)

        elif env['PATH_INFO'] == '/align/cancel': 
            try:
                align_id = body["align_id"]
            except: 
                return error_400(start_response, "Missing mandatory parameters/keys." )

            ret = tL.tLint_align_cancel(align_id, lec_hash)
            if ret.scode != 0:
                if ret.scode == 2:
                    Dresp['rcode'] = 1
                    Dresp['rcode_description'] = "Align ID [ %s ] for media ID [ %s ] does not exist." % (align_id, extid)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data]
                else:
                    report_error(CFG, env, "[TLP][WS][/align] tL.tLint_align_get() returned %s" % (ret.scode))
                    return error_500(start_response)

            Dresp['rcode'] = 0
            Dresp['rcode_description'] = "Align process successfully cancelled."
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        else: 
            return error_400(start_response, 'Unknown operation')



 
    #--------------------------------------------#
    #----------------- /ingest ------------------#
    #--------------------------------------------#

    elif env['PATH_INFO'].find('/ingest') == 0: 

        opc = None
        if env['PATH_INFO'] == '/ingest/new': opc = 0
        elif env['PATH_INFO'] == '/ingest/update': opc = 1
        elif env['PATH_INFO'] == '/ingest/delete': opc = 2
        elif env['PATH_INFO'] == '/ingest/cancel': opc = 3
        else: return error_400(start_response, 'Unknown operation')

        pkg_file = None
        if opc in [0,1]:

            if env['REQUEST_METHOD'] != 'POST':
                return error_400(start_response, "This interface only supports POST requests")
             
            if env['CONTENT_TYPE'].find('application/json') >= 0:
                length = int(env.get('CONTENT_LENGTH', '0'))
                if length == 0:
                    return error_400(start_response, "Content-Length must be greater than 0")
                try:
                    data = env['wsgi.input'].read(length)
                except:
                    return error_400(start_response, "Cannot read POST data.")
    
            elif env['CONTENT_TYPE'].find('multipart/form-data') >= 0: 
                try:
                    fs = FieldStorage(fp=env['wsgi.input'],environ=env)
                except:
                    report_error(CFG, env, "[TLP][WS][/ingest] Exception building FieldStorage()")
                    return error_500(start_response)
                if "json" not in fs.keys():
                    #return error_400(start_response, str(fs.keys()))
                    return error_400(start_response, "missing 'json' field")
                if fs['json'].type != "application/json":
                    #return error_400(start_response, "'json' field Content-Type must be 'application/json'")
                    pass
                data = fs.getvalue("json")

            else:
                return error_400(start_response, "Unsupported Content-Type.")

            try:
                body = json.loads(data)
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
            try:
                manif = body["manifest"]
            except:
                return error_400(start_response, "'manifest' parameter missing.")
            ok, req_pkg, err = check_manifest(manif, opc)
            if not(ok):
                err_code, err_info = err
                if err_code == 400:
                    return error_400(start_response, err_info)
                elif err_code == 500:
                    return error_500(start_response)
            if req_pkg:
                if env['CONTENT_TYPE'].find('multipart/form-data') < 0:
                    return error_400(start_response, 'According to manifest data, you must attach too a media package (Content-Type: multipart/form-data)')
                if "pkg" not in fs.keys():
                    return error_400(start_response, "'pkg' field missing.")
                if fs['pkg'].type != "application/zip":
                    return error_400(start_response, "'pkg' field Content-Type must be 'application/zip'")
                pkg_file = fs["pkg"].file
   

        elif opc in [2,3]:
            encoded_params = True
            if env['REQUEST_METHOD'] == 'GET':
                queries = parse_qs(env['QUERY_STRING'])
                if "data" in queries:
                    try:
                        data = base64.b64decode(queries["data"][0])
                    except: 
                        return error_400(start_response, "Could not parse 'data' GET Parameter" )
                else:
                    encoded_params = False
    
            else: # POST
                if env.get('CONTENT_TYPE', None) != 'application/json':
                    return error_400(start_response, "Content-Type must be 'application/json'")
                length = int(env.get('CONTENT_LENGTH', '0'))
                if length == 0:
                    return error_400(start_response, "Content-Length must be greater than 0")
                try:
                    data = env['wsgi.input'].read(length)
                except:
                    return error_400(start_response, "Cannot read POST data.")
    
            if encoded_params:
                try:
                    body = json.loads(data, encoding="UTF-8")
                except:
                    return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
            else:
                body = { k:queries[k][0] for k in queries.keys() }
     

            try:
                extid = body["id"]
            except: 
                return error_400(start_response, "Missing mandatory parameters/keys." )

            manif = {'metadata':{'external_id':extid}}
            if opc == 2:
                del_mode = body.get("mode", "hard")
                manif['delete_mode'] = del_mode
     
        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        su = body.get("su", None)
        debug = body.get("debug", False)

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/ingest] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/ingest] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")
                
        lec_hash = get_hash(manif['metadata']['external_id'], user_id)

        WSSdata = tL.tLint_ingest_storeMP(opc, manif, lec_hash, user_id, pkg_file=pkg_file, debug_mode=debug)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/ingest] tL.tLint_ingest_storeMP() returned %s" % WSSdata.scode)
            return error_500(start_response)
        upload_id = WSSdata.data[0]['upload_id']
        Dresp = {'rcode':0, 'rcode_description':'Ingestion complete.', 'id':str(upload_id), 'hash':lec_hash}
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
   

    #--------------------------------------------#
    #----------------- /status ------------------#
    #--------------------------------------------#

    elif env['PATH_INFO'] == '/status': # See transcription status

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            upid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )
    
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
       
        Dresp = {}
        
        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/status] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/status] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_auth_is_owner_upload(user_id, upid)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_auth_is_owner_upload() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Upload ID [ %s ] does not exist.' % upid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_400(start_response, "Authentication error: user '%s' does not own upload ID '%s'." % (str(user), str(upid)))

        WSSdata = tL.tLint_status_getStatus(upid) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_status_getStatus() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = "Upload ID [ %s ] does not exist." % upid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        elif WSSdata.results == 1:
            res = WSSdata.data[0]
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = "Upload ID exists."
            Dresp['status_code'] = res['upload_status']
            Dresp['object_id'] = res['object_id']
            Dresp['status_info'] = res['info']
            Dresp['error_code'] = res['error_code']
            Dresp['uploaded_on'] = str(res['up_ts'])
            Dresp['last_update'] = str(res['last_check'])
            Dresp['estimated_time_complete'] = str(res['eta'])
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        else:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_status_getStatus() returned more than one row")
            return error_500(start_response)


    #-------------------------------------------------#
    #----------------- /metadata ------------------#
    #-------------------------------------------------#
  
    elif env['PATH_INFO'] == '/metadata':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/metadata] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/metadata] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_lecturedata_getMedia(extid, lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_lecturedata_getMedia() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist or has no media.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        else:
            res = WSSdata.data
            media = []
            for r in res:
                if not(r['is_url']) and CFG['URL']['USE']:
                    location = "%s/%s" % (CFG['URL']['BASEURL'], r['location'])
                    is_url = True
                else:
                    location = r['location']
                    is_url = r['is_url']
                media.append({ 'type_code': r['type_code'], 'media_format': r['media_format'], 'location': location , 'is_url':is_url})

        WSSdata = tL.tLint_lecturedata_getAttachments(extid, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_lecturedata_getAttachments() returned %d" % WSSdata.scode)
            return error_500(start_response)
        else:
            res = WSSdata.data
            attachments = []
            for r in res:
                if not(r['is_url']) and CFG['URL']['USE']:
                    location = "%s/%s" % (CFG['URL']['BASEURL'], r['location'])
                    is_url = True
                else:
                    location = r['location']
                    is_url = r['is_url']
                attachments.append({ 'type_code': r['type_code'], 'media_format': r['media_format'], 'location': location , 'is_url':is_url})

        WSSdata = tL.tLint_lecturedata_getLectureinfo(extid, lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_lecturedata_getLectureInfo() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist or has no metadata.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        else:
            res = WSSdata.data
            lecinfo = { 'language': res[0]['language'], 'duration': res[0]['duration'], 'title': res[0]['title'], 'category': res[0]['topic'] }

        WSSdata = tL.tLint_lecturedata_getSpeakers(extid, lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_lecturedata_getSpeakers() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            #Dresp['rcode'] = 1
            #Dresp['rcode_description'] = 'Media ID [ %s ] does not exist or has no metadata.' % extid
            #data = json.dumps(Dresp, indent=4)
            #status = '200 OK'
            #response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            #start_response(status, response_headers + CACHE_RESP_HEADERS)
            #return [data]
            lecinfo['speakers'] = []
        else:
            res = WSSdata.data
            lecinfo['speakers'] = [ { 'name':r['name'] } for r in res ]

        WSSdata = tL.tLint_lecturedata_getAudiotracks(extid, lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_lecturedata_getAudioTracks() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            audiotracks = []
        else:
            res = WSSdata.data
            audiotracks = []
            for r in res:
                if not(r['is_url']) and CFG['URL']['USE']:
                    location = "%s/%s" % (CFG['URL']['BASEURL'], r['location'])
                    is_url = True
                else:
                    location = r['location']
                    is_url = r['is_url']
                audiotracks.append( { 'id':r['id'], 'lang':r['lang'], 'voice_type':r['voice_type'], 'voice_gender':r['voice_gender'], 'audio_type':r['audio_type'], 'description':r['desc'], 'sub_type':r['sub_type'], 'media_format': r['media_format'], 'location': location , 'is_url':is_url})
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Media list and info available.'
        Dresp['mediainfo'] = lecinfo
        Dresp['media'] = media
        Dresp['audiotracks'] = audiotracks
        Dresp['attachments'] = attachments
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #-------------------------------------------#
    #----------------- /langs ------------------#
    #-------------------------------------------#
  
    elif env['PATH_INFO'] == '/langs':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None, 'media_lang':None, 'langs':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/langs] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/langs] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/langs] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_langs_getLangs(extid, lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/langs] tL.tLint_langs_getLangs() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist or has no captions.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        else:
            # Get audiotracks
            atresdata = tL.tLint_lecturedata_getAudiotracks(extid, lec_hash)
            if atresdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/langs] tL.tLint_lecturedata_getAudiotracks() returned %s" % (atresdata.scode))
                return error_500(start_response)
            # Prepare data and return
            atd = {}
            for d in atresdata.data:
                try:
                    atd[d["lang"]].append({ "aid":d["id"], "voice_type":d["voice_type"], "voice_gender":d["voice_gender"] })    
                except KeyError:
                    atd[d["lang"]] = [{ "aid":d["id"], "voice_type":d["voice_type"], "voice_gender":d["voice_gender"] }]
            res = WSSdata.data
            media_lang = res[0]['media_lang']
            locked = res[0]['locked']
            langs = [ { 'lang_code': r['code'], 'lang_name': r['value'] , 'sup_status':r['type'], 
                        'rev_time': r['rev_time'], 'rev_rtf': r['rev_rtf'], 'system_time': r['system_time'], 'system_rtf': r['system_rtf'],
                         'sup_ratio':r['sup_ratio'], 'audiotracks':atd.get(r['code'], []) } for r in res ]
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Language list available.'
            Dresp['media_lang'] = media_lang
            Dresp['subs_locked'] = locked
            Dresp['langs'] = langs
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

 
    #------------------------------------------#
    #----------------- /get ------------------#
    #------------------------------------------#

    elif env['PATH_INFO'] == '/get': # Get dfxp file

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            lang = body["lang"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        try:
            policy = int(body["sel_data_policy"])
        except:
            policy = 2
        try:
            seg_filt_pol = int(body["seg_filt_policy"])
        except:
            seg_filt_pol = 1
        try:
            form = int(body["format"])
        except:
            form = 0


        session_id = body.get("session_id", None)
        if session_id != None:
            try:
                session_id = int(session_id)
            except:
                return error_400(start_response, "'session_id' must be a integer")

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/subs] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/subs] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/subs] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_dfxp_getDFXP(lec_hash, lang) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/subs] tL.tLint_dfxp_getDFXP() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            status = '200 OK'
            data = json.dumps({'rcode':1,'rcode_description':"ID '%s' does not exist or captions for language '%s' not available"%(extid,lang)}, indent=4)
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        elif WSSdata.results == 1:
            dfxp_fn = WSSdata.data[0]['fp']
        else:
            report_error(CFG, env, "[TLP][WS][/subs] tL.tLint_dfxp_getDFXP() returned more that 1 row.")
            return error_500(start_response)
        pdfxp = parseDfxp(dfxp_fn)

        if session_id != None:
            WSSdata = tL.tLint_get_mod_session(session_id)
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/subs] tL.tLint_get_mod_revision_mods() returned %d" % WSSdata.scode)
                return error_500(start_response)
            if WSSdata.results == 0:
                status = '200 OK'
                data = json.dumps({'rcode':2,'rcode_description':"Provided session ID %d does not exist" % (session_id)}, indent=4)
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            sess_data = WSSdata.data[0]

            if sess_data['object_id'] != lec_hash:
                Dresp = {}
                Dresp['rcode'] = 3
                Dresp['rcode_description'] = 'Provided session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            if sess_data['mods'] != None: 

                mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8") 

                if lang in mods.keys():

                    mod = mods[lang]

                    try:
                        res = updateDfxp(pdfxp, mod, sess_data['author_id'], sess_data['author_conf'], "human", sess_data['last_update'], highlight=True)
                    except:
                        report_error(CFG, env, "[TLP][WS][/subs] Exception on updateDfxp()", extra_data=pdfxp.toXml())
                        return error_500(start_response)
 
        data = convert_dfxp(pdfxp, form, policy, seg_filt_pol)
        status = '200 OK'
        if form in [0,1]:
            response_headers = [
                ('Content-Type','application/ttml+xml;'),
                ('Content-Disposition','inline; filename="%s.%s.dfxp"' % (extid.encode('utf-8'), lang.encode('utf-8'))),
                ('Content-Length', str(len(data)))]
        elif form == 2:
            response_headers = [
                ('Content-Type','application/x-subrip;'),
                ('Content-Disposition','inline; filename="%s.%s.srt"' % (extid.encode('utf-8'), lang.encode('utf-8'))),
                ('Content-Length', str(len(data)))]
        elif form == 3:
            response_headers = [
                ('Content-Type','text/vtt;'),
                ('Content-Disposition','inline; filename="%s.%s.vtt"' % (extid.encode('utf-8'), lang.encode('utf-8'))),
                ('Content-Length', str(len(data)))]
        elif form == 4:
            response_headers = [
                ('Content-Type','text/plain;'),
                ('Content-Disposition','inline; filename="%s.%s.txt"' % (extid.encode('utf-8'), lang.encode('utf-8'))),
                ('Content-Length', str(len(data)))]
        elif form == 5:
            response_headers = [
                ('Content-Type','application/json;'),
                ('Content-Disposition','inline; filename="%s.%s.json"' % (extid.encode('utf-8'), lang.encode('utf-8'))),
                ('Content-Length', str(len(data)))]

        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
 
    #------------------------------------------------#
    #----------------- /audiotrack ------------------#
    #------------------------------------------------#

    elif env['PATH_INFO'] == '/audiotrack': # Get dfxp file

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            lang = body["lang"]
            aid = body["aid"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        try:
            aid = int(aid)
        except:
            return error_400(start_response, "'aid' parameter must be integer" )

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/audiotrack] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/audiotrack] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/audiotrack] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_audiotrack_getAudiotrack(extid, lec_hash, lang, aid) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/audiotrack] tL.tLint_audiotrack_getAudiotrack() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            status = '200 OK'
            data = json.dumps({'rcode':1,'rcode_description':"Audiotrack ID '%d' or Media ID '%s' does not exist"%(aid, extid)}, indent=4)
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        elif WSSdata.results == 1:
            at_fp = WSSdata.data[0]['fp']
        else:
            report_error(CFG, env, "[TLP][WS][/audiotrack] tL.tLint_audiotrack_getAudiotrack() returned more than one row")
            return error_500(start_response)
        fd = open(at_fp, 'rb')
        data = fd.read()
        fd.close()
        response_headers = [
          ('Content-Type','application/octet-stream'),
          ('Content-Length', str(len(data))),
          ('Content-Disposition', 'attachment; filename=%s' % (os.path.basename(at_fp)))]
        status = '200 OK'
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #--------------------------------------------#
    #---------- /start_session --------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/start_session': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
  
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        sess_type = body.get("type", "player")
   
        if sess_type not in [ "player", "transcriber" ]:
            return error_400(start_response, "'type' parameter value must be one of these: ['player', 'transcriber'].")
 
        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        author_name = body.get("author_name", None)

        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_langs_getLangs(extid, lec_hash)

        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Media ID has no subtitles.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        subs_locked = WSSdata.data[0]['locked']
        if subs_locked and author_conf < 100:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Subtitles for this Media ID have been locked. Only authors are allowed to modify them.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
   
        WSSdata = tL.tLint_check_open_mod_sessions(lec_hash)

        if WSSdata.results > 0:

            sess_data = WSSdata.data[0]
            Dresp = { 'session_id':sess_data['session_id'], 'session_type':sess_data['session_type'], 'author_id':sess_data['author_id'], 'author_conf':sess_data['author_conf'], 
                      'author_name':sess_data['author_name'], 'author_type':sess_data['author_type'], 'started_at':str(sess_data['started_at']), 'last_update':str(sess_data['last_update']) }

            if user_id == sess_data['user_id'] and author_id == sess_data['author_id'] and author_conf == sess_data['author_conf'] and sess_type == sess_data['session_type']: # same user that started mod session before

                WSSdata = tL.tLint_restart_session(sess_data['session_id'])

                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session restarted.'
            else:
                if user_id != sess_data['user_id'] or author_id != sess_data['author_id'] or author_conf != sess_data['author_conf']:
                    Dresp['rcode'] = 4
                    Dresp['rcode_description'] = 'Cannot start session: there exists an open session for the given media ID.'
                elif sess_type != sess_data['session_type']:
                    Dresp['rcode'] = 5
                    Dresp['rcode_description'] = 'Cannot start session: there exists an open session on "%s"' % sess_data['session_type']
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        WSSdata = tL.tLint_check_previous_unrevised_session(lec_hash, author_id)

        if WSSdata.results > 0:
            sess_data = WSSdata.data[0]
            previous_session_id = sess_data['session_id']
            previous_mods = sess_data['mods']
            previous_mods_stats = sess_data['mods_stats']
        else:
            previous_session_id = None
            previous_mods = None
            previous_mods_stats = None

        WSSdata = tL.tLint_start_session(lec_hash, user_id, sess_type, author_id, author_conf, author_name=author_name, 
                                             prev_session_id=previous_session_id, prev_mods=previous_mods, 
                                             prev_mods_stats=previous_mods_stats)

        sess_data = WSSdata.data[0]
        Dresp = { 'session_id':sess_data['session_id'], 'session_type':sess_data['session_type'], 'author_id':sess_data['author_id'], 'author_conf':sess_data['author_conf'], 
                  'author_name':sess_data['author_name'], 'started_at':str(sess_data['started_at']), 'last_update':str(sess_data['last_update']) }

        if previous_session_id != None:

            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session started, initialized with editions from a previous unrevised session.'
        else:
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session started.'

        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
   
    #--------------------------------------------#
    #------------------ /session_status --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/session_status': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            session_id = body["session_id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        alive = body.get("alive", 0)
        try:
            alive = bool(int(alive))
        except:
            return error_400(start_response, "'alive' parameter must be an integer value [0,1]." )
   
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['object_id'] != lec_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ] (%s != %s).' % (session_id, extid, sess_data['object_id'], lec_hash)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if author_id != sess_data['author_id'] or author_conf != sess_data['author_conf'] or user_id != sess_data['user_id']:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'You are not allowed to check status of this session ID.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        Dresp = OrderedDict()
        Dresp['rcode'] = None
        Dresp['rcode_description'] = None
        Dresp['type'] = sess_data['session_type']
        Dresp['started_at'] = str(sess_data['started_at'])
        Dresp['last_update'] = str(sess_data['last_update'])
        Dresp['last_alive'] = str(sess_data['last_alive'])
        Dresp['ended_at'] = str(sess_data['ended_at'])
        Dresp['ended_by_id'] = sess_data['ended_by_id']
        Dresp['ended_by_name'] = sess_data['ended_by_name']
        Dresp['ended_by_type'] = sess_data['ended_by_type']

        if sess_data['ended_at'] != None:
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Session ID %d is closed.' % sess_data['session_id']
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        now = datetime.now()
        if alive:
            WSSdata = tL.tLint_update_last_session_alive(session_id, now) 
        else:
            WSSdata = tL.tLint_update_last_session_update(session_id, now) 

        Dresp['last_update'] = str(now)
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Session alive.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


   
    #--------------------------------------------#
    #------------------ /mod --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/mod': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            session_id = body["session_id"]
            mods = body["mods"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")


        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        if not(encoded_params):
            try:
                mods = json.loads(base64.b64decode(mods), encoding="UTF-8")
            except:
                error_400(start_response, "Could not decode 'mod' GET parameter (should be base64-encoded).")
   
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        session_alive = True
        if sess_data['ended_at'] != None:
            session_alive = False
 
        if sess_data['object_id'] != lec_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if author_id != sess_data['author_id'] or author_conf != sess_data['author_conf'] or user_id != sess_data['user_id']:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'You are not allowed to send modifications under this session ID.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
       
        if len(mods.keys()) == 0: 
            Dresp = {}
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Empty mods dict.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]


        WSSdata = tL.tLint_langs_getLangs(extid, lec_hash)
        langlist = [ x['code'] for x in WSSdata.data ]

        errors = False
        rinfo = []
        for language in mods.keys():
            if language not in langlist:
                mods.pop(language, None)
                rinfo.append({"language":language, "rcode":10, "rcode_description":"The given media ID does not have '%s' subtitles." % language })
                errors = True
                continue
            mod = mods[language]
            ret = check_mod_format(mod)
            if ret[0] != 0:
                mods.pop(language, None)
                rinfo.append({"language":language, "rcode":ret[0], "rcode_description":ret[1] })
                errors = True
                continue
            #if sess_data['session_type'] == 'transcriber':
            #    if len(mod['txt']) != 2:
            #        mods.pop(language, None)
            #        rinfo.append({"language":language, "rcode":20, "rcode_description":"Under 'transcriber' mode requires two segments (given: %d)" % len(mods['txt']) })
            #        errors = True
            #        continue
            rinfo.append({"language":language, "rcode":0, "rcode_description":"Changes successfully saved." })
                
        if len(mods.keys()) == 0: 
            Dresp = {}
            Dresp['rcode'] = 5
            Dresp['rcode_description'] = 'Format errors where found in all subtitle language modifications.'
            Dresp['details'] = rinfo
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

    
        if sess_data['mods'] == None:
            sess_mods = {}
        else:
            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8") 

        media_length = tL.get_object_length(lec_hash)

        mods_stats = compute_mods_stats(mods, media_length)

        new_session_mods = merge_mods(mods, sess_mods)
        
        new_session_mods_stats = compute_mods_stats(new_session_mods, media_length)
 
        b64_mods = base64.b64encode(json.dumps(mods))
        b64_mods_stats = base64.b64encode(json.dumps(mods_stats))
        b64_session_mods = base64.b64encode(json.dumps(new_session_mods))
        b64_session_mods_stats = base64.b64encode(json.dumps(new_session_mods_stats))

        WSSdata = tL.tLint_store_mod(session_id, language, b64_mods, b64_mods_stats, b64_session_mods, b64_session_mods_stats, session_alive) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_store_mod() returned %d" % WSSdata.scode)
            return error_500(start_response)

        if not(session_alive):
            Dresp = {}
            if errors:
                Dresp['rcode'] = 6
                Dresp['rcode_description'] = 'Session ID [ %d ] is closed, changes have been backuped. Also some errors were found (see \'details\' key). Please contact system administrator.' % session_id
                Dresp['details'] = rinfo
            else:
                Dresp['rcode'] = 7
                Dresp['rcode_description'] = 'Session ID [ %d ] is closed, changes have been backuped. Please contact system administrator.' % session_id
            
        else:
            Dresp = {}
            if errors:
                Dresp['rcode'] = 8
                Dresp['rcode_description'] = 'Modifications for some subtitle languages were successfully saved, but other failed. Please see \'details\' key.'
                Dresp['details'] = rinfo
            else:
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Changes successfully saved.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

    #--------------------------------------------#
    #--------- /end_session -----------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/end_session': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            session_id = body["session_id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        author_name = body.get("author_name", None)

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        regenerate = body.get("regenerate", 1)
        try:
            regenerate = bool(int(regenerate))
        except:
            return error_400(start_response, "'regenerate' parameter must be an integer value [0,1]." )
 
        force = body.get("force", 0)
        try:
            force = bool(int(force))
        except:
            return error_400(start_response, "force parameter must be [0,1]." )
  
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['ended_at'] != None:
            Dresp = {}
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session ID [ %d ] already closed (closed at %s).' % (session_id, str(sess_data['ended_at']))
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if sess_data['object_id'] != lec_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
  
        if sess_data['author_type'] == 'computer':
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Session ID [ %d ] is owned by the Ingest Service and it cannot be manually ended.' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if user_id != sess_data['user_id']: # Different API user
            if not(force):
                Dresp = {}
                Dresp['rcode'] = 4
                Dresp['rcode_description'] = 'You are not allowed to close this session.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
            elif not(is_admin):
                WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash, other_perms=False)
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
                    return error_500(start_response)
                if WSSdata.results == 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] Unexpected result of tL.tLint_auth_is_owner_object()")
                    return error_500(start_response)
                if WSSdata.data == False:
                    Dresp = {}
                    Dresp['rcode'] = 4
                    Dresp['rcode_description'] = 'You are not allowed to close this session.'
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data]
        else: # Same API user
            if (author_id != sess_data['author_id'] or author_conf != sess_data['author_conf']) and (author_conf < 100 or not(force)):
                Dresp = {}
                Dresp['rcode'] = 4
                Dresp['rcode_description'] = 'You are not allowed to close this session.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

        if sess_data['author_conf'] < 100:
            requires_revision = (sess_data['mods'] != None)
            revised = False
            requires_postprocess = False
        elif sess_data['author_conf'] == 100:
            requires_revision = False
            revised = None
            requires_postprocess = (sess_data['mods'] != None)

        if not(requires_revision) and sess_data['mods'] != None: # store changes in DFXP file

            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8")

            for lang in sess_mods.keys():

                mod = sess_mods[lang]

                WSSdata = tL.tLint_dfxp_getDFXP(lec_hash, lang) 
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_dfxp_getDFXP() returned %d" % WSSdata.scode)
                    return error_500(start_response)
                if WSSdata.results == 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_dfxp_getDFXP() returned nothing" % WSSdata.scode)
                    return error_500(start_response)
                elif WSSdata.results == 1:
                    dfxp_fn = WSSdata.data[0]['fp']
                else:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_dfxp_getDFXP() returned more that 1 row.")
                    return error_500(start_response)
                pdfxp = parseDfxp(dfxp_fn)
                #if sess_data['session_type'] == 'player':
                res = updateDfxp(pdfxp, mod, sess_data['author_id'], sess_data['author_conf'], "human", datetime.now())
                #elif sess_data['session_type'] == 'transcriber':
                #    sup_part = mod['txt'][0]
                #    unsup_part = mod['txt'][1]
                #    if unsup_part['t'].strip() == "":
                #        unsup_part = None
                #    res = overwriteDfxp(pdfxp, sup_part, unsup_part, sess_data['author_id'], sess_data['author_conf'], datetime.now())
                #else:
                #    report_error(CFG, env, "[TLP][WS][/end_session] unexpected session_type value: '%s'" % sess_data['session_type'])
                #    return error_500(start_response)
                if res != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] updateDfxp() returned %s" % res, extra_data=pdfxp.toXml())
                    return error_500(start_response)
                data = pdfxp.toXml()
                WSSdata = tL.tLint_mod_storeDFXP(lec_hash, lang, data, pdfxp.status)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_mod_storeDFXP() returned %d" % WSSdata.scode)
                    return error_500(start_response)
    
        WSSdata = tL.tLint_end_session(session_id, author_id, requires_revision, requires_postprocess, ended_by_name=author_name, do_regenerate=regenerate) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_end_session() returned %d" % WSSdata.scode)
            return error_500(start_response)

        Dresp = {}
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Session succesfully closed.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #--------------------------------------------#
    #------------------ /lock_subs --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/lock': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            lock = body["lock"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            lock = bool(int(lock))
        except:
            return error_400(start_response, "lock parameter must be [0,1].")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order to lock/unlock subtitles.")

        if not(encoded_params):
            try:
                mods = json.loads(base64.b64decode(mods), encoding="UTF-8")
            except:
                error_400(start_response, "Could not decode 'mod' GET parameter (should be base64-encoded).")
   
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        tL.tLint_lock_subtitles(lec_hash, lock)

        Dresp = {}
        Dresp['rcode'] = 0
        if lock:
            Dresp['rcode_description'] = 'Subtitles successfully locked.'
        else:
            Dresp['rcode_description'] = 'Subtitles successfully unlocked.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]





    #-----------------------------------------------#
    #----------------- /revisions ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/revisions':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_get_all_pending_revisions(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_get_pending_user_mod_revisions() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        else:
            Lresp = []
            for r in WSSdata.data:
                Dresp = OrderedDict()
                Dresp['media_id'] = r['media_id']
                Dresp['session_id'] = r['session_id']
                Dresp['session_type'] = r['session_type']
                Dresp['author_id'] = r['author_id']
                Dresp['author_conf'] = r['author_conf']
                Dresp['author_name'] = r['author_name']
                Dresp['timestamp'] = str(r['timestamp']) if r['timestamp'] != None else None
                Dresp['edit_stats'] = json.loads(base64.b64decode(r['edit_stats']), encoding="UTF-8") if r['edit_stats'] != None else None
                Lresp.append(Dresp)
            data = json.dumps(Lresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

    #------------------------------------------------#
    #------------------- /mark_revised --------------#
    #------------------------------------------------#
  
    elif env['PATH_INFO'] == '/mark_revised': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
  
        try:
            extid = body["id"]
            session_id = body["session_id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order to mark edition as revised.")

        author_name = body.get("author_name", None)

        rev_session_id = body.get("revision_session_id", None)
        if rev_session_id != None:
            try:
                rev_session_id = int(rev_session_id)
            except:
                return error_400(start_response, "revision_session_id parameter must be an integer value.")
 
        if session_id == rev_session_id:   
            return error_400(start_response, "session_id and revision_session_id parameters cannot be equal.")
  
        unmark = body.get("unmark", 0)
        try:
            unmark = bool(int(unmark))
        except:
            return error_400(start_response, "'unmark' parameter must be an integer value [0,1]." )
         
 
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['ended_at'] == None:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Session ID [ %d ] has not been closed yet.' % (session_id)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if sess_data['object_id'] != lec_hash:
            Dresp = {}
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if not(sess_data['requires_revision']):
            Dresp = {}
            Dresp['rcode'] = 5
            Dresp['rcode_description'] = 'Session ID [ %d ] did not require revision.' % (session_id)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if unmark:

            if not(sess_data['revised']):
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session ID [ %d ] has not been revised yet.' % (session_id)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            WSSdata = tL.tLint_unmark_revised(session_id)
    
            if WSSdata.scode == 0:  # OK
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Revised mark successfully removed.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
 
        else:

            if sess_data['revised']:
                Dresp = {}
                Dresp['rcode'] = 6
                Dresp['rcode_description'] = 'Session ID [ %d ] was already revised before.' % (session_id)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            if rev_session_id != None:
                WSSdata = tL.tLint_get_mod_session(rev_session_id)

                if WSSdata.results == 0:
                    Dresp = {}
                    Dresp['rcode'] = 7
                    Dresp['rcode_description'] = 'The provided "revision_session_id" %d does not exist.' % (rev_session_id)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data]

                sess_data = WSSdata.data[0]

                if sess_data['object_id'] != lec_hash:
                    Dresp = {}
                    Dresp['rcode'] = 8
                    Dresp['rcode_description'] = 'The provided "revision_session_id" %d does not involve Media ID %s.' % (rev_session_id, extid)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data] 
 
                if sess_data['author_id'] != author_id or sess_data['user_id'] != user_id or sess_data['author_conf'] != author_conf:
                    Dresp = {}
                    Dresp['rcode'] = 9
                    Dresp['rcode_description'] = "The provided 'revision_session_id' %d is not owned by the author ID that is requesting this session ID to be marked as revised." % (rev_session_id)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data] 

   
            WSSdata = tL.tLint_mark_revised(session_id, author_id, author_name=author_name, rev_session_id=rev_session_id)
    
            if WSSdata.scode == 0:  # OK
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session ID marked as revised.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
   

    #-----------------------------------------------#
    #----------------- /edit_history ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/edit_history':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")
 
        su = body.get("su", None)
        lec_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if lec_hash == None:
            lec_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, lec_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Media ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_get_all_media_editions(lec_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_get_mod_revisions() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        else:
            Dresp = OrderedDict()
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Edit history available.'
            Dresp['edit_history'] = []
            for r in WSSdata.data:
                d = OrderedDict()
                d['session_id'] = r['session_id']
                d['session_type'] = r['session_type']
                d['author_id'] = r['author_id']
                d['author_conf'] = r['author_conf']
                d['author_name'] = r['author_name']
                d['timestamp'] = str(r['timestamp']) if r['timestamp'] != None else None
                d['requires_revision'] = r['requires_revision']
                d['revised'] = r['revised']
                d['revised_at'] = str(r['revised_at']) if r['revised_at'] != None else None
                d['revised_by_id'] = r['revised_by_id']
                d['revised_by_name'] = r['revised_by_name']
                d['revised_via'] = r['revised_via']
                d['revised_in_session_id'] = r['revised_in_session_id']
                d['edit_stats'] = json.loads(base64.b64decode(r['edit_stats']), encoding="UTF-8") if r['edit_stats'] != None else None
                Dresp['edit_history'].append(d)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
    #-----------------------------------------------#
    #----------------- /reject ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/reject':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            id_param = str(body["id"])
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order accept/decline changes.")

        author_name = body.get("author_name", None)
 
        try:
            sess_ids = map(lambda x: int(x), id_param.split(","))
        except:
            return error_400(start_response, '"id" parameter must be a string with comma-separated Session IDs (integers).')
 
        if len(sess_ids) > 100:
            return error_400(start_response, 'Exceeded limit of Session IDs to be accepted (max 100).')

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, id_param, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

 
        res = []

        for session_id in sess_ids: 
            WSSdata = tL.tLint_get_mod_session(session_id) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
                return error_500(start_response)
            
            if WSSdata.results == 0:
                res.append({'session_id':session_id, 'rcode':1, 'rcode_description':'Session ID does not exist.'})
                continue
 
            sess_data = WSSdata.data[0]
    
            if user_id != sess_data['user_id']:
                res.append({'session_id':session_id, 'rcode':2, 'rcode_description':'You are not allowed to modify the status of this Session ID.'})
                continue
   
            if sess_data['ended_at'] == None:
                res.append({'session_id':session_id, 'rcode':3, 'rcode_description':'Session ID has not been ended yet.'})
                continue
            if not(sess_data['requires_revision']):
                res.append({'session_id':session_id, 'rcode':4, 'rcode_description':'Session ID did not require revision.'})
                continue

            if sess_data['revised']:
                res.append({'session_id':session_id, 'rcode':5, 'rcode_description':'Session ID already revised.'})
                continue
 
            WSSdata = tL.tLint_reject(session_id, author_id, author_name)

            res.append({'session_id':session_id, 'rcode':0, 'rcode_description':'Changes from Session ID successfully rejected.'})
 
        data = json.dumps(res, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #-----------------------------------------------#
    #----------------- /accept ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/accept':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            id_param = str(body["id"])
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order accept/decline changes.")

        author_name = body.get("author_name", None)
 
        try:
            sess_ids = map(lambda x: int(x), id_param.split(","))
        except:
            return error_400(start_response, '"id" parameter must be a string with comma-separated Session IDs (integers).')

        if len(sess_ids) > 100:
            return error_400(start_response, 'Exceeded limit of Session IDs to be accepted (max 100).')
 
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, id_param, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

 
        res = []

        for session_id in sess_ids: 
            WSSdata = tL.tLint_get_mod_session(session_id) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
                return error_500(start_response)
            
            if WSSdata.results == 0:
                res.append({'session_id':session_id, 'rcode':1, 'rcode_description':'Session ID does not exist.'})
                continue
 
            sess_data = WSSdata.data[0]
    
            if user_id != sess_data['user_id']:
                res.append({'session_id':session_id, 'rcode':2, 'rcode_description':'You are not allowed to modify the status of this Session ID.'})
                continue
   
            if sess_data['ended_at'] == None:
                res.append({'session_id':session_id, 'rcode':3, 'rcode_description':'Session ID has not been ended yet.'})
                continue
            if not(sess_data['requires_revision']):
                res.append({'session_id':session_id, 'rcode':4, 'rcode_description':'Session ID did not require revision.'})
                continue

            if sess_data['revised']:
                res.append({'session_id':session_id, 'rcode':5, 'rcode_description':'Session ID already revised.'})
                continue

            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8")

            for lang in sess_mods.keys():

                mod = sess_mods[lang]
                lec_hash = sess_data['object_id']

                try:
                    WSSdata = tL.tLint_dfxp_getDFXP(lec_hash, lang) 
                except:
                    report_error(CFG, env, "[TLP][WS][/accept] Exception on tL.tLint_dfxp_getDFXP()")
                    continue
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/accept] tL.tLint_dfxp_getDFXP() returned %d" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                if WSSdata.results == 0:
                    report_error(CFG, env, "[TLP][WS][/accept] tL.tLint_dfxp_getDFXP() returned nothing" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                elif WSSdata.results == 1:
                    try:
                        dfxp_fn = WSSdata.data[0]['fp']
                    except:
                        report_error(CFG, env, "[TLP][WS][/accept] Exception when getting DFXP file path")
                        res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                        continue
                else:
                    report_error(CFG, env, "[TLP][WS][/accept] tL.tLint_dfxp_getDFXP() returned more that 1 row.")
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                try:
                    pdfxp = parseDfxp(dfxp_fn)
                except: 
                    report_error(CFG, env, "[TLP][WS][/accept] Exception on parseDfxp().")
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                try:
                    ret = updateDfxp(pdfxp, mod, sess_data['author_id'], sess_data['author_conf'], "human", datetime.now())
                except:
                    report_error(CFG, env, "[TLP][WS][/accept] Exception on updateDfxp()", extra_data=pdfxp.toXml())
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
        
                if ret != 0:
                    report_error(CFG, env, "[TLP][WS][/accept] updateDfxp() returned %s" % res, extra_data=pdfxp.toXml())
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                try:
                    data = pdfxp.toXml()
                except:
                    report_error(CFG, env, "[TLP][WS][/accept] Exception on pdfxp.toXml()", extra_data="File path: %s" % dfxp_fn)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                try:
                    WSSdata = tL.tLint_mod_storeDFXP(lec_hash, lang, data, pdfxp.status)
                except:
                    report_error(CFG, env, "[TLP][WS][/accept] Exception on tL.tLint_mod_storeDFXP()", extra_data=data)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/accept] tL.tLint_mod_storeDFXP() returned %d" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into subtitles file.'})
                    continue
 
            WSSdata = tL.tLint_accept(session_id, author_id, author_name)

            res.append({'session_id':session_id, 'rcode':0, 'rcode_description':'Changes from Session ID successfully accepted.'})
 
        data = json.dumps(res, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
 

 
    else: # query not known, bad request
        status = '400 Bad Request'
        data = CFG['MISC']['HTML_MSG']
        response_headers = [('Content-Type', 'text/html'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
    

def application(env, start_response):
    try:
        tL = tLWSSLayer(CFG['STG']['DB_NAME'], CFG['STG']['DB_USER'], CFG['STG']['DB_HOST'], tLWSSLayer.OBJ_SPEECH, passwd=CFG['STG']['DB_PASS'])
        tlpauth = TLPAuth(CFG['AUT']['KEY'])
        return my_application(env, start_response, tL, tlpauth)
    except:
        report_error(CFG, env, "[TLP][WS] Exception")
        return error_500(start_response)


#############################################################
#----------------- Error return functions ------------------#
#############################################################

def error_400(start_response, text=None):
    status = '400 Bad Request'
    response_headers = [('Content-Type', 'text/plain')]
    if text == None:
        data = "Bad Request"
    else:
        data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_401(start_response, text):
    status = '401 Unauthorized'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_419(start_response, text):
    status = '419 Authentication Timeout'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_500(start_response):
    status = '500 Internal Error'
    data = "Internal Server Error"
    response_headers = [('Content-Type', 'text/plain'),('Content-Length', str(len(data)))]
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]

ENV_KEYS = [ 'CONTENT_TYPE', 'SCRIPT_NAME', 'REQUEST_METHOD', 'PATH_INFO', 'REQUEST_URI', 'QUERY_STRING', 'wsgi.url_scheme', 'SERVER_NAME', 'SERVER_PORT', 'SCRIPT_FILENAME', 'HTTP_HOST', 'REMOTE_ADDR', 'REMOTE_PORT' ]

def report_error(CFG, env, subject, extra_data=None):
    if sys.exc_type != None:
        excpt = True
        excpt_str = traceback.format_exc()
        sys.stderr.write("\n************ API EXCEPTION ************\n\n")
        sys.stderr.write("%s\n" % excpt_str)
        #traceback.print_exc()
    else:
        excpt = False

    if CFG['MAI']['ENABLED'] and CFG['MAI']['TO'] != None:
        try:
            out = cStringIO.StringIO()
            out.write("DATE: %s\n" % datetime.now().strftime("%d/%m/%Y %H:%M:%S %Z"))
            out.write("EXCEPTION: %s\n" % str(excpt))
            if excpt:
                out.write("\n+++++++++++++++++++++++++++++++++ EXCEPTION ++++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % excpt_str)
            out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
            print_log(out, env)
            print_env(out, env, space=1)
            if env['PATH_INFO'] == '/ingest': 
              print_input_get_params(out, env, space=1)
            else:
              if env['REQUEST_METHOD'] == 'GET':
                  print_input_get_params(out, env, space=1)
              elif env['REQUEST_METHOD'] == 'POST':
                  print_icont(out, env)

            if extra_data != None:
                out.write("\n++++++++++++++++++++++++++++++++ EXTRA DATA +++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % extra_data)
                    
            content = out.getvalue() 
            send_mail(CFG['MAI']['SMTP_SERVER'], CFG['MAI']['FROM'], subject, content.decode('utf-8'), CFG['MAI']['TO'])
        except:
            sys.stderr.write("\n************ SEND_MAIL EXCEPTION ************\n")
            traceback.print_exc()
    
def send_mail(smtp_server, src_mail, subject, content, to, cc=None, cco=None):
    # Create e-mail data
    msg = MIMEText(content.encode('UTF-8'))
    msg['Subject'] = subject.encode('UTF-8')
    msg['From'] = src_mail
    msg['To'] = ",".join(to)
    addrl = []
    addrl = addrl + to
    if cc != None: 
        msg['cc'] = ",".join(cc)
        addrl = addrl + cc
    if cco != None: 
        msg['cco'] = ",".join(cco)
        addrl = addrl + cco
    # Send e-mail
    smtp = smtplib.SMTP(smtp_server)
    smtp.sendmail(src_mail, addrl, msg.as_string())







#####################################################
#----------------- Misc functions ------------------#
#####################################################


def check_manifest(manif, operation):
    if 'metadata' not in manif.keys():
        return (False, None, (400, "Missing 'metadata' key in Manifest JSON"))
    if 'external_id' not in manif['metadata']:
        return (False, None, (400, "Missing 'external_id' key in Manifest JSON"))
    requires_mpkg = False
    if operation == 0: # New
        if 'media' not in  manif:
            return (False, None, (400, "Missing 'media' key in Manifest JSON"))
        if 'url' not in manif['media']:
            requires_mpkg = True
        att = manif.get('attachments', [])
        if not(isinstance(att, list)):
            return (False, None, (400, "'attachments' key must be an array of dictionaries"))
        if len(att) > 0:
            requires_mpkg = True
        if 'title' not in manif['metadata']:
            return (False, None, (400, "Missing 'title' key in Manifest JSON"))
        if 'language' not in manif['metadata']:
            return (False, None, (400, "Missing 'language' key in Manifest JSON"))
        #if 'speakers' not in manif['metadata']:
        #    return (False, None, (400, "Missing 'speakers' key in Manifest JSON"))
    elif operation == 1: # Update
        att = manif.get('attachments', [])
        if 'media' in manif:
            if 'url' not in manif['media']:
                requires_mpkg = True
        if len(att) > 0: 
            requires_mpkg = True
    return (True, requires_mpkg, None)

 
def compute_mod_stats(mod, media_duration):
    stats = {}
    stats['del_segs'] = len(mod.get('del',[]))
    stats['edit_segs'] = len(mod.get('txt', []))
    stats['edit_time'] = sum(map(lambda x: x['e']-x['b'], mod['txt']))
    stats['edit_time_percent'] = stats['edit_time']/float(media_duration)*100
    stats['rev_time'] = mod.get('rev_time', None)
    if stats['rev_time'] != None:
        stats['rev_rtf'] = mod['rev_time']/float(media_duration)
    else:
        stats['rev_rtf'] = None
    return stats
    

def compute_mods_stats(mods, media_duration):
    stats = {}
    for l in mods.keys():
        lmod = mods[l]
        stats[l] = compute_mod_stats(lmod, media_duration)
    return stats
         

def check_mod_format(moddata):

    if 'txt' not in moddata:
        return 1, "'txt' key not found in 'mod'."

    mod = moddata['txt']
    delsegl = moddata.get('del', [])
    revt = moddata.get('rev_time', 0)
    try:
        revt = int(revt)
    except:
        return 10, "'rev_time' must be an integer (seconds)"

    if not(isinstance(mod, list)) or not(isinstance(delsegl, list)):
        return 2, "'txt' and 'del' keys must be a list"

    if len(mod) == 0 and len(delsegl) == 0:
        return 3, "'txt' and 'del' keys are empty lists"

    if len(mod) > 0:
        # Format and safety checks
        mod_sId_lst = []
        for cm in mod:
            if ('sI' not in cm) or ('b' not in cm) or ('e' not in cm) or ('t' not in cm): 
                return 4, "'txt' elements must include the following keys: 'sI', 'b', 'e', 't'"
            try:
                cm['sI'] = int(cm['sI'])
            except:
                return 5, "Invalid format of 'txt' elements. Must be: 'sI':<int>"
            try:
                cm['b'] = float(cm['b'])
            except:
                return 5, "Invalid format of 'txt' elements. Must be: 'b':<float/int>"
            try:
                cm['e'] = float(cm['e'])
            except:
                return 5, "Invalid format of 'txt' elements. Must be: 'e':<float/int>"
            if cm['e'] - cm['b'] < 0.01: 
                return 6, "Segment end 'e' must be greater than segment begin 'b' (on segment ID %d)" % cm['sI']
            if cm['sI'] in mod_sId_lst: 
                return 7, "Segment ID %d provided twice" % cm['sI']
            mod_sId_lst.append(cm['sI'])
    
        slmod = sorted(mod, key=lambda x: x['b'], reverse=True)
        if slmod[-1]['b'] < 0: 
            return 8, "First Segment (ID %d) begin 'b' must be greater than 0" % cm['sI']
    
    if len(delsegl) > 0:    
        for i in xrange(len(delsegl)):
            try:
                delsegl[i] = int(delsegl[i])
            except: 
                return 9, "'del' elements must be integers"

    return 0, ""

def segment_overlapping(rb, re, hb, he):
  return (hb < rb and he > rb + 0.1) or (hb >= rb and he <= re) or (hb < re - 0.1 and he > re)

def merge_mods(mods, session_mods):
    for language in mods.keys():

        mod = mods[language]
        new_txt = mod.get('txt', [])
        new_del = list(set(mod.get('del', [])))
        new_rev_time = mod.get('rev_time', None)
        if new_rev_time != None:
            new_rev_time = int(new_rev_time)
    
        if language in session_mods:
    
            lang_mod = session_mods[language]
            lang_txt = lang_mod['txt']
            lang_del = lang_mod['del']
            lang_rev_time = lang_mod['rev_time']
    
            # Process mod's rev_time key
            if lang_rev_time != None and new_rev_time != None:
                lang_mod['rev_time'] += new_rev_time
                

            # Process mod's del key
            if len(new_del) > 0:
                new_lang_txt = []
                for seg in lang_txt: 
                    if seg['sI'] not in new_del: # Delete session segments => do not append it to the new list
                        new_lang_txt.append(seg)
                lang_del = list(set(lang_del + new_del)) # Update language session del list
                lang_mod['del'] = lang_del
                lang_mod['txt'] = new_lang_txt
                lang_txt = lang_mod['txt']
    
            # Process mod's txt key
            if len(new_txt) > 0:
    
                # 1. Drop edited segment IDs
                edited_segIDs = [ seg['sI'] for seg in new_txt ]
                new_lang_txt = []
                for seg in lang_txt: 
                    if seg['sI'] not in edited_segIDs: 
                        new_lang_txt.append(seg)
                lang_mod['txt'] = new_lang_txt
                lang_txt = lang_mod['txt']
    
                # 2. DO the magic stuff
                slmod = sorted(new_txt, key=lambda x: float(x['b']), reverse=True)
                cm = slmod.pop()
                ins = False
                new_lang_txt = []
                for seg in lang_txt:
    
                    while len(slmod) > 0 and float(cm['e']) < seg['b'] + 0.1:
                        if not(ins):
                            new_lang_txt.append(cm)
                        cm = slmod.pop()
                        ins = False
            
                    if len(slmod) == 0 and float(cm['e']) < seg['b'] + 0.1 and not(ins):
                        new_lang_txt.append(cm)
                        ins = True
            
                    if segment_overlapping(float(cm['b']), float(cm['e']), seg['b'], seg['e']) and not(ins):
                        new_lang_txt.append(cm)
                        ins = True
                    else:
                        new_lang_txt.append(seg)
            
                if not(ins):
                    new_lang_txt.append(cm)
              
                while len(slmod) > 0:
                    cm = slmod.pop()
                    new_lang_txt.append(cm)
                
                lang_mod['txt'] = new_lang_txt
            
        else:
    
            session_mods[language] = { 'txt':new_txt, 'del':new_del, 'rev_time':new_rev_time }
    
    return session_mods

def check_auth_token(tL, tlpauth, user_name, given_token, oid, expire, author=None, author_conf=None, allow_request_key=True):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None, None, None
    if WSSdata.results == 0:
        return 2, None, None, None
    user_id = WSSdata.data[0]['id']
    user_pass = WSSdata.data[0]['password']
    is_admin = WSSdata.data[0]['is_admin']
    secret_key = tlpauth.get_secret_key(user_name.encode('utf-8'), user_pass)
    if secret_key == given_token:
        return 0, user_id, is_admin, False
    elif allow_request_key:
        if expire == None: 
            return 3, None, None, None
        request_key = tlpauth.get_request_key(user_name.encode('utf-8'), user_pass, oid.encode('utf-8'), expire, author, author_conf)
        if author == None:
            given_token = given_token[:tlpauth.request_subkey_length]
        if request_key == given_token:
            ts = int((datetime.utcnow() - datetime(1970,01,01)).total_seconds())
            if ts > expire:
                return 4, None, None, None
            return 0, user_id, is_admin, True
        else:
            return 3, None, None, None
    else:
        return 3, None, None, None

def get_user_id(tL, user_name):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None
    if WSSdata.results == 0:
        return 2, None
    user_id = WSSdata.data[0]['id']
    return 0, user_id

def get_hash(obj_id, user_id):
    h = hashlib.sha256()
    if isinstance(obj_id, unicode):
      obj_id = obj_id.encode('utf-8')
    if user_id != None:
        h.update("%d|%s" % (user_id, obj_id))
    else:
        h.update("%s" % obj_id)
    return h.hexdigest()


def parse_config(cfg_fd): 

    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}

    # General
    CFG['AUT'] = {}
    CFG['AUT']['KEY'] = config.get('authentication', 'key_generator')
    CFG['STG'] = {}
    CFG['STG']['DB_NAME'] = config.get('storage', 'db_name')
    CFG['STG']['DB_USER'] = config.get('storage', 'db_user')
    CFG['STG']['DB_HOST'] = config.get('storage', 'db_host')
    if CFG['STG']['DB_HOST'] != None: 
      if CFG['STG']['DB_HOST'].strip() == "": CFG['STG']['DB_HOST'] = None
    CFG['STG']['DB_PASS'] = config.get('storage', 'db_passwd')
    if CFG['STG']['DB_PASS'] != None: 
      if CFG['STG']['DB_PASS'].strip() == "": CFG['STG']['DB_PASS'] = None
    CFG['URL'] = {}
    CFG['URL']['USE'] = config.getboolean('media_urls', 'use_urls')
    CFG['URL']['BASEURL'] = config.get('media_urls', 'base_url')
    CFG['MAI'] = {}
    CFG['MAI']['ENABLED'] = config.getboolean('mailing', 'enabled')
    CFG['MAI']['FROM'] = config.get('mailing', 'from_address')
    CFG['MAI']['SMTP_SERVER'] = config.get('mailing', 'smtp_server')
    try:
        CFG['MAI']['TO'] = config.get('mailing', 'to_address').replace(' ', '').split(',')
    except:
        CFG['MAI']['TO'] = None
    CFG['MISC'] = {}
    CFG['MISC']['HTML_MSG'] = config.get('misc', 'html_msg')

    return CFG


# Load Config file
with open(CFG_FILE) as cfg_fd:
    CFG = parse_config(cfg_fd)


def get_mod_stats(mod):
    nsegs = 0
    accdur = 0
    for m in mod['txt']:
        st = float(m['b'])
        et = float(m['e'])
        dur = et-st
        accdur += dur
        nsegs += 1
    return nsegs, int(round(accdur,0))

##################################################
#----------- Logging WSGI Middelware ------------#
##################################################

import threading


def print_key_value(out, k, v, space=30):
    out.write("%s:%s%s\n" % (str(k), ' ' * max((space-len(k)),1), str(v)))

def print_width(out, data, width=80):
    ndata=''; i = 0; c = 0
    while i < len(data):
        ndata += data[i]
        if c == width: 
            ndata += '\n'
            c = 0
        else: c += 1
        i += 1
    out.write("%s\n" % ndata)

def print_log(out, env):
  out.write('PATH: %s\n' % env['SCRIPT_NAME'])
  out.write('INTERFACE: %s\n' % env['PATH_INFO'])

def print_env(out, environ, space=30):
  out.write('\n#------- INPUT ENVIRONMENT -------#\n\n')
  fenv = {}
  for k in ENV_KEYS:
    print_key_value(out, k, environ.get(k, "-"), space=space)

def print_input_get_params(out, env, space=30):    
  out.write('\n#------- INPUT DATA (GET) -------#\n\n')
  encoded_params = True
  queries = parse_qs(env['QUERY_STRING'])
  if "data" in queries:
      try:
          data = base64.b64decode(queries["data"][0])
      except: 
         out.write("Could not parse 'data' GET Parameter\n")
  else:
      encoded_params = False
  
  if encoded_params:
      try:
          body = json.loads(data, encoding="UTF-8")
      except:
         out.write("Cannot parse JSON data (Invalid JSON).\n")
      else:
         out.write("%s\n" % json.dumps(body, indent=4))
  else:
      body = { k:queries[k][0] for k in queries.keys() }
      for k in body:
           print_key_value(out, k, body.get(k, "-"), space=space)

  out.write("\n-- Get URL:\n\n")
  url = "http://%s%s" % (env['SERVER_NAME'], env['REQUEST_URI'])
  out.write("%s\n" % url)

def print_icont(out, environ):
   
  length = int(environ.get('CONTENT_LENGTH', '0'))
  
  out.write('\n#------- INPUT DATA (POST) -------#\n\n')
  if length > 0 and length < 2147483648: # 2GB
      icontent_fp = cStringIO.StringIO()
      icontent_fp.write(environ['wsgi.input'].read(length))
      if environ.get('CONTENT_TYPE', '').find('application/json') >= 0:
          icont = icontent_fp.getvalue()
          try:
              ret = str(icont)
          except:
              return
      elif environ.get('CONTENT_TYPE','').find('multipart/form-data') >= 0: 
          icontent_fp.seek(0, os.SEEK_SET)
          try:
              fs = FieldStorage(fp=icontent_fp,environ=environ)
          except:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if "json" not in fs.keys():
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if fs['json'].type.find("application/json") <  0:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          #out.write("%s\n" % icontent_fp.getvalue())
          ret = fs.getvalue("json")

      try:
          jret = json.loads(ret)
          jtret = json.dumps(jret, indent=4)
          encurl = urllib.urlencode(jret)
      except:
         out.write("\n-- Text:\n")
         out.write("%s\n" % ret)
      else:
         out.write("\n-- Text:\n\n")
         out.write("%s\n" % jtret)
         out.write("\n-- Get URL:\n\n")
         url = "http://%s%s?%s" % (environ['SERVER_NAME'], environ['REQUEST_URI'], encurl)
         out.write("%s\n" % url)

      icontent_fp.seek(0, os.SEEK_SET)
      environ['wsgi.input'] = icontent_fp

  else:
     out.write("None\n")


class LoggingInstance:
  def __init__(self, start_response):
    self.__start_response = start_response
    self.__print_content = True
    self.__content_type = ""
    self.__out = cStringIO.StringIO()

  def __call__(self, status, headers, *args):
    self.__out.write('\n#------- OUTPUT HEADERS -------#\n\n')
    print_key_value(self.__out, 'Response-Code', status)
    ct = None
    for h in headers:
        if h[0] == 'Content-Type': 
            ct = h[1]
        print_key_value(self.__out, h[0], h[1])

    self.__content_type = ct
    if ct.find('application/json') == -1 and ct.find('text/plain') == -1:
      self.__print_content = False

    self.__out.write('\n#------- OUTPUT DATA -------#\n\n')

    self.__write = self.__start_response(status, headers, *args)
    return self.write

  def __iter__(self):
    return self

  def write(self, data):
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return self.__write(data)

  def next(self):
    data = self.__iterable.next()
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return data

  def close(self):
    if hasattr(self.__iterable, 'close'):
      self.__iterable.close()
    self.__out.write('\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n\n\n')
    print self.__out.getvalue()

  def link(self, iterable):
    self.__iterable = iter(iterable)

class LoggingMiddleware:

  def __init__(self, application):
    self.__application = application
    self.__lock = threading.Lock()

  def __call__(self, environ, start_response):
    self.__lock.acquire()
    date = datetime.now()
    out = cStringIO.StringIO()
    out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
    print_log(out, environ)
    print_env(out, environ)
    if environ['PATH_INFO'] in ['/ingest/new', '/ingest/update']: 
      print_icont(out, environ)
    else:
      if environ['REQUEST_METHOD'] == 'GET':
          print_input_get_params(out, environ)
      elif environ['REQUEST_METHOD'] == 'POST':
          print_icont(out, environ)
    print out.getvalue() 
    iterable = LoggingInstance(start_response)
    iterable.link(self.__application(environ, iterable))
    self.__lock.release()
    return iterable

try:
    application = LoggingMiddleware(application)
except:
    print '\n++++++++++++++++++++++++++++++++++ LOGGING EXCEPTION!! ++++++++++++++++++++++++++++++++++++++++++\n'
    traceback.print_exc()

