# -*- coding: utf-8 -
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

####---- Standard Module imports ----####

import sys
import os
import json
import hashlib
import base64
from cgi import parse_qs, FieldStorage
from datetime import datetime, timedelta
from imp import load_source
from ConfigParser import ConfigParser
from collections import OrderedDict
import mimetypes
from HTMLParser import HTMLParser

import smtplib
from email.mime.text import MIMEText
from datetime import datetime
import traceback
from cStringIO import StringIO
import pprint
import urllib
import bs4
import re

####---- tL libraries + Config file ----####

WS_PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, WS_PATH)

from lib.libdtlx import *
from lib.libtlpauth import TLPAuth
from lib.libwsstorage import tLWSSLayer, tLWSSData
from lib.dtlx2html import regen_html_from_dtlx

CFG_FILE = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))

####---- Cache control ----####

CACHE_CONTROL="no-cache, no-store, private, must-revalidate, max-stale=0"
PRAGMA="no-cache"
EXPIRES="Tue, 01 Jan 1980 12:00:00 GMT"
CACHE_RESP_HEADERS = [('Cache-Control', CACHE_CONTROL), ('Pragma', PRAGMA), ('Expires', EXPIRES)]


#################################################
#--------------- Webservice API ----------------#
#################################################

def my_application(env, start_response, tL, tlpauth):


 
    #---------------------------------------------#
    #----------------- /systems ------------------#
    #---------------------------------------------#
  
    if env['PATH_INFO'] == '/systems':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Dresp = {}

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/systems] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/systems] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_systems_getUserSystems(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/systems] tL.tLint_systems_getUserSystems() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Dresp['mt']  = [ { 'id':s['id'], 'source_lang':s['lang1'], 'target_lang':s['lang2'], 'name':s['name'], 'description':s['description'], 'topic_adaptation':s['features_tadapt'], 'processing_time':s['const_time'] } for s in WSSdata.data if s['type'] == 'mt' ]
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #-------------------------------------------------#
    #----------------- /uploadslist ------------------#
    #-------------------------------------------------#
  
    if env['PATH_INFO'] == '/uploadslist':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        object_id = body.get("object_id", None)
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Lresp = []

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/uploadslist] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/uploadslist] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_uploadslist_getUploads(user_id, object_id)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/uploadslist] tL.tLint_uploadslists_getUploads() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Lresp = [ { 'id':s['id'], 'status_code':s['status_id'], 'object_id':s['object_id'], 'last_update':str(s['last_update']), 'uploaded_on':str(s['uploaded_on']) } for s in WSSdata.data ]
        data = json.dumps(Lresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #---------------------------------------------#
    #----------------- /list ------------------#
    #---------------------------------------------#
  
    if env['PATH_INFO'] == '/list':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/list] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/list] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.text_list_getList(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/list] tL.text_list_getList() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        Lresp = [ { 'id':s['id'], 'title':s['title'], 'language':s['lang'], 'format':s['format'], 'sentences':s['sentences'], 'words':s['words'], 'characters':s['chars'] } for s in WSSdata.data ]
        data = json.dumps(Lresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


 
    #--------------------------------------------#
    #----------------- /ingest ------------------#
    #--------------------------------------------#

    elif env['PATH_INFO'].find('/ingest') == 0: 

        opc = None
        if env['PATH_INFO'] == '/ingest/new': opc = 0
        elif env['PATH_INFO'] == '/ingest/update': opc = 1
        elif env['PATH_INFO'] == '/ingest/delete': opc = 2
        elif env['PATH_INFO'] == '/ingest/cancel': opc = 3
        else: return error_400(start_response, 'Unknown operation')

        pkg_file = None
        if opc in [0,1]:

            if env['REQUEST_METHOD'] != 'POST':
                return error_400(start_response, "This interface only supports POST requests")
             
            if env['CONTENT_TYPE'].find('application/json') >= 0:
                length = int(env.get('CONTENT_LENGTH', '0'))
                if length == 0:
                    return error_400(start_response, "Content-Length must be greater than 0")
                try:
                    data = env['wsgi.input'].read(length)
                except:
                    return error_400(start_response, "Cannot read POST data.")
    
            elif env['CONTENT_TYPE'].find('multipart/form-data') >= 0: 
                try:
                    fs = FieldStorage(fp=env['wsgi.input'],environ=env)
                except:
                    report_error(CFG, env, "[TLP][WS][/ingest] Exception building FieldStorage()")
                    return error_500(start_response)
                if "json" not in fs.keys():
                    #return error_400(start_response, str(fs.keys()))
                    return error_400(start_response, "missing 'json' field")
                if fs['json'].type != "application/json":
                    #return error_400(start_response, "'json' field Content-Type must be 'application/json'")
                    pass
                data = fs.getvalue("json")

            else:
                return error_400(start_response, "Unsupported Content-Type.")

            try:
                body = json.loads(data)
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
            try:
                manif = body["manifest"]
            except:
                return error_400(start_response, "'manifest' parameter missing.")
            ok, req_pkg, err = check_manifest(manif, opc)
            if not(ok):
                err_code, err_info = err
                if err_code == 400:
                    return error_400(start_response, err_info)
                elif err_code == 500:
                    return error_500(start_response)
            if req_pkg:
                if env['CONTENT_TYPE'].find('multipart/form-data') < 0:
                    return error_400(start_response, 'According to manifest data, you must attach too a media package (Content-Type: multipart/form-data)')
                if "pkg" not in fs.keys():
                    return error_400(start_response, "'pkg' field missing.")
                if fs['pkg'].type != "application/zip":
                    return error_400(start_response, "'pkg' field Content-Type must be 'application/zip'")
                pkg_file = fs["pkg"].file
   

        elif opc in [2,3]:
            encoded_params = True
            if env['REQUEST_METHOD'] == 'GET':
                queries = parse_qs(env['QUERY_STRING'])
                if "data" in queries:
                    try:
                        data = base64.b64decode(queries["data"][0])
                    except: 
                        return error_400(start_response, "Could not parse 'data' GET Parameter" )
                else:
                    encoded_params = False
    
            else: # POST
                if env.get('CONTENT_TYPE', None) != 'application/json':
                    return error_400(start_response, "Content-Type must be 'application/json'")
                length = int(env.get('CONTENT_LENGTH', '0'))
                if length == 0:
                    return error_400(start_response, "Content-Length must be greater than 0")
                try:
                    data = env['wsgi.input'].read(length)
                except:
                    return error_400(start_response, "Cannot read POST data.")
    
            if encoded_params:
                try:
                    body = json.loads(data, encoding="UTF-8")
                except:
                    return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
            else:
                body = { k:queries[k][0] for k in queries.keys() }
     

            if "id" not in body.keys():
                return error_400(start_response, "Missing mandatory parameters/keys." )
    
            if opc == 2:
                extids = body["id"].split(",")
                if len(extids) == 0 and extids[0].strip() == "":
                    return error_400(start_response, "Empty IDs list" )
                manif = { 'documents':[] }
                for extid in extids:
                    manif['documents'].append({'external_id':extid.strip()})
    
        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        su = body.get("su", None)
        debug = body.get("debug", False)

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/ingest] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/ingest] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")
        if opc == 3:
            upload_id = body["id"]
            WSSdata = tL.tLint_auth_is_owner_upload(user_id, upload_id)
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/ingest] tL.tLint_auth_is_owner_upload() returned %s" % WSSdata.scode)
                return error_500(start_response)
            if WSSdata.results == 0:
                Dresp = {'rcode':1, 'rcode_description':'Upload ID %s does not exist.' % upload_id}
            if not(WSSdata.data) and not(is_admin):
                Dresp = {'rcode':2, 'rcode_description':'You are not allowed to Cancel Upload ID %s.' % upload_id}
            else:
                WSSdata = tL.tLint_ingest_cancelUpload(upload_id, debug_mode=debug)
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/ingest] tL.tLint_ingest_cancelUpload() returned %s" % WSSdata.scode)
                    return error_500(start_response)
                Dresp = {'rcode':0, 'rcode_description':'Upload cancelled.'}
        else: 
            doc_hash = '0'
            WSSdata = tL.tLint_ingest_storeMP(opc, manif, doc_hash, user_id, pkg_file=pkg_file, debug_mode=debug)
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/ingest] tL.tLint_ingest_storeMP() returned %s" % WSSdata.scode)
                return error_500(start_response)
            upload_id = WSSdata.data[0]['upload_id']
            Dresp = {'rcode':0, 'rcode_description':'Ingestion complete.', 'id':str(upload_id)}
    
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
   

    #--------------------------------------------#
    #----------------- /status ------------------#
    #--------------------------------------------#

    elif env['PATH_INFO'] == '/status': # See transcription status

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            upid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )
    
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
       
        Dresp = {}
        
        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/status] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/status] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_auth_is_owner_upload(user_id, upid)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_auth_is_owner_upload() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Upload ID [ %s ] does not exist.' % upid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_400(start_response, "Authentication error: user '%s' does not own upload ID '%s'." % (str(user), str(upid)))

        WSSdata = tL.tLint_status_getStatus(upid) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_status_getStatus() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = "Upload ID [ %s ] does not exist." % upid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        elif WSSdata.results == 1:
            res = WSSdata.data[0]
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = "Upload ID exists."
            Dresp['status_code'] = res['upload_status']
            Dresp['object_id'] = res['object_id']
            Dresp['status_info'] = res['info']
            Dresp['error_code'] = res['error_code']
            Dresp['uploaded_on'] = str(res['up_ts'])
            Dresp['last_update'] = str(res['last_check'])
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        else:
            report_error(CFG, env, "[TLP][WS][/status] tL.tLint_status_getStatus() returned more than one row")
            return error_500(start_response)


    #-------------------------------------------------#
    #----------------- /metadata ------------------#
    #-------------------------------------------------#
  
    elif env['PATH_INFO'] == '/metadata':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/metadata] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/metadata] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.text_metadata_getMetadata(extid, doc_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/metadata] tL.text_metadata_getMetadata() returned %d" % WSSdata.scode)
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        ret = WSSdata.data[0]
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Document info available.'
        Dresp['info'] = { "title":ret["title"], "language":ret["language"], "format":ret["format"], "sentences":ret["sentences"], "words":ret["words"], "characters":ret["characters"] }
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #-------------------------------------------#
    #----------------- /langs ------------------#
    #-------------------------------------------#
  
    elif env['PATH_INFO'] == '/langs':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )
        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None, 'locked':None, 'document_lang':None, 'langs':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/langs] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/langs] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/langs] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.text_langs_getLangs(extid, doc_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/langs] tL.text_langs_getLangs() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist or has no translations.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        res = WSSdata.data
        document_lang = res[0]['document_lang']
        locked = res[0]['locked']
        langs = [ { 'lang_code': r['code'], 'lang_name': r['value'] , 'sup_status':r['status'], 'sup_ratio':r['sup_ratio'] } for r in res ]
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Language list available.'
        Dresp['document_lang'] = document_lang
        Dresp['locked'] = locked
        Dresp['langs'] = langs
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #------------------------------------------#
    #----------------- /get ------------------#
    #------------------------------------------#

    elif env['PATH_INFO'] == '/get': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            lang = body["lang"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )

        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        try:
            form = int(body.get("format",0))
        except:
            return error_400(start_response, "'format' parameter must be an integer value." )
        if form not in (0,1,2,3):
            return error_400(start_response, "Unknown 'format' value." )

        try:
            policy = int(body.get("sel_data_policy",2))
        except:
            return error_400(start_response, "'sel_data_policy' parameter must be an integer value." )
        if policy not in (0,1,2):
            return error_400(start_response, "Unknown 'sel_data_policy' value." )
        
        session_id = body.get("session_id", None)
        if session_id != None:
            try:
                session_id = int(session_id)
            except:
                return error_400(start_response, "'session_id' must be a integer")

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/subs] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/subs] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/subs] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.text_langs_getLangs(extid, doc_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/get] tL.text_langs_getLangs() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist or has no translations.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        res = WSSdata.data
        document_lang = res[0]['document_lang']
        available_langs = [ r['code'] for r in res ]

        if lang == document_lang:
            WSSdata = tL.text_get_getSrcDocument(doc_hash, form) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/get] tL.text_get_getSrcDocument() returned %s" % (WSSdata.scode))
                return error_500(start_response)
            data = WSSdata.data[0]['data']
            ff =  WSSdata.data[0]['ff']
            status = '200 OK'
            response_headers = [
                ('Content-Disposition','inline; filename="%s.%s.%s"' % (extid.encode('utf-8'), lang.encode('utf-8'), ff.encode('utf-8'))),
                ('Content-Length', str(len(data)))]
            if form == 0:
                mimetypes.init()
                ct = mimetypes.types_map.get('.%s'%(ff), 'text/plain').encode('utf-8')
                response_headers.append(('Content-Type', ct))
            else:
                response_headers.append(('Content-Type', 'text/html'))
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]       
        else: # other languages
            if lang not in available_langs:
                Dresp['rcode'] = 2
                Dresp['rcode_description'] = 'Document ID [ %s ] does not feature %s translations.' % (extid, lang)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
            WSSdata = tL.text_get_getTrgDocument(doc_hash, lang, form=form) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/get] tL.text_get_getTrgDocument() returned %s" % (WSSdata.scode))
                return error_500(start_response)
            data = WSSdata.data[0]['data']
            ff =  WSSdata.data[0]['ff']

            if form in (0,1): # original, html
                status = '200 OK'
                response_headers = [
                    ('Content-Disposition','inline; filename="%s.%s.%s"' % (extid.encode('utf-8'), lang.encode('utf-8'), ff.encode('utf-8'))),
                    ('Content-Length', str(len(data)))]
                if form == 0:
                    mimetypes.init()
                    ct = mimetypes.types_map.get('.%s'%(ff), 'text/plain').encode('utf-8')
                    response_headers.append(('Content-Type', ct))
                else:
                    response_headers.append(('Content-Type','text/html'))
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            else: # 2,3: from DTLX
       
                dtlx = Parser().parseString(data)

                if session_id != None:
                    WSSdata = tL.tLint_get_mod_session(session_id)
                    if WSSdata.scode != 0:
                        report_error(CFG, env, "[TLP][WS][/get] tL.tLint_get_mod_revision_mods() returned %d" % WSSdata.scode)
                        return error_500(start_response)
                    if WSSdata.results == 0:
                        status = '200 OK'
                        data = json.dumps({'rcode':2,'rcode_description':"Provided session ID %d does not exist" % (session_id)}, indent=4)
                        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                        start_response(status, response_headers + CACHE_RESP_HEADERS)
                        return [data]
        
                    sess_data = WSSdata.data[0]
        
                    if sess_data['object_id'] != doc_hash:
                        Dresp = {}
                        Dresp['rcode'] = 3
                        Dresp['rcode_description'] = 'Provided session ID [ %d ] does not involve document ID [ %s ].' % (session_id, extid)
                        data = json.dumps(Dresp, indent=4)
                        status = '200 OK'
                        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                        start_response(status, response_headers + CACHE_RESP_HEADERS)
                        return [data]
        
                    if sess_data['mods'] != None: 
        
                        mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8") 
                        if lang in mods.keys():
                            mod = mods[lang]
                            dtlxmod = sess_mods2dtlx_mods(mod)
                            try:
                                dtlx.update(dtlxmod, sess_data['author_id'], sess_data['author_conf'], "human", 1.0, datetime.now(), highlight=True)
                            except:
                                report_error(CFG, env, "[TLP][WS][/end_session] Could not update DTLX file")
                                return error_500(start_response)
                            dtlx.updateStatus()

                if policy == 0:
                    dtlx.filter(Document.FILTER_NO_FILTER)
                    data = dtlx.toXML(filtered=False)
                elif policy == 1:
                    dtlx.filter(Document.FILTER_GET_FORMER)
                    data = dtlx.toXML(filtered=True)
                elif policy == 2:
                    dtlx.filter(Document.FILTER_GET_BEST)
                    data = dtlx.toXML(filtered=True)

                status = '200 OK'
              
                response_headers = [
                    ('Content-Disposition','inline; filename="%s.%s.xml"'%(extid.encode('utf-8'), lang.encode('utf-8'))),
                    ('Content-Length', str(len(data)))]
                if form == 2:
                    response_headers.append(('Content-Type','application/xml'))
                else:
                    response_headers.append(('Content-Type','plain/text'))
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

    #--------------------------------------------#
    #---------- /start_session --------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/start_session': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
  
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        author_name = body.get("author_name", None)

        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.text_langs_getLangs(extid, doc_hash)

        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Document ID has no translations.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        subs_locked = WSSdata.data[0]['locked']
        if subs_locked and author_conf < 100:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Subtitles for this Document ID have been locked. Only authors are allowed to modify them.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
   
        WSSdata = tL.tLint_check_open_mod_sessions(doc_hash)

        if WSSdata.results > 0:

            sess_data = WSSdata.data[0]
            Dresp = { 'session_id':sess_data['session_id'], 'author_id':sess_data['author_id'], 'author_conf':sess_data['author_conf'], 
                      'author_name':sess_data['author_name'], 'author_type':sess_data['author_type'], 'started_at':str(sess_data['started_at']), 'last_update':str(sess_data['last_update']) }

            if user_id == sess_data['user_id'] and author_id == sess_data['author_id'] and author_conf == sess_data['author_conf']: # same user that started mod session before

                WSSdata = tL.tLint_restart_session(sess_data['session_id'])

                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session restarted.'
            else:
                Dresp['rcode'] = 4
                Dresp['rcode_description'] = 'Cannot start mod session: there exist an open mod session for the given media ID.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        WSSdata = tL.tLint_check_previous_unrevised_session(doc_hash, author_id)

        if WSSdata.results > 0:
            sess_data = WSSdata.data[0]
            previous_session_id = sess_data['session_id']
            previous_mods = sess_data['mods']
            previous_mods_stats = sess_data['mods_stats']
        else:
            previous_session_id = None
            previous_mods = None
            previous_mods_stats = None

        WSSdata = tL.tLint_start_session(doc_hash, user_id, author_id, author_conf, author_name=author_name, 
                                             prev_session_id=previous_session_id, prev_mods=previous_mods, 
                                             prev_mods_stats=previous_mods_stats)

        sess_data = WSSdata.data[0]
        Dresp = { 'session_id':sess_data['session_id'], 'author_id':sess_data['author_id'], 'author_conf':sess_data['author_conf'], 
                  'author_name':sess_data['author_name'], 'started_at':str(sess_data['started_at']), 'last_update':str(sess_data['last_update']) }

        if previous_session_id != None:

            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session started, initialized with editions from a previous unrevised session.'
        else:
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session started.'

        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
   
    #--------------------------------------------#
    #------------------ /session_status --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/session_status': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            session_id = body["session_id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        alive = body.get("alive", 0)
        try:
            alive = bool(int(alive))
        except:
            return error_400(start_response, "'alive' parameter must be an integer value [0,1]." )
   
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['object_id'] != doc_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ] (%s != %s).' % (session_id, extid, sess_data['object_id'], doc_hash)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if author_id != sess_data['author_id'] or author_conf != sess_data['author_conf'] or user_id != sess_data['user_id']:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'You are not allowed to check status of this session ID.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        Dresp = OrderedDict()
        Dresp['rcode'] = None
        Dresp['rcode_description'] = None
        Dresp['started_at'] = str(sess_data['started_at'])
        Dresp['last_update'] = str(sess_data['last_update'])
        Dresp['last_alive'] = str(sess_data['last_alive'])
        Dresp['ended_at'] = str(sess_data['ended_at'])
        Dresp['ended_by_id'] = sess_data['ended_by_id']
        Dresp['ended_by_name'] = sess_data['ended_by_name']
        Dresp['ended_by_type'] = sess_data['ended_by_type']

        if sess_data['ended_at'] != None:
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Session ID %d is closed.' % sess_data['session_id']
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

        now = datetime.now()
        if alive:
            WSSdata = tL.tLint_update_last_session_alive(session_id, now) 
        else:
            WSSdata = tL.tLint_update_last_session_update(session_id, now) 

        Dresp['last_update'] = str(now)
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Session alive.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


   
    #--------------------------------------------#
    #------------------ /mod --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/mod': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            session_id = body["session_id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            mods = body["mods"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        if not(encoded_params):
            try:
                mods = json.loads(base64.b64decode(mods), encoding="UTF-8")
            except:
                error_400(start_response, "Could not decode 'mod' GET parameter (should be base64-encoded).")
   
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        session_alive = True
        if sess_data['ended_at'] != None:
            session_alive = False
 
        if sess_data['object_id'] != doc_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if author_id != sess_data['author_id'] or author_conf != sess_data['author_conf'] or user_id != sess_data['user_id']:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'You are not allowed to send modifications under this session ID.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
       
        if len(mods.keys()) == 0: 
            Dresp = {}
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Empty mods dict.'
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]


        WSSdata = tL.text_langs_getLangs(extid, doc_hash)
        langlist = [ x['code'] for x in WSSdata.data ]

        errors = False
        rinfo = []
        dmods = {}
        for language in mods.keys():
            if language not in langlist:
                mods.pop(language, None)
                rinfo.append({"language":language, "rcode":10, "rcode_description":"The given document ID does not have '%s' translation." % language })
                errors = True
                continue
            modsl = mods[language]
            ret = check_mod_format(modsl)
            if ret[0] != 0:
                mods.pop(language, None)
                rinfo.append({"language":language, "rcode":ret[0], "rcode_description":ret[1] })
                errors = True
                continue
            dmods[language] = {}
            for s in modsl['txt']:
                sI = s['sI']
                t = s['t']
                dmods[language][sI] = t
            rinfo.append({"language":language, "rcode":0, "rcode_description":"Changes successfully saved." })
                
        if len(mods.keys()) == 0: 
            Dresp = {}
            Dresp['rcode'] = 5
            Dresp['rcode_description'] = 'Format errors where found in all subtitle language modifications.'
            Dresp['details'] = rinfo
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

    
        if sess_data['mods'] == None:
            sess_mods = {}
        else:
            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8") 

        doc_words = tL.get_object_length(doc_hash)

        mods_stats = compute_mods_stats(dmods, doc_words)

        new_session_mods = merge_mods(dmods, sess_mods)
        
        new_session_mods_stats = compute_mods_stats(new_session_mods, doc_words)
 
        b64_mods = base64.b64encode(json.dumps(mods))
        b64_mods_stats = base64.b64encode(json.dumps(mods_stats))
        b64_session_mods = base64.b64encode(json.dumps(new_session_mods))
        b64_session_mods_stats = base64.b64encode(json.dumps(new_session_mods_stats))

        WSSdata = tL.tLint_store_mod(session_id, language, b64_mods, b64_mods_stats, b64_session_mods, b64_session_mods_stats, session_alive) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_store_mod() returned %d" % WSSdata.scode)
            return error_500(start_response)

        if not(session_alive):
            Dresp = {}
            if errors:
                Dresp['rcode'] = 6
                Dresp['rcode_description'] = 'Session ID [ %d ] is closed, changes have been backuped. Also some errors were found (see \'details\' key). Please contact system administrator.' % session_id
                Dresp['details'] = rinfo
            else:
                Dresp['rcode'] = 7
                Dresp['rcode_description'] = 'Session ID [ %d ] is closed, changes have been backuped. Please contact system administrator.' % session_id
            
        else:
            Dresp = {}
            if errors:
                Dresp['rcode'] = 8
                Dresp['rcode_description'] = 'Modifications for some subtitle languages were successfully saved, but other failed. Please see \'details\' key.'
                Dresp['details'] = rinfo
            else:
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Changes successfully saved.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

    #--------------------------------------------#
    #--------- /end_session -----------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/end_session': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            session_id = body["session_id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        author_name = body.get("author_name", None)

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        force = body.get("force", 0)
        try:
            force = bool(int(force))
        except:
            return error_400(start_response, "force parameter must be [0,1]." )
  
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['ended_at'] != None:
            Dresp = {}
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Session ID [ %d ] already closed (closed at %s).' % (session_id, str(sess_data['ended_at']))
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if sess_data['object_id'] != doc_hash:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
  
        if sess_data['author_type'] == 'computer':
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Session ID [ %d ] is owned by the Ingest Service and it cannot be manually ended.' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if user_id != sess_data['user_id']: # Different API user
            if not(force):
                Dresp = {}
                Dresp['rcode'] = 4
                Dresp['rcode_description'] = 'You are not allowed to close this session.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
            elif not(is_admin):
                WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash, other_perms=False)
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
                    return error_500(start_response)
                if WSSdata.results == 0:
                    return error_500(start_response)
                if WSSdata.data == False:
                    Dresp = {}
                    Dresp['rcode'] = 4
                    Dresp['rcode_description'] = 'You are not allowed to close this session.'
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data]
        else: # Same API user
            if (author_id != sess_data['author_id'] or author_conf != sess_data['author_conf']) and (author_conf < 100 or not(force)):
                Dresp = {}
                Dresp['rcode'] = 4
                Dresp['rcode_description'] = 'You are not allowed to close this session.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

        if sess_data['author_conf'] < 100:
            requires_revision = (sess_data['mods'] != None)
            revised = False
            requires_postprocess = False
        elif sess_data['author_conf'] == 100:
            requires_revision = False
            revised = None
            requires_postprocess = (sess_data['mods'] != None)

        if not(requires_revision) and sess_data['mods'] != None: # store changes in DFXP file

            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8")

            for lang in sess_mods.keys():

                mod = sess_mods[lang]
                dtlxmod = sess_mods2dtlx_mods(mod)

                WSSdata = tL.text_get_getTrgDocument(doc_hash, lang, form=2) 
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_get_getTrgDocument() returned %s" % (WSSdata.scode))
                    return error_500(start_response)
                data = WSSdata.data[0]['data']
                dtlx = Parser().parseString(data)
                try:
                    dtlx.update(dtlxmod, sess_data['author_id'], sess_data['author_conf'], "human", 1.0, datetime.now())
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] Could not update DTLX file")
                    return error_500(start_response)
        
                dtlx.updateStatus()
                data = dtlx.toXML(filtered=False)
                WSSdata = tL.text_endsession_storeTranslation(doc_hash, lang, data, dtlx.status, dtlx.sup_ratio)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeTranslation() returned %d" % WSSdata.scode)
                    return error_500(start_response)

                WSSdata = tL.text_get_getSrcDocument(doc_hash, 1) 
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_get_getTrgDocument() returned %s" % (WSSdata.scode))
                    return error_500(start_response)
                html = WSSdata.data[0]['data']
                try:
                    regen_html = regen_html_from_dtlx(dtlx, html)
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] could not regenerate HTML from DTLX")
                    return error_500(start_response)
                WSSdata = tL.text_endsession_storeDocument(doc_hash, lang, regen_html, form=1)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeDocument() returned %d" % WSSdata.scode)
                    return error_500(start_response)

                try:
                    clean_regen_html = clean_myhtml(regen_html)
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] Exception on clean_myhtml()")
                    return error_500(start_response)

                WSSdata = tL.text_endsession_storeDocument(doc_hash, lang, clean_regen_html, form=0)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeDocument() returned %d" % WSSdata.scode)
                    return error_500(start_response)

   
        WSSdata = tL.tLint_end_session(session_id, author_id, requires_revision, requires_postprocess, ended_by_name=author_name) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_end_session() returned %d" % WSSdata.scode)
            return error_500(start_response)

        Dresp = {}
        Dresp['rcode'] = 0
        Dresp['rcode_description'] = 'Session succesfully closed.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]


    #--------------------------------------------#
    #------------------ /lock_subs --------------------#
    #--------------------------------------------#
  
    elif env['PATH_INFO'] == '/lock': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        try:
            extid = body["id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
            lock = body["lock"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            lock = bool(int(lock))
        except:
            return error_400(start_response, "lock parameter must be [0,1].")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order to lock/unlock subtitles.")

        if not(encoded_params):
            try:
                mods = json.loads(base64.b64decode(mods), encoding="UTF-8")
            except:
                error_400(start_response, "Could not decode 'mod' GET parameter (should be base64-encoded).")
   
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)
 
        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        tL.tLint_lock_translations(doc_hash, lock)

        Dresp = {}
        Dresp['rcode'] = 0
        if lock:
            Dresp['rcode_description'] = 'Subtitles successfully locked.'
        else:
            Dresp['rcode_description'] = 'Subtitles successfully unlocked.'
        data = json.dumps(Dresp, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]





    #-----------------------------------------------#
    #----------------- /revisions ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/revisions':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
 
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None, allow_request_key=False)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        WSSdata = tL.tLint_get_all_pending_revisions(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_get_pending_user_mod_revisions() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        else:
            Lresp = []
            for r in WSSdata.data:
                Dresp = OrderedDict()
                Dresp['media_id'] = r['media_id']
                Dresp['session_id'] = r['session_id']
                Dresp['author_id'] = r['author_id']
                Dresp['author_conf'] = r['author_conf']
                Dresp['author_name'] = r['author_name']
                Dresp['timestamp'] = str(r['timestamp']) if r['timestamp'] != None else None
                Dresp['edit_stats'] = json.loads(base64.b64decode(r['edit_stats']), encoding="UTF-8") if r['edit_stats'] != None else None
                Lresp.append(Dresp)
            data = json.dumps(Lresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]

    #------------------------------------------------#
    #------------------- /mark_revised --------------#
    #------------------------------------------------#
  
    elif env['PATH_INFO'] == '/mark_revised': 

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
  
        try:
            extid = body["id"]
            session_id = body["session_id"]
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            session_id = int(session_id)
        except:
            return error_400(start_response, "session_id parameter must be an integer value.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order to mark edition as revised.")

        author_name = body.get("author_name", None)

        rev_session_id = body.get("revision_session_id", None)
        if rev_session_id != None:
            try:
                rev_session_id = int(rev_session_id)
            except:
                return error_400(start_response, "revision_session_id parameter must be an integer value.")
 
        if session_id == rev_session_id:   
            return error_400(start_response, "session_id and revision_session_id parameters cannot be equal.")
  
        unmark = body.get("unmark", 0)
        try:
            unmark = bool(int(unmark))
        except:
            return error_400(start_response, "'unmark' parameter must be an integer value [0,1]." )
         
 
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/start_session] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/start_session] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/start_session] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))
 
        WSSdata = tL.tLint_get_mod_session(session_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/end_session] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
            return error_500(start_response)
        
        if WSSdata.results == 0:
            Dresp = {}
            Dresp['rcode'] = 2
            Dresp['rcode_description'] = 'Session ID [ %d ] does not exist.' % session_id
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        
        sess_data = WSSdata.data[0]

        if sess_data['ended_at'] == None:
            Dresp = {}
            Dresp['rcode'] = 3
            Dresp['rcode_description'] = 'Session ID [ %d ] has not been closed yet.' % (session_id)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if sess_data['object_id'] != doc_hash:
            Dresp = {}
            Dresp['rcode'] = 4
            Dresp['rcode_description'] = 'Session ID [ %d ] does not involve media ID [ %s ].' % (session_id, extid)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if not(sess_data['requires_revision']):
            Dresp = {}
            Dresp['rcode'] = 5
            Dresp['rcode_description'] = 'Session ID [ %d ] did not require revision.' % (session_id)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
        if unmark:

            if not(sess_data['revised']):
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session ID [ %d ] has not been revised yet.' % (session_id)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            WSSdata = tL.tLint_unmark_revised(session_id)
    
            if WSSdata.scode == 0:  # OK
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Revised mark successfully removed.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
 
        else:

            if sess_data['revised']:
                Dresp = {}
                Dresp['rcode'] = 6
                Dresp['rcode_description'] = 'Session ID [ %d ] was already revised before.' % (session_id)
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]

            if rev_session_id != None:
                WSSdata = tL.tLint_get_mod_session(rev_session_id)

                if WSSdata.results == 0:
                    Dresp = {}
                    Dresp['rcode'] = 7
                    Dresp['rcode_description'] = 'The provided "revision_session_id" %d does not exist.' % (rev_session_id)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data]

                sess_data = WSSdata.data[0]

                if sess_data['object_id'] != doc_hash:
                    Dresp = {}
                    Dresp['rcode'] = 8
                    Dresp['rcode_description'] = 'The provided "revision_session_id" %d does not involve Document ID %s.' % (rev_session_id, extid)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data] 
 
                if sess_data['author_id'] != author_id or sess_data['user_id'] != user_id or sess_data['author_conf'] != author_conf:
                    Dresp = {}
                    Dresp['rcode'] = 9
                    Dresp['rcode_description'] = "The provided 'revision_session_id' %d is not owned by the author ID that is requesting this session ID to be marked as revised." % (rev_session_id)
                    data = json.dumps(Dresp, indent=4)
                    status = '200 OK'
                    response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                    start_response(status, response_headers + CACHE_RESP_HEADERS)
                    return [data] 

   
            WSSdata = tL.tLint_mark_revised(session_id, author_id, author_name=author_name, rev_session_id=rev_session_id)
    
            if WSSdata.scode == 0:  # OK
                Dresp = {}
                Dresp['rcode'] = 0
                Dresp['rcode_description'] = 'Session ID marked as revised.'
                data = json.dumps(Dresp, indent=4)
                status = '200 OK'
                response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
                start_response(status, response_headers + CACHE_RESP_HEADERS)
                return [data]
   

    #-----------------------------------------------#
    #----------------- /edit_history ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/edit_history':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            extid = body["id"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")
 
        su = body.get("su", None)
        doc_hash = body.get("hash", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, extid, expire, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

        if doc_hash == None:
            doc_hash = get_hash(extid, user_id)

        WSSdata = tL.tLint_auth_is_owner_object(user_id, doc_hash)
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_auth_is_owner_object() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        if WSSdata.results == 0:
            Dresp['rcode'] = 1
            Dresp['rcode_description'] = 'Document ID [ %s ] does not exist.' % extid
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
        if WSSdata.data == False and is_admin == False:
            return error_401(start_response, "Authentication error: user '%s' does not own media ID '%s'." % (str(user), str(extid)))

        WSSdata = tL.tLint_get_all_media_editions(doc_hash) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/revisions] tL.tLint_get_mod_revisions() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        else:
            Dresp = OrderedDict()
            Dresp['rcode'] = 0
            Dresp['rcode_description'] = 'Edit history available.'
            Dresp['edit_history'] = []
            for r in WSSdata.data:
                d = OrderedDict()
                d['session_id'] = r['session_id']
                d['author_id'] = r['author_id']
                d['author_conf'] = r['author_conf']
                d['author_name'] = r['author_name']
                d['timestamp'] = str(r['timestamp']) if r['timestamp'] != None else None
                d['requires_revision'] = r['requires_revision']
                d['revised'] = r['revised']
                d['revised_at'] = str(r['revised_at']) if r['revised_at'] != None else None
                d['revised_by_id'] = r['revised_by_id']
                d['revised_by_name'] = r['revised_by_name']
                d['revised_via'] = r['revised_via']
                d['revised_in_session_id'] = r['revised_in_session_id']
                d['edit_stats'] = json.loads(base64.b64decode(r['edit_stats']), encoding="UTF-8") if r['edit_stats'] != None else None
                Dresp['edit_history'].append(d)
            data = json.dumps(Dresp, indent=4)
            status = '200 OK'
            response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
            start_response(status, response_headers + CACHE_RESP_HEADERS)
            return [data]
 
    #-----------------------------------------------#
    #----------------- /reject ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/reject':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            id_param = str(body["id"])
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order accept/decline changes.")

        author_name = body.get("author_name", None)
 
        try:
            sess_ids = map(lambda x: int(x), id_param.split(","))
        except:
            return error_400(start_response, '"id" parameter must be a string with comma-separated Session IDs (integers).')
 
        if len(sess_ids) > 100:
            return error_400(start_response, 'Exceeded limit of Session IDs to be accepted (max 100).')

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, id_param, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

 
        res = []

        for session_id in sess_ids: 
            WSSdata = tL.tLint_get_mod_session(session_id) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
                return error_500(start_response)
            
            if WSSdata.results == 0:
                res.append({'session_id':session_id, 'rcode':1, 'rcode_description':'Session ID does not exist.'})
                continue
 
            sess_data = WSSdata.data[0]
    
            if user_id != sess_data['user_id']:
                res.append({'session_id':session_id, 'rcode':2, 'rcode_description':'You are not allowed to modify the status of this Session ID.'})
                continue
   
            if sess_data['ended_at'] == None:
                res.append({'session_id':session_id, 'rcode':3, 'rcode_description':'Session ID has not been ended yet.'})
                continue
            if not(sess_data['requires_revision']):
                res.append({'session_id':session_id, 'rcode':4, 'rcode_description':'Session ID did not require revision.'})
                continue

            if sess_data['revised']:
                res.append({'session_id':session_id, 'rcode':5, 'rcode_description':'Session ID already revised.'})
                continue
 
            WSSdata = tL.tLint_reject(session_id, author_id, author_name)

            res.append({'session_id':session_id, 'rcode':0, 'rcode_description':'Changes from Session ID successfully rejected.'})
 
        data = json.dumps(res, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]

 
    #-----------------------------------------------#
    #----------------- /accept ------------------#
    #-----------------------------------------------#
  
    elif env['PATH_INFO'] == '/accept':

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }
      
        try:
            id_param = str(body["id"])
            author_id = body["author_id"]
            author_conf = body["author_conf"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys.")

        try:
            author_conf = int(author_conf)
        except:
            return error_400(start_response, "author_conf parameter must be an integer value.")

        if author_conf < 0 or author_conf > 100:
            return error_400(start_response, "author_conf parameter must be ranged in [0-100].")

        if author_conf != 100:
            return error_400(start_response, "author_conf parameter must be 100 in order accept/decline changes.")

        author_name = body.get("author_name", None)
 
        try:
            sess_ids = map(lambda x: int(x), id_param.split(","))
        except:
            return error_400(start_response, '"id" parameter must be a string with comma-separated Session IDs (integers).')

        if len(sess_ids) > 100:
            return error_400(start_response, 'Exceeded limit of Session IDs to be accepted (max 100).')
 
        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        expire = body.get("expire", None)
        if expire != None: 
            try:
                expire = int(expire)
            except:
                return error_400(start_response, "'expire' parameter must be an integer value." )

        Dresp = { 'rcode':None, 'rcode_description':None }

        scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, id_param, expire, author=author_id, author_conf=author_conf, allow_request_key=True)
        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/revisions] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/revisions] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_401(start_response, "Cannot su: admin privileges are required.")

 
        res = []

        for session_id in sess_ids: 
            WSSdata = tL.tLint_get_mod_session(session_id) 
            if WSSdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/mod] tL.tLint_get_mod_session() returned %d" % WSSdata.scode)
                return error_500(start_response)
            
            if WSSdata.results == 0:
                res.append({'session_id':session_id, 'rcode':1, 'rcode_description':'Session ID does not exist.'})
                continue
 
            sess_data = WSSdata.data[0]
    
            if user_id != sess_data['user_id']:
                res.append({'session_id':session_id, 'rcode':2, 'rcode_description':'You are not allowed to modify the status of this Session ID.'})
                continue
   
            if sess_data['ended_at'] == None:
                res.append({'session_id':session_id, 'rcode':3, 'rcode_description':'Session ID has not been ended yet.'})
                continue
            if not(sess_data['requires_revision']):
                res.append({'session_id':session_id, 'rcode':4, 'rcode_description':'Session ID did not require revision.'})
                continue

            if sess_data['revised']:
                res.append({'session_id':session_id, 'rcode':5, 'rcode_description':'Session ID already revised.'})
                continue

            sess_mods = json.loads(base64.b64decode(sess_data['mods']), encoding="UTF-8")

            for lang in sess_mods.keys():

                mod = sess_mods[lang]
                dtlxmod = sess_mods2dtlx_mods(mod)
                doc_hash = sess_data['object_id']

                WSSdata = tL.text_get_getTrgDocument(doc_hash, lang, form=2) 
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_get_getTrgDocument() returned %s" % (WSSdata.scode))
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue
                data = WSSdata.data[0]['data']
                dtlx = Parser().parseString(data)
                try:
                    dtlx.update(dtlxmod, sess_data['author_id'], sess_data['author_conf'], "human", 1.0, datetime.now())
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] Could not update DTLX file")
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue
        
                dtlx.updateStatus()
                data = dtlx.toXML(filtered=False)
                WSSdata = tL.text_endsession_storeTranslation(doc_hash, lang, data, dtlx.status, dtlx.sup_ratio)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeTranslation() returned %d" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue

                WSSdata = tL.text_get_getSrcDocument(doc_hash, 1) 
                if WSSdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_get_getTrgDocument() returned %s" % (WSSdata.scode))
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue
                html = WSSdata.data[0]['data']
                try:
                    regen_html = regen_html_from_dtlx(dtlx, html)
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] could not regenerate HTML from DTLX")
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue
                WSSdata = tL.text_endsession_storeDocument(doc_hash, lang, regen_html, form=1)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeDocument() returned %d" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue

                try:
                    clean_regen_html = clean_myhtml(regen_html)
                except:
                    report_error(CFG, env, "[TLP][WS][/end_session] Exception on clean_myhtml()")
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue

                WSSdata = tL.text_endsession_storeDocument(doc_hash, lang, clean_regen_html, form=0)
                if WSSdata.scode != 0:           
                    report_error(CFG, env, "[TLP][WS][/end_session] tL.text_endsession_storeDocument() returned %d" % WSSdata.scode)
                    res.append({'session_id':session_id, 'rcode':6, 'rcode_description':'Could not save Session edits into translation file.'})
                    continue

            WSSdata = tL.tLint_accept(session_id, author_id, author_name)

            res.append({'session_id':session_id, 'rcode':0, 'rcode_description':'Changes from Session ID successfully accepted.'})
 
        data = json.dumps(res, indent=4)
        status = '200 OK'
        response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
 

 
    else: # query not known, bad request
        status = '400 Bad Request'
        data = CFG['MISC']['HTML_MSG']
        response_headers = [('Content-Type', 'text/html'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS)
        return [data]
    

def application(env, start_response):
    try:
        tL = tLWSSLayer(CFG['STG']['DB_NAME'], CFG['STG']['DB_USER'], CFG['STG']['DB_HOST'], tLWSSLayer.OBJ_TEXT, passwd=CFG['STG']['DB_PASS'])
        tlpauth = TLPAuth(CFG['AUT']['KEY'])
        return my_application(env, start_response, tL, tlpauth)
    except:
        report_error(CFG, env, "[TLP][WS] Exception")
        return error_500(start_response)


#############################################################
#----------------- Error return functions ------------------#
#############################################################

def error_400(start_response, text=None):
    status = '400 Bad Request'
    response_headers = [('Content-Type', 'text/plain')]
    if text == None:
        data = "Bad Request"
    else:
        data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_401(start_response, text):
    status = '401 Unauthorized'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_419(start_response, text):
    status = '419 Authentication Timeout'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_500(start_response):
    status = '500 Internal Error'
    data = "Internal Server Error"
    response_headers = [('Content-Type', 'text/plain'),('Content-Length', str(len(data)))]
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]

ENV_KEYS = [ 'CONTENT_TYPE', 'SCRIPT_NAME', 'REQUEST_METHOD', 'PATH_INFO', 'REQUEST_URI', 'QUERY_STRING', 'wsgi.url_scheme', 'SERVER_NAME', 'SERVER_PORT', 'SCRIPT_FILENAME', 'HTTP_HOST', 'REMOTE_ADDR', 'REMOTE_PORT' ]

def report_error(CFG, env, subject, extra_data=None):
    if sys.exc_type != None:
        excpt = True
        excpt_str = traceback.format_exc()
        sys.stderr.write("\n************ API EXCEPTION ************\n\n")
        sys.stderr.write("%s\n" % excpt_str)
        #traceback.print_exc()
    else:
        excpt = False

    if CFG['MAI']['ENABLED'] and CFG['MAI']['TO'] != None:
        try:
            out = cStringIO.StringIO()
            out.write("DATE: %s\n" % datetime.now().strftime("%d/%m/%Y %H:%M:%S %Z"))
            out.write("EXCEPTION: %s\n" % str(excpt))
            if excpt:
                out.write("\n+++++++++++++++++++++++++++++++++ EXCEPTION ++++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % excpt_str)
            out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
            print_log(out, env)
            print_env(out, env, space=1)
            if env['PATH_INFO'] == '/ingest': 
              print_input_get_params(out, env, space=1)
            else:
              if env['REQUEST_METHOD'] == 'GET':
                  print_input_get_params(out, env, space=1)
              elif env['REQUEST_METHOD'] == 'POST':
                  print_icont(out, env)

            if extra_data != None:
                out.write("\n++++++++++++++++++++++++++++++++ EXTRA DATA +++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % extra_data)
                    
            content = out.getvalue() 
            send_mail(CFG['MAI']['SMTP_SERVER'], CFG['MAI']['FROM'], subject, content.decode('utf-8'), CFG['MAI']['TO'])
        except:
            sys.stderr.write("\n************ SEND_MAIL EXCEPTION ************\n")
            traceback.print_exc()
    
def send_mail(smtp_server, src_mail, subject, content, to, cc=None, cco=None):
    # Create e-mail data
    msg = MIMEText(content.encode('UTF-8'))
    msg['Subject'] = subject.encode('UTF-8')
    msg['From'] = src_mail
    msg['To'] = ",".join(to)
    addrl = []
    addrl = addrl + to
    if cc != None: 
        msg['cc'] = ",".join(cc)
        addrl = addrl + cc
    if cco != None: 
        msg['cco'] = ",".join(cco)
        addrl = addrl + cco
    # Send e-mail
    smtp = smtplib.SMTP(smtp_server)
    smtp.sendmail(src_mail, addrl, msg.as_string())







#####################################################
#----------------- Misc functions ------------------#
#####################################################

class MLStripper(HTMLParser):
    def __init__(self):
        self.reset()
        self.fed = []
    def handle_data(self, d):
        self.fed.append(d)
    def get_data(self):
        return ''.join(self.fed)

def strip_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()


def has_span_dtlx_attr(tag):
  return tag.name == "span" and tag.has_attr('data-dtlx-id')

def has_dtlx_attr(tag): # Compatibility for old parsings
  return tag.has_attr('data-dtlx-id')

def clean_myhtml(data):
  s = bs4.BeautifulSoup(data)
  for i in s.find_all(has_span_dtlx_attr):
    i.unwrap()
  for i in s.find_all(has_dtlx_attr): # Compatibility for old parsings
    del i['data-dtlx-id']
  return s.prettify().encode('utf-8')

def c(s, m=0):
    if m == 0: # general
        return (re.subn(r'[\r\b\n$&`´|{}()\[\];/\\"\'#*~^<>=+!?¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 1: # filenames, allow ( ) +
        return (re.subn(r'[\r\b\n$&`´|{}\[\];/\\"\'#*~^<>=!?¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 2: # urls
        return (re.subn(r'[\r\b\n`´|{}\\"\'^<>¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 3: # titles, allow ?¿¡! ( )
        return (re.subn(r'[\r\b\n$&`´|{}\[\];/\\"\'#*~^<>=+]|--', '', unicode(s))[1] > 0)

def check_manifest(manif, operation):
    if 'documents' not in manif.keys():
        return (False, None, (400, "Manifest JSON: Missing 'documents' key."))
    for k in manif.keys():
        if k == "documents":
            if not(isinstance(manif['documents'], list)):
                return (False, None, (400, "Manifest JSON: 'documents' must be an array"))
            if len(manif['documents']) == 0:
                return (False, None, (400, "Manifest JSON: 'documents' must be a non-empty array of dicts"))
            if len(manif['documents']) > 500:
                return (False, None, (400, "Manifest JSON: 'documents' length exceeded from the maximum (500)"))
            for d in manif['documents']:
                if not(isinstance(d, dict)):
                    return (False, None, (400, "Manifest JSON: Each 'documents' item must be a dict"))
                if 'external_id' not in d.keys():
                    return (False, None, (400, "Manifest JSON: Missing 'documents'->'external_id' key"))
                for dk in d.keys():
                   if dk == "filename":
                       if c(d[dk], m=1):
                           return (False, None, (400, "Manifest JSON: Invalid 'documents'->'%s' key value (contains forbidden characters)" % (dk.encode("utf-8"))))
                   elif dk == "url":
                       if c(d[dk], m=2):
                           return (False, None, (400, "Manifest JSON: Invalid 'documents'->'%s' key value (contains forbidden characters)" % (dk.encode("utf-8"))))
                   elif dk == "title":
                       if c(d[dk], m=3):
                           return (False, None, (400, "Manifest JSON: Invalid 'documents'->'%s' key value (contains forbidden characters)" % (dk.encode("utf-8"))))
                   else:
                       if c(d[dk]):
                           return (False, None, (400, "Manifest JSON: Invalid 'documents'->'%s' key value (contains forbidden characters)" % (dk.encode("utf-8"))))
        elif k == "language":
            if c(manif[k]):
                return (False, None, (400, "Manifest JSON: Invalid '%s' key value (contains forbidden characters)" % (k.encode("utf-8"))))
        elif k == "requested_langs":
            if not(isinstance(manif[k], dict)):
                return (False, None, (400, "Manifest JSON: '%s' key value must be a dict" % (k.encode("utf-8"))))
            rl = manif[k]
            for lang in rl.keys():
                if c(lang):
                    return (False, None, (400, "Manifest JSON: Invalid language code in 'requested_langs' (contains forbidden characters)"))
                ld = rl[lang]
                if not(isinstance(ld, dict)):
                    return (False, None, (400, "Manifest JSON: 'requested_langs'->'%s' value must be a dict" % lang.encode("utf-8")))
                for opt in ld.keys():
                    if opt == "sid":
                        if not(isinstance(ld[opt], int)):
                            return (False, None, (400, "Manifest JSON: 'requested_langs'->'%s'->'sid' key value must be an integer" % (lang.encode("utf-8"))))
                    elif opt == "tlpath":
                        if not(isinstance(ld[opt], list)):
                            return (False, None, (400, "Manifest JSON: 'requested_langs'->'%s'->'tlpath' key value must be a list" % (lang.encode("utf-8"))))
                        for i in ld[opt]:
                            if not(isinstance(i, dict)):
                                return (False, None, (400, "Manifest JSON: 'requested_langs'->'%s'->'tlpath' items must be dicts" % (lang.encode("utf-8"))))
                            if "l" not in i.keys(): 
                                return (False, None, (400, "Manifest JSON: 'requested_langs'->'%s'->'tlpath' items must contain at least a 'l' key" % (lang.encode("utf-8"))))
                            for ik in i.keys():
                                if ik == "l":
                                    if c(i[ik]):
                                        return (False, None, (400, "Manifest JSON: Invalid 'l' key value in 'requested_langs'->'%s'->'tlpath' dict item (contains forbidden characters)" % (lang.encode("utf-8"))))
                                elif ik == "sid":
                                    if not(isinstance(i[ik], int)):
                                        return (False, None, (400, "Manifest JSON: 'sid' key value in 'requested_langs'->'%s'->'tlpath' item must be an integer" % (lang.encode("utf-8"))))
                                elif ik == "skip":
                                    if not(isinstance(i[ik], bool)):
                                        return (False, None, (400, "Manifest JSON: 'skip' key value in 'requested_langs'->'%s'->'tlpath' item must be a boolean" % (lang.encode("utf-8"))))
                                else:
                                    return (False, None, (400, "Manifest JSON: Unknown key '%s' provided in 'requested_langs'->'%s'->'tlpath' dict item" % (ik.encode("utf-8"), lang.encode("utf-8"))))
                    else:
                        return (False, None, (400, "Manifest JSON: Unknown option '%s' provided in 'requested_langs'->'%s' dict" % (opt.encode("utf-8"), lang.encode("utf-8"))))
        elif k in ("generate", "re-generate", "force", "test_mode", "from_webui"):
            if not(isinstance(manif[k], bool)):
                return (False, None, (400, "Manifest JSON: '%s' key value must be a boolean" % (k.encode("utf-8"))))
                
    requires_mpkg = False
    if operation == 0: # New
        if 'language' not in manif.keys():
            return (False, None, (400, "Missing 'language' key in Manifest JSON"))
        for d in manif['documents']:
            if 'title' not in d.keys():
                return (False, None, (400, "Missing 'documents->title' key in Manifest JSON"))
            if 'url' not in d.keys():
                requires_mpkg = True
                if 'filename' not in d.keys():
                    return (False, None, (400, "Missing 'documents->filename' key in Manifest JSON"))
                if 'fileformat' not in d.keys():
                    return (False, None, (400, "Missing 'documents->fileformat' key in Manifest JSON"))
                if 'md5' not in d.keys():
                    return (False, None, (400, "Missing 'documents->md5' key in Manifest JSON"))
    elif operation == 1: # Update
        for d in manif['documents']:
             if 'url' not in d.keys() and 'filename' in d.keys():
                requires_mpkg = True
                if 'fileformat' not in d.keys():
                    return (False, None, (400, "Missing 'documents->fileformat' key in Manifest JSON"))
                if 'md5' not in d.keys():
                    return (False, None, (400, "Missing 'documents->md5' key in Manifest JSON"))
    return (True, requires_mpkg, None)

 
def compute_mod_stats(mod, doc_words):
    stats = {}
    stats['edit_sents'] = len(mod.keys())
    stats['edit_words'] = sum(map(lambda x: len(''.join(c if c.isalnum() else ' ' for c in strip_tags(mod[x])).split()), mod.keys()))
    stats['edit_percent'] = stats['edit_words']/float(doc_words)*100
    return stats
    

def compute_mods_stats(mods, doc_words):
    stats = {}
    for l in mods.keys():
        lmods = mods[l]
        stats[l] = compute_mod_stats(lmods, doc_words)
    return stats
         

def check_mod_format(moddata):

    if 'txt' not in moddata:
        return 1, "'txt' key not found in 'mod'."

    mod = moddata['txt']

    if not(isinstance(mod, list)):
        return 2, "'txt' key must be a list"

    if len(mod) == 0:
        return 3, "'txt' is an empty list"

    if len(mod) > 0:
        # Format and safety checks
        mod_sId_lst = []
        for cm in mod:
            if ('sI' not in cm) or ('t' not in cm): 
                return 4, "'txt' elements must include the following keys: 'sI', 't'"
            try: 
                cm['sI'] = int(cm['sI'])
            except:
                return 5, "Invalid format of 'txt' elements. Must be: 'sI':<int>, 't':<str>"
            if cm['sI'] in mod_sId_lst: 
                return 7, "Segment ID %d provided more than once" % cm['sI']
            mod_sId_lst.append(cm['sI'])
    
    return 0, ""

def merge_mods(dmods, session_mods):
    for language in dmods.keys():
        dmodsl = dmods[language]
        if language not in session_mods:
            session_mods[language] = dmodsl
        else:
            for sI in dmodsl:
                session_mods[language][sI] = dmodsl[sI]
    return session_mods

def sess_mods2dtlx_mods(mod):
    dtlxmod = {}
    for sI in mod.keys():
        ft = mod[sI]
        rt = strip_tags(ft)
        dtlxmod[sI] = { 't':{'r':rt, 'f':ft} }
    return dtlxmod

def check_auth_token(tL, tlpauth, user_name, given_token, oid, expire, author=None, author_conf=None, allow_request_key=True):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None, None, None
    if WSSdata.results == 0:
        return 2, None, None, None
    user_id = WSSdata.data[0]['id']
    user_pass = WSSdata.data[0]['password']
    is_admin = WSSdata.data[0]['is_admin']
    secret_key = tlpauth.get_secret_key(user_name.encode('utf-8'), user_pass)
    if secret_key == given_token:
        return 0, user_id, is_admin, False
    elif allow_request_key:
        if expire == None: 
            return 3, None, None, None
        request_key = tlpauth.get_request_key(user_name.encode('utf-8'), user_pass, oid.encode('utf-8'), expire, author, author_conf)
        if author == None:
            given_token = given_token[:tlpauth.request_subkey_length]
        if request_key == given_token:
            ts = int((datetime.utcnow() - datetime(1970,01,01)).total_seconds())
            if ts > expire:
                return 4, None, None, None
            return 0, user_id, is_admin, True
        else:
            return 3, None, None, None
    else:
        return 3, None, None, None

def get_user_id(tL, user_name):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None
    if WSSdata.results == 0:
        return 2, None
    user_id = WSSdata.data[0]['id']
    return 0, user_id

def get_hash(obj_id, user_id):
    h = hashlib.sha256()
    if isinstance(obj_id, unicode):
      obj_id = obj_id.encode('utf-8')
    if user_id != None:
        h.update("%d|%s" % (user_id, obj_id))
    else:
        h.update("%s" % obj_id)
    return h.hexdigest()


def parse_config(cfg_fd): 

    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}

    # General
    CFG['AUT'] = {}
    CFG['AUT']['KEY'] = config.get('authentication', 'key_generator')
    CFG['STG'] = {}
    CFG['STG']['DB_NAME'] = config.get('storage', 'db_name')
    CFG['STG']['DB_USER'] = config.get('storage', 'db_user')
    CFG['STG']['DB_HOST'] = config.get('storage', 'db_host')
    if CFG['STG']['DB_HOST'] != None: 
      if CFG['STG']['DB_HOST'].strip() == "": CFG['STG']['DB_HOST'] = None
    CFG['STG']['DB_PASS'] = config.get('storage', 'db_passwd')
    if CFG['STG']['DB_PASS'] != None: 
      if CFG['STG']['DB_PASS'].strip() == "": CFG['STG']['DB_PASS'] = None
    CFG['URL'] = {}
    CFG['URL']['USE'] = config.getboolean('media_urls', 'use_urls')
    CFG['URL']['BASEURL'] = config.get('media_urls', 'base_url')
    CFG['MAI'] = {}
    CFG['MAI']['ENABLED'] = config.getboolean('mailing', 'enabled')
    CFG['MAI']['FROM'] = config.get('mailing', 'from_address')
    CFG['MAI']['SMTP_SERVER'] = config.get('mailing', 'smtp_server')
    try:
        CFG['MAI']['TO'] = config.get('mailing', 'to_address').replace(' ', '').split(',')
    except:
        CFG['MAI']['TO'] = None
    CFG['MISC'] = {}
    CFG['MISC']['HTML_MSG'] = config.get('misc', 'html_msg')

    return CFG


# Load Config file
with open(CFG_FILE) as cfg_fd:
    CFG = parse_config(cfg_fd)


def get_mod_stats(mod):
    nsegs = 0
    accdur = 0
    for m in mod['txt']:
        st = float(m['b'])
        et = float(m['e'])
        dur = et-st
        accdur += dur
        nsegs += 1
    return nsegs, int(round(accdur,0))

##################################################
#----------- Logging WSGI Middelware ------------#
##################################################

import threading


def print_key_value(out, k, v, space=30):
    out.write("%s:%s%s\n" % (str(k), ' ' * max((space-len(k)),1), str(v)))

def print_width(out, data, width=80):
    ndata=''; i = 0; c = 0
    while i < len(data):
        ndata += data[i]
        if c == width: 
            ndata += '\n'
            c = 0
        else: c += 1
        i += 1
    out.write("%s\n" % ndata)

def print_log(out, env):
  out.write('PATH: %s\n' % env['SCRIPT_NAME'])
  out.write('INTERFACE: %s\n' % env['PATH_INFO'])

def print_env(out, environ, space=30):
  out.write('\n#------- INPUT ENVIRONMENT -------#\n\n')
  fenv = {}
  for k in ENV_KEYS:
    print_key_value(out, k, environ.get(k, "-"), space=space)

def print_input_get_params(out, env, space=30):    
  out.write('\n#------- INPUT DATA (GET) -------#\n\n')
  encoded_params = True
  queries = parse_qs(env['QUERY_STRING'])
  if "data" in queries:
      try:
          data = base64.b64decode(queries["data"][0])
      except: 
         out.write("Could not parse 'data' GET Parameter\n")
  else:
      encoded_params = False
  
  if encoded_params:
      try:
          body = json.loads(data, encoding="UTF-8")
      except:
         out.write("Cannot parse JSON data (Invalid JSON).\n")
      else:
         out.write("%s\n" % json.dumps(body, indent=4))
  else:
      body = { k:queries[k][0] for k in queries.keys() }
      for k in body:
           print_key_value(out, k, body.get(k, "-"), space=space)

  out.write("\n-- Get URL:\n\n")
  url = "http://%s%s" % (env['SERVER_NAME'], env['REQUEST_URI'])
  out.write("%s\n" % url)

def print_icont(out, environ):
   
  length = int(environ.get('CONTENT_LENGTH', '0'))
  
  out.write('\n#------- INPUT DATA (POST) -------#\n\n')
  if length > 0 and length < 2147483648: # 2GB
      icontent_fp = cStringIO.StringIO()
      icontent_fp.write(environ['wsgi.input'].read(length))
      if environ.get('CONTENT_TYPE', '').find('application/json') >= 0:
          icont = icontent_fp.getvalue()
          try:
              ret = str(icont)
          except:
              return
      elif environ.get('CONTENT_TYPE','').find('multipart/form-data') >= 0: 
          icontent_fp.seek(0, os.SEEK_SET)
          try:
              fs = FieldStorage(fp=icontent_fp,environ=environ)
          except:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if "json" not in fs.keys():
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if fs['json'].type.find("application/json") <  0:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          #out.write("%s\n" % icontent_fp.getvalue())
          ret = fs.getvalue("json")

      try:
          jret = json.loads(ret)
          jtret = json.dumps(jret, indent=4)
          encurl = urllib.urlencode(jret)
      except:
         out.write("\n-- Text:\n")
         out.write("%s\n" % ret)
      else:
         out.write("\n-- Text:\n\n")
         out.write("%s\n" % jtret)
         out.write("\n-- Get URL:\n\n")
         url = "http://%s%s?%s" % (environ['SERVER_NAME'], environ['REQUEST_URI'], encurl)
         out.write("%s\n" % url)

      icontent_fp.seek(0, os.SEEK_SET)
      environ['wsgi.input'] = icontent_fp

  else:
     out.write("None\n")



class LoggingInstance:
  def __init__(self, start_response):
    self.__start_response = start_response
    self.__print_content = True
    self.__content_type = ""
    self.__out = cStringIO.StringIO()

  def __call__(self, status, headers, *args):
    self.__out.write('\n#------- OUTPUT HEADERS -------#\n\n')
    print_key_value(self.__out, 'Response-Code', status)
    ct = None
    for h in headers:
        if h[0] == 'Content-Type': 
            ct = h[1]
        print_key_value(self.__out, h[0], h[1])

    self.__content_type = ct
    if ct == None:
        self.__print_content = False
    else:
        if ct.find('application/json') == -1 and ct.find('text/plain') == -1:
            self.__print_content = False

    self.__out.write('\n#------- OUTPUT DATA -------#\n\n')

    self.__write = self.__start_response(status, headers, *args)
    return self.write

  def __iter__(self):
    return self

  def write(self, data):
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return self.__write(data)

  def next(self):
    data = self.__iterable.next()
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return data

  def close(self):
    if hasattr(self.__iterable, 'close'):
      self.__iterable.close()
    self.__out.write('\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n\n\n')
    print self.__out.getvalue()

  def link(self, iterable):
    self.__iterable = iter(iterable)

class LoggingMiddleware:

  def __init__(self, application):
    self.__application = application
    self.__lock = threading.Lock()

  def __call__(self, environ, start_response):
    self.__lock.acquire()
    date = datetime.now()
    out = cStringIO.StringIO()
    out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
    print_log(out, environ)
    print_env(out, environ)
    if environ['PATH_INFO'] in ['/ingest/new', '/ingest/update']: 
      print_icont(out, environ)
    else:
      if environ['REQUEST_METHOD'] == 'GET':
          print_input_get_params(out, environ)
      elif environ['REQUEST_METHOD'] == 'POST':
          print_icont(out, environ)
    print out.getvalue() 
    iterable = LoggingInstance(start_response)
    iterable.link(self.__application(environ, iterable))
    self.__lock.release()
    return iterable

try:
    application = LoggingMiddleware(application)
except:
    print '\n++++++++++++++++++++++++++++++++++ LOGGING EXCEPTION!! ++++++++++++++++++++++++++++++++++++++++++\n'
    traceback.print_exc()

