# -*- coding: utf-8 -
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

####---- Standard Module imports ----####

import sys
import os
import json
import hashlib
import base64
from cgi import parse_qs, FieldStorage
from datetime import datetime, timedelta
from imp import load_source
from ConfigParser import ConfigParser
from collections import OrderedDict


import smtplib
from email.mime.text import MIMEText
from datetime import datetime
import traceback
from cStringIO import StringIO
import pprint
import urllib

####---- tL libraries + Config file ----####

WS_PATH = os.path.dirname(os.path.realpath(__file__))
sys.path.insert(0, WS_PATH)

from lib.libdfxp import *
from lib.libtlpauth import TLPAuth
from lib.libwsstorage import tLWSSLayer, tLWSSData

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
CFG_FILE = "%s/config.ini" % SCRIPT_DIR

####---- Cache control ----####

CACHE_CONTROL="no-cache, no-store, private, must-revalidate, max-stale=0"
PRAGMA="no-cache"
EXPIRES="Tue, 01 Jan 1980 12:00:00 GMT"
CACHE_RESP_HEADERS = [('Cache-Control', CACHE_CONTROL), 
                      ('Pragma', PRAGMA), 
                      ('Expires', EXPIRES)]
CROSSDOMAIN_RESP_HEADERS = [ ('Access-Control-Allow-Origin', '*'), 
                      ('Access-Control-Allow-Methods', 'GET, PUT, POST, DELETE, OPTIONS'), 
                      ('Access-Control-Allow-Headers', 'Content-Type, Authorization, X-Requested-With')]


#################################################
#--------------- Webservice API ----------------#
#################################################

def my_application(env, start_response, tL, tlpauth):

    if env['REQUEST_METHOD'] == 'OPTIONS':
        return answer_text("", start_response)

    if env['PATH_INFO'] in ('/translate', '/amend'):

        if env['REQUEST_METHOD'] != 'POST':
            return error_400(start_response, "This method only accepts POST requests" )

        if env.get('CONTENT_TYPE', None) != 'application/json':
            return error_400(start_response, "Content-Type must be 'application/json'")

        length = int(env.get('CONTENT_LENGTH', '0'))
        if length == 0:
            return error_400(start_response, "Content-Length must be greater than 0")

        try:
            data = env['wsgi.input'].read(length)
        except:
            return error_400(start_response, "Cannot read POST data.")

        try:
            body = json.loads(data, encoding="UTF-8")
        except:
            return error_400(start_response, "Cannot parse input JSON data (Invalid JSON).")
 
        try:
             text = body["text"]
             key = body["key"]
             language = body["language"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )
    
        if not(isinstance(text, dict)):
            return error_400(start_response, "'text' parameter must be a JSON dict." )

        if len(text.keys()) == 0:
            return error_400(start_response, "'text' parameter cannot be an empty dict." )

        resp = {}
        ref = env.get('HTTP_REFERER', None)
        if ref == None:
            resp['rcode'] = 1
            resp['rcode_description'] = 'Could not get HTTP Referer.'
            return answer_json(resp, start_response)

        wssdata = tL.web_get_website(key)
        if wssdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/web] tL.web_get_website() returned %s" % wssdata.scode)
            return error_500(start_response)
        if wssdata.results == 0:
            resp['rcode'] = 2
            resp['rcode_description'] = 'Unknown web key \'%s\'.' % key
            return answer_json(resp, start_response)
        websitedata = wssdata.data[0]
 
        if not(is_allowed_url(ref, websitedata["url_pattern"])):
            resp['rcode'] = 3
            resp['rcode_description'] = 'Unallowed HTTP Referer URL \'%s\' for key \'%s\'' % (ref, key)
            return answer_json(resp, start_response)

    elif env['PATH_INFO'] == '/getjs':

        if env['REQUEST_METHOD'] != 'GET':
            return error_400(start_response, "This method only accepts GET requests" )
        queries = parse_qs(env['QUERY_STRING'])
        try:
            key = queries['key'][0]
        except:
            return error_400(start_response, "Missing 'key' parameter.")

    else:

        allow_request_key = False
        if env['PATH_INFO'][0:7] == '/editor':
            allow_request_key = True

        encoded_params = True
        if env['REQUEST_METHOD'] == 'GET':
            queries = parse_qs(env['QUERY_STRING'])
            if "data" in queries:
                try:
                    data = base64.b64decode(queries["data"][0])
                except: 
                    return error_400(start_response, "Could not parse 'data' GET Parameter" )
            else:
                encoded_params = False

        else: # POST
            if env.get('CONTENT_TYPE', None) != 'application/json':
                return error_400(start_response, "Content-Type must be 'application/json'")
            length = int(env.get('CONTENT_LENGTH', '0'))
            if length == 0:
                return error_400(start_response, "Content-Length must be greater than 0")
            try:
                data = env['wsgi.input'].read(length)
            except:
                return error_400(start_response, "Cannot read POST data.")

        if encoded_params:
            try:
                body = json.loads(data, encoding="UTF-8")
            except:
                return error_400(start_response, "Cannot parse JSON data (Invalid JSON).")
        else:
            body = { k:queries[k][0] for k in queries.keys() }

        su = body.get("su", None)

        try:
            user = body["user"]
            auth_token = body["auth_token"]
        except:
            return error_400(start_response, "Missing required keys for user authentication in JSON data." )

        if allow_request_key:
            try:
                oid = body["id"]
            except:
                return error_400(start_response, "Missing required keys.")

            expire = body.get("expire", None)
            if expire != None: 
                try:
                    expire = int(expire)
                except:
                    return error_400(start_response, "'expire' parameter must be an integer value." )
            scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, oid, expire, allow_request_key=True)
 
        else:   
            scode, user_id, is_admin, is_request_key = check_auth_token(tL, tlpauth, user, auth_token, None, None)

        if scode == 1:
            report_error(CFG, env, "[TLP][WS][/editor] check_auth_token() returned %s" % (scode))
            return error_500(start_response)
        elif scode == 2:
            return error_401(start_response, "Authentication error: unknown user.")
        elif scode == 3:
            return error_401(start_response, "Authentication error: invalid auth token.")
        elif scode == 4:
            return error_419(start_response, "Authentication error: auth token expired.")

        if su != None:
            if is_admin:
                scode, su_id = get_user_id(tL, su)
                if scode == 1:
                    report_error(CFG, env, "[TLP][WS][/editor] get_user_id() returned %d" % scode)
                    return error_500(start_response)
                elif scode == 2:
                    return error_400(start_response, "Cannot su: unknown user.")
                user_id = su_id
            else:
                return error_400(start_response, "Cannot su: admin privileges are required.")

        resp = {}


    if env['PATH_INFO'] == '/translate':

        wssdata = tL.web_get_translations(websitedata['id'], text, language)
        if wssdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_get_translations() returned %s" % wssdata.scode)
            return error_500(start_response)
        resp['rcode'] = 0
        resp['rcode_description'] = 'Returning available translations.' 
        resp['translations'] = wssdata.data
        return answer_json(resp, start_response)

    elif env['PATH_INFO'] == '/amend':

        try:
             passwd = body["pass"]
        except: 
            return error_400(start_response, "Missing mandatory parameters/keys." )
        
        wssdata = tL.web_check_passwd(websitedata['id'], passwd)
        if wssdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_process_texts() returned %s" % wssdata.scode)
            return error_500(start_response)
        allowed = wssdata.data
        if not(allowed):
            resp['rcode'] = 4
            resp['rcode_description'] = 'Invalid password.'
            return answer_json(resp, start_response)

        if len(text.keys()) > 1: # Bulk update
            wssdata = tL.web_amend_translations(websitedata['id'], text, language)
            if wssdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_amend_translations() returned %s" % wssdata.scode)
                return error_500(start_response)
            n = wssdata.data
        elif len(text.keys()) == 1: # Single update
            src_hash = text.keys()[0]
            tl_text = text[src_hash]
            wssdata = tL.web_amend_translation(websitedata['id'], language, src_hash, tl_text)
            if wssdata.scode != 0:
                report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_amend_translation() returned %s" % wssdata.scode)
                return error_500(start_response)
            n = wssdata.data
        else:
            n = 0
        resp['rcode'] = 0
        resp['rcode_description'] = 'Updated %d translations.' % n
        resp['translations'] = n
        return answer_json(resp, start_response)

    elif env['PATH_INFO'] == '/getjs':

        api_url = "%s://%s%s/" % (env['wsgi.url_scheme'], env['SERVER_NAME'], env['SCRIPT_NAME'])
        ret = tL.web_generate_js(CFG['WEB']['JS'], key, api_url)
        if ret.scode != 0:
            report_error(CFG, env, "[TLP][WS][/web] web_generate_browser_js() returned %s" % ret.scode)
            return error_500(start_response)
        data = ret.data
        status = '200 OK'
        response_headers = [('Content-Type', 'text/javascript'),('Content-Length', str(len(data)))]
        start_response(status, response_headers + CACHE_RESP_HEADERS + CROSSDOMAIN_RESP_HEADERS)
        return [data]

    elif env['PATH_INFO'][0:7] == '/editor':

        try:
            web_key = body["id"]
        except:
            return error_400(start_response, "Missing required keys.")

        ret = tL.web_get_user_website(web_key, user_id)
        if ret.scode != 0: 
            report_error(CFG, env, "[TLP][WS][/web] web_get_user_website() returned %d" % ret.scode)
            return error_500(start_response)
        if ret.results == 0:
            resp['rcode'] = 1
            resp['rcode_description'] = 'Website key "%s" does not exist' % web_key
            return answer_json(resp, start_response)

        webdata = ret.data[0]

        if env['PATH_INFO'][7:] == '/langs':
            try:
                langs = webdata["options"]["requested_langs"].keys()
            except:
                report_error(CFG, env, "[TLP][WS][/web] Could not get requested_langs option")
                return error_500(start_response)
            langs.append(webdata['language'])
            ret = tL.web_get_language_names(langs)
            if ret.scode != 0:
                report_error(CFG, env, "[TLP][WS][/web] web_get_language_names() returned %d" % ret.scode)
                return error_500(start_response)
            langs = [ { "lang_code":c, "lang_name":ret.data[c] } for c in ret.data if c != webdata['language'] ] 
            resp['rcode'] = 0
            resp['rcode_description'] = 'Returning language list'
            resp['web_lang_code'] = webdata['language']
            resp['web_lang_name'] = ret.data[webdata['language']]
            resp['langs'] = langs
            return answer_json(resp, start_response)
               
        elif env['PATH_INFO'][7:] == '/get':

            try:
                language = body["language"]
            except:
                return error_400(start_response, "Missing required keys.")

            if webdata["language"] == language:
                ret = tL.web_get_all_source_texts(webdata["id"])
                texts = { h: { 'text':ret.data[h][0], 'status':None, 'added':str(ret.data[h][1]), 'last':str(ret.data[h][2]) } for h in ret.data.keys() } 
            else:
                ret = tL.web_get_all_translations(webdata["id"], language)
                texts = { h: { 'text':ret.data[h][0], 'status':ret.data[h][1], 'added':None, 'last':str(ret.data[h][2]) } for h in ret.data.keys() } 
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_get_all_source_texts() or web_get_all_translations() returned %d" % ret.scode)
                return error_500(start_response)
            resp['rcode'] = 0
            resp['rcode_description'] = 'Returning texts'
            resp['texts'] = texts
            return answer_json(resp, start_response)

        elif env['PATH_INFO'][7:] == '/mod':
            try:
                language = body["language"]
                text = body["text"]
            except:
                return error_400(start_response, "Missing required keys.")

            if not(isinstance(text, dict)):
                return error_400(start_response, "'text' must be a dict.")

            if len(text.keys()) > 1: # Bulk update
                wssdata = tL.web_amend_translations(webdata['id'], text, language)
                if wssdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_amend_translations() returned %s" % wssdata.scode)
                    return error_500(start_response)
                n = wssdata.data
            elif len(text.keys()) == 1: # Single update
                src_hash = text.keys()[0]
                tl_text = text[src_hash]
                wssdata = tL.web_amend_translation(webdata['id'], language, src_hash, tl_text)
                if wssdata.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/web/translate] tL.web_amend_translation() returned %s" % wssdata.scode)
                    return error_500(start_response)
                n = wssdata.data
            else:
                n = 0
            resp['rcode'] = 0
            resp['rcode_description'] = 'Updated %d translations.' % n
            resp['translations'] = n
            return answer_json(resp, start_response)

        else:
            return error_400_API(start_response)

    elif env['PATH_INFO'][0:7] == '/manage':

        if env['PATH_INFO'][7:] == '/list':

            ret = tL.web_get_websites_list(user_id)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_get_websites_list() returned %d" % ewt.scode)
                return error_500(start_response)
            resp = [ { "name":x["name"], "key":x["key"], "language":x["language"], "url_pattern":x["url_pattern"], "options":x["options"], "active":x["active"] } for x in ret.data ]
            return answer_json(resp, start_response)

        elif env['PATH_INFO'][7:] == '/add':

            try:
                name = body["name"]
                lang = body["language"]
                url_pattern = body["url_pattern"]
                req_langs = body["requested_langs"]
            except:
                return error_400(start_response, "Missing required keys.")

            if not(isinstance(url_pattern, unicode)):
                return error_400(start_response, "'url_pattern' must be a string")

            if not(isinstance(lang, unicode)):
                return error_400(start_response, "'language' must be a string")

            if not(isinstance(name, unicode)):
                return error_400(start_response, "'name' must be a string")

            ok, err = check_requested_langs_format(req_langs)
            if not(ok):
                err_code, err_info = err
                if err_code == 400:
                    return error_400(start_response, err_info)
                elif err_code == 500:
                    return error_500(start_response)

            # Check language code
            ret = tL.check_language_code(lang)
            if ret.scode != 0:
                report_error(CFG, env, "[TLP][WS][/web] check_language_code() returned %d" % ret.scode)
                return error_500(start_response)
            if not(ret.data):
                resp['rcode'] = 1
                resp['rcode_description'] = 'Unknown language code "%s"' % lang
                return answer_json(resp, start_response)

            # Check url_pattern
            ret = tL.web_get_user_websites_by_url_pattern(url_pattern, user_id)
            if ret.scode != 0:
                report_error(CFG, env, "[TLP][WS][/web] web_get_websites_by_url_pattern() returned %d" % ret.scode)
                return error_500(start_response)
            if ret.results > 0:
                keys_str = ", ".join([ d['key'] for d in ret.data ])
                resp['rcode'] = 2
                resp['rcode_description'] = 'Provided URL pattern conflicting with existing websites: %s' % keys_str
                return answer_json(resp, start_response)

            ret = tL.web_add_website(name, lang, url_pattern, req_langs, user_id)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_add_website() returned %d" % ret.scode)
                return error_500(start_response)
    
            resp['rcode'] = 0
            resp['rcode_description'] = 'New website successfully added'
            resp['id'] = ret.data
            return answer_json(resp, start_response)

        elif env['PATH_INFO'][7:] == '/edit':

            try:
                web_key = body["id"]
            except:
                return error_400(start_response, "Missing required keys.")

            enabled = body.get("enabled", None)
            req_langs = body.get("requested_langs", None)
            url_pattern = body.get("url_pattern", None)

            if enabled == None and req_langs == None and url_pattern == None:
                return error_400(start_response, "You must provide at least one option: 'enabled', 'requested_langs', 'url_pattern'.")

            if enabled != None:
                if not(isinstance(enabled, bool)):
                    return error_400(start_response, "'enabled' must be a boolean")

            if req_langs != None:
                ok, err = check_requested_langs_format(req_langs)
                if not(ok):
                    err_code, err_info = err
                    if err_code == 400:
                        return error_400(start_response, err_info)
                    elif err_code == 500:
                        return error_500(start_response)

            if url_pattern != None:
                if not(isinstance(url_pattern, unicode)):
                    return error_400(start_response, "'url_pattern' must be a string")

            ret = tL.web_get_user_website(web_key, user_id)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_get_user_website() returned %d" % ret.scode)
                return error_500(start_response)
            if ret.results == 0:
                resp['rcode'] = 1
                resp['rcode_description'] = 'Website key "%s" does not exist' % web_key
                return answer_json(resp, start_response)

            if url_pattern != None:
                ret = tL.web_get_user_websites_by_url_pattern(url_pattern, user_id)
                if ret.scode != 0:
                    report_error(CFG, env, "[TLP][WS][/web] web_get_websites_by_url_pattern() returned %d" % ret.scode)
                    return error_500(start_response)
                if ret.results > 0:
                    keys_str = ", ".join([ d['key'] for d in ret.data ])
                    resp['rcode'] = 2
                    resp['rcode_description'] = 'Provided URL pattern conflicting with existing websites: %s' % keys_str
                    return answer_json(resp, start_response)
    
            ret = tL.web_edit_website(web_key, enabled, url_pattern, req_langs)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_edit_website() returned %d" % ret.scode)
                return error_500(start_response)
 
            resp['rcode'] = 0
            resp['rcode_description'] = 'Website successfully edited'
            return answer_json(resp, start_response)

        elif env['PATH_INFO'][7:] == '/delete':

            if not(is_admin):
                resp['rcode'] = 1
                resp['rcode_description'] = 'Operation only permitted to admin users.'
                return answer_json(resp, start_response)

            try:
                web_key = body["id"]
            except:
                return error_400(start_response, "Missing required keys.")

            purge = body.get("purge", False)

            ret = tL.web_get_user_website(web_key, user_id)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_get_user_website() returned %d" % ret.scode)
                return error_500(start_response)
            if ret.results == 0:
                resp['rcode'] = 1
                resp['rcode_description'] = 'Website key "%s" does not exist' % web_key
                return answer_json(resp, start_response)

            web_id = ret.data[0]['id']

            ret = tL.web_delete_website(web_id, purge)
            if ret.scode != 0: 
                report_error(CFG, env, "[TLP][WS][/web] web_delete_website() returned %d" % ret.scode)
                return error_500(start_response)
 
            resp['rcode'] = 0
            resp['rcode_description'] = 'Website successfully deleted'
            return answer_json(resp, start_response)

        else:

            return error_400_API(start_response)

    elif env['PATH_INFO'] == '/systems':

        WSSdata = tL.tLint_systems_getUserSystems(user_id) 
        if WSSdata.scode != 0:
            report_error(CFG, env, "[TLP][WS][/systems] tL.tLint_systems_getUserSystems() returned %s" % (WSSdata.scode))
            return error_500(start_response)
        resp['mt']  = [ { 'id':s['id'], 'source_lang':s['lang1'], 'target_lang':s['lang2'], 'name':s['name'], 'description':s['description'], 'topic_adaptation':s['features_tadapt'], 'processing_time':s['const_time'] } for s in WSSdata.data if s['type'] == 'mt' ]
        return answer_json(resp, start_response)


    else: # query not known, bad request
        return error_400_API(start_response)
    

def application(env, start_response):
    try:
        tL = tLWSSLayer(CFG['STG']['DB_NAME'], CFG['STG']['DB_USER'], CFG['STG']['DB_HOST'], tLWSSLayer.OBJ_SPEECH, passwd=CFG['STG']['DB_PASS'])
        tlpauth = TLPAuth(CFG['AUT']['KEY'])
        return my_application(env, start_response, tL, tlpauth)
    except:
        report_error(CFG, env, "[TLP][WS] Exception")
        return error_500(start_response)

def answer_json(Dresp, start_response):
   data = json.dumps(Dresp, indent=4)
   status = '200 OK'
   response_headers = [('Content-Type', 'application/json'),('Content-Length', str(len(data)))]
   start_response(status, response_headers + CACHE_RESP_HEADERS + CROSSDOMAIN_RESP_HEADERS)
   return [data]

def answer_text(data, start_response):
   status = '200 OK'
   response_headers = [('Content-Type', 'text/plain'),('Content-Length', str(len(data)))]
   start_response(status, response_headers + CACHE_RESP_HEADERS + CROSSDOMAIN_RESP_HEADERS)
   return [data]


#############################################################
#----------------- Error return functions ------------------#
#############################################################

def error_400(start_response, text=None):
    status = '400 Bad Request'
    response_headers = [('Content-Type', 'text/plain')]
    if text == None:
        data = "Bad Request"
    else:
        data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]

def error_400_API(start_response):
    status = '400 Bad Request'
    data = CFG['MISC']['HTML_MSG']
    response_headers = [('Content-Type', 'text/html'),('Content-Length', str(len(data)))]
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]



def error_401(start_response, text):
    status = '401 Unauthorized'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_419(start_response, text):
    status = '419 Authentication Timeout'
    response_headers = [('Content-Type', 'text/plain')]
    data = text
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]


def error_500(start_response):
    status = '500 Internal Error'
    data = "Internal Server Error"
    response_headers = [('Content-Type', 'text/plain'),('Content-Length', str(len(data)))]
    start_response(status, response_headers + CACHE_RESP_HEADERS)
    return [data]

ENV_KEYS = [ 'HTTP_REFERER', 'CONTENT_TYPE', 'SCRIPT_NAME', 'REQUEST_METHOD', 'PATH_INFO', 'REQUEST_URI', 'QUERY_STRING', 'wsgi.url_scheme', 'SERVER_NAME', 'SERVER_PORT', 'SCRIPT_FILENAME', 'HTTP_HOST', 'REMOTE_ADDR', 'REMOTE_PORT' ]

def report_error(CFG, env, subject, extra_data=None):
    if sys.exc_type != None:
        excpt = True
        excpt_str = traceback.format_exc()
        sys.stderr.write("\n************ API EXCEPTION ************\n\n")
        sys.stderr.write("%s\n" % excpt_str)
        #traceback.print_exc()
    else:
        excpt = False

    if CFG['MAI']['ENABLED'] and CFG['MAI']['TO'] != None:
        try:
            out = cStringIO.StringIO()
            out.write("DATE: %s\n" % datetime.now().strftime("%d/%m/%Y %H:%M:%S %Z"))
            out.write("EXCEPTION: %s\n" % str(excpt))
            if excpt:
                out.write("\n+++++++++++++++++++++++++++++++++ EXCEPTION ++++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % excpt_str)
            out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
            print_log(out, env)
            print_env(out, env, space=1)
            if env['PATH_INFO'] == '/ingest': 
              print_input_get_params(out, env, space=1)
            else:
              if env['REQUEST_METHOD'] == 'GET':
                  print_input_get_params(out, env, space=1)
              elif env['REQUEST_METHOD'] == 'POST':
                  print_icont(out, env)

            if extra_data != None:
                out.write("\n++++++++++++++++++++++++++++++++ EXTRA DATA +++++++++++++++++++++++++++++++++++++++++\n\n")
                out.write("%s\n" % extra_data)
                    
            content = out.getvalue() 
            send_mail(CFG['MAI']['SMTP_SERVER'], CFG['MAI']['FROM'], subject, content.decode('utf-8'), CFG['MAI']['TO'])
        except:
            sys.stderr.write("\n************ SEND_MAIL EXCEPTION ************\n")
            traceback.print_exc()
    
def send_mail(smtp_server, src_mail, subject, content, to, cc=None, cco=None):
    # Create e-mail data
    msg = MIMEText(content.encode('UTF-8'))
    msg['Subject'] = subject.encode('UTF-8')
    msg['From'] = src_mail
    msg['To'] = ",".join(to)
    addrl = []
    addrl = addrl + to
    if cc != None: 
        msg['cc'] = ",".join(cc)
        addrl = addrl + cc
    if cco != None: 
        msg['cco'] = ",".join(cco)
        addrl = addrl + cco
    # Send e-mail
    smtp = smtplib.SMTP(smtp_server)
    smtp.sendmail(src_mail, addrl, msg.as_string())



#####################################################
#----------------- Misc functions ------------------#
#####################################################

def check_auth_token(tL, tlpauth, user_name, given_token, oid, expire, author=None, author_conf=None, allow_request_key=True):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None, None, None
    if WSSdata.results == 0:
        return 2, None, None, None
    user_id = WSSdata.data[0]['id']
    user_pass = WSSdata.data[0]['password']
    is_admin = WSSdata.data[0]['is_admin']
    secret_key = tlpauth.get_secret_key(user_name.encode('utf-8'), user_pass)
    if secret_key == given_token:
        return 0, user_id, is_admin, False
    elif allow_request_key:
        if expire == None: 
            return 3, None, None, None
        request_key = tlpauth.get_request_key(user_name.encode('utf-8'), user_pass, oid.encode('utf-8'), expire, author, author_conf)
        if author == None:
            given_token = given_token[:tlpauth.request_subkey_length]
        if request_key == given_token:
            ts = int((datetime.utcnow() - datetime(1970,01,01)).total_seconds())
            if ts > expire:
                return 4, None, None, None
            return 0, user_id, is_admin, True
        else:
            return 3, None, None, None
    else:
        return 3, None, None, None

def get_user_id(tL, user_name):
    WSSdata = tL.tLint_auth_get_user_by_username(user_name)
    if WSSdata.scode != 0:
        return 1, None
    if WSSdata.results == 0:
        return 2, None
    user_id = WSSdata.data[0]['id']
    return 0, user_id


def is_allowed_url(url, url_pattern):
    myurl = re.sub('https://|http://|www\.', '', str(url))
    mypat = re.sub('https://|http://|www\.', '', re.sub('\*', '.*', url_pattern))
    if re.match(mypat, myurl):
        return True
    else:
        return False

def c(s, m=0):
    if m == 0: # general
        return (re.subn(r'[\r\b\n$&`´|{}()\[\];/\\"\'#*~^<>=+!?¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 1: # filenames, allow ( ) +
        return (re.subn(r'[\r\b\n$&`´|{}\[\];/\\"\'#*~^<>=!?¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 2: # urls
        return (re.subn(r'[\r\b\n`´|{}\\"\'^<>¡¿]|--', '', unicode(s))[1] > 0)
    elif m == 3: # titles, allow ?¿¡! ( )
        return (re.subn(r'[\r\b\n$&`´|{}\[\];/\\"\'#*~^<>=+]|--', '', unicode(s))[1] > 0)


def check_requested_langs_format(rl):
    if not(isinstance(rl, dict)):
        return (False, (400, "'requested_langs' option: must be a dict"))
    for lang in rl.keys():
        if c(lang):
            return (False, (400, "'requested_langs' option: Invalid language code in 'requested_langs' (contains forbidden characters)"))
        ld = rl[lang]
        if not(isinstance(ld, dict)):
            return (False, (400, "'requested_langs' option: 'requested_langs'->'%s' value must be a dict" % lang.encode("utf-8")))
        for opt in ld.keys():
            if opt == "sid":
                if not(isinstance(ld[opt], int)):
                    return (False, (400, "'requested_langs' option: 'requested_langs'->'%s'->'sid' key value must be an integer" % (lang.encode("utf-8"))))
            elif opt == "tlpath":
                if not(isinstance(ld[opt], list)):
                    return (False, (400, "'requested_langs' option: 'requested_langs'->'%s'->'tlpath' key value must be a list" % (lang.encode("utf-8"))))
                for i in ld[opt]:
                    if not(isinstance(i, dict)):
                        return (False, (400, "'requested_langs' option: 'requested_langs'->'%s'->'tlpath' items must be dicts" % (lang.encode("utf-8"))))
                    if "l" not in i.keys(): 
                        return (False, (400, "'requested_langs' option: 'requested_langs'->'%s'->'tlpath' items must contain at least a 'l' key" % (lang.encode("utf-8"))))
                    for ik in i.keys():
                        if ik == "l":
                            if c(i[ik]):
                                return (False, (400, "'requested_langs' option: Invalid 'l' key value in 'requested_langs'->'%s'->'tlpath' dict item (contains forbidden characters)" % (lang.encode("utf-8"))))
                        elif ik == "sid":
                            if not(isinstance(i[ik], int)):
                                return (False, (400, "'requested_langs' option: 'sid' key value in 'requested_langs'->'%s'->'tlpath' item must be an integer" % (lang.encode("utf-8"))))
                        elif ik == "skip":
                            if not(isinstance(i[ik], bool)):
                                return (False, (400, "'requested_langs' option: 'skip' key value in 'requested_langs'->'%s'->'tlpath' item must be a boolean" % (lang.encode("utf-8"))))
                        else:
                            return (False, (400, "'requested_langs' option: Unknown key '%s' provided in 'requested_langs'->'%s'->'tlpath' dict item" % (ik.encode("utf-8"), lang.encode("utf-8"))))
            else:
                return (False, (400, "'requested_langs' option: Unknown option '%s' provided in 'requested_langs'->'%s' dict" % (opt.encode("utf-8"), lang.encode("utf-8"))))
    return (True, None)




def parse_config(cfg_fd): 

    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}

    # General
    CFG['AUT'] = {}
    CFG['AUT']['KEY'] = config.get('authentication', 'key_generator')
    CFG['STG'] = {}
    CFG['STG']['DB_NAME'] = config.get('storage', 'db_name')
    CFG['STG']['DB_USER'] = config.get('storage', 'db_user')
    CFG['STG']['DB_HOST'] = config.get('storage', 'db_host')
    if CFG['STG']['DB_HOST'] != None: 
      if CFG['STG']['DB_HOST'].strip() == "": CFG['STG']['DB_HOST'] = None
    CFG['STG']['DB_PASS'] = config.get('storage', 'db_passwd')
    if CFG['STG']['DB_PASS'] != None: 
      if CFG['STG']['DB_PASS'].strip() == "": CFG['STG']['DB_PASS'] = None
    CFG['URL'] = {}
    CFG['URL']['USE'] = config.getboolean('media_urls', 'use_urls')
    CFG['URL']['BASEURL'] = config.get('media_urls', 'base_url')
    CFG['MAI'] = {}
    CFG['MAI']['ENABLED'] = config.getboolean('mailing', 'enabled')
    CFG['MAI']['FROM'] = config.get('mailing', 'from_address')
    CFG['MAI']['SMTP_SERVER'] = config.get('mailing', 'smtp_server')
    CFG['WEB'] = {}
    CFG['WEB']['JS'] = "%s/wt.js" % SCRIPT_DIR
    try:
        CFG['MAI']['TO'] = config.get('mailing', 'to_address').replace(' ', '').split(',')
    except:
        CFG['MAI']['TO'] = None
    CFG['MISC'] = {}
    CFG['MISC']['HTML_MSG'] = config.get('misc', 'html_msg')

    return CFG


# Load Config file
with open(CFG_FILE) as cfg_fd:
    CFG = parse_config(cfg_fd)


##################################################
#----------- Logging WSGI Middelware ------------#
##################################################

import threading


def print_key_value(out, k, v, space=30):
    out.write("%s:%s%s\n" % (str(k), ' ' * max((space-len(k)),1), str(v)))

def print_width(out, data, width=80):
    ndata=''; i = 0; c = 0
    while i < len(data):
        ndata += data[i]
        if c == width: 
            ndata += '\n'
            c = 0
        else: c += 1
        i += 1
    out.write("%s\n" % ndata)

def print_log(out, env):
  out.write('PATH: %s\n' % env['SCRIPT_NAME'])
  out.write('INTERFACE: %s\n' % env['PATH_INFO'])

def print_env(out, environ, space=30):
  out.write('\n#------- INPUT ENVIRONMENT -------#\n\n')
  fenv = {}
  for k in ENV_KEYS:
    print_key_value(out, k, environ.get(k, "-"), space=space)

def print_input_get_params(out, env, space=30):    
  out.write('\n#------- INPUT DATA (GET) -------#\n\n')
  encoded_params = True
  queries = parse_qs(env['QUERY_STRING'])
  if "data" in queries:
      try:
          data = base64.b64decode(queries["data"][0])
      except: 
         out.write("Could not parse 'data' GET Parameter\n")
  else:
      encoded_params = False
  
  if encoded_params:
      try:
          body = json.loads(data, encoding="UTF-8")
      except:
         out.write("Cannot parse JSON data (Invalid JSON).\n")
      else:
         out.write("%s\n" % json.dumps(body, indent=4))
  else:
      body = { k:queries[k][0] for k in queries.keys() }
      for k in body:
           print_key_value(out, k, body.get(k, "-"), space=space)

  out.write("\n-- Get URL:\n\n")
  url = "http://%s%s" % (env['SERVER_NAME'], env['REQUEST_URI'])
  out.write("%s\n" % url)

def print_icont(out, environ):
   
  length = int(environ.get('CONTENT_LENGTH', '0'))
  
  out.write('\n#------- INPUT DATA (POST) -------#\n\n')
  if length > 0 and length < 2147483648: # 2GB
      ret = ""
      icontent_fp = cStringIO.StringIO()
      icontent_fp.write(environ['wsgi.input'].read(length))
      if environ.get('CONTENT_TYPE', '').find('application/json') >= 0:
          icont = icontent_fp.getvalue()
          try:
              ret = str(icont)
          except:
              return
      elif environ.get('CONTENT_TYPE','').find('multipart/form-data') >= 0: 
          icontent_fp.seek(0, os.SEEK_SET)
          try:
              fs = FieldStorage(fp=icontent_fp,environ=environ)
          except:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if "json" not in fs.keys():
              out.write("%s\n" % icontent_fp.getvalue())
              return
          if fs['json'].type.find("application/json") <  0:
              out.write("%s\n" % icontent_fp.getvalue())
              return
          #out.write("%s\n" % icontent_fp.getvalue())
          ret = fs.getvalue("json")

      try:
          jret = json.loads(ret)
          jtret = json.dumps(jret, indent=4)
          encurl = urllib.urlencode(jret)
      except:
         out.write("\n-- Text:\n")
         out.write("%s\n" % ret)
      else:
         out.write("\n-- Text:\n\n")
         out.write("%s\n" % jtret)
         out.write("\n-- Get URL:\n\n")
         url = "http://%s%s?%s" % (environ['SERVER_NAME'], environ['REQUEST_URI'], encurl)
         out.write("%s\n" % url)

      icontent_fp.seek(0, os.SEEK_SET)
      environ['wsgi.input'] = icontent_fp

  else:
     out.write("None\n")


class LoggingInstance:
  def __init__(self, start_response):
    self.__start_response = start_response
    self.__print_content = True
    self.__content_type = ""
    self.__out = cStringIO.StringIO()

  def __call__(self, status, headers, *args):
    self.__out.write('\n#------- OUTPUT HEADERS -------#\n\n')
    print_key_value(self.__out, 'Response-Code', status)
    ct = None
    for h in headers:
        if h[0] == 'Content-Type': 
            ct = h[1]
        print_key_value(self.__out, h[0], h[1])

    self.__content_type = ct
    if ct.find('application/json') == -1 and ct.find('text/plain') == -1:
      self.__print_content = False

    self.__out.write('\n#------- OUTPUT DATA -------#\n\n')

    self.__write = self.__start_response(status, headers, *args)
    return self.write

  def __iter__(self):
    return self

  def write(self, data):
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return self.__write(data)

  def next(self):
    data = self.__iterable.next()
    if self.__print_content:
      self.__out.write("%s\n" % data)
    else:
      self.__out.write("[NOT PRINTED]\n")
    return data

  def close(self):
    if hasattr(self.__iterable, 'close'):
      self.__iterable.close()
    self.__out.write('\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n\n\n')
    print self.__out.getvalue()

  def link(self, iterable):
    self.__iterable = iter(iterable)

class LoggingMiddleware:

  def __init__(self, application):
    self.__application = application
    self.__lock = threading.Lock()

  def __call__(self, environ, start_response):
    self.__lock.acquire()
    date = datetime.now()
    out = cStringIO.StringIO()
    out.write('\n++++++++++++++++++++++++++++++++++ API CALL ++++++++++++++++++++++++++++++++++++++++++\n\n')
    print_log(out, environ)
    print_env(out, environ)
    if environ['PATH_INFO'] in ['/ingest/new', '/ingest/update']: 
      print_icont(out, environ)
    else:
      if environ['REQUEST_METHOD'] == 'GET':
          print_input_get_params(out, environ)
      elif environ['REQUEST_METHOD'] == 'POST':
          print_icont(out, environ)
    print out.getvalue() 
    iterable = LoggingInstance(start_response)
    iterable.link(self.__application(environ, iterable))
    self.__lock.release()
    return iterable

try:
    application = LoggingMiddleware(application)
except:
    print '\n++++++++++++++++++++++++++++++++++ LOGGING EXCEPTION!! ++++++++++++++++++++++++++++++++++++++++++\n'
    traceback.print_exc()

