var TLP_API_URL = 'https://ttp.mllp.upv.es/api/v3/web/editor/';
jQuery(document).ready(function() {
  /* Plug-in variables */
  var g_langs = null;
  var g_source_data = {};
  var g_target_lang_code = null;
  var g_target_lang_name = '';
  var g_target_data = {};

  /* Auxiliar functions */
  var getUrlParameter = function getUrlParameter(sParam) {
    var sPageURL = decodeURIComponent(window.location.search.substring(1)),
      sURLVariables = sPageURL.split('&'),
      sParameterName,
      i;

    for (i=0; i<sURLVariables.length; i++) {
      sParameterName = sURLVariables[i].split('=');
      if (sParameterName[0] === sParam) {
        return sParameterName[1] === undefined ? true : sParameterName[1];
      }
    }
  };

  var getLangName = function getLangName(lang_code) {
    var lang_name = lang_code;
    jQuery.each(g_langs["langs"], function() {
      if (this["lang_code"] == lang_code) lang_name = this["lang_name"];
    });
    return lang_name;
  };

  var fatalError = function fatalError(message) {
    jQuery('.container').empty().append('<div class="alert alert-danger" style="margin-top: 10%;"><strong>Error loading Web Translation Editor:</strong> '+message+'</div>');
    console.log(message);
  };

  var warningError = function warningError(message) {
    console.log(message);
  }

  /* API functions */
  var apiLangs = function apiLangs(onsuccess) {
    jQuery.ajax({
      url: TLP_API_URL+'langs?id='+TLP_REQUEST["id"]+'&auth_token='+TLP_REQUEST["request_key"]+'&user='+TLP_REQUEST["api_user"]+("expire" in TLP_REQUEST ? '&expire='+TLP_REQUEST["expire"] : ''),
      dataType: "json",
      cache: false
    }).done(function(response) {
      if (response["rcode"] == 0) {
        g_langs = response;
        jQuery.each(response["langs"], function() {
          jQuery('#webtl_target_language').append('<option value="'+this["lang_code"]+'">'+this["lang_name"]+'</option>');
        });
        if (typeof onsuccess !== 'undefined') onsuccess();
      }
      else {
        fatalError('Could not retrieve available languages from the server: '+response["rcode_description"]);
      }
    }).fail(function() {
      fatalError('Could not retrieve available languages from the server.');
    });
  }

  var apiGet = function apiGet(lang_code, onsuccess) {
    if (g_langs == null) {
      warningError('Could not load translations data until available languages have been retrieved.');
      return;
    }
    jQuery.ajax({
      url: TLP_API_URL+'get?id='+TLP_REQUEST["id"]+'&auth_token='+TLP_REQUEST["request_key"]+'&user='+TLP_REQUEST["api_user"]+("expire" in TLP_REQUEST ? '&expire='+TLP_REQUEST["expire"] : '')+'&language='+lang_code,
      dataType: "json",
      cache: false
    }).done(function(response) {
      if (response["rcode"] == 0) {
        if (lang_code == g_langs["web_lang_code"]) g_source_data = response;
        else {
          g_target_lang_code = lang_code;
          g_target_data = response;
          // Set filtering options
          var n_auto = 0;
          var n_revised = 0;
          jQuery.each(g_target_data["texts"], function(hash_value) {
            if (this["status"] == "automatic") n_auto++;
            else if (this["status"] == "revised") n_revised++;
          });
          jQuery('#webtl_filter').empty();
          jQuery('#webtl_filter').append('<option value="all">All ('+(n_auto+n_revised)+')</option><option value="auto">Automatic ('+n_auto+')</option><option value="revised">Revised ('+n_revised+')</option>');
          // Set completion percentage
          jQuery('#webtl_progress_bar .progress-bar').attr('data-valuenow', Math.round((n_revised/(n_auto+n_revised))*100)).css('width', Math.round((n_revised/(n_auto+n_revised))*1000)/10+'%').text(Math.round((n_revised/(n_auto+n_revised))*1000)/10+'%');
        }
        if (typeof onsuccess !== 'undefined') onsuccess();
      }
      else {
        warningError('Could not retrieve "'+lang_code+'" language data from the server: '+response["rcode_description"]);
      }
    }).fail(function() {
      warningError('Could not retrieve "'+lang_code+'" language data from the server.');
    });
  }

  var apiMod = function apiMod(lang_code, hash_value, text, onsuccess, onerror) {
    var dataJson = {"id": TLP_REQUEST["id"], "auth_token": TLP_REQUEST["request_key"], "user": TLP_REQUEST["api_user"], "language": lang_code, "text": {}};
    if ("expire" in TLP_REQUEST) dataJson["expire"] = TLP_REQUEST["expire"];
    dataJson["text"][hash_value] = text;
    jQuery.ajax({
      url: TLP_API_URL+'mod',
      method: "POST",
      data: JSON.stringify(dataJson),
      contentType: "application/json",
      dataType: "json",
      crossDomain: true,
      cache: false
    }).done(function(response) {
      if (response["rcode"] == 0) if (typeof onsuccess !== 'undefined') onsuccess();
      else warningError('Could not save translation changes: '+response["rcode_description"]);
    }).fail(function() {
      warningError('Could not save translation changes: there was no response from the server.');
      if (typeof onerror !== 'undefined') onerror();
    });
  }

  /* UI functions */
  var uiFillTranslationsTable = function uiFillTranslationsTable(filter) {
    if (typeof filter === 'undefined') {
      if (typeof getUrlParameter('filter') !== 'undefined') filter = getUrlParameter('filter');
      else filter = "all";
    }
    jQuery('#webtl_translations_table tbody').empty();
    if ("texts" in g_source_data && "texts" in g_target_data) {
      // Fill translations table
      jQuery('#webtl_translations_table thead tr th:first-child').text(g_langs["web_lang_name"]);
      jQuery('#webtl_translations_table thead tr th:nth-child(2)').text(getLangName(g_target_lang_code));
      jQuery.each(g_target_data["texts"], function(hash_value) {
        if (filter == "all" || (filter == "auto" && this["status"] == "automatic") || (filter == "revised" && this["status"] == "revised")) jQuery('#webtl_translations_table tbody').append('<tr data-hash="'+hash_value+'" class="'+(this["status"] == "revised" ? "success" : "")+'"><td>'+g_source_data["texts"][hash_value]["text"]+'</td><td contenteditable="true">'+this["text"]+'</td></tr>');
      });
      jQuery(window).trigger("resize");
    }
  };

  jQuery('#webtl_target_language').on("change", function() {
    apiGet(jQuery(this).val(), uiFillTranslationsTable);
  });

  jQuery('#webtl_filter').on("change", function() {
    uiFillTranslationsTable(jQuery(this).val());
  });

  jQuery('#webtl_translations_table').on("focus", "tbody tr td:nth-child(2)", function() {
    jQuery(this).parent().removeClass('success');
  });

  jQuery('#webtl_translations_table').on("blur", "tbody tr td:nth-child(2)", function() {
    var thisText = jQuery(this).text();
    jQuery(this).parent().removeClass('success').addClass('info text-muted');
    jQuery(this).attr('contenteditable', 'false').append('<div class="webtl_status_icon" style="float: right; animation: spin 2s ease infinite;"><i class="material-icons">sync</i></div>');
    var thisElem = this;
    apiMod(g_target_lang_code, jQuery(this).parent().attr('data-hash'), thisText, function() {
      jQuery(thisElem).parent().removeClass('info text-muted').addClass('success');
      jQuery(thisElem).attr('contenteditable', 'true');
      jQuery(thisElem).children('.webtl_status_icon').remove();
    }, function() {
      jQuery(thisElem).parent().removeClass('info text-muted');
      jQuery(thisElem).attr('contenteditable', 'true');
      jQuery(thisElem).children('.webtl_status_icon').remove();
    });
  });

  jQuery(window).on("resize", function() {
    var tableHeight = jQuery(window).height()-jQuery('#webtl_topbar').outerHeight()-20;
    jQuery('.pre-scrollable').css('max-height', tableHeight+'px');
  });

  /* Main execution */
  var TLP_REQUEST_B64STRING = getUrlParameter('request');
  var TLP_REQUEST = {};

  if (typeof TLP_REQUEST_B64STRING === 'undefined') {
    fatalError('Missing "request" parameter in URL');
  }
  else {
    try {
      TLP_REQUEST = JSON.parse(window.atob(decodeURIComponent(TLP_REQUEST_B64STRING)));
    } catch (e) {
      fatalError('Wrong request format: unable to parse request string.');
      return;
    }
    if (Object.keys(TLP_REQUEST).length == 0) {
      fatalError('Wrong request format: unable to parse request string.');
    }
    else if (!("request_key" in TLP_REQUEST)) {
      console.log(TLP_REQUEST);
      fatalError('Missing request "request_key" parameter');
    }
    else if (!("api_user" in TLP_REQUEST)) {
      fatalError('Missing request "api_user" parameter');
    }
    else if (!("id" in TLP_REQUEST)) {
      fatalError('Missing request "id" parameter');
    }
    else {
      apiLangs(function() {
        apiGet(g_langs["web_lang_code"], uiFillTranslationsTable);
        if ("lang" in TLP_REQUEST) {
          jQuery('#webtl_target_language').val(TLP_REQUEST["lang"]);
        }
        jQuery('#webtl_target_language').trigger("change");
      });
    }
  }
});
